<?xml version="1.0" encoding="UTF-8"?>
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/cmyk"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/cmyk"
             exclude-result-prefixes="this xs math map array errors point util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <!--
    : this:cmyk()
    :-->
   <xsl:function name="this:cmyk" as="map(xs:string, item()*)">
      <xsl:param name="c" as="xs:double"/>
      <xsl:param name="m" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      point:point($c, $m, $y, $k)
    '/>
   </xsl:function>

   <!--
    : this:c()
    :-->
   <xsl:function name="this:c" as="xs:double">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($cmyk)
    '/>
   </xsl:function>

   <!--
    : this:m()
    :-->
   <xsl:function name="this:m" as="xs:double">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($cmyk)
    '/>
   </xsl:function>

   <!--
    : this:y()
    :-->
   <xsl:function name="this:y" as="xs:double">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($cmyk)
    '/>
   </xsl:function>

   <!--
    : this:k()
    :-->
   <xsl:function name="this:k" as="xs:double">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pw($cmyk)
    '/>
   </xsl:function>

   <!--
    : this:saturate()
    :-->
   <xsl:function name="this:saturate" as="map(xs:string, item()*)">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:cmyk(
        util:clamp(this:c($cmyk), 0.0, 1.0),
        util:clamp(this:m($cmyk), 0.0, 1.0),
        util:clamp(this:y($cmyk), 0.0, 1.0),
        util:clamp(this:k($cmyk), 0.0, 1.0)
      )
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "cmyk("||
        this:c($cmyk)||","||
        this:m($cmyk)||","||
        this:y($cmyk)||","||
        this:k($cmyk)||
      ")"
    '/>
   </xsl:function>

</xsl:package>
<!--cmyk.xsl-->
