<?xml version="1.0" encoding="UTF-8"?>
<!--:-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/space"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:cmyk="http://mathling.com/colour/cmyk"
             xmlns:hsl="http://mathling.com/colour/hsl"
             xmlns:lab="http://mathling.com/colour/lab"
             xmlns:lch="http://mathling.com/colour/lch"
             xmlns:luv="http://mathling.com/colour/luv"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:rgb="http://mathling.com/colour/rgb"
             xmlns:stop="http://mathling.com/colour/stop"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xyz="http://mathling.com/colour/xyz"
             version="3.0"
             name="http://mathling.com/colour/space"
             exclude-result-prefixes="this xs math map array util point rgb hsl xyz luv lch lab cmyk stop">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/rgb" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/hsl" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/xyz" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/luv" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/lch" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/lab" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/cmyk" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/stop" package-version="*"/>

   <xsl:variable name="this:π" as="xs:double">
      <xsl:sequence select='math:pi()'/>
   </xsl:variable>
   <xsl:variable name="this:π_div_180" as="xs:double">
      <xsl:sequence select='$this:π div 180'/>
   </xsl:variable>
   <xsl:variable name="this:inv_π_div_180" as="xs:double">
      <xsl:sequence select='180 div $this:π'/>
   </xsl:variable>

   <xsl:variable name="this:REF-UV" as="map(xs:string,item()*)">
      <xsl:sequence select='luv:luv(0E0, 0.19783000664283680764E0, 0.46831999493879100370E0)'/>
   </xsl:variable>
   <xsl:variable name="this:κ" as="xs:double">
      <xsl:sequence select='903.29629629629629629630E0'/>
   </xsl:variable>
   <xsl:variable name="this:ε" as="xs:double">
      <xsl:sequence select='0.00885645167903563082E0'/>
   </xsl:variable>
   <xsl:variable name="this:δ" as="xs:double">
      <xsl:sequence select='0.00000001E0'/>
   </xsl:variable>

   <!--
    : For XYX=>RGB 
    :-->
   <xsl:variable name="this:M" as="map(xs:string,item()*)*">
      <xsl:sequence select='(
        point:point(3.24096994190452134377E0, -1.53738317757009345794E0, -0.49861076029300328366E0),
        point:point(-0.96924363628087982613E0, 1.87596750150772066772E0, 0.04155505740717561247E0),
        point:point(0.05563007969699360846E0, -0.20397695888897656435E0, 1.05697151424287856072E0)
    )'/>
   </xsl:variable>

   <!--
    : For RGB=>XYZ
    :-->
   <xsl:variable name="this:M-INV" as="map(xs:string,item()*)*">
      <xsl:sequence select='(
      point:point(0.41239079926595948129E0, 0.35758433938387796373E0, 0.18048078840183428751E0),
      point:point(0.21263900587151035754E0, 0.71516867876775592746E0, 0.07219231536073371500E0),
      point:point(0.01933081871559185069E0, 0.11919477979462598791E0, 0.95053215224966058086E0)
    )'/>
   </xsl:variable>

   <!--
    : XYZ => Lab
    :-->
   <xsl:variable name="this:LM" as="map(xs:string,item()*)*">
      <xsl:sequence select='(
      point:point(0.4121656120E0, 0.5362752080E0, 0.0514575653E0),
      point:point(0.2118591070E0, 0.6807189584E0, 0.1074065790E0),
      point:point(0.0883097947E0, 0.2818474174E0, 0.6302613616E0),
      point:point(0.2104542553E0, 0.7936177850E0, -0.0040720468E0),
      point:point(1.9779984951E0, -2.4285922050E0, 0.4505937099E0),
      point:point(0.0259040371E0, 0.7827717662E0, -0.8086757660E0)
    )'/>
   </xsl:variable>

   <!--
    : Lab => XYZ
    :-->
   <xsl:variable name="this:LM_INV" as="map(xs:string,item()*)*">
      <xsl:sequence select='(
      point:point(1E0, 0.3963377774E0, 0.2158037573E0),
      point:point(1E0, -0.1055613458E0, -0.0638541728E0),
      point:point(1E0, -0.0894841775E0, -1.2914855480E0),
      point:point(4.0767245293E0, -3.3072168827E0, 0.2307590544E0),
      point:point(-1.2681437731E0, 2.6093323231E0, -0.3411344290E0),
      point:point(-0.0041119885E0, -0.7034763098E0, 1.7068625689E0)
    )'/>
   </xsl:variable>

   <!--
    : this:line()
    :-->
   <xsl:function name="this:line" as="map(xs:string, item())">
      <xsl:param name="slope" as="xs:double"/>
      <xsl:param name="intercept" as="xs:double"/>
      <xsl:sequence select='
      map {
        "slope": $slope,
        "intercept": $intercept
      }
    '/>
   </xsl:function>

   <!--
    : this:slope()
    :-->
   <xsl:function name="this:slope" as="xs:double">
      <xsl:param name="line" as="map(xs:string, item())"/>
      <xsl:sequence select='
      $line("slope")
    '/>
   </xsl:function>

   <!--
    : this:intercept()
    :-->
   <xsl:function name="this:intercept" as="xs:double">
      <xsl:param name="line" as="map(xs:string, item())"/>
      <xsl:sequence select='
      $line("intercept")
    '/>
   </xsl:function>

   <!--
    : this:hsluv-to-rgb()
    : Construct an RGB value from an HSL one. RGB given in that order, values [0.0, 1.0]
    :
    : $h: Hue [0.0, 360.0]
    : $s: Saturation [0.0, 100.0]
    : $l: Lightness [0.0, 100.0]
    :-->
   <xsl:function name="this:hsluv-to-rgb" as="xs:double*">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="s" as="xs:double"/>
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      let $rgb := this:hsluv-to-rgb(hsl:hsl($h, $s, $l))
      return (
        rgb:r($rgb), rgb:g($rgb), rgb:b($rgb)
      )
    '/>
   </xsl:function>

   <!--
    : this:hsluv-to-rgb()
    :-->
   <xsl:function name="this:hsluv-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:hsluv-to-lch($hsl)=>
        this:lch-to-luv()=>
        this:luv-to-xyz()=>
        this:xyz-to-rgb()
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-hsluv()
    : Construct an HSL value from an RGB one. HSL given in that order, values
    : ([0.0, 360.0], [0.0, 100.0], [0.0, 100.0])
    :
    : $r: Red [0.0, 1.0]
    : $g: Green [0.0, 1.0]
    : $b: Blue [0.0, 1.0]
    :-->
   <xsl:function name="this:rgb-to-hsluv" as="xs:double*">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="g" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $hsl := this:rgb-to-hsluv(rgb:rgb($r, $g, $b))
      return (
        hsl:h($hsl), hsl:s($hsl), hsl:l($hsl)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-hsluv()
    :-->
   <xsl:function name="this:rgb-to-hsluv" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:rgb-to-xyz($rgb)=>
        this:xyz-to-luv()=>
        this:luv-to-lch()=>
        this:lch-to-hsluv()
    '/>
   </xsl:function>

   <!--
    : this:hpluv-to-rgb()
    : Construct an RGB value from an HSL one. RGB given in that order, values [0.0, 1.0]
    :
    : $h: Hue [0.0, 360.0]
    : $s: Saturation [0.0, 100.0]
    : $l: Lightness [0.0, 100.0]
    :-->
   <xsl:function name="this:hpluv-to-rgb" as="xs:double*">
      <xsl:param name="h" as="xs:double"/>
      <xsl:param name="s" as="xs:double"/>
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      let $rgb := this:hpluv-to-rgb(hsl:hsl($h, $s, $l))
      return (
        rgb:r($rgb), rgb:g($rgb), rgb:b($rgb)
      )
    '/>
   </xsl:function>

   <!--
    : this:hpluv-to-rgb()
    :-->
   <xsl:function name="this:hpluv-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:hpluv-to-lch($hsl)=>
        this:lch-to-luv()=>
        this:luv-to-xyz()=>
        this:xyz-to-rgb()
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-hpluv()
    : Construct an HPL value from a RGB one. HSL given in that order, values
    : ([0.0, 360.0], [0.0, 100.0], [0.0, 100.0])
    : 
    : HPLuv does not contain all the colors of RGB, so converting
    : arbitrary RGB to it may generate invalid HPLuv colors.
    :
    : $r: Red [0.0, 1.0]
    : $g: Green [0.0, 1.0]
    : $b: Blue [0.0, 1.0]
    :-->
   <xsl:function name="this:rgb-to-hpluv" as="xs:double*">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="g" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $hsl := this:rgb-to-hpluv(rgb:rgb($r, $g, $b))
      return (
        hsl:h($hsl), hsl:s($hsl), hsl:l($hsl)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-hpluv()
    :-->
   <xsl:function name="this:rgb-to-hpluv" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:rgb-to-xyz($rgb)=>
        this:xyz-to-luv()=>
        this:luv-to-lch()=>
        this:lch-to-hpluv()
    '/>
   </xsl:function>

   <!--
    : this:lch-to-hpluv()
    :-->
   <xsl:function name="this:lch-to-hpluv" as="map(xs:string, item()*)">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := lch:l($lch) return
      let $c := lch:c($lch) return
      let $h := lch:h($lch) 
      return
        hsl:hsl(
          (: Greys: disambiguate hue :)
          if ($c &lt; $this:δ) then 0E0 else $h,
          (: White and black: disambiguate saturation :)
          if ($l > 100E0 - $this:δ or $l &lt; $this:δ)
          then 0E0
          else $c div this:max_safe_chroma_for_l($l) * 100E0,
          if ($l > 100E0 - $this:δ) then 100E0
          else if ($l &lt; $this:δ) then 0E0
          else $l
        )
    '/>
   </xsl:function>

   <!--
    : this:hpluv-to-lch()
    :-->
   <xsl:function name="this:hpluv-to-lch" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $h := hsl:h($hsl) return
      let $s := hsl:s($hsl) return
      let $l := hsl:l($hsl)
      return
        lch:lch(
          $l,
          (: White and black: disambiguate chroma :)
          if ($l > 100 - $this:δ or $l &lt; $this:δ)
          then 0E0
          else (this:max_safe_chroma_for_l($l) div 100E0) * $s,
          (: Greys: disambiguate hue :)
          if ($s &lt; $this:δ) then 0E0 else $h
        )
    '/>
   </xsl:function>

   <!--
    : this:lch-to-hsluv()
    :-->
   <xsl:function name="this:lch-to-hsluv" as="map(xs:string, item()*)">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := lch:l($lch) return
      let $c := lch:c($lch) return
      let $h := lch:h($lch)
      return
        hsl:hsl(
          (: Greys: disambiguate hue :)
          if ($c &lt; $this:δ) then 0E0 else $h,
          (: White and black: disambiguate saturation :)
          if ($l > 100E0 - $this:δ or $l &lt; $this:δ)
          then 0E0
          else ($c div this:max-chroma-for-lh($l, $h)) * 100E0,
          (: $l :)
          if ($l > 100E0 - $this:δ) then 100E0
          else if ($l &lt; $this:δ) then 0E0
          else $l
        )
    '/>
   </xsl:function>

   <!--
    : this:hsluv-to-lch()
    :-->
   <xsl:function name="this:hsluv-to-lch" as="map(xs:string, item()*)">
      <xsl:param name="hsl" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $h := hsl:h($hsl) return
      let $s := hsl:s($hsl) return
      let $l := hsl:l($hsl)
      return
        lch:lch(
          if ($l > 100E0 - $this:δ) then 100E0
          else if ($l &lt; $this:δ) then 0E0
          else $l,
          (: White and black: disambiguate chroma :)
          if ($l > 100E0 - $this:δ or $l &lt; $this:δ)
          then 0E0
          else (this:max-chroma-for-lh($l, $h) div 100E0) * $s,
          (: Greys: disambiguate hue :)
          if ($s &lt; $this:δ) then 0E0 else $h
        )
    '/>
   </xsl:function>

   <!--
    : this:lch-to-luv()
    :-->
   <xsl:function name="this:lch-to-luv" as="map(xs:string, item()*)">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := lch:l($lch) return
      let $c := lch:c($lch) return
      let $h := lch:h($lch) return
      let $hrad := $h * $this:π_div_180 return
      let $u := math:cos($hrad) * $c return
      let $v := math:sin($hrad) * $c
      return
        luv:luv($l, $u, $v)    
    '/>
   </xsl:function>

   <!--
    : this:luv-to-lch()
    :-->
   <xsl:function name="this:luv-to-lch" as="map(xs:string, item()*)">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := luv:l($luv) return
      let $u := luv:u($luv) return
      let $v := luv:v($luv) return
      let $c := math:sqrt($u * $u + $v * $v) return
      let $h := 
        (: Greys: disambiguate hue :)
        if ($c &lt; $this:δ) then (
          0E0
        ) else (
          let $h := math:atan2($v, $u) * $this:inv_π_div_180
          return
            if ($h &lt; 0E0)
            then $h + 360E0
            else $h
        )
      return
        point:point($l, $c, $h)
    '/>
   </xsl:function>

   <!--
    : this:luv-to-xyz()
    :-->
   <xsl:function name="this:luv-to-xyz" as="map(xs:string, item()*)">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := luv:l($luv) return
      let $u := luv:u($luv) return
      let $v := luv:v($luv) 
      return (
        if ($l &lt;= $this:δ) then (
          (: Black will create a divide-by-zero error. :)
          xyz:xyz(0E0, 0E0, 0E0)
        ) else (
          let $var-u := $u div (13E0 * $l) + luv:u($this:REF-UV) return
          let $var-v := $v div (13E0 * $l) + luv:v($this:REF-UV) return
          let $y := this:l2y($l) return
          let $x := -(9E0 * $y * $var-u) div (($var-u - 4E0) * $var-v - $var-u * $var-v) return
          let $z := (9E0 * $y - (15E0 * $var-v * $y) - ($var-v * $x)) div (3E0 * $var-v)
          return
            xyz:xyz($x, $y, $z)
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:xyz-to-luv()
    :-->
   <xsl:function name="this:xyz-to-luv" as="map(xs:string, item()*)">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $x := xyz:x($xyz) return
      let $y := xyz:y($xyz) return
      let $z := xyz:z($xyz) return
      let $var-u := (4E0 * $x) div ($x + (15E0 * $y) + (3E0 * $z)) return
      let $var-v := (9E0 * $y) div ($x + (15E0 * $y) + (3E0 * $z)) return
      let $l := this:y2l($y) return
      let $u := 13E0 * $l * ($var-u - luv:u($this:REF-UV)) return
      let $v := 13E0 * $l * ($var-v - luv:v($this:REF-UV))
      return
        if ($l &lt; $this:δ)
        then luv:luv($l, 0E0, 0E0)
        else luv:luv($l, $u, $v)
    '/>
   </xsl:function>

   <!--
    : this:y2l()
    : https://en.wikipedia.org/wiki/CIELUV
    : In these formulas, Yn refers to the reference white point. We are using
    : illuminant D65, so Yn (see refY in Maxima file) equals 1. The formula is
    : simplified accordingly.
    :-->
   <xsl:function name="this:y2l" as="xs:double">
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      if ($y &lt;= $this:ε)
      then $y * $this:κ
      else 116E0 * math:pow($y, 1 div 3) - 16E0  (: cbrt :)
    '/>
   </xsl:function>

   <!--
    : this:l2y()
    :-->
   <xsl:function name="this:l2y" as="xs:double">
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      if ($l &lt;= 8E0) then (
        $l div $this:κ
      ) else (
        let $x := ($l + 16E0) div 116E0
        return ($x * $x * $x)
      )
    '/>
   </xsl:function>

   <!--
    : this:xyz-to-rgb()
    :-->
   <xsl:function name="this:xyz-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $r := util:clamp(this:from-linear(point:dot($this:M[1], $xyz)), 0E0, 1E0) return
      let $g := util:clamp(this:from-linear(point:dot($this:M[2], $xyz)), 0E0, 1E0) return
      let $b := util:clamp(this:from-linear(point:dot($this:M[3], $xyz)), 0E0, 1E0)
      return (
        if ($r != $r or $g != $g or $b != $b) then (
           trace((), "XYZ=>RGB=>NaN: "||util:quote($xyz))
        ) else (),
        rgb:rgb($r, $g, $b)
      )
    '/>
   </xsl:function>

   <!--
    : this:xyza-to-rgba()
    :-->
   <xsl:function name="this:xyza-to-rgba" as="map(xs:string, item()*)">
      <xsl:param name="xyza" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $r := util:clamp(this:from-linear(point:dot($this:M[1], $xyza)), 0E0, 1E0) return
      let $g := util:clamp(this:from-linear(point:dot($this:M[2], $xyza)), 0E0, 1E0) return
      let $b := util:clamp(this:from-linear(point:dot($this:M[3], $xyza)), 0E0, 1E0) return
      let $a := xyz:alpha($xyza)
      return rgb:rgba($r, $g, $b, $a)
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-xyz()
    :-->
   <xsl:function name="this:rgb-to-xyz" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $r := rgb:r($rgb) return
      let $g := rgb:g($rgb) return
      let $b := rgb:b($rgb) return
      let $rgbl := point:point(this:to-linear($r), this:to-linear($g), this:to-linear($b)) return
      let $x := point:dot($this:M-INV[1], $rgbl) return
      let $y := point:dot($this:M-INV[2], $rgbl) return
      let $z := point:dot($this:M-INV[3], $rgbl)
      return (
        if ($x != $x or $y != $y or $z != $z) then (
          trace((), "RGB=>XYZ=>NaN: "||util:quote($rgb))
        ) else (),
        xyz:xyz($x, $y, $z)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgba-to-xyza()
    :-->
   <xsl:function name="this:rgba-to-xyza" as="map(xs:string, item()*)">
      <xsl:param name="rgba" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $r := rgb:r($rgba) return
      let $g := rgb:g($rgba) return
      let $b := rgb:b($rgba) return
      let $a := rgb:alpha($rgba) return
      let $rgbl := point:point(this:to-linear($r), this:to-linear($g), this:to-linear($b)) return
      let $x := point:dot($this:M-INV[1], $rgbl) return
      let $y := point:dot($this:M-INV[2], $rgbl) return
      let $z := point:dot($this:M-INV[3], $rgbl)
      return
        xyz:xyza($x, $y, $z, $a)
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-lab()
    : Construct an Lab value from an RGB one.
    :
    : $r: Red [0.0, 1.0]
    : $g: Green [0.0, 1.0]
    : $b: Blue [0.0, 1.0]
    :-->
   <xsl:function name="this:rgb-to-lab" as="xs:double*">
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="g" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $hsl := this:rgb-to-lab(rgb:rgb($r, $g, $b))
      return (
        hsl:h($hsl), hsl:s($hsl), hsl:l($hsl)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-lab()
    :-->
   <xsl:function name="this:rgb-to-lab" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:rgb-to-xyz($rgb)=>this:xyz-to-lab()
    '/>
   </xsl:function>

   <!--
    : this:xyz-to-lab()
    :-->
   <xsl:function name="this:xyz-to-lab" as="map(xs:string, item()*)">
      <xsl:param name="xyz" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := point:dot($this:LM[1], $xyz) return
      let $m := point:dot($this:LM[2], $xyz) return
      let $s := point:dot($this:LM[3], $xyz) return
      let $lms := (: cbrt :)
        point:point(
          math:pow($l, 1 div 3),
          math:pow($m, 1 div 3),
          math:pow($s, 1 div 3)
        )
      return
        point:point(
          point:dot($lms, $this:LM[4]),
          point:dot($lms, $this:LM[5]),
          point:dot($lms, $this:LM[6])
        )
    '/>
   </xsl:function>

   <!--
    : this:lab-to-rgb()
    : Construct an RGB value from an LaB one. RGB given in that order, values [0.0, 1.0]
    :
    : $l: Perceived lightness
    : $a: how green/red [-1.0,1.0]
    : $b: how blue/yellow [-1.0,1.0]
    :-->
   <xsl:function name="this:lab-to-rgb" as="xs:double*">
      <xsl:param name="l" as="xs:double"/>
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      let $rgb := this:lab-to-rgb(lab:lab($l, $a, $b))
      return (
        rgb:r($rgb), rgb:g($rgb), rgb:b($rgb)
      )
    '/>
   </xsl:function>

   <!--
    : this:lab-to-rgb()
    :-->
   <xsl:function name="this:lab-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:lab-to-xyz($lab)=>this:xyz-to-rgb()
    '/>
   </xsl:function>

   <!--
    : this:lab-to-xyz()
    :-->
   <xsl:function name="this:lab-to-xyz" as="map(xs:string, item()*)">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := point:dot($this:LM_INV[1], $lab) return
      let $m := point:dot($this:LM_INV[2], $lab) return
      let $s := point:dot($this:LM_INV[3], $lab) return
      let $lms := 
        point:point(
          math:pow($l, 3),
          math:pow($m, 3),
          math:pow($s, 3)
        )
      return
        point:point(
          point:dot($lms, $this:LM_INV[4]),
          point:dot($lms, $this:LM_INV[5]),
          point:dot($lms, $this:LM_INV[6])
        )
    '/>
   </xsl:function>

   <!-- For RGB conversions -->

   <!--
    : this:from-linear()
    :-->
   <xsl:function name="this:from-linear" as="xs:double">
      <xsl:param name="c" as="xs:double"/>
      <xsl:sequence select='
      if ($c &lt;= 0.0031308E0)
      then 12.92E0 * $c
      else 1.055E0 * math:pow($c, 1E0 div 2.4E0) - 0.055E0
    '/>
   </xsl:function>

   <!--
    : this:to-linear()
    :-->
   <xsl:function name="this:to-linear" as="xs:double">
      <xsl:param name="c" as="xs:double"/>
      <xsl:sequence select='
      if ($c > 0.04045E0)
      then math:pow(($c + 0.055E0) div 1.055E0, 2.4E0)
      else $c div 12.92E0
    '/>
   </xsl:function>

   <!--
    : this:dot-product()
    :-->
   <xsl:function name="this:dot-product" as="xs:double">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($p1)*point:px($p2) +
      point:py($p1)*point:py($p2) + 
      point:pz($p1)*point:pz($p2)
    '/>
   </xsl:function>

   <!--
    : this:max_safe_chroma_for_l()
    :-->
   <xsl:function name="this:max_safe_chroma_for_l" as="xs:double">
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      let $bounds := this:get-bounds($l) return
      let $min-len-squared :=
        min((
          xs:double("INF"),
          for $i in 1 to 6 return
          let $m1 := this:slope($bounds[$i]) return
          let $b1 := this:intercept($bounds[$i]) return
          (: x where line intersects with perpendicular running though (0, 0) :)
          let $line2 := this:line(-1E0 div $m1, 0E0) return
          let $x := this:intersect-line-line($bounds[$i], $line2)
          return this:dist-from-pole-squared($x, $b1 + $x * $m1)
        ))
      return math:sqrt($min-len-squared)
    '/>
   </xsl:function>

   <!--
    : this:max-chroma-for-lh()
    :-->
   <xsl:function name="this:max-chroma-for-lh" as="xs:double">
      <xsl:param name="l" as="xs:double"/>
      <xsl:param name="h" as="xs:double"/>
      <xsl:sequence select='
     let $hrad := $h * $this:π_div_180 return
     let $bounds := this:get-bounds($l)
     return
       min((
         xs:double("INF"),
         for $i in 1 to 6 return
         let $len := this:ray-length-until-intersect($hrad, $bounds[$i])
         return if ($len >= 0) then $len else ()
       ))
    '/>
   </xsl:function>

   <!--
    : this:intersect-line-line()
    :-->
   <xsl:function name="this:intersect-line-line" as="xs:double">
      <xsl:param name="line1" as="map(xs:string, item()*)"/>
      <xsl:param name="line2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (this:intercept($line1) - this:intercept($line2)) div (this:slope($line2) - this:slope($line1))
    '/>
   </xsl:function>

   <!--
    : this:dist-from-pole-squared()
    :-->
   <xsl:function name="this:dist-from-pole-squared" as="xs:double">
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      $x * $x + $y * $y
    '/>
   </xsl:function>

   <!--
    : this:ray-length-until-intersect()
    :-->
   <xsl:function name="this:ray-length-until-intersect" as="xs:double">
      <xsl:param name="θ" as="xs:double"/>
      <xsl:param name="line" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:intercept($line) div (math:sin($θ) - this:slope($line) * math:cos($θ))
    '/>
   </xsl:function>

   <!--
    : this:get-bounds()
    :-->
   <xsl:function name="this:get-bounds" as="map(xs:string, item()*)*">
      <xsl:param name="l" as="xs:double"/>
      <xsl:sequence select='
      let $sub1 := math:pow($l + 16.0E0, 3) div 1560896E0 return
      let $sub2 := if ($sub1 > $this:ε) then $sub1 else ($l div $this:κ) return
      for $channel in 1 to 3 return
      let $m1 := point:px($this:M[$channel]) return
      let $m2 := point:py($this:M[$channel]) return
      let $m3 := point:pz($this:M[$channel]) return
      for $t in 0 to 1 return
      let $top1 := (284517E0 * $m1 - 94839E0 * $m3) * $sub2 return
      let $top2 := (838422E0 * $m3 + 769860E0 * $m2 + 731718E0 * $m1) * $l * $sub2 -  769860E0 * $t * $l return
      let $bottom := (632260E0 * $m3 - 126452E0 * $m2) * $sub2 + 126452E0 * $t
      return (
        this:line($top1 div $bottom, $top2 div $bottom)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-cmyk()
    : Construct a CMYK point from an RGB one. 
    :
    : $rgb: RGB point
    :-->
   <xsl:function name="this:rgb-to-cmyk" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $k := 1E0 - max((rgb:r($rgb), rgb:g($rgb), rgb:b($rgb))) return
      let $one-k := 1E0 - $k
      return (
        if ($one-k = 0) then (
          cmyk:cmyk(
            1E0 - rgb:r($rgb),
            1E0 - rgb:g($rgb),
            1E0 - rgb:b($rgb),
            $k
          )
        ) else (
          cmyk:cmyk(
            ($one-k - rgb:r($rgb)) div $one-k,
            ($one-k - rgb:g($rgb)) div $one-k,
            ($one-k - rgb:b($rgb)) div $one-k,
            $k
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:cmyk-to-rgb()
    : Construct am RGB point from a CMYK one. 
    :
    : $cmyk: CMYK point
    :-->
   <xsl:function name="this:cmyk-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="cmyk" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $one-k := 1E0 - cmyk:k($cmyk)
      return (
        if ($one-k = 0) then (
          rgb:rgb(
            1E0 - cmyk:c($cmyk),
            1E0 - cmyk:m($cmyk),
            1E0 - cmyk:y($cmyk)
          )
        ) else (
          rgb:rgb(
            util:clamp((1E0 - cmyk:c($cmyk)) * $one-k, 0, 1),
            util:clamp((1E0 - cmyk:m($cmyk)) * $one-k, 0, 1),
            util:clamp((1E0 - cmyk:y($cmyk)) * $one-k, 0, 1)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:greyscale-value-to-rgb()
    :-->
   <xsl:function name="this:greyscale-value-to-rgb" as="map(xs:string, item()*)">
      <xsl:param name="grey" as="xs:double"/>
      <xsl:sequence select='
      rgb:rgb($grey, $grey, $grey)
    '/>
   </xsl:function>

   <!--
    : this:greyscale-value-to-xyz()
    :-->
   <xsl:function name="this:greyscale-value-to-xyz" as="map(xs:string, item()*)">
      <xsl:param name="grey" as="xs:double"/>
      <xsl:sequence select='
      let $greyl := this:to-linear($grey) return
      let $rgbl := point:point($greyl, $greyl, $greyl) return
      let $x := point:dot($this:M-INV[1], $rgbl) return
      let $y := point:dot($this:M-INV[2], $rgbl) return
      let $z := point:dot($this:M-INV[3], $rgbl)
      return (
        if ($x != $x or $y != $y or $z != $z) then (
          trace((), "RGB=>XYZ=>NaN: "||util:quote($grey))
        ) else (),
        xyz:xyz($x, $y, $z)
      )
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-greyscale()
    :-->
   <xsl:function name="this:rgb-to-greyscale" as="map(xs:string, item()*)">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $l := util:clamp(this:rgb-to-greyscale-value($rgb), 0.0, 1.0)
      return rgb:rgb($l, $l, $l)
    '/>
   </xsl:function>

   <!--
    : this:rgb-to-greyscale-value()
    :-->
   <xsl:function name="this:rgb-to-greyscale-value" as="xs:double">
      <xsl:param name="rgb" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: Rec601Luma :)
      rgb:r($rgb) * 0.298839 + rgb:g($rgb) * 0.586811 + rgb:b($rgb) * 0.114350
    '/>
   </xsl:function>

   <!--
    : this:alpha-blend()
    : Blend colours channel by channel: assumes we have RGBA space
    :-->
   <xsl:function name="this:alpha-blend" as="map(xs:string, item()*)">
      <xsl:param name="rgba1" as="map(xs:string, item()*)"/>
      <xsl:param name="rgba2" as="map(xs:string, item()*)"/>
      <xsl:param name="alpha" as="xs:double"/>
      <xsl:sequence select='
      this:xyza-to-rgba(
        xyz:alpha-blend(
          this:rgba-to-xyza($rgba1),
          this:rgba-to-xyza($rgba2),
          $alpha
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:interpolate-colour()
    : Get the interpolated colour for the given location in the gradient.
    : We use linear interpolation over the XYZA space and we're assuming here
    : the gradient points are RGBA. 
    : TODO: space selection
    : 
    : $t: fraction of the curve to interpolate in range [0,1]
    : $gradient-points: points in the gradient to interpolate
    :-->
   <xsl:function name="this:interpolate-colour" as="map(xs:string, item()*)">
      <xsl:param name="t" as="xs:double"/>
      <xsl:param name="gradient-points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $n := count($gradient-points) return
      let $t := util:clamp($t, stop:offset($gradient-points[1]), stop:offset($gradient-points[$n])) return
      
      (: Find the first control point has a value larger than $t; $n+1 if none :)
      let $index-pos :=
        (
          for $i in 1 to count($gradient-points) return (
            if (stop:offset($gradient-points[$i]) > $t) then $i else ()
          )
          (: for $pt at $i in $gradient-points
          where stop:offset($pt) > $t
          return $i:)
          ,
          $n + 1
        )[1] return
      (: Find the two nearest control points to interpolate between :)
      let $index0 := util:clamp($index-pos - 1, 1, $n) return
      let $index1 := util:clamp($index-pos, 1, $n )
      return (
        (: 
         : They may be the same if we are of the edges of the range
         :)
        if ($index0 = $index1) then (
          stop:point($gradient-points[$index1])
        ) else (
          (: Compute the alpha value used for linear interpolation :)
          let $input0 := stop:offset($gradient-points[$index0]) return
          let $input1 := stop:offset($gradient-points[$index1]) return
          let $alpha := ($t - $input0) div ($input1 - $input0)
          (: Perform linear interpolation and return :)
          return
            this:alpha-blend(
              stop:point($gradient-points[$index0]),
              stop:point($gradient-points[$index1]),
              $alpha
            )
        )
      )
    '/>
   </xsl:function>

</xsl:package>
<!--colour-space.xsl-->
