<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Lab points
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/lab"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/lab"
             exclude-result-prefixes="this xs math map array point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <!--
    : this:lab()
    : Make an Lab colour point (see https://bottosson.github.io/posts/oklab/)
    :
    : $l: perceived lightness [??]
    : $a: how green/red [-1.0,1.0]
    : $b: how blue/yellow [-1.0,1.0]
    :-->
   <xsl:function name="this:lab" as="map(xs:string, item()*)">
      <xsl:param name="l" as="xs:double"/>
      <xsl:param name="a" as="xs:double"/>
      <xsl:param name="b" as="xs:double"/>
      <xsl:sequence select='
      point:point($l, $a, $b)
    '/>
   </xsl:function>

   <!--
    : this:l()
    : Accessor for L value of the point
    :
    : $lab The point
    :-->
   <xsl:function name="this:l" as="xs:double">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($lab)
    '/>
   </xsl:function>

   <!--
    : this:a()
    : Accessor for a value of the point
    :
    : $lab The point
    :-->
   <xsl:function name="this:a" as="xs:double">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($lab)
    '/>
   </xsl:function>

   <!--
    : this:b()
    : Accessor for b value of the point
    :
    : $lab The point
    :-->
   <xsl:function name="this:b" as="xs:double">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($lab)
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert LAB point value to string representation suitable for debugging
    :
    : $lab The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="lab" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "lab("||
        this:l($lab)||","||
        this:a($lab)||","||
        this:b($lab)||
      ")"
    '/>
   </xsl:function>

</xsl:package>
<!--lab.xsl-->
