<?xml version="1.0" encoding="UTF-8"?>
<!--
 : LCH points (used for internal colour-space conversions)
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/lch"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/lch"
             exclude-result-prefixes="this xs math map array point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>

   <!--
    : this:lch()
    : Make an LCH colour point
    :
    : $l: Lightness [0.0, 1.0]
    : $c: Chroma [0.0, 360.0]
    : $h: Hue [0.0,INF]
    :-->
   <xsl:function name="this:lch" as="map(xs:string, item()*)">
      <xsl:param name="l" as="xs:double"/>
      <xsl:param name="c" as="xs:double"/>
      <xsl:param name="h" as="xs:double"/>
      <xsl:sequence select='
      point:point($l, $c, $h)
    '/>
   </xsl:function>

   <!--
    : this:l()
    : Accessor for luminance value of the point
    :
    : $lch The point
    :-->
   <xsl:function name="this:l" as="xs:double">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($lch)
    '/>
   </xsl:function>

   <!--
    : this:c()
    : Accessor for c value of the point
    :
    : $lch The point
    :-->
   <xsl:function name="this:c" as="xs:double">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($lch)
    '/>
   </xsl:function>

   <!--
    : this:h()
    : Accessor for h value of the point
    :
    : $lch The point
    :-->
   <xsl:function name="this:h" as="xs:double">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($lch)
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert LCH point value to string representation suitable for debugging
    : or use in CSS for browsers that support it
    :
    : $lch The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="lch" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "lch("||
        100*this:l($lch)||"% "||
        this:c($lch)||" "||
        this:h($lch)||
      ")"
    '/>
   </xsl:function>

</xsl:package>
<!--lch.xsl-->
