<?xml version="1.0" encoding="UTF-8"?>
<!--
 : LUV points
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/luv"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/luv"
             exclude-result-prefixes="this xs math map array point">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>

   <!--
    : this:luv()
    : Make an LUV colour point (see https://en.wikipedia.org/wiki/CIELUV)
    :
    : $l: [0.0,1.0]
    : $u: [-1.0,1.0] ("typically")
    : $v: [-1.0,1.0] ("typically")
    :-->
   <xsl:function name="this:luv" as="map(xs:string, item()*)">
      <xsl:param name="l" as="xs:double"/>
      <xsl:param name="u" as="xs:double"/>
      <xsl:param name="v" as="xs:double"/>
      <xsl:sequence select='
      point:point($l, $u, $v)
    '/>
   </xsl:function>

   <!--
    : this:l()
    : Accessor for L value of the point
    :
    : $luv The point
    :-->
   <xsl:function name="this:l" as="xs:double">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:px($luv)
    '/>
   </xsl:function>

   <!--
    : this:u()
    : Accessor for u value of the point
    :
    : $luv The point
    :-->
   <xsl:function name="this:u" as="xs:double">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:py($luv)
    '/>
   </xsl:function>

   <!--
    : this:v()
    : Accessor for v value of the point
    :
    : $luv The point
    :-->
   <xsl:function name="this:v" as="xs:double">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:pz($luv)
    '/>
   </xsl:function>

   <!--
    : this:to-string()
    : Convert LUV point value to string representation suitable for debugging
    :
    : $luv The point
    :-->
   <xsl:function name="this:to-string" as="xs:string">
      <xsl:param name="luv" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "luv("||
        this:l($luv)||","||
        this:u($luv)||","||
        this:v($luv)||
      ")"
    '/>
   </xsl:function>

</xsl:package>
<!--luv.xsl-->
