<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Colour stops: for when we convert gradients into points in colourspace
 : for computing interpolations and blends.
 :
 : Copyright© Mary Holstege 2021-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/colour/stop"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/colour/stop"
             exclude-result-prefixes="this xs math map array util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <!--
    : this:stop()
    :-->
   <xsl:function name="this:stop" as="map(xs:string, item())">
      <xsl:param name="space" as="xs:string"/>
      <xsl:param name="offset" as="xs:double"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:assert($space=("xyz","rgb","hsl","luv","lab","lch"),"Unknown colour space"),
      map {
        "space": $space,
        "offset": $offset,
        "point": $point
      }
    '/>
   </xsl:function>

   <!--
    : this:space()
    :-->
   <xsl:function name="this:space" as="xs:string">
      <xsl:param name="stop" as="map(xs:string, item())"/>
      <xsl:sequence select='
      $stop("space")
    '/>
   </xsl:function>

   <!--
    : this:offset()
    :-->
   <xsl:function name="this:offset" as="xs:double">
      <xsl:param name="stop" as="map(xs:string, item())"/>
      <xsl:sequence select='
      $stop("offset")
    '/>
   </xsl:function>

   <!--
    : this:point()
    :-->
   <xsl:function name="this:point" as="map(xs:string, item()*)">
      <xsl:param name="stop" as="map(xs:string, item())"/>
      <xsl:sequence select='
      $stop("point")
    '/>
   </xsl:function>

</xsl:package>
<!--stop.xsl-->
