<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Error catalogues. Port of errors.xqy.
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 !-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
             xmlns:this="http://mathling.com/core/errors"
             name="http://mathling.com/core/errors"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"  
             xmlns:errors="http://mathling.com/errors"
             exclude-result-prefixes="this xs map array errors"
             version="3.0">

<xsl:expose component="function" names="this:*" visibility="public"/>
<xsl:expose component="variable" names="this:*" visibility="public"/>
 
<xsl:variable name="this:ERRORNS" as="xs:string">http://mathling.com/errors</xsl:variable>
<xsl:variable name="this:MESSAGES" as="xs:string">../Messages/</xsl:variable>
<xsl:variable name="this:LANG" as="xs:string">en</xsl:variable>
<xsl:variable name="this:QUOTE-LIMIT" as="xs:integer">100</xsl:variable>

<xsl:function name="this:error">
  <xsl:param name="code" as="xs:string"/>

  <xsl:sequence select='this:error($code, ())'/>
</xsl:function>

<xsl:function name="this:error">
  <xsl:param name="code" as="xs:string"/>
  <xsl:param name="data" as="item()*"/>

  <xsl:variable name="catalog" as="document-node()?">
    <xsl:if test='doc-available($this:MESSAGES||substring-before($code,"-")||"-"||$this:LANG||".xml")'>
      <xsl:sequence select='doc($this:MESSAGES||substring-before($code,"-")||"-"||$this:LANG||".xml")'/>
    </xsl:if>
  </xsl:variable>
  <xsl:variable name="error" as="element(errors:error)?">
    <xsl:sequence select='$catalog/errors:errors/errors:error[errors:code=$code]'/>
  </xsl:variable>
  <xsl:sequence select='
  if (empty($error))
  then error(QName($this:ERRORNS,"ml:"||$code), "Error("||this:quote($data, $this:QUOTE-LIMIT)||")")
  else error(QName($this:ERRORNS,"ml:"||$code), this:format($error, $data))
  '/>
</xsl:function>

<xsl:function name="this:format" as="xs:string">
  <xsl:param name="error" as="element(errors:error)"/>
  <xsl:param name="data" as="item()*"/>
  <xsl:sequence select='
  fold-left(1 to count($data), string($error/errors:format),
    function($string as xs:string, $i as xs:integer) as xs:string {
      replace($string, "%"||$i, this:quote($data[$i], $this:QUOTE-LIMIT))
    }
  )
  '/>
</xsl:function>

<xsl:function name="this:quote" as="xs:string">
  <xsl:param name="data" as="item()*"/>
  <xsl:sequence select='string-join(this:quote-items($data),",")'/>
</xsl:function>

<xsl:function name="this:quote-items" as="xs:string*" visibility="private">
  <xsl:param name="data" as="item()*"/>

  <xsl:for-each select='$data'>
    <xsl:variable name="val" select="." as="item()"/>
    <xsl:choose>
      <xsl:when test='$val instance of xs:QName'>
        <xsl:variable name="prefix" as="xs:string?" select='prefix-from-QName($val)'/>
        <xsl:variable name="uri" as="xs:string?" select='namespace-uri-from-QName($val)'/>
        <xsl:variable name="lc" as="xs:string" select='local-name-from-QName($val)'/>
        <xsl:sequence select='
        if (empty($prefix)) then (
          if (empty($uri)) then $lc
          else "{"||$uri||"}"||$lc
        ) else (
          $prefix||":"||$lc
        )
        '/>
      </xsl:when>
      <xsl:when test='$val instance of map(*)'>
        <xsl:sequence select='
        if ($val("describe") instance of function(*))
        then $val("describe")($val)
        else
        "{"||(
          string-join(
            map:for-each($val, function($k,$v) {$k||":"||this:quote($v)}),
            ","
          )
        )||"}"
        '/>
      </xsl:when>
      <xsl:when test='$val instance of array(*)'>
        <xsl:sequence select='
        "["||(
          string-join(array:for-each($val, this:quote#1), ",")
        )||"]"
        '/>
      </xsl:when>
      <xsl:when test='$val instance of function(*)'>
        <xsl:sequence select='
        "function "||this:quote(function-name($val))||"("||(
          string-join(
            for $i in 1 to function-arity($val) return "v"||$i,
            ","
          )
        )||")"
        '/>
      </xsl:when>
      <xsl:when test='$val instance of document-node()'>
        <xsl:sequence select='
        "document {"||this:quote($val/node())||"}"
        '/>
      </xsl:when>
      <xsl:when test='$val instance of node()'>
        <xsl:sequence select='serialize($val)'/>
      </xsl:when>
      <xsl:otherwise>
        <xsl:sequence select='string($val)'/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:for-each>
</xsl:function>

<xsl:function name="this:quote" as="xs:string">
  <xsl:param name="data" as="item()*"/>
  <xsl:param name="limit" as="xs:integer"/>

  <xsl:variable name="str" as="xs:string" select='this:quote($data)'/>
  <xsl:sequence select='
  substring($str, 1, $limit)||(
    if (string-length($str) > $limit) then "..." else ""
  )
  '/>
</xsl:function>
</xsl:package>
<!-- errors.xsl -->
