<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Vectors of numbers
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:this="http://mathling.com/core/vector"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:config="http://mathling.com/core/config"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             exclude-result-prefixes="this config util errors xs math map array"
             version="3.0"
             name="http://mathling.com/core/vector">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

   <!--
    : this:as-dimension()
    :-->
   <xsl:function name="this:as-dimension" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:sequence select='
      ($v, for $i in 1 to $dim return 0E0)[position() le $dim]
      '/>
   </xsl:function>

   <!--
    : this:unquote()
    :-->
   <xsl:function name="this:unquote" as="xs:double*">
      <xsl:param name="string" as="xs:string"/>
      <xsl:sequence select='
      tokenize($string, "[ ]*,[ ]*")!xs:double(.)
      '/>
   </xsl:function>

   <!--
    : this:dimension()
    :-->
   <xsl:function name="this:dimension" as="xs:integer">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='count($v)'/>
   </xsl:function>

   <!--
    : this:map()
    :-->
   <xsl:function name="this:map" as="xs:double*">
      <xsl:param name="f" as="(function(xs:double) as xs:double)"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='for-each($v, $f)'/>
   </xsl:function>

   <!--
    : this:map()
    :-->
   <xsl:function name="this:map" as="xs:double*">
      <xsl:param name="f" as="(function(xs:double) as xs:double)"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      for-each(this:as-dimension($v, $d), $f)
      '/>
   </xsl:function>

   <!--
    : this:map2()
    :-->
   <xsl:function name="this:map2" as="xs:double*">
      <xsl:param name="f" as="(function(xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:sequence select='util:zip($f, $p1, $p2)'/>
   </xsl:function>

   <!--
    : this:map2()
    :-->
   <xsl:function name="this:map2" as="xs:double*">
      <xsl:param name="f" as="(function(xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      util:zip($f, this:as-dimension($p1, $d), this:as-dimension($p2, $d))
      '/>
   </xsl:function>

   <!--
    : this:map3()
    :-->
   <xsl:function name="this:map3" as="xs:double*">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:param name="p3" as="xs:double*"/>
      <xsl:sequence select='
      util:zip(
        function ($a, $fc) {$fc($a)},
        $p3,
        util:zip(
          function($a, $b) { $f($a, $b, ?) },
          $p1, $p2
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:map3()
    :-->
   <xsl:function name="this:map3" as="xs:double*">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:param name="p3" as="xs:double*"/>
      <xsl:param name="d" as="xs:integer"/>
      <xsl:sequence select='
      util:zip(
        function ($a, $fc) {$fc($a)},
        this:as-dimension($p3, $d),
        util:zip(
          function($a, $b) { $f($a, $b, ?) },
          this:as-dimension($p1, $d), this:as-dimension($p2, $d)
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:map4()
    :-->
   <xsl:function name="this:map4" as="xs:double*">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:param name="p3" as="xs:double*"/>
      <xsl:param name="p4" as="xs:double*"/>
      <xsl:sequence select='
      util:zip(
        function($d, $fd) {$fd($d)},
        $p4,
        util:zip(
          function ($c, $fc) {$fc($c, ?)},
          $p3,
          util:zip(
            function($a, $b) { $f($a, $b, ?, ?) },
            $p1, $p2
          )
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:map4()
    :-->
   <xsl:function name="this:map4" as="xs:double*">
      <xsl:param name="f"
                 as="(function(xs:double, xs:double, xs:double, xs:double) as xs:double)"/>
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:param name="p3" as="xs:double*"/>
      <xsl:param name="p4" as="xs:double*"/>
      <xsl:param name="dim" as="xs:integer"/>
      <xsl:sequence select='
      util:zip(
        function($d, $fd) {$fd($d)},
        this:as-dimension($p4, $dim),
        util:zip(
          function ($c, $fc) {$fc($c, ?)},
          this:as-dimension($p3, $dim),
          util:zip(
            function($a, $b) { $f($a, $b, ?, ?) },
            this:as-dimension($p1, $dim), this:as-dimension($p2, $dim)
          )
        )
      )
      '/>
   </xsl:function>

   <!--
    : this:px()
    :-->
   <xsl:function name="this:px" as="xs:double">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[1],0E0))'/>
   </xsl:function>

   <!--
    : this:py()
    :-->
   <xsl:function name="this:py" as="xs:double">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[2],0E0))'/>
   </xsl:function>

   <!--
    : this:pz()
    :-->
   <xsl:function name="this:pz" as="xs:double">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[3],0E0))'/>
   </xsl:function>

   <!--
    : this:pw()
    :-->
   <xsl:function name="this:pw" as="xs:double">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[4],0E0))'/>
   </xsl:function>

   <!--
    : this:pxy()
    :-->
   <xsl:function name="this:pxy" as="xs:double*">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[1],0E0)), head(($vector[1],0E0))'/>
   </xsl:function>

   <!--
    : this:pxz()
    :-->
   <xsl:function name="this:pxz" as="xs:double*">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='head(($vector[1],0E0)), head(($vector[3],0E0))'/>
   </xsl:function>

   <!--
    : this:x()
    :-->
   <xsl:function name="this:x" as="xs:integer">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='
      let $x := head(($vector[1],0))
      return (
        if ($x = xs:double("INF")) then $util:UINT64_MAX 
        else if ($x = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($x) cast as xs:integer
      )
      '/>
   </xsl:function>

   <!--
    : this:y()
    :-->
   <xsl:function name="this:y" as="xs:integer">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='
      let $y := head(($vector[2],0))
      return (
        if ($y = xs:double("INF")) then $util:UINT64_MAX 
        else if ($y = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($y) cast as xs:integer
      )
      '/>
   </xsl:function>

   <!--
    : this:z()
    :-->
   <xsl:function name="this:z" as="xs:integer">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='
      let $z := head(($vector[3],0))
      return (
        if ($z = xs:double("INF")) then $util:UINT64_MAX 
        else if ($z = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($z) cast as xs:integer
      )
      '/>
   </xsl:function>

   <!--
    : this:w()
    :-->
   <xsl:function name="this:w" as="xs:integer">
      <xsl:param name="vector" as="xs:double*"/>
      <xsl:sequence select='
      let $w := head(($vector[4],0))
      return (
        if ($w = xs:double("INF")) then $util:UINT64_MAX 
        else if ($w = xs:double("-INF")) then -$util:UINT64_MAX 
        else round-half-to-even($w) cast as xs:integer
      )
      '/>
   </xsl:function>

   <!--
    : this:magnitude()
    :-->
   <xsl:function name="this:magnitude" as="xs:double">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      math:sqrt(
        sum(for $c in $v return $c*$c)
      )
      '/>
   </xsl:function>

   <!--
    : this:dot()
    :-->
   <xsl:function name="this:dot" as="xs:double">
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:sequence select='
      sum(
        this:map2(function ($a,$b) {$a * $b}, $p1, $p2)
      )
      '/>
   </xsl:function>

   <!--
    : this:dot2()
    :-->
   <xsl:function name="this:dot2" as="xs:double">
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:sequence select='
      sum(
        this:map(function ($a) {$a * $a}, $p1)
      )
      '/>
   </xsl:function>

   <!--
    : this:ndot()
    : Negated dot: useful for some of the SDF functions 
    :-->
   <xsl:function name="this:ndot" as="xs:double">
      <xsl:param name="p1" as="xs:double*"/>
      <xsl:param name="p2" as="xs:double*"/>
      <xsl:sequence select='
      sum(
        this:map2(function ($a,$b) {$a * -$b}, $p1, $p2)
      )
      '/>
   </xsl:function>

   <!--
    : this:mix()
    :-->
   <xsl:function name="this:mix" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="f" as="xs:double"/>
      <xsl:sequence select='
      (: Logically $a + ($b - $a) * $f but better endpoint behaviour :)
      this:map2(function($a, $b) {$a * (1 - $f) + $b * $f}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:determinant()
    : 2D only
    :-->
   <xsl:function name="this:determinant" as="xs:double">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:px($u)*this:py($v) - this:py($u)*this:px($v)
      '/>
   </xsl:function>

   <!--
    : this:cross()
    : 3D only
    :-->
   <xsl:function name="this:cross" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:py($u)*this:pz($v) - this:pz($u)*this:py($v),
      this:pz($u)*this:px($v) - this:px($u)*this:pz($v),
      this:px($u)*this:py($v) - this:py($u)*this:px($v)
      '/>
   </xsl:function>

   <!--
    : this:round()
    :-->
   <xsl:function name="this:round" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map(function ($c) {round($c)}, $v)
      '/>
   </xsl:function>

   <!--
    : this:times()
    :-->
   <xsl:function name="this:times" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select=' 
      this:map(function ($c) {$k * $c}, $v)
     '/>
   </xsl:function>

   <!--
    : this:multiply()
    :-->
   <xsl:function name="this:multiply" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function ($a,$b) {$a * $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:div()
    :-->
   <xsl:function name="this:div" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:map(function ($c) {($c div $k) cast as xs:double}, $v)
      '/>
   </xsl:function>

   <!--
    : this:divide()
    :-->
   <xsl:function name="this:divide" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {$a div $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:mod()
    :-->
   <xsl:function name="this:mod" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:map(function ($c) {$c mod $k}, $v)
      '/>
   </xsl:function>

   <!--
    : this:modulo()
    :-->
   <xsl:function name="this:modulo" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function ($a,$b) {$a mod $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:add()
    :-->
   <xsl:function name="this:add" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {$a + $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:plus()
    :-->
   <xsl:function name="this:plus" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:map(function($a) {$a + $k}, $v)
      '/>
   </xsl:function>

   <!--
    : this:minus()
    :-->
   <xsl:function name="this:minus" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map(function($a) {-$a}, $v)
      '/>
   </xsl:function>

   <!--
    : this:minus()
    :-->
   <xsl:function name="this:minus" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="k" as="xs:double"/>
      <xsl:sequence select='
      this:map(function($a) {$a - $k}, $v)
      '/>
   </xsl:function> 

   <!--
    : this:sub()
    :-->
   <xsl:function name="this:sub" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {$a - $b}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:abs()
    :-->
   <xsl:function name="this:abs" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map(function($a) {abs($a)}, $v)
      '/>
   </xsl:function>

   <!--
    : this:sign()
    :-->
   <xsl:function name="this:sign" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map(
        function($a) {
          if ($a eq 0) then 0E0
          else if ($a lt 0) then -1E0
          else 1E0
        },
        $v
      )
      '/>
   </xsl:function>

   <!--
    : this:min()
    :-->
   <xsl:function name="this:min" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {min(($a,$b))}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:max()
    :-->
   <xsl:function name="this:max" as="xs:double*">
      <xsl:param name="u" as="xs:double*"/>
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      this:map2(function($a,$b) {max(($a,$b))}, $u, $v)
      '/>
   </xsl:function>

   <!--
    : this:at-min()
    :-->
   <xsl:function name="this:at-min" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="val" as="xs:double"/>
      <xsl:sequence select='
      this:map(function($a) {min(($a, $val))}, $v)
      '/>
   </xsl:function>

   <!--
    : this:at-max()
    :-->
   <xsl:function name="this:at-max" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="val" as="xs:double"/>
      <xsl:sequence select='
      this:map(function($a) {max(($a, $val))}, $v)
      '/>
   </xsl:function>

   <!--
    : this:normalize()
    :-->
   <xsl:function name="this:normalize" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      let $l := this:magnitude($v)
      return
        if ($l = 0) then $v
        else this:map(function ($c) {$c div $l}, $v)
      '/>
   </xsl:function>

   <!--
    : this:clamp()
    :-->
   <xsl:function name="this:clamp" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="min" as="xs:double"/>
      <xsl:param name="max" as="xs:double"/>
      <xsl:sequence select='
      this:map(function($a) {util:clamp($a, $min, $max)}, $v)
      '/>
   </xsl:function>

   <!--
    : this:clampv()
    :-->
   <xsl:function name="this:clampv" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="minv" as="xs:double*"/>
      <xsl:param name="maxv" as="xs:double*"/>
      <xsl:sequence select='
      this:map3(function($a, $min, $max) {util:clamp($a, $min, $max)}, $v, $minv, $maxv)
      '/>
   </xsl:function>

   <!--
    : this:perpendicular()
    : 2D perpendicular vector: point rotated 90 degrees
    :-->
   <xsl:function name="this:perpendicular" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      (this:py($v), -this:px($v))
      '/>
   </xsl:function>

   <!--
    : this:snap()
    :-->
   <xsl:function name="this:snap" as="xs:integer*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      for-each($v, function ($c) {round($c) cast as xs:integer})
      '/>
   </xsl:function>

   <!--
    : this:floor()
    :-->
   <xsl:function name="this:floor" as="xs:integer*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      for-each($v, function($c) {floor($c) cast as xs:integer})
      '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal" as="xs:double*">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for-each($v, function($c) {util:decimal($c, $digits)})
      '/>
   </xsl:function>

   <!--
    : this:quote()
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="v" as="xs:double*"/>
      <xsl:sequence select='
      string-join($v!string(.),",")
      '/>
   </xsl:function>

   <!--
    : this:same()
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="xs:double*"/>
      <xsl:param name="other" as="xs:double*"/>
      <xsl:sequence select='
      deep-equal($this, $other)
      '/>
   </xsl:function>

   <!--
    : this:distance()
    : Euclidean distance between points
    :-->
   <xsl:function name="this:distance" as="xs:double">
      <xsl:param name="v1" as="xs:double*"/>
      <xsl:param name="v2" as="xs:double*"/>
      <xsl:sequence select='
      let $d := max((this:dimension($v1), this:dimension($v2)))
      return (
        math:sqrt(
          sum(
            util:zip(
              function($c1 as xs:double, $c2 as xs:double) as xs:double {
                ($c1 - $c2)*($c1 - $c2)
              },
              this:as-dimension($v1, $d),
              this:as-dimension($v2, $d)
            )
          )
        )
      )
      '/>
   </xsl:function>
</xsl:package>
<!--vector.xsl-->
