<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Ellipses (including circles)
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric/ellipse"
             xmlns:affine="http://mathling.com/geometric/affine"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/box"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/geometric/ellipse"
             exclude-result-prefixes="this xs math map array errors config util point box affine">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/box" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/affine" package-version="*"/>

   <xsl:variable name="this:RESERVED" as="xs:string*">,
      <xsl:sequence select=' ("kind", "center", "rx", "ry", "rotation")'/>
   </xsl:variable>

   <!--
    : this:ellipse()
    : Construct an ellipse. 
    :
    : $center: center of ellipse
    : $rx: axis in x (before rotation)
    : $ry: axis in y (before rotation)
    : $rotation-degrees: how much to rotate ellipse around center
    : $properties: additional properties
    :-->
   <xsl:function name="this:ellipse" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="rx" as="xs:double"/>
      <xsl:param name="ry" as="xs:double"/>
      <xsl:param name="rotation-degrees" as="xs:double"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  util:merge-into($properties,
    map {
      "kind": "ellipse",
      "center": $center,
      "rx": $rx,
      "ry": $ry,
      "rotation": $rotation-degrees
    }
  )
'/>
   </xsl:function>

   <!--
    : this:ellipse()
    :-->
   <xsl:function name="this:ellipse" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="rx" as="xs:double"/>
      <xsl:param name="ry" as="xs:double"/>
      <xsl:param name="rotation-degrees" as="xs:double"/>
      <xsl:sequence select='
  map {
    "kind": "ellipse",
    "center": $center,
    "rx": $rx,
    "ry": $ry,
    "rotation": $rotation-degrees
  }
'/>
   </xsl:function>

   <!--
    : this:ellipse()
    :-->
   <xsl:function name="this:ellipse" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="rx" as="xs:double"/>
      <xsl:param name="ry" as="xs:double"/>
      <xsl:sequence select='
  this:ellipse($center, $rx, $ry, 0)
'/>
   </xsl:function>

   <!--
    : this:ellipse()
    :-->
   <xsl:function name="this:ellipse" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
  this:ellipse($center, $r, $r, 0)
'/>
   </xsl:function>

   <!--
    : this:circle()
    :-->
   <xsl:function name="this:circle" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
  this:ellipse($center, $r, $r, 0)
'/>
   </xsl:function>

   <!--
    : this:circle()
    :-->
   <xsl:function name="this:circle" as="map(xs:string, item()*)">
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:ellipse($center, $r, $r, 0, $properties)
'/>
   </xsl:function>

   <!--
    : this:circle-from()
    : Return unique circle that intersects three points
    :-->
   <xsl:function name="this:circle-from" as="map(xs:string, item()*)?">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:param name="r" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  let $unshifted-center :=
    this:circle-center-of($q=>point:sub($p), $r=>point:sub($p)) return
  let $center := if (exists($unshifted-center)) then $unshifted-center=>point:add($p) else ()
  return (
    if (exists($center)) 
    then this:circle($center, point:distance($center, $p))
    else ()
  )
'/>
   </xsl:function>

   <!--
    : this:circle-from()
    : Smallest circle that intersects 2 points
    :-->
   <xsl:function name="this:circle-from" as="map(xs:string, item()*)">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:circle( point:midpoint($p, $q), point:distance($p, $q) div 2 )
'/>
   </xsl:function>

   <!--
    : this:circle-center-of()
    : Return the center of the circle passing through three points, if there is 
    : one; empty sequence otherwise. 2D.
    :
    : $a, $b, $c: the points
    :-->
   <xsl:function name="this:circle-center-of" as="map(xs:string, item()*)?">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (point:magnitude(point:sub($b,$c)) &lt; 0) then ()
  else (
    point:add(
      point:midpoint($a, $b),
      point:times(
        point:perpendicular(point:sub($b,$a)),
        point:dot(point:sub($c,$b), point:sub($c,$a)) div
          (-2 * point:determinant(point:sub($b,$a), point:sub($c,$a)))
      )
    )
  )
'/>
   </xsl:function>

   <!--
    : this:circle-center-of()
    : Function to get the center of a circle between two points. 2D
    :-->
   <xsl:function name="this:circle-center-of" as="map(xs:string, item()*)?">
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:param name="c" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  let $B := point:px($b) * point:px($b) + point:py($b) * point:py($b) return
  let $C := point:px($c) * point:px($c) + point:py($c) * point:py($c) return
  let $D := point:px($b) * point:py($c) - point:py($b) * point:px($c)
  return (
    if ($D = 0) then () else
    point:point(
      (point:py($c) * $B - point:py($b) * $C) div (2 * $D),
      (point:px($b) * $C - point:px($c) * $B) div (2 * $D)
    )
  )
'/>
   </xsl:function>

   <!--
    : this:kind()
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  $ellipse("kind")
'/>
   </xsl:function>

   <!--
    : this:is-circle()
    :-->
   <xsl:function name="this:is-circle" as="xs:boolean">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($circle) = "ellipse") then (this:rx($circle)=this:ry($circle))
  else errors:error("GEOM-BADREGION", ($circle, "is-circle"))
'/>
   </xsl:function>

   <!--
    : this:center()
    :-->
   <xsl:function name="this:center" as="map(xs:string, item()*)">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($ellipse) = "ellipse") then $ellipse("center")
  else errors:error("GEOM-BADREGION", ($ellipse, "center"))
'/>
   </xsl:function>

   <!--
    : this:radius()
    :-->
   <xsl:function name="this:radius" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:is-circle($ellipse))
  then $ellipse("rx")
  else errors:error("GEOM-BADREGION", ($ellipse, "radius"))
'/>
   </xsl:function>

   <!--
    : this:rx()
    :-->
   <xsl:function name="this:rx" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($ellipse) = "ellipse") then $ellipse("rx")
  else errors:error("GEOM-BADREGION", ($ellipse, "rx"))
'/>
   </xsl:function>

   <!--
    : this:ry()
    :-->
   <xsl:function name="this:ry" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($ellipse) = "ellipse") then $ellipse("ry")
  else errors:error("GEOM-BADREGION", ($ellipse, "ry"))
'/>
   </xsl:function>

   <!--
    : this:rotation()
    :-->
   <xsl:function name="this:rotation" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($ellipse) = "ellipse")  then $ellipse("rotation")
  else errors:error("GEOM-BADREGION", ($ellipse, "rotation"))
'/>
   </xsl:function>

<!--
 :======================================================================
 : Property management
 :======================================================================
 :-->

   <!--
    : this:property-map()
    : Return the annotation properties of the point as a map. Check whether this
    : is actually an ellipse.
    :
    : $region: the region
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($region) = "ellipse") then util:exclude($region, $this:RESERVED)
  else map {}
'/>
   </xsl:function>

   <!--
    : this:properties()
    : Return the names of the annotation properties of the ellipse.
    : Check whether this is actually an ellipse.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($region) = "ellipse") then ($region=>map:keys())[not(. = $this:RESERVED)]
  else ()
'/>
   </xsl:function>

   <!--
    : this:with-properties()
    : Annotate the ellipse with some new properties and return the new ellipse.
    : Will not touch any of the core properties. Will override existing properties
    : with the same keys but leave properties with different keys in place.
    : Raises an error if this is not actually an ellipse.
    : 
    : $region: the region
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:kind($region) = "ellipse") then
    util:merge-into($region, util:exclude($properties,$this:RESERVED))
  else errors:error("GEOM-BADREGION", ($region, "with-properties"))
'/>
   </xsl:function>

<!--
 :======================================================================
 : Operations
 :======================================================================
 :-->

   <!--
    : this:snap()
    : Snap the coordinates of the center, returning the ellipse with snapped
    : (i.e. integer) center coordinates and radii.
    :
    : $ellipses: the ellipses to snap
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="ellipses" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
  for $ellipse in $ellipses return 
  if (this:kind($ellipse) = "ellipse") then
    this:ellipse(
      point:snap(this:center($ellipse)),
      round-half-to-even(this:rx($ellipse)) cast as xs:integer,
      round-half-to-even(this:ry($ellipse)) cast as xs:integer,
      this:rotation($ellipse),
      this:property-map($ellipse)
    )
  else $ellipse
'/>
   </xsl:function>

   <!--
    : this:decimal()
    : Perform decimal rounding on all the point coordinates (see util:decimal).
    : and radii.
    :
    : $ellipses: the ellipses to round
    : $digits: how many digits after the decimal point to keep
    :-->
   <xsl:function name="this:decimal" as="map(xs:string, item()*)*">
      <xsl:param name="ellipses" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
  for $ellipse in $ellipses return 
  if (this:kind($ellipse) = "ellipse") then
    this:ellipse(
      point:decimal(this:center($ellipse), $digits),
      util:decimal(this:rx($ellipse), $digits),
      util:decimal(this:ry($ellipse), $digits),
      this:rotation($ellipse),
      this:property-map($ellipse)
    )
  else $ellipse
'/>
   </xsl:function>

   <!--
    : this:quote()
    : Return a string value for the ellipses, suitable for debugging.
    : 
    : $ellipses: the ellipses to quote
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="ellipses" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
  string-join(
    for $ellipse in $ellipses return
    if (this:kind($ellipse) = "ellipse") then
      if (this:is-circle($ellipse)) then (
        "@"||$ellipse("rx")||" "||point:quote($ellipse("center"))
      ) else (
        "@@"||($ellipse("rx"))||"x"||($ellipse("ry"))||" "||point:quote($ellipse("center"))||"∠"||$ellipse("rotation")
      )
    else errors:quote($ellipse)
    ,
    " "
  )
'/>
   </xsl:function>

   <!--
    : this:same()
    : Equality comparison for ellipses, ignoring annotation properties.
    : Return true() if they have equal coordinates.
    : 
    : $this: one ellipse
    : $other: the ellipse to compare it to
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  let $this-kind := this:kind($this) return
  let $other-kind := this:kind($other) return
  (
    $this-kind=$other-kind and (
      if ($this-kind = "ellipse") then 
        point:same($this("center"),$other("center")) and
        (this:rx($this)=this:rx($other)) and
        (this:ry($this)=this:ry($other)) and
        (this:rotation($this)=this:rotation($other))
    else deep-equal($this,$other)
    )
  )
'/>
   </xsl:function>

   <!--
    : this:mutate() 
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutate" as="(function(item()) as map(xs:string, item()*))"/>
      <xsl:sequence select='
  for $region in $regions return 
  if (this:kind($region) = "ellipse") then
    let $center := $mutate(this:center($region))
    return
      this:ellipse($center, this:rx($region), this:ry($region),
                   this:rotation($region), this:property-map($region))
  else $region
'/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
  if (this:is-circle($ellipse)) then (
    point:distance(this:center($ellipse), $point) &lt;= this:radius($ellipse) + $tolerance
  ) else (
    let $α := util:radians(this:rotation($ellipse)) return
    let $a := this:rx($ellipse) return
    let $b := this:ry($ellipse) return
    let $xp := point:px($point) return
    let $yp := point:py($point) return
    let $x0 := point:px(this:center($ellipse)) return
    let $y0 := point:py(this:center($ellipse))
    return (
      some $t1 in distinct-values((0, $tolerance, -$tolerance)) satisfies (
        some $t2 in distinct-values((0, $tolerance, -$tolerance)) satisfies (
          let $f1 := math:cos($α)*($xp + $t1 - $x0) + math:sin($α)*($yp + $t2 - $y0) return
          let $f2 := math:sin($α)*($xp + $t1 - $x0) + math:cos($α)*($yp + $t2 - $y0)
          return ( ($f1*$f1) div ($a*$a) + ($f2*$f2) div ($b*$b) ) &lt;= 1
        )
      )
    )
  )
'/>
   </xsl:function>

   <!--
    : this:contains-point()
    :-->
   <xsl:function name="this:contains-point" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:contains-point($ellipse, $point, $config:tolerance)
'/>
   </xsl:function>

   <!--
    : this:perimeter()
    : Perimeter of the ellipse. Approximately.
    : circle= 2πr
    : ellipse≈ 2π√(a²+b²)
    :-->
   <xsl:function name="this:perimeter" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  if (this:is-circle($ellipse))
  then 2*math:pi()*this:radius($ellipse)
  else 2*math:pi()*math:sqrt(this:rx($ellipse)*this:rx($ellipse) + this:ry($ellipse)*this:ry($ellipse))
'/>
   </xsl:function>

   <!--
    : this:area()
    : Area of the ellipse. 
    : circle= πr²
    : ellipse= πab
    :-->
   <xsl:function name="this:area" as="xs:double">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  math:pi()*this:rx($ellipse)*this:ry($ellipse)
'/>
   </xsl:function>

   <!--
    : this:circle-point()
    : Return a point at the given fraction along the circle, starting at
    : 0 degrees increasing to 360.
    :
    : $circle: the circle
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
   <xsl:function name="this:circle-point" as="map(xs:string, item()*)*">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
  if (this:is-circle($circle)) then (
    point:destination(this:center($circle), $t * 360, this:radius($circle))
  ) else errors:error("GEOM-BADREGION", ($circle, "circle-point"))
'/>
   </xsl:function>

   <!--
    : this:ellipse-point()
    : Return a point at the given fraction along the ellipse, starting at
    : 0 degrees increasing to 360. We compute this by computing the point
    : at the circle centered on ellipse center with radius equal to the 
    : maximum radius of the ellipse and then scaling and rotating.
    :
    : $cubic: the edge
    : $t: fraction of distance from start to end (t) [0,1]
    :-->
   <xsl:function name="this:ellipse-point" as="map(xs:string, item()*)*">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
  if (this:is-circle($ellipse)) then (
    this:circle-point($ellipse, $t)
  ) else (
    let $max-r := max((this:rx($ellipse), this:ry($ellipse))) return
    let $min-r := min((this:rx($ellipse), this:ry($ellipse))) return
    let $flip := this:rx($ellipse) > this:ry($ellipse) return
    let $circle-point := this:circle-point(this:circle(this:center($ellipse), $max-r), $t) return
    let $scaling := $min-r div $max-r return
    let $rotation := this:rotation($ellipse) return
    let $affine :=
      affine:compose2( 
        affine:rotation2(this:center($ellipse), -$rotation),
        if ($flip)
        then affine:scaling2(this:center($ellipse), 1, $scaling) 
        else affine:scaling2(this:center($ellipse), $scaling, 1)
      )
    return (
      point:vector(affine:affine2(point:pcoordinates($circle-point), $affine))
    )
  )
'/>
   </xsl:function>

   <!--
    : this:circle-circle-intersection()
    n: Points of intersection, if any, between two circles
    :-->
   <xsl:function name="this:circle-circle-intersection" as="map(xs:string, item()*)*">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="that" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
  if (not(this:is-circle($this))) then errors:error("GEOM-BADREGION", ($this, "circle-circle-intersection"))
  else if (not(this:is-circle($that))) then errors:error("GEOM-BADREGION", ($that, "circle-circle-intersection"))
  else (),
  (: https://math.stackexchange.com/questions/256100/how-can-i-find-the-points-at-which-two-circles-intersect :)
  let $d := point:distance(this:center($this), this:center($that)) return
  let $d_2 := $d * $d return
  let $r1_2 := this:radius($this) * this:radius($this) return
  let $r2_2 := this:radius($that) * this:radius($that) return
  let $x1 := point:px(this:center($this)) return
  let $y1 := point:py(this:center($this)) return
  let $x2 := point:px(this:center($that)) return
  let $y2 := point:py(this:center($that))
  return (
    if ($d > this:radius($this) + this:radius($that)) then (
      if ($d > this:radius($this) + this:radius($that) + $tolerance) then () else (
        this:circle-circle-intersection(
          this:circle(this:center($this), this:radius($this) + $tolerance),
          $that
        )
      )
    ) else if ($d &lt; abs(this:radius($this) - this:radius($that))) then (
      if ($d &lt; abs(this:radius($this) - this:radius($that)) - $tolerance) then () else (
        this:circle-circle-intersection(
          this:circle(this:center($this), this:radius($this) - $tolerance),
          $that
        )
      )
    ) else (
      let $a := 0.5 return
      let $b := ($r1_2 - $r2_2) div (2 * $d_2) return
      let $c :=
        0.5 * 
        math:sqrt(
          (2 * ($r1_2 + $r2_2) div $d_2) -
          ((($r1_2 - $r2_2)*($r1_2 - $r2_2)) div ($d_2 * $d_2)) -
          1
        ) 
      return (
        point:point(
          $a * ($x1 + $x2) + $b * ($x2 - $x1) + $c * ($y2 - $y1),
          $a * ($y1 + $y2) + $b * ($y2 - $y1) + $c * ($x1 - $x2)
        ),
        if ($c = 0) then () else (
          point:point(
            $a * ($x1 + $x2) + $b * ($x2 - $x1) - $c * ($y2 - $y1),
            $a * ($y1 + $y2) + $b * ($y2 - $y1) - $c * ($x1 - $x2)
          )
        )
      )
    )
  )
'/>
   </xsl:function>

   <!--
    : this:circle-circle-intersection()
    :-->
   <xsl:function name="this:circle-circle-intersection" as="map(xs:string, item()*)*">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="that" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:circle-circle-intersection($this, $that, $config:tolerance)
'/>
   </xsl:function>

   <!--
    : this:ellipse-intersects-ellipse()
    :-->
   <xsl:function name="this:ellipse-intersects-ellipse" as="xs:boolean">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
  if ($region=>this:is-circle()) then (
    if ($other=>this:is-circle()) then (
      point:distance(this:center($region), this:center($other)) &lt;=
      this:radius($region) + this:radius($other) + $tolerance
    ) else (
      errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects-ellipse",$other))
    )
  ) else (
    errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects-ellipse",$region))
  )
'/>
   </xsl:function>

   <!--
    : this:bounding-box()
    : Minimum box surrounding the region.
    :-->
   <xsl:function name="this:bounding-box" as="map(xs:string, item()*)">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
  let $boxes := (
    for $region in $regions return 
    if (this:kind($region) = "ellipse") then (
      if ($region=>this:is-circle()) then (
        let $c := this:center($region) return
        let $r := this:radius($region)
        return
          box:box(
            point:px($c) - $r, point:py($c) - $r,
            point:px($c) + $r, point:py($c) + $r
          )
      ) else (
        let $c := this:center($region) return
        let $rx := this:rx($region) return
        let $ry := this:rx($region) return
        let $a := this:rotation($region) return
        let $min-pt := point:point(point:px($c) - $rx, point:py($c) - $rx) return
        let $max-pt := point:point(point:px($c) + $ry, point:py($c) + $ry)
        return
          box:box(
            affine:rotate($min-pt, $a, $c), affine:rotate($max-pt, $a, $c)
          )
      )
    )
    else errors:error("GEOM-BADREGION", ($region, "bounding-box"))
  )
  return (
    if (empty($boxes)) then box:box(0,0,0,0)
    else if (empty(tail($boxes))) then $boxes
    else (
      let $pts := $boxes!box:points(.)
      return
        box:box(
          min($pts!point:px(.)), min($pts!point:py(.)),
          max($pts!point:px(.)), max($pts!point:py(.))
        )
    )
  )
'/>
   </xsl:function>

   <!--
    : this:tangent-angle()
    :-->
   <xsl:function name="this:tangent-angle" as="xs:double">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
  if (this:is-circle($region)) then (
    360 * $t  - 90
  ) else (
    let $c := this:center($region) return
    let $p := $region=>this:ellipse-point($t) 
    return point:angle($c, $p) - 90
  )
'/>
   </xsl:function>

<!--
 :======================================================================
 : Interpolations
 :======================================================================
 :-->

   <!--
    : this:interpolate-circle-using()
    : Return a sequence of points formed by interpolating the path of the circle.
    :
    : $n: how many points to return
    : $circle: the base circle
    :-->
   <xsl:function name="this:interpolate-circle-using" as="map(xs:string, item()*)*">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
      <xsl:sequence select='
  if (this:is-circle($circle)) then (
    let $center := this:center($circle) return
    let $radius := this:radius($circle) return
    for $t in sort($divisions($circle))
    return (
      point:destination($center, $t * 360, $radius)
    )
  ) else errors:error("GEOM-BADREGION", ($circle, "interpolate-circle-using"))
'/>
   </xsl:function>

   <!--
    : this:interpolate-circle()
    : Return a sequence of points formed by interpolating the path of the circle.
    :
    : $n: how many points to return
    : $circle: the base circle
    :-->
   <xsl:function name="this:interpolate-circle" as="map(xs:string, item()*)*">
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:interpolate-circle-using($circle,
    function($region as map(xs:string,item()*)) as xs:double* {
      util:linspace($n, 0, 1, false())
    }
  )
'/>
   </xsl:function>

   <!--
    : this:interpolate-using()
    : Return a sequence of points formed by interpolating the path of the ellipse.
    :
    : $n: how many points to return
    : $ellipse: the base ellipse
    :-->
   <xsl:function name="this:interpolate-using" as="map(xs:string, item()*)*">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
      <xsl:sequence select='
  if (this:is-circle($ellipse)) then (
    this:interpolate-circle-using($ellipse, $divisions)
  ) else (
    let $max-r := max((this:rx($ellipse), this:ry($ellipse))) return
    let $min-r := min((this:rx($ellipse), this:ry($ellipse))) return
    let $flip := this:rx($ellipse) > this:ry($ellipse) return
    let $circle :=
      this:interpolate-circle-using(
        this:circle(this:center($ellipse), $max-r),
        $divisions
      ) 
    return
    let $scaling := $min-r div $max-r return 
    let $rotation := this:rotation($ellipse) return
    let $affine :=
      affine:compose2(
        affine:rotation2(-$rotation),
        if ($flip)
        then affine:scaling2(this:center($ellipse), 1, $scaling)
        else affine:scaling2(this:center($ellipse), $scaling, 1)
      )
    return
    for $point in $circle return (
      point:vector(affine:affine2(point:pcoordinates($point), $affine))
    )
  )
'/>
   </xsl:function>

   <!--
    : this:interpolate()
    : Return a sequence of points formed by interpolating the path of the ellipse.
    :
    : $n: how many points to return
    : $ellipse: the base ellipse
    :-->
   <xsl:function name="this:interpolate" as="map(xs:string, item()*)*">
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
  this:interpolate-using($ellipse,
    function ($region as map(xs:string,item()*)) { util:linspace($n, 0, 1) }
  )
'/>
   </xsl:function>

   <xsl:function name="this:draw-ellipse" as="element()" visibility="private">
     <xsl:param name="cx" as="xs:numeric"/>
     <xsl:param name="cy" as="xs:numeric"/>
     <xsl:param name="rx" as="xs:numeric"/>
     <xsl:param name="ry" as="xs:numeric"/>
     <xsl:param name="rotation" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <art:ellipse cx="{$cx}" cy="{$cy}" rx="{$rx}" ry="{$ry}" rotate="{$rotation}">
       <xsl:sequence select='util:as-attributes($style-properties)'/>
      </art:ellipse>
   </xsl:function>

   <xsl:function name="this:draw-svg-ellipse" as="element()" visibility="private">
     <xsl:param name="cx" as="xs:numeric"/>
     <xsl:param name="cy" as="xs:numeric"/>
     <xsl:param name="rx" as="xs:numeric"/>
     <xsl:param name="ry" as="xs:numeric"/>
     <xsl:param name="rotation" as="xs:numeric"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <svg:ellipse cx="{$cx}" cy="{$cy}" rx="{$rx}" ry="{$ry}">
       <xsl:if test='empty($rotation) or $rotation=0'>
         <xsl:attribute name="transform">
           <xsl:sequence select='"rotate(",$rotation,",",$cx,",",$cy,")"'/>
         </xsl:attribute>
       </xsl:if>
       <xsl:sequence select='util:as-attributes($style-properties)'/>
      </svg:ellipse>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item),("r","type")))) return
            let $center := $item=>this:center() return
            let $rx := round-half-to-even($item=>this:rx()) return
            let $ry := round-half-to-even($item=>this:ry()) return
            let $rotation := $item=>this:rotation()
            return (
              this:draw-ellipse(point:x($center), point:y($center), $rx, $ry, $rotation, $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($properties, util:exclude(this:property-map($item),("r","type")))) return
            let $center := $item=>this:center() return
            let $rx := $item=>this:rx() return
            let $ry := $item=>this:ry() return
            let $rotation := $item=>this:rotation()
            return (
              this:draw-ellipse(point:px($center), point:py($center), $rx, $ry, $rotation, $style-properties)
            )
          }
        )
      ) else (
        if ($config:DRAW-SNAPPED) then (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(util:exclude(this:property-map($item),("r","type"))))) return
            let $center := $item=>this:center() return
            let $rx := round-half-to-even($item=>this:rx()) return
            let $ry := round-half-to-even($item=>this:ry()) return
            let $rotation := $item=>this:rotation()
            return (
              this:draw-svg-ellipse(point:x($center), point:y($center), $rx, $ry, $rotation, $style-properties)
            )
          }
        ) else (
          function($item, $properties, $drawing) {
            let $style-properties :=
              util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(util:exclude(this:property-map($item),("r","type"))))) return
            let $center := $item=>this:center() return
            let $rx := $item=>this:rx() return
            let $ry := $item=>this:ry() return
            let $rotation := $item=>this:rotation()
            return (
              this:draw-svg-ellipse(point:px($center), point:py($center), $rx, $ry, $rotation, $style-properties)
            )
          }
        )
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>

<!--ellipse.xsl-->
