<?xml version="1.0" encoding="UTF-8"?>
<!--
 : SDF annotations
 : Since we can't use function annotations in Saxon-JS because we don't have
 : saxon:xquery, and it is awkward to do so in XSLT; use callable wrappers. They're
 : more portable, but less type-safe. Drat.
 :
 : Copyright© Mary Holstege 2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/sdf/annotations"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:callable="http://mathling.com/core/callable"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/sdf/annotations"
             exclude-result-prefixes="this xs math map array errors util callable">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>

   <!--
    : this:is-sdf()
    : Type checking
    :-->
   <xsl:function name="this:is-sdf" as="xs:boolean">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      ($f instance of map(*) and util:kind($f)="callable" and this:is-sdf(callable:function($f))) or
      ($f instance of function(xs:double*) as xs:double*)
    '/>
   </xsl:function>

   <!--
    : this:is-renderer()
    : Type checking
    :-->
   <xsl:function name="this:is-renderer" as="xs:boolean">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      ($f instance of map(*) and util:kind($f)="callable" and this:is-renderer(callable:function($f))) or
      ($f instance of function(xs:double*, xs:double*, xs:double*, item()) as xs:double*)
    '/>
   </xsl:function>

   <!--
    : this:is-caster()
    : Type checking
    :-->
   <xsl:function name="this:is-caster" as="xs:boolean">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      ($f instance of map(*) and util:kind($f)="callable" and this:is-caster(callable:function($f))) or
      ($f instance of function(xs:double*, xs:double*, item()) as xs:double*)
    '/>
   </xsl:function>

   <!--
    : this:is-distance-function()
    : Type checking
    :-->
   <xsl:function name="this:is-distance-function" as="xs:boolean">
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      ($f instance of map(*) and util:kind($f)="callable" and this:is-distance-function(callable:function($f))) or
      ($f instance of function(xs:double) as xs:double)
    '/>
   </xsl:function>

   <!--
    : this:f()
    : Make an SDF callable with a given name
    :-->
   <xsl:function name="this:f" as="map(*)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="f" as="item()"/>
      <xsl:sequence select='
      if (not(callable:is-callable($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $prior-name := util:function-name($f)
      return (
        if ($prior-name = ("[anon]", "", $name))
        then callable:named($name, callable:function($f))
        else callable:named($name||"("||$prior-name||")", callable:function($f))
      )
    '/>
   </xsl:function>

   <!--
    : this:f()
    : Make an SDF callable with an expanded name with parameter tracking
    :-->
   <xsl:function name="this:f" as="map(*)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="f" as="item()"/>
      <xsl:param name="parms" as="item()*"/>
      <xsl:sequence select='
      if (not(callable:is-callable($f))) then errors:error("ML-BADARGS", ("f", $f)) else (),
      let $name := util:function-name($f)||"("||string-join($parms!util:quote(.),",")||")"
      return callable:named($name, callable:function($f))
    '/>
   </xsl:function>

</xsl:package>
<!--annotations.xsl-->
