<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Module with functions providing some colouring support at the
 : algorithm level.
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/svg/gradients"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:cs="http://mathling.com/colour/space"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:ellipse="http://mathling.com/geometric/ellipse"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:path="http://mathling.com/geometric/path"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:rand="http://mathling.com/core/random"
             xmlns:rgb="http://mathling.com/colour/rgb"
             xmlns:spline="http://mathling.com/geometric/spline"
             xmlns:stop="http://mathling.com/colour/stop"
             xmlns:tone="http://mathling.com/colour/tonemap"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:xyz="http://mathling.com/colour/xyz"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:xlink="http://www.w3.org/1999/xlink"
             version="3.0"
             name="http://mathling.com/svg/gradients"
             exclude-result-prefixes="this xs math map array util rand point edge path ellipse spline cs rgb xyz tone stop">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/random" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/spline" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/space" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/rgb" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/xyz" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/tonemap" package-version="*"/>
   <xsl:use-package name="http://mathling.com/colour/stop" package-version="*"/>

   <!--
    : $this:NAMED-GRADIENTS
    :-->
   <xsl:variable name="this:NAMED-GRADIENTS" as="xs:string*">
      <xsl:sequence select='(
      "acton", "bam", "bamako", "batlow", "berlin", "bilbao", "broc", "buda", "bukavu",
      "cork", "davos", "devon", "fes", "grayC", "hawaii", "imola", "lajolla", "lapaz",
      "lisbon", "nuuk", "oleron", "oslo", "roma", "tofino", "tokyo", "turku",
      "vanimo", "vik",
      "actonS", "bamakoS", "batlowS", (:"berlinS",:) "bilbaoS", (:"brocS",:) "budaS",
      (: "corkS", :) "davosS", "devonS", "grayCS", "hawaiiS", "imolaS", "lajollaS",
      "lapazS", (:"lisbonS",:) "nuukS", (:"oleronS",:) "osloS", (:"romaS",:) (:"tofinoS",:)
      "tokyoS", "turkuS", (:"vikS",:)
      "bamO", "brocO", "corkO", "romaO", "vikO",
      "batlowK", "batlowW",
      "inferno", "magma", "plasma", "viridis",
      "cividis",
      "twilight", "twilight-blue", "twilight-red",
      "sawtooth","azimuth",
      "cube1_0","cubeYF","Linear_L",
      "bathymetry","bathymetryO","CDOM","chlorophyll","density",
      "freesurface-red","freesurface-blue","oxygen","phase2","PAR","salinity",
      "temperature","turbidity","velocity-blue","velocity-green",
      "vorticity-pink","vorticity-turquoise",
    
      "autumn","blackbody","bluered","bone","cold","cool","copper",
      "cubehelix","earth","greens","greys","hot","hsv",
      "jet","picnic","portland","rainbow-soft","rainbow","rdbu",
      "spring","summer","warm","winter","yignbu","yiorrd",
      "expanded-bone","expanded-copper",
    
      "CET-R3","CET-R2","CET-R1","CET-L9","CET-L8","CET-L7","CET-L6","CET-L5",
      "CET-L4","CET-L3","CET-L2","CET-L1","CET-L19","CET-L18","CET-L17",
      "CET-L16","CET-L15","CET-L14","CET-L13","CET-L12","CET-L11","CET-L10",
      "CET-I3","CET-I2","CET-I1","CET-D9","CET-D8","CET-D7","CET-D6","CET-D4",
      "CET-D3","CET-D2","CET-D1","CET-D1A","CET-D13","CET-D12","CET-D11",
      "CET-D10","CET-CBTL2","CET-CBTL1","CET-CBTD1","CET-CBTC2","CET-CBTC1",
      "CET-CBL2","CET-CBL1","CET-CBD1","CET-CBC2","CET-CBC1","CET-C5s",
      "CET-C5","CET-C4s","CET-C4","CET-C3s","CET-C3","CET-C2s","CET-C2",
      "CET-C1s","CET-C1",
    
      "algae", "amp", "balance", "curl", "deep", "delta", "dense",
      "diff", "cm-gray", "haline", "ice", "matter", "oxy", "phase",
      "rain", "solar", "speed", "tarn", "tempo", "thermal", "topo",
      "turbid",
      
      "simpleBlack", "simpleBlue", "simpleGreen", "simpleIce", "simpleWhite"
    )'/>
   </xsl:variable>

   <!--
    : $this:ALIASES
    :-->
   <xsl:variable name="this:ALIASES" as="map(xs:string,item()*)">
      <xsl:sequence select='
      map {
        "bluegold": "CET-CBL2",
        "tampa": "CET-I2",
        "burn": "CET-L4",
        "jungle": "CET-L5",
        "sea": "CET-L6",
        "blossom": "CET-L7",
        "oasis": "CET-L11",
        "sky": "CET-L12",
        "blood": "CET-L13",
        "forest": "CET-L14",
        "deepsea": "CET-L15",
        "warmth": "CET-L18",
        "watermelon": "CET-D3"
      }
    '/>
   </xsl:variable>

   <!--
    : $this:NAMED-ALIASES
    :-->
   <xsl:variable name="this:NAMED-ALIASES" as="xs:string*">
      <xsl:sequence select='$this:ALIASES=>map:keys()'/>
   </xsl:variable>

   <!--
    : $this:WHITES
    :-->
   <xsl:variable name="this:WHITES" as="xs:string*">
      <xsl:sequence select='(
      "snow", "seashell", "white", "whitesmoke", "oldlace", "mintcream",
      "ivory", "ghostwhite", "floralwhite", "cornsilk", "gainsboro", "lemonchiffon"
    )'/>
   </xsl:variable>

   <!--
    : $this:PRETTY-PALETTES
    : PrettyCols from https://nrennie.github.io/PrettyCols/ CC0 :)
    :-->
   <xsl:variable name="this:PRETTY-PALETTES" as="map(xs:string,item()*)">
      <xsl:sequence select='
      map {
      (: sequential palettes :)
      "Blues": (
        "#436F85", "#4C7D96", "#548BA7", "#6497B1", "#75A2BA", "#86AEC2", "#97B9CB"
      ),
      "Purples": (
        "#432263", "#502876", "#5D2F89", "#6A359C", "#773BAF", "#8444C0", "#9057C6"
      ),
      "Tangerines": (
        "#DE7A00", "#F28500", "#FF9B21", "#FFB04F", "#FFC47D", "#FFD6A3", "#FFE1BD"
      ),
      "Greens": (
        "#416322", "#4E7628", "#5A892F", "#679C35", "#74AF3B", "#80C044", "#8DC657"
      ),
      "Pinks": (
        "#860A4D", "#9E0C5B", "#B50E68", "#CD1076", "#E51284", "#EE2290", "#F03A9C"
      ),
      "Teals": (
        "#004C4C", "#006666", "#008080", "#329999", "#66B2B2", "#99CCCC", "#CCE5E5"
      ),
      "Yellows": (
        "#E6B400", "#E6C700", "#E8D119", "#EBD632", "#F0E066", "#F2E57F", "#F7EFB2"
      ),
      "Reds": (
        "#B53737", "#BE5151", "#C76B6B", "#D08585", "#D99F9F", "#E3B9B9", "#ECD3D3"
      ),
      (: diverging palettes :)
      "PurpleGreens": (
        "#420F75", "#7640A9", "#AD72D6", "#E7A8FB", "#F5F5F5", "#99CE64", "#659A32", "#326812", "#033800"
      ),
      "PinkGreens": (
        "#7F0038", "#C31E6E", "#EF5FAF", "#FCAADE", "#F5F5F5", "#99CE64", "#659A32", "#326812", "#033800"
      ),
      "TangerineBlues": (
        "#552000", "#8A4D00", "#C17D17", "#F8B150", "#F5F5F5", "#93C6E1", "#5F93AC", "#2E627A", "#00344A"
      ),
      "PurpleTangerines": (
        "#420F75", "#7640A9", "#AD72D6", "#E7A8FB", "#F5F5F5", "#F8B150", "#C17D17", "#8A4D00", "#552000"),
      "PurplePinks": (
        "#420F75", "#7640A9", "#AD72D6", "#E7A8FB", "#F5F5F5", "#FCAADE", "#EF5FAF", "#C31E6E", "#7F0038"),
      "TealGreens": (
        "#00393A", "#0A6969", "#2D9C9C", "#6DCFCF", "#F5F5F5", "#99CE64", "#659A32", "#326812", "#033800"),
      "PurpleYellows": (
        "#420F75", "#7640A9", "#AD72D6", "#E7A8FB", "#F5F5F5", "#F2E8C4", "#EED682", "#EAC541", "#E6B400"
      ),
      "RedBlues": (
        "#B53737", "#C66969", "#D79C9C", "#E8CFCF", "#F5F5F5", "#93C6E1", "#5F93AC", "#2E627A", "#00344A"
      ),
      (: qualitative palettes :)
      "Bold": (
        "#6497B1", "#6A359C", "#FFB04F", "#679C35", "#CD1076"
      ),
      "Dark": (
        "#436F85", "#432263", "#DE7A00", "#416322", "#860A4D"
      ),
      "Light": (
        "#97B9CB", "#9057C6", "#FFE1BD", "#8DC657", "#F03A9C"
      ),
      "Neon": (
        "#FF9062", "#FD6598", "#CB64C0", "#3294DD", "#75FB8A", "#d0eb60"
      ),
      "Summer": (
        "#398DB2", "#D8B31E", "#2C350B", "#829625", "#867112", "#5D761E",
        "#6293A7", "#3E5A5E", "#AC5C05", "#FFA300", "#A47DB9", "#EC94CA"
      ),
      "Autumn": (
        "#774762", "#BA6E1D", "#D6BB3B", "#755028", "#F2DD78", "#205F4B",
        "#913914", "#585854", "#F0A430", "#768048", "#800000", "#1B3A54"
      ),
      "Winter": (
        "#446C84", "#C0CBDC", "#746E6F", "#C6DCF0", "#596D80", "#B9BFFF",
        "#A0C4E1", "#897340", "#E1E3E7", "#313C45", "#9BA7B2", "#CAE9F5"
      ),
      "Rainbow": (
        "#E51E32", "#FF782A", "#FDA805", "#E2CF04", "#B1CA05", "#98C217",
        "#779815", "#029E77", "#09989C", "#059CCD", "#3F64CE", "#7E2B8E"
      ),
      "Beach": (
        "#0E7C7B", "#17BEBB", "#D4F4DD", "#D62246", "#4B1D3F"
      ),
      "Fun": (
        "#134074", "#BFAB25", "#4EA699", "#EFB0A1", "#DF2935"
      ),
      "Sea": (
        "#86CB92", "#71B48D", "#404E7C", "#3A3559", "#260F26"
      ),
      "Bright": (
        "#462255", "#FF8811", "#9DD9D2", "#046E8F", "#D44D5C"
      ),
      "Relax": (
        "#4B3F72", "#CBB3BF", "#FFC857", "#119DA4", "#19647E"
      ),
      "Lucent": (
        "#E01A4F", "#F15946", "#F9C22E", "#53B3CB", "#7DCFB6"
      ),
      "Lively": (
        "#413C58", "#D1495B", "#EDAE49", "#00798C", "#003D5B"
      ),
      "Joyful": (
        "#80A1C1", "#C94277", "#EEE3AB", "#274C77", "#5E8C61"
      )
      }
      '/>
   </xsl:variable>
   
   <!--
    : $this:PALETTES
    : Werner colours from Werner's Nomenclature of Colour
    :-->
   <xsl:variable name="this:WERNER-PALETTES" as="map(xs:string,item()*)">
      <xsl:sequence select='
      map {
        "werner_whites": (
          "SnowWhite", "ReddishWhite", "PurplishWhite", "YellowishWhite",
          "OrangeColouredWhite", "GreenishWhite", "SkimmedMilkWhite",
          "GreyishWhite"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_greys": (
          "AshGrey", "SmokeGrey", "FrenchGrey", "PearlGrey", "YellowishGrey",
          "BluishGrey", "GreenishGrey", "BlackishGrey"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_blacks": (
          "GreyishBlack", "BluishBlack", "GreenishBlack",
          "PitchOrBrownishBlack", "ReddishBlack", "InkBlack", "VelvetBlack"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_blues": (
          "ScotchBlue", "PrussianBlue", "ChinaBlue", "IndigoBlue", 
          "AzureBlue", "UltramarineBlue", "FlaxFlowerBlue", 
          "VerditterBlue", "GreenishBlue", "BerlinBlue", "GreyishBlue"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_purples": (
          "BluishLilacPurple", "RedLilacPurple", "BluishPurple", "LavenderPurple",
          "CampanulaPurple", "ImperialPurple", "AuriculaPurple",
          "PaleBlackishPurple", "PlumPurple", "VioletPurple", "PansyPurple"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_greens": (
          "SiskinGreen", "AsparagusGreen", "CeladineGreen", "AppleGreen",
          "EmeraldGreen", "BluishGreen", "MountainGreen", "VerdigrisGreen",
          "LeekGreen", "PistachioGreen", "OilGreen", "GrassGreen", "SapGreen",
          "OliveGreen", "BlackishGreen", "DuckGreen"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_yellows": (
          "CreamYellow", "SiennaYellow", "StrawYellow", "OchreYellow",
          "PrimroseYellow", "KingsYellow", "GambogeYellow", "LemonYellow",
          "WineYellow", "SaffronYellow", "SulphurYellow", "WaxYellow",
          "HoneyYellow", "GallstoneYellow"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_oranges": (
          "DutchOrange", "BuffOrange", "OrpimentOrange", 
          "ReddishOrange", "DeepReddishOrange", "BrownishOrange"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_reds": (
          "PurplishRed", "VeinousBloodRed", "ChocolateRed", "BrownishRed",
          "ArterialBloodRed", "ScarletRed", "TileRed", "VermilionRed",
          "HyacinthRed", "AuroraRed", "FleshRed", "RoseRed", "PeachBlossomRed",
          "CrimsonRed", "LakeRed", "CarmineRed", "BrownishPurpleRed",
          "CochinealRed"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "werner_browns": (
          "DeepOrangeColouredBrown", "DeepReddishBrown", "UmberBrown",
          "ChestnutBrown", "YellowishBrown", "WoodBrown", "HairBrown",
          "BroccoliBrown", "CloveBrown", "LiverBrown", "BlackishBrown"
        )!(rgb:named-colour(.)=>rgb:to-string())
      }
      '/>
   </xsl:variable>
   
   <!--
    : $this:PALETTES
    :-->
   <xsl:variable name="this:PALETTES" as="map(xs:string,item()*)">
     <xsl:sequence select='
     util:merge-into((
       map {
        "whites": $this:WHITES,
        "browns": (
          "#671905", "#833012", "#974B1A", "#AA6925", "#B6872D", "#BFA73B"
        ),
        "olives": (
          "#51351F", "#6D5332", "#8C7448", "#AC9561", "#CCB778", "#EEDA91"
        ),
        "sunsetland": (
          "#51351F", "#846841", "#B6A067", "#EEB46A", "#EA8C42", "#DF631D"
        ),
        "october": (
          "#695851", "#87837A", "#A4B0A4", "#FED32F", "#FEC52E", "#FCB72C"
        ),
        "farhills": (
          "#235240", "#426D5A", "#608977", "#80A594", "#A1C1B2", "#C4E0D2"
        ),
        "springhills": (
          "#636B32", "#778536", "#8E9F41", "#A7B84E", "#C4D263", "#E1EB7A"
        ),
        "mountains": (
          "#7E507E", "#9C7296", "#BB98AE", "#86963D", "#B0C258", "#E2EB78"
        ),
        "wotw": (
          "#3D3D1B", "#4F4E20", "#675C2E", "#FC6456", "#EC432E", "#D12615"
        ),
        "carnival": (
          "#2E294E", "#541388", "#F1E9DA", "#FFD400", "#D90368"
        ),
        "leaves": (
          "#24400C", "#2C510B", "#396C0D", "#4B8016", "#6BB121", "#8CD01F"
        ),
        "lake": (
          "#173A3F", "#0C5450", "#10615B", "#166862", "#30817A", "#41918F"
        ),
        "high-tea": (
          "MiddleBlueGreen", "TeaGreen", "Cream", "MediumChampagne", "DarkByzantium"
        )!(rgb:named-colour(.)=>rgb:to-string()),
        "plant-dyes": (
         "RedHotPoker", "HawthornLeaf", "BogMyrtle", "Silverweed", "GorseFlower",
         "ReedFlower", "NettleLeaf", "ColtsFoot", "Acorn", "IrisRoot"
         )!(rgb:named-colour(.)=>rgb:to-string())
       },
       $this:WERNER-PALETTES,
       $this:PRETTY-PALETTES
    ))
    '/>
   </xsl:variable>

   <!--
    : $this:NAMED-PALETTES
    :-->
   <xsl:variable name="this:NAMED-PALETTES" as="xs:string*">
      <xsl:sequence select='$this:PALETTES=>map:keys()'/>
   </xsl:variable>

   <!--
    : $this:SMALL-GRADIENTS
    :-->
   <xsl:variable name="this:SMALL-GRADIENTS" as="xs:string*">
      <xsl:sequence select='(
      $this:NAMED-PALETTES,
      "spring", "summer", "autumn", "winter",
      "bone", "copper", "earth", "electric", "portland", "rainbow-soft", "rainbow",
      "cold", "cool", "hot",
      "blackbody", "bluered", "greens", "greys", "hsv", "jet", "mars", "rdbu",
      "yignbu", "yiorrd",
      "simpleBlack", "simpleBlue", "simpleGreen", "simpleIce", "simpleWhite"
    )'/>
   </xsl:variable>

   <!--
    : $this:SPLIT-GRADIENTS
    :-->
   <xsl:variable name="this:SPLIT-GRADIENTS" as="xs:string*">
      <xsl:sequence select='(
      "bukavu", "fes", "oleron", "topo"
    )'/>
   </xsl:variable>

   <!--
    : $this:CYCLIC-GRADIENTS
    :-->
   <xsl:variable name="this:CYCLIC-GRADIENTS" as="xs:string*">
      <xsl:sequence select='(
      "bamO", "bathymetryO", "brocO", "corkO", "romaO", "vikO",
    
      "CET-CBTC2", "CET-CBTC1", "CET-CBC2", "CET-CBC1"," CET-C5s",
      "CET-C5", "CET-C4s", "CET-C4", "CET-C3s", "CET-C3", "CET-C2s", "CET-C2",
      "CET-C1s", "CET-C1",
    
      "azimuth", "hsv", "phase", "phase2", "rainbow-soft", "sunlight", "twilight"
    )'/>
   </xsl:variable>

   <!--
    : $this:LINEAR-GRADIENTS
    :-->
   <xsl:variable name="this:LINEAR-GRADIENTS" as="xs:string*">
      <xsl:sequence select='(
      "acton", "bamako", "batlow", "bilbao", "buda",
      "davos", "devon", "grayC", "hawaii", "imola", "lajolla", "lapaz",
      "nuuk", "oslo", "tokyo", "turku",
      "batlowK", "batlowW",
      "inferno", "magma", "plasma", "viridis",
      "cividis",
      "twilight-blue", "twilight-red",
      "cube1_0", "cubeYF", "Linear_L",
      "bathymetry", "CDOM", "chlorophyll", "density",
      "freesurface-red", "freesurface-blue", "oxygen", "PAR", "salinity",
      "temperature", "turbidity", "velocity-blue", "velocity-green",
      "vorticity-pink", "vorticity-turquoise",
    
      "autumn", "blackbody", "bluered", "bone", "cold", "cool", "copper",
      "cubehelix", "earth", "greens", "greys", "hot",
      "jet", "rainbow", "rdbu",
      "spring", "summer", "warm", "winter", "yignbu", "yiorrd",
      "expanded-bone", "expanded-copper",
    
      "CET-R3", "CET-R2", "CET-R1",
      "CET-L9", "CET-L8", "CET-L7", "CET-L6", "CET-L5",
      "CET-L4", "CET-L3", "CET-L2", "CET-L1", "CET-L19", "CET-L18", "CET-L17",
      "CET-L16", "CET-L15", "CET-L14", "CET-L13", "CET-L12", "CET-L11", "CET-L10",
      "CET-I3", "CET-I2", "CET-I1", 
      "CET-CBTL2", "CET-CBTL1", 
      "CET-CBL2", "CET-CBL1",
    
      "algae", "amp", "deep", "dense",
      "cm-gray", "haline", "ice", "matter", 
      "rain", "solar", "speed", "tempo", "thermal", "topo",
      "turbid",
      
      "simpleBlack", "simpleBlue", "simpleGreen", "simpleIce", "simpleWhite",

      "Blues", "Purples", "Tangerines", "Greens", "Pinks", "Teals",
      "Yellows", "Reds"
    )'/>
   </xsl:variable>

   <!--
    : this:starts-darker()
    : Does this gradient start with darker colours? i.e. dark-to-light
    : Return false for those that are dark-light-dark or light-dark-light or
    : scrambled
    :-->
   <xsl:function name="this:starts-darker" as="xs:boolean">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      ($gradient = (
      "acton", "bamako", "batlow", "buda", 
      "davos", "devon", "hawaii", "imola", "lapaz", 
      "nuuk", "oleron", "oslo", "tokyo", "turku", 
      "actonS", "bamakoS", "batlowS", (:"berlinS",:) "bilbaoS", (:"brocS",:) "budaS",
      "inferno", "magma", "plasma", "viridis",
      "cividis",
      "twilight-blue", "twilight-red",
      "sawtooth","azimuth",
      "cube1_0","cubeYF","Linear_L",
      "bathymetry","bathymetryO","CDOM","chlorophyll","density",
      "freesurface-red","freesurface-blue","oxygen","PAR","salinity",
      "temperature","turbidity","velocity-blue","velocity-green",
      "vorticity-pink","vorticity-turquoise",
    
      "autumn","blackbody","bluered","bone","cool","copper",
      "cubehelix","earth","electric","greens","greys","hot",
      "jet","picnic","portland","rainbow",
      "summer","warm","winter","yignbu","yiorrd",
      "expanded-bone","expanded-copper",
    
      "CET-L9","CET-L8","CET-L7","CET-L6","CET-L5",
      "CET-L4","CET-L3","CET-L2","CET-L1","CET-L18",
      "CET-L16","CET-L15","CET-L14","CET-L13","CET-L11","CET-L10",
      "CET-CBTL2","CET-CBTL1",
      "CET-CBL2","CET-CBL1",
    
      "bluegold", "burn", "jungle", "sea", "blossom", "oasis", "blood", "forest",
      "deepsea",
    
      "cm-gray", "haline", "ice", "oxy",
      "solar", "thermal", "topo",
      "turbid",
      
      "simpleBlack",
    
      "browns", "olives", "sunsetland", "october", "farhills", "springhills",
      "mountains", "wotw", "carnival", "greens",
    
      "werner_reds", "werner_blues", "werner_browns",

      "Blues", "Purples", "Tangerines", "Greens", "Pinks", "Teals", "Yellows", "Reds"
      ))
      or
      (contains($gradient,"-reverse") and
       not(this:starts-darker(replace($gradient,"-reverse","")))
      )
    '/>
   </xsl:function>

   <!--
    : this:random-gradient()
    : Return then name of a random well-known base gradient or its reverse.
    :-->
   <xsl:function name="this:random-gradient" as="xs:string" new-each-time="yes">
      <xsl:sequence select='
      let $exclusion := "(^simple)" return
      let $valid-names :=
        ($this:NAMED-GRADIENTS,$this:NAMED-ALIASES,$this:NAMED-PALETTES)[
          not(matches(.,$exclusion))
        ]
      return
      let $gradient := rand:select-random($valid-names)
      return (
        if (empty($gradient) or $gradient="") then util:log("Empty gradient? "||string-join($valid-names,"+")) else (),
        util:assert(not(empty($gradient)), "__empty__"),
        util:assert($gradient ne "", "__empty string__"),
        if (rand:flip(50))
        then $gradient
        else $gradient||"-reverse"
      )
    '/>
   </xsl:function>

   <!--
    : this:random-gradient()
    : Return then name of a random well-known base gradient or its reverse.
    :
    : $exclusion: regular expression of gradient names to exclude in addition
    :   to simpleXXX
    :-->
   <xsl:function name="this:random-gradient" as="xs:string" new-each-time="yes">
      <xsl:param name="exclusion" as="xs:string"/>
      <xsl:sequence select='
      let $exclusion := "("||$exclusion||")|(^simple)" return
      let $valid-names :=
        ($this:NAMED-GRADIENTS,$this:NAMED-ALIASES,$this:NAMED-PALETTES)[
          not(matches(.,$exclusion))
        ]
      return
      let $gradient := rand:select-random($valid-names)
      return (
        if (rand:flip(50))
        then $gradient
        else $gradient||"-reverse"
      )
    '/>
   </xsl:function>

   <!--
    : this:fetch-gradient()
    : Get the gradient definition of a known named gradient or its alias.
    : Empty otherwise.
    :-->
   <xsl:function name="this:fetch-gradient" as="element()?" new-each-time="no">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:choose>
        <xsl:when test='$gradient = $this:NAMED-GRADIENTS'>
          <!-- SLEAZY HACK ALERT  :-->
          <xsl:variable name="def" as="element()?">
            <xsl:try select='doc("../COLOURS/"||$gradient||".xsl")//xsl:param/*'>
              <xsl:catch errors="*"/>
              <xsl:fallback/>
            </xsl:try>
          </xsl:variable>
          <xsl:choose>
            <xsl:when test='exists($def)'><xsl:sequence select="$def"/></xsl:when>
            <xsl:otherwise>
              <xsl:try select='doc("../COLOURS/"||$gradient||".svg")//xsl:defs/*'>
                <xsl:catch errors="*"/>
                <xsl:fallback/>
              </xsl:try>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:when>
        <xsl:when test='($this:ALIASES($gradient) = $this:NAMED-GRADIENTS)'>
          <xsl:sequence select='this:fetch-gradient($this:ALIASES($gradient))'/>
        </xsl:when>
        <xsl:otherwise/>
      </xsl:choose>
   </xsl:function>

   <!--
    : this:get-gradient-definition()
    : Get the gradient definition of the name, allowing for variants for
    : flips, inversions, samples, etc.
    :-->
   <xsl:function name="this:get-gradient-definition" as="element()?" new-each-time="yes">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:variable name="gradient-pieces" as="xs:string*">
        <xsl:sequence select='tokenize($gradient, "[·+]")'/> <!-- middle dot  :-->
      </xsl:variable>
      <xsl:variable name="do-interleave" as="xs:boolean">
        <xsl:sequence select='contains($gradient,"+")'/>
      </xsl:variable>
      <xsl:variable name="gradients" as="element()*">
        <xsl:for-each select="$gradient-pieces">
          <xsl:variable name="gradient" select='.'/>
          <xsl:choose>
            <xsl:when test='$gradient="random"'>
              <xsl:sequence select='this:gradient(this:random-gradient())'/>
            </xsl:when>
            <xsl:when test='($gradient = $this:NAMED-GRADIENTS)'>
              <xsl:sequence select='this:fetch-gradient($gradient)'/>
            </xsl:when>
            <xsl:when test='($this:ALIASES($gradient) = $this:NAMED-GRADIENTS)'>
              <xsl:variable name="def" as="element()?">
                <xsl:sequence select='this:fetch-gradient($this:ALIASES($gradient))'/>
              </xsl:variable>
              <svg:linearGradient id="{$gradient}">
                <xsl:copy-of select='$def/(@* except @id)'/>
                <xsl:copy-of select='$def/*'/>
              </svg:linearGradient>
            </xsl:when>
            <xsl:when test='($gradient = $this:NAMED-PALETTES)'>
              <xsl:sequence select='this:linear-gradient-definition($gradient, "linear", $this:PALETTES($gradient))'/>
            </xsl:when>
            <xsl:when test='matches($gradient,"(-reverse|-random|-flip|-invert|-full|-anti|-outflow|-inflow|-slant|-fade|-[0-9]+)$")'>
              <xsl:variable name="root" as="xs:string">
                <xsl:sequence select='replace($gradient,"(-reverse|-random|-flip|-invert|-full|-anti|-outflow|-inflow|-slant|-fade|-[0-9]+)$","")'/>
              </xsl:variable>
              <xsl:sequence select='
              let $definition := this:get-gradient-definition($root) return
              let $operation := substring-after($gradient, $root||"-")
              return 
              if ($operation = "random") then this:gradient-random($definition)=>this:gradient-name($gradient)
              else if ($operation = "reverse") then this:gradient-reverse($definition)=>this:gradient-name($gradient)
              else if ($operation = "flip") then this:linear-flip($definition)=>this:gradient-name($gradient)
              else if ($operation = "invert") then this:linear-invert($definition)=>this:gradient-name($gradient)
              else if ($operation = "full") then this:linear-full($definition)=>this:gradient-name($gradient)
              else if ($operation = "anti") then this:linear-anti($definition)=>this:gradient-name($gradient)
              else if ($operation = "outflow") then this:radial-outflow($definition)=>this:gradient-name($gradient)
              else if ($operation = "inflow") then this:radial-inflow($definition)=>this:gradient-name($gradient)
              else if ($operation = "slant") then this:linear-slant($definition)=>this:gradient-name($gradient)
              else if ($operation = "fade") then this:gradient-fade($definition, 0.0, 1.0)=>this:gradient-name($gradient)
              else (: one of the samples :)  (
                let $n := xs:integer($operation)
                return this:gradient-sample($definition, $n)=>this:gradient-name($gradient)
              )
              '/>
            </xsl:when>
            <xsl:when test='starts-with($gradient,"black-to-")'>
              <xsl:variable name="colour" as="xs:string">
                <xsl:sequence select='this:colour(substring-after($gradient,"black-to-"))'/>
              </xsl:variable>
              <svg:linearGradient id="{this:safe($gradient)}" gradientUnits="objectBoundingBox" spreadMethod="pad" x1="0%" x2="100%" y1="0%" y2="0%">      
                <svg:stop offset="0%" stop-color="#000000"/>
                <svg:stop offset="100%" stop-color="{$colour}"/>
              </svg:linearGradient>
            </xsl:when>
            <xsl:when test='starts-with($gradient,"white-to-")'>
              <xsl:variable name="colour" as="xs:string">
                <xsl:sequence select='this:colour(substring-after($gradient,"white-to-"))'/>
              </xsl:variable>
              <svg:linearGradient id="{this:safe($gradient)}" gradientUnits="objectBoundingBox" spreadMethod="pad" x1="0%" x2="100%" y1="0%" y2="0%">      
                <svg:stop offset="0%" stop-color="#FFFFFF"/>
                <svg:stop offset="100%" stop-color="{$colour}"/>
              </svg:linearGradient>
            </xsl:when>
            <xsl:otherwise>
              <!-- util:log("gradient="||$gradient||" no definition") -->
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </xsl:variable>
      <xsl:sequence select='
      if (count($gradients) le 1) then $gradients
      else if (count($gradients) = count($gradient-pieces)) then (
        this:merge($gradient, $gradients, $do-interleave)
      ) else (
      )
      '/>
   </xsl:function>

   <!--
    : this:gradient()
    :-->
   <xsl:function name="this:gradient" as="element()?">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      this:get-gradient-definition($gradient)
    '/>
   </xsl:function>

   <!--
    : this:random-colour()
    :-->
   <xsl:function name="this:random-colour" as="xs:string" new-each-time="yes">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:sequence select='
      rand:select-random(this:colours($gradients))
    '/>
   </xsl:function>

   <!--
    : this:random-colour()
    :-->
   <xsl:function name="this:random-colour" as="xs:string" new-each-time="yes">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:param name="luminance" as="xs:double"/>
      <xsl:sequence select='
      rand:select-random(this:colours($gradients))=>
        rgb:rgb()=>cs:rgb-to-xyz()=>
        tone:luminate($luminance)=>
        cs:xyz-to-rgb()=>rgb:to-string()
    '/>
   </xsl:function>

   <!--
    : this:random-colour()
    :-->
   <xsl:function name="this:random-colour" as="xs:string" new-each-time="yes">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:param name="low-luminance" as="xs:double"/>
      <xsl:param name="high-luminance" as="xs:double"/>
      <xsl:sequence select='
      let $colour := rand:select-random(this:colours($gradients)) return
      let $xyz := $colour=>rgb:rgb()=>cs:rgb-to-xyz() return
      let $luminance := tone:luminance($xyz)
      return (
        if (util:twixt($luminance, $low-luminance, $high-luminance))
        then $colour
        else (
          $xyz=>tone:luminate(rand:uniform($low-luminance, $high-luminance))=>cs:xyz-to-rgb()=>rgb:to-string()
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:colours()
    : Return the colour strings from the named gradients in order.
    :-->
   <xsl:function name="this:colours" as="xs:string*">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:sequence select='
      for $gradient in $gradients return (
        let $definition := this:get-gradient-definition($gradient) return
        let $colours :=
          if (empty($definition)) then $gradient
          else $definition/svg:stop/@stop-color
        return
          if (matches($gradient,"-reverse")) then (
            reverse($colours)
          ) else (
            $colours
          )
      )
    '/>
   </xsl:function>

   <!--
    : this:colours()
    : Return the colour strings from the named gradients in order, but reluminated
    : to ensure their luminance is in the given range.
    :-->
   <xsl:function name="this:colours" as="xs:string*">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:param name="low" as="xs:double"/>
      <xsl:param name="high" as="xs:double"/>
      <xsl:sequence select='
      let $low := util:clamp-some($low, 0, ()) return
      let $high := util:clamp-some($high, (), 1) return
      let $colours := this:colours($gradients)
      return (
        if ($low &lt;= 0 and $high >= 1) then $colours else (
          for $c in $colours return
          let $xyz := rgb:rgb($c)=>cs:rgb-to-xyz() return
          let $luminance := tone:luminance($xyz)
          return (
            if ($luminance &lt; $low)
            then tone:luminate($xyz, $low)=>cs:xyz-to-rgb()=>rgb:to-string()
            else if ($luminance > $high)
            then tone:luminate($xyz, $high)=>cs:xyz-to-rgb()=>rgb:to-string()
            else $c
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:gradient-colours()
    :-->
   <xsl:function name="this:gradient-colours" as="xs:string*">
      <xsl:param name="gradients" as="xs:string*"/>
      <xsl:sequence select='
      this:colours($gradients)
    '/>
   </xsl:function>

   <!--
    : this:gradient-stops()
    : Return all the gradient stops, as svg:stop elements, from the given gradient.
    :-->
   <xsl:function name="this:gradient-stops" as="element(svg:stop)*">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      this:get-gradient-definition($gradient)/svg:stop
    '/>
   </xsl:function>

   <!--
    : this:n-gradient-colours()
    :-->
   <xsl:function name="this:n-gradient-colours" as="xs:integer">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      let $n := count(this:gradient-stops($gradient))
      return
        if ($n=0) then 1
        else $n
    '/>
   </xsl:function>

   <!--
    : this:n-colours()
    :-->
   <xsl:function name="this:n-colours" as="xs:integer">
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      let $n := count(this:gradient-stops($gradient))
      return
        if ($n=0) then 1
        else $n
    '/>
   </xsl:function>

   <!--
    : this:gradient-points()
    : Return the points in the gradient as gradient points. This assumes
    : the stop values are in "rgb(r,g,b)" or hex format. Opacity gives the
    : alpha channel.
    : 
    : A gradient point has an offset and a colour point. The set of
    : gradient points can be used for interpolations and suchlike.
    : The colour points returned from this are RGBA points.
    : TODO: select output colour space
    :-->
   <xsl:function name="this:gradient-points"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="gradient" as="xs:string"/>
      <xsl:sequence select='
      for $stop in this:gradient-stops($gradient) return
      let $offset :=
        if (ends-with($stop/@offset,"%"))
        then xs:double(substring-before($stop/@offset,"%")) div 100
        else xs:double($stop/@offset)
      return
      let $alpha :=
        if (empty($stop/@stop-opacity)) then 1
        else if (ends-with($stop/@stop-opacity,"%"))
        then xs:double(substring-before($stop/@stop-opacity,"%")) div 100
        else xs:double($stop/@stop-opacity)
      return
      let $rgba := rgb:rgba(rgb:rgb($stop/@stop-color), $alpha)
      return stop:stop("rgb", $offset, $rgba)
    '/>
   </xsl:function>

   <!--
    : this:id()
    :-->
   <xsl:function name="this:id" as="xs:string">
      <xsl:param name="gradient" as="element()"/>
      <xsl:sequence select='
      $gradient/@id
    '/>
   </xsl:function>

   <!--
    : this:merge()
    :-->
   <xsl:function name="this:merge" as="element()">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="gradients" as="element()+"/>
      <xsl:param name="do-interleave" as="xs:boolean"/>
      <xsl:sequence select='
      let $kind :=
        if (head($gradients) instance of element(svg:radialGradient)) then "outflow"
        else "linear"
      return
      let $max-n := max(for $gradient in $gradients return count($gradient//svg:stop))
      return
      let $stops :=
        if ($do-interleave) then (
          for $i in 1 to $max-n return
          for $gradient in $gradients
          return (
            ($gradient//svg:stop)[$i]
          )
        ) else (
          for $gradient in $gradients return $gradient//svg:stop
        )
      return (
        this:gradient-definition($name, $kind,
          for $stop in $stops return $stop/@stop-opacity,
          for $stop in $stops return $stop/@stop-color
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:sample()
    :-->
   <xsl:function name="this:sample" as="xs:string*">
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:sequence select='
      if (empty($colours) or $num &lt;= 1) then () else
      let $sep := count($colours) div $num return
      for $i in 1 to $num
      return $colours[1 + ceiling($sep*($i - 1))]
    '/>
   </xsl:function>

   <!--
    : this:sample()
    :-->
   <xsl:function name="this:sample" as="xs:string*">
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:param name="above" as="xs:double"/>
      <xsl:param name="below" as="xs:double"/>
      <xsl:sequence select='
      this:sample($colours, count($colours), $above, $below)
    '/>
   </xsl:function>

   <!--
    : this:sample()
    :-->
   <xsl:function name="this:sample" as="xs:string*">
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:param name="above" as="xs:double"/>
      <xsl:param name="below" as="xs:double"/>
      <xsl:sequence select='
      if (empty($colours) or $num &lt;= 1) then () else
      let $n := count($colours) return
      let $min-n := round($n * $above) cast as xs:integer return
      let $max-n := round($n * $below) cast as xs:integer return
      (: Asymmetry here because we want to allow 0 for $above and 1 for $below :)
      let $colours := $colours[position() > $min-n and position() &lt;= $max-n] return
      let $sep := count($colours) div $num return
      for $i in 1 to $num
      return $colours[1 + ceiling($sep*($i - 1))]
    '/>
   </xsl:function>

   <!--
    : this:sample-stops()
    :-->
   <xsl:function name="this:sample-stops" as="element(svg:stop)*">
      <xsl:param name="stops" as="element(svg:stop)*"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:sequence select='
      if (empty($stops) or $num &lt;= 1) then () else
      let $sep := count($stops) div $num return
      for $i in 1 to $num
      return $stops[1 + ceiling($sep*($i - 1))]
    '/>
   </xsl:function>

   <!--
    : this:circularize()
    :-->
   <xsl:function name="this:circularize" as="xs:string*">
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      tail($colours), tail(reverse($colours))
    '/>
   </xsl:function>

   <!--
    : this:luminance-range()
    :-->
   <xsl:function name="this:luminance-range" as="xs:string*">
      <xsl:param name="colour" as="xs:string"/>
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="delta" as="xs:double"/>
      <xsl:sequence select='
      let $rgb := rgb:rgb($colour) return
      let $lum := tone:luminance($rgb)
      return (
        for $l in util:linspace($n, $lum - $delta, $lum, true())
        return tone:luminate($rgb, util:clamp($l, 0.0, 1.0))=>rgb:to-string(),
        $colour,
        for $l in util:linspace($n, $lum, $lum + $delta)=>tail()
        return tone:luminate($rgb, util:clamp($l, 0.0, 1.0))=>rgb:to-string()
      )
    '/>
   </xsl:function>

   <!--
    : this:colour()
    :-->
   <xsl:function name="this:colour" as="xs:string">
      <xsl:param name="name" as="xs:string"/>
      <xsl:sequence select='
      let $known := $rgb:KNOWN-COLOURS($name)
      return (
        if (empty($known)) then $name
        else $known
      )
    '/>
   </xsl:function>

   <!--
    : this:ref()
    :-->
   <xsl:function name="this:ref" as="xs:string">
      <xsl:param name="name" as="xs:string"/>
      <xsl:sequence select='
      let $known := $rgb:KNOWN-COLOURS($name)
      return (
        if (exists($known)) then $known
        else if ($name="none") then $name
        else if (matches($name, "^(rgb|url|#)")) then $name
        else "url(#"||this:safe($name)||")"
      )
    '/>
   </xsl:function>

   <!--
    : this:safe()
    :-->
   <xsl:function name="this:safe" as="xs:string">
      <xsl:param name="name" as="xs:string"/>
      <xsl:sequence select='
      replace(replace($name, "[(),#]", "-"), " ", "")
    '/>
   </xsl:function>

   <!--
    : this:average-luminance()
    : Average luminance: a way of determining relative brightness of a set
    : of colours.
    :-->
   <xsl:function name="this:average-luminance" as="xs:double">
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      avg(
        for $colour in $colours return (
          tone:luminance(cs:rgb-to-xyz(rgb:rgb($colour)))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:linear-gradient-definition()
    :-->
   <xsl:function name="this:linear-gradient-definition" as="element(svg:linearGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="layout" as="xs:integer*"/>
      <xsl:param name="opacities" as="xs:double*"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:variable name="n" as="xs:integer" select='count($colours)'/>
      <xsl:variable name="x1" as="xs:integer" select='($layout[1],0)[1]'/>
      <xsl:variable name="x2" as="xs:integer" select='($layout[2],100)[1]'/>
      <xsl:variable name="y1" as="xs:integer" select='($layout[3],0)[1]'/>
      <xsl:variable name="y2" as="xs:integer" select='($layout[4],0)[1]'/>
      <xsl:variable name="start-opacity" as="xs:double" select='($opacities[1],1)[1]'/>
      <xsl:variable name="end-opacity" as="xs:double" select='($opacities[2],1)[1]'/>
      <xsl:choose>
        <xsl:when test='$start-opacity=$end-opacity'>
          <svg:linearGradient id="{$name}" gradientUnits="objectBoundingBox" spreadMethod="pad" x1="{$x1}%" x2="{$x2}%" y1="{$y1}%" y2="{$y2}%">
            <svg:stop offset="0.00%" stop-color="{$colours[1]}" stop-opacity="{$start-opacity}"/>
            <xsl:choose>
              <xsl:when test='$n gt 100'>
                <xsl:for-each select="1 to $n">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
                  <xsl:variable name="offset" as="xs:double">
                    <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                    <!-- percent to 2 decimals :-->
                  </xsl:variable>
                  <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$start-opacity}"/>
                  <xsl:if test='exists($colours[$i + 1])'>
                    <svg:stop offset="{$offset}%" stop-color="{$colours[$i + 1]}" stop-opacity="{$start-opacity}"/>
                  </xsl:if>
                </xsl:for-each>
              </xsl:when>
              <xsl:otherwise>
                <xsl:for-each select="1 to $n">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
                  <xsl:variable name="offset" as="xs:double">
                    <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                    <!-- percent to 2 decimals :-->
                  </xsl:variable>
                  <xsl:if test='$i gt 1 and $i lt $n'>
                    <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$start-opacity}"/>
                  </xsl:if>
                </xsl:for-each>
              </xsl:otherwise>
            </xsl:choose>
            <svg:stop offset="100.00%" stop-color="{$colours[last()]}" stop-opacity="{$start-opacity}"/>
          </svg:linearGradient>
        </xsl:when>
        <xsl:otherwise>
          <xsl:variable name="delta">
            <xsl:sequence select='($end-opacity - $start-opacity) div $n'/>
          </xsl:variable>
          <svg:linearGradient id="{$name}" gradientUnits="objectBoundingBox" spreadMethod="pad" x1="{$x1}%" x2="{$x2}%" y1="{$y1}%" y2="{$y2}%">
            <svg:stop offset="0.00%" stop-color="{$colours[1]}" stop-opacity="{$start-opacity}"/>
            <xsl:for-each select="1 to $n">
              <xsl:variable name="i" as="xs:integer" select="."/>
              <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
              <xsl:variable name="offset" as="xs:double">
                <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                <!-- percent to 2 decimals :-->
              </xsl:variable>
              <xsl:variable name="opacity">
                <xsl:sequence select='util:decimal($start-opacity + $delta*($i - 1), 2)'/>
              </xsl:variable>
              <xsl:choose>
                <xsl:when test='$n gt 100'>
                  <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$opacity}"/>
                  <xsl:if test='exists($colours[$i + 1])'>
                    <svg:stop offset="{$offset}%" stop-color="{$colours[$i + 1]}" stop-opacity="{$opacity}"/>
                  </xsl:if>
                </xsl:when>
                <xsl:otherwise>
                  <xsl:if test='$i gt 1 and $i lt $n'>
                    <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$opacity}"/>
                  </xsl:if>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:for-each>
            <svg:stop offset="100.00%" stop-color="{$colours[last()]}" stop-opacity="{$end-opacity}"/>
          </svg:linearGradient>
        </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <!--
    : this:linear-standard-layout()
    :-->
   <xsl:function name="this:linear-standard-layout" as="xs:integer*">
      <xsl:param name="kind" as="xs:string"/>
      <xsl:sequence select='
      if ($kind="reverse") then (100,0,0,0)
      else if ($kind="flip") then (0,0,0,100)
      else if ($kind="invert") then (0,0,100,0)
      else if ($kind="full") then (0,100,0,100)
      else if ($kind="anti") then (100,0,100,0)
      else if ($kind="slant") then (0,50,0,90)
      else (0,100,0,0)
    '/>
   </xsl:function>

   <!--
    : this:linear-gradient-definition()
    :-->
   <xsl:function name="this:linear-gradient-definition" as="element(svg:linearGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="kind" as="xs:string"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      this:linear-gradient-definition($name, this:linear-standard-layout($kind), (1,1), $colours)
    '/>
   </xsl:function>

   <!--
    : this:radial-gradient-definition()
    :-->
   <xsl:function name="this:radial-gradient-definition" as="element(svg:radialGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="layout" as="xs:integer*"/>
      <xsl:param name="opacities" as="xs:double*"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:variable name="n" as="xs:integer" select='count($colours)'/>
      <xsl:variable name="r" as="xs:integer" select='($layout[1],50)[1]'/>
      <xsl:variable name="cx" as="xs:integer" select='($layout[2],50)[1]'/>
      <xsl:variable name="cy" as="xs:integer" select='($layout[3],$cx)[1]'/>
      <xsl:variable name="start-opacity" as="xs:double" select='($opacities[1],1)[1]'/>
      <xsl:variable name="end-opacity" as="xs:double" select='($opacities[2],1)[1]'/>
      <xsl:choose>
        <xsl:when test='$start-opacity=$end-opacity'>
          <svg:radialGradient id="{$name}" gradientUnits="objectBoundingBox" spreadMethod="pad" cx="{$cx}%" cy="{$cy}%" r="{$r}%">      
            <svg:stop offset="0.00%" stop-color="{$colours[1]}" stop-opacity="{$start-opacity}"/>
            <xsl:choose>
              <xsl:when test='$n gt 100'>
                <xsl:for-each select="1 to $n">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
                  <xsl:variable name="offset" as="xs:double">
                    <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                    <!-- percent to 2 decimals :-->
                  </xsl:variable>
                  <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$start-opacity}"/>
                  <xsl:if test='exists($colours[$i + 1])'>
                    <svg:stop offset="{$offset}%" stop-color="{$colours[$i + 1]}" stop-opacity="{$start-opacity}"/>
                  </xsl:if>
                </xsl:for-each>
              </xsl:when>
              <xsl:otherwise>
                <xsl:for-each select="1 to $n">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
                  <xsl:variable name="offset" as="xs:double">
                    <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                    <!-- percent to 2 decimals :-->
                  </xsl:variable>
                  <xsl:if test='$i gt 1 and $i lt $n'>
                    <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$start-opacity}"/>
                  </xsl:if>
                </xsl:for-each>
              </xsl:otherwise>
            </xsl:choose>
            <svg:stop offset="100.00%" stop-color="{$colours[last()]}" stop-opacity="{$start-opacity}"/>
          </svg:radialGradient>
        </xsl:when>
        <xsl:otherwise>
          <xsl:variable name="delta">
            <xsl:sequence select='($end-opacity - $start-opacity) div $n'/>
          </xsl:variable>
          <svg:radialGradient id="{$name}" gradientUnits="objectBoundingBox" spreadMethod="pad" cx="{$cx}%" cy="{$cy}%" r="{$r}%">      
            <svg:stop offset="0.00%" stop-color="{$colours[1]}" stop-opacity="{$start-opacity}"/>
            <xsl:for-each select="1 to $n">
              <xsl:variable name="i" as="xs:integer" select="."/>
              <xsl:variable name="colour" as="xs:string" select="$colours[$i]"/>
              <xsl:variable name="offset" as="xs:double">
                <xsl:sequence select='round(100* 100*$i div $n) div 100'/>
                <!-- percent to 2 decimals :-->
              </xsl:variable>
              <xsl:variable name="opacity">
                <xsl:sequence select='util:decimal($start-opacity + $delta*($i - 1), 2)'/>
              </xsl:variable>
              <xsl:choose>
                <xsl:when test='$n gt 100'>
                  <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$opacity}"/>
                  <xsl:if test='exists($colours[$i + 1])'>
                    <svg:stop offset="{$offset}%" stop-color="{$colours[$i + 1]}" stop-opacity="{$opacity}"/>
                  </xsl:if>
                </xsl:when>
                <xsl:otherwise>
                  <xsl:if test='$i gt 1 and $i lt $n'>
                    <svg:stop offset="{$offset}%" stop-color="{$colour}" stop-opacity="{$opacity}"/>
                  </xsl:if>
                </xsl:otherwise>
              </xsl:choose>
            </xsl:for-each>
            <svg:stop offset="100.00%" stop-color="{$colours[last()]}" stop-opacity="{$end-opacity}"/>
          </svg:radialGradient>
        </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <!--
    : this:radial-standard-layout()
    :-->
   <xsl:function name="this:radial-standard-layout"
                 as="xs:integer*"
                >
      <xsl:param name="kind" as="xs:string"/>
      <xsl:sequence select='
      if ($kind="stargate") then (50,50,80) else (50,50,50)
    '/>
   </xsl:function>

   <!--
    : this:radial-colours()
    :-->
   <xsl:function name="this:radial-colours" as="xs:string*">
      <xsl:param name="kind" as="xs:string"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      if ($kind="inflow") then reverse($colours) else $colours
    '/>
   </xsl:function>

   <!--
    : this:radial-gradient-definition()
    :-->
   <xsl:function name="this:radial-gradient-definition" as="element(svg:radialGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="kind" as="xs:string"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      this:radial-gradient-definition($name, this:radial-standard-layout($kind), (1,1), this:radial-colours($kind, $colours))
    '/>
   </xsl:function>

   <!--
    : this:gradient-definition()
    :-->
   <xsl:function name="this:gradient-definition" as="element()">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="kind" as="xs:string"/>
      <xsl:param name="opacities" as="xs:double*"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      if ($kind=("inflow","outflow","stargate")) then (
        this:radial-gradient-definition($name, this:radial-standard-layout($kind), $opacities, this:radial-colours($kind, $colours))
      ) else (
        this:linear-gradient-definition($name, this:linear-standard-layout($kind), $opacities, $colours)
      )
    '/>
   </xsl:function>

   <!--
    : this:gradient-definition()
    :-->
   <xsl:function name="this:gradient-definition" as="element()">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="kind" as="xs:string"/>
      <xsl:param name="colours" as="xs:string*"/>
      <xsl:sequence select='
      this:gradient-definition($name, $kind, (1,1), $colours)
    '/>
   </xsl:function>

   <!--
    : this:is-radial()
    :-->
   <xsl:function name="this:is-radial" as="xs:boolean">
      <xsl:param name="def" as="element()"/>
      <xsl:sequence select='
      local-name($def)="radialGradient"
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Gradient manipulations
 :======================================================================
 :-->

   <!--
    : this:gradient-name()
    :-->
   <xsl:function name="this:gradient-name" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="name" as="xs:string"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-name($base-gradient, $name)
      else this:linear-name($base-gradient, $name)
    '/>
   </xsl:function>

   <!--
    : this:gradient-reverse()
    :-->
   <xsl:function name="this:gradient-reverse" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-reverse($base-gradient)
      else this:linear-reverse($base-gradient)
    '/>
   </xsl:function>

   <!--
    : this:gradient-random()
    :-->
   <xsl:function name="this:gradient-random" as="element()" new-each-time="yes">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-random($base-gradient)
      else this:linear-random($base-gradient)
    '/>
   </xsl:function>


   <!--
    : this:gradient-sample()
    :-->
   <xsl:function name="this:gradient-sample" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-sample($base-gradient, $num)
      else this:linear-sample($base-gradient, $num)
    '/>
   </xsl:function>

   <!--
    : this:gradient-fade()
    :-->
   <xsl:function name="this:gradient-fade" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="start-opacity" as="xs:double"/>
      <xsl:param name="end-opacity" as="xs:double"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-fade($base-gradient, $start-opacity, $end-opacity)
      else this:linear-fade($base-gradient, $start-opacity, $end-opacity)
    '/>
   </xsl:function>

   <!--
    : this:gradient-transform()
    :-->
   <xsl:function name="this:gradient-transform" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="transform" as="xs:string"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-transform($base-gradient, $transform)
      else this:linear-transform($base-gradient, $transform)
    '/>
   </xsl:function>

   <!--
    : this:gradient-units()
    :-->
   <xsl:function name="this:gradient-units" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="objectBoundingBox" as="xs:boolean"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-units($base-gradient, $objectBoundingBox)
      else this:linear-units($base-gradient, $objectBoundingBox)
    '/>
   </xsl:function>

   <!--
    : this:gradient-spread()
    :-->
   <xsl:function name="this:gradient-spread" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="method" as="xs:string"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-spread($base-gradient, $method)
      else this:linear-spread($base-gradient, $method)
    '/>
   </xsl:function>

   <!--
    : this:gradient-layout()
    :-->
   <xsl:function name="this:gradient-layout" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="layout" as="xs:integer*"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-layout($base-gradient, $layout)
      else this:linear-layout($base-gradient, $layout)
    '/>
   </xsl:function>

   <!--
    : this:gradient-interleave()
    :-->
   <xsl:function name="this:gradient-interleave" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="interleave" as="xs:string"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-interleave($base-gradient, $interleave)
      else this:linear-interleave($base-gradient, $interleave)
    '/>
   </xsl:function>

   <!--
    : this:gradient-opacity()
    :-->
   <xsl:function name="this:gradient-opacity" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="opacity" as="(function(xs:integer) as xs:double)"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-opacity($base-gradient, $opacity)
      else this:linear-opacity($base-gradient, $opacity)
    '/>
   </xsl:function>

   <!--
    : this:gradient-colour()
    :-->
   <xsl:function name="this:gradient-colour" as="element()">
      <xsl:param name="base-gradient" as="element()"/>
      <xsl:param name="colour" as="(function(xs:integer) as xs:string)"/>
      <xsl:sequence select='
      if (this:is-radial($base-gradient))
      then this:radial-colour($base-gradient, $colour)
      else this:linear-colour($base-gradient, $colour)
    '/>
   </xsl:function>

   <!--
    : this:gradient-ref()
    :-->
   <xsl:function name="this:gradient-ref" as="element()">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="base-gradient-name" as="xs:string"/>
      <xsl:sequence select='
      let $def := this:gradient($base-gradient-name)
      return (
        if (exists($def) and this:is-radial($def))
        then this:radial-ref($name, $base-gradient-name)
        else this:linear-ref($name, $base-gradient-name)
      )
    '/>
   </xsl:function>

   <!--
    : this:linear-name()
    :-->
   <xsl:function name="this:linear-name" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="name" as="xs:string"/>
      <svg:linearGradient id="{this:safe($name)}">
        <xsl:copy-of select="$base-gradient/(@* except @id)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-reverse()
    :-->
   <xsl:function name="this:linear-reverse" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient x1="{$base-gradient/@x2}" x2="{$base-gradient/@x1}"
                          y1="{$base-gradient/@y2}" y2="{$base-gradient/@y1}">
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-flip()
    :-->
   <xsl:function name="this:linear-flip" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient x1="{$base-gradient/@y1}" x2="{$base-gradient/@y2}"
                          y1="{$base-gradient/@x1}" y2="{$base-gradient/@x2}"
                          >
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-invert()
    :-->
   <xsl:function name="this:linear-invert" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient x1="{$base-gradient/@y1}" x2="{$base-gradient/@y2}"
                          y1="{$base-gradient/@x2}" y2="{$base-gradient/@x1}"
                          >
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-full()
    :-->
   <xsl:function name="this:linear-full" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient y1="{$base-gradient/@x1}" y2="{$base-gradient/@x2}">
        <xsl:copy-of select="$base-gradient/(@* except (@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-anti()
    :-->
   <xsl:function name="this:linear-anti" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient x1="{$base-gradient/@x2}" x2="{$base-gradient/@x1}"
                          y1="{$base-gradient/@x2}" y2="{$base-gradient/@x1}"
                          >
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-slant()
    :-->
   <xsl:function name="this:linear-slant" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:linearGradient x1="0%" x2="50%"
                          y1="0%" y2="90%"
                          >
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-adjust()
    :-->
   <xsl:function name="this:linear-adjust" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="xys" as="xs:integer*"/>
      <svg:linearGradient x1="{$xys[1]}%" x2="{$xys[3]}%"
                          y1="{$xys[2]}%" y2="{$xys[4]}%"
                          >
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-random()
    :-->
   <xsl:function name="this:linear-random" as="element(svg:linearGradient)" new-each-time="yes">
      <xsl:param name="base-gradient"
                 as="element(svg:linearGradient)"/>
      <xsl:variable name="stops" as="element()*" select='rand:shuffle($base-gradient/svg:stop)'/>
      <xsl:variable name="percentages" as="xs:double*" select='util:linspace(count($stops), 0, 100)!util:decimal(., 2)'/>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()" select='$stops[$i]'/>
          <svg:stop offset="{$percentages[$i]}%">
            <xsl:copy-of select="$stop/(@* except @offset)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-interleave()
    :-->
   <xsl:function name="this:linear-interleave" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="interleave" as="xs:string"/>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*" select='$base-gradient/svg:stop'/>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()" select='$stops[$i]'/>
          <xsl:choose>
            <xsl:when test='$i mod 2 = 0'> 
              <xsl:sequence select='$stop'/>
            </xsl:when>
            <xsl:otherwise>
              <svg:stop stop-color="{$interleave}">
                <xsl:copy-of select="$stop/(@* except @stop-color)"/>
              </svg:stop>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-opacity()
    :-->
   <xsl:function name="this:linear-opacity" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="opacity" as="(function(xs:integer) as xs:double)"/>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*" select='$base-gradient/svg:stop'/>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()" select='$stops[$i]'/>
          <xsl:variable name="stop-opacity" as="xs:double">
            <xsl:sequence select='util:decimal(util:clamp($opacity($i),0,1),3)'/>
          </xsl:variable>
          <svg:stop stop-opacity="{$stop-opacity}">
            <xsl:copy-of select="$stop/(@* except @stop-opacity)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-colour()
    :-->
   <xsl:function name="this:linear-colour" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="colour" as="(function(xs:integer) as xs:string)"/>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*" select='$base-gradient/svg:stop'/>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()" select='$stops[$i]'/>
          <xsl:variable name="stop-colour" as="xs:string" select='$colour($i)'/>
          <svg:stop stop-color="{$stop-colour}">
            <xsl:copy-of select="$stop/(@* except @stop-color)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-sample()
    :-->
   <xsl:function name="this:linear-sample" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:choose>
        <xsl:when test='$num le 1'>
          <!-- An error, really: just return the gradient back -->
          <xsl:sequence select='$base-gradient'/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:variable name="stops" as="element()*" select='$base-gradient/svg:stop'/>
          <xsl:variable name="percentages" as="xs:double*">
            <xsl:sequence select='util:linspace($num, 0, 100)!util:decimal(., 2)'/>
          </xsl:variable>
          <xsl:variable name="sep" as="xs:double" select='count($stops) div $num'/>
          <xsl:choose>
            <xsl:when test='count($stops) le $num'>
              <xsl:sequence select='$base-gradient'/>
            </xsl:when>
            <xsl:otherwise>
              <svg:linearGradient>
                <xsl:copy-of select="$base-gradient/@*"/>
                <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
                <xsl:for-each select="1 to $num">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="stop" as="element()">
                    <xsl:sequence select='$stops[1 + ceiling($sep*($i - 1))]'/>
                  </xsl:variable>
                  <svg:stop offset="{$percentages[$i]}%">
                    <xsl:copy-of select="$stop/(@* except @offset)"/>
                  </svg:stop>
                </xsl:for-each>
              </svg:linearGradient>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <!--
    : this:linear-fade()
    :-->
   <xsl:function name="this:linear-fade" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="start-opacity" as="xs:double"/>
      <xsl:param name="end-opacity" as="xs:double"/>
      <xsl:variable name="n" as="xs:integer">
        <xsl:sequence select='count($base-gradient/svg:stop)'/>
      </xsl:variable>
      <xsl:variable name="delta" as="xs:double">
        <xsl:sequence select='($end-opacity - $start-opacity) div ($n - 1)'/>
      </xsl:variable>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:variable name="opacity" as="xs:double">
            <xsl:sequence select='util:decimal($start-opacity + $delta*($i - 1), 3)'/>
          </xsl:variable>
          <svg:stop stop-opacity="{$opacity}">
            <xsl:copy-of select="$stop/(@* except @opacity)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-symmetric-fade()
    :-->
   <xsl:function name="this:linear-symmetric-fade" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="start-opacity" as="xs:double"/>
      <xsl:param name="peak-opacity" as="xs:double"/>
      <xsl:param name="end-opacity" as="xs:double"/>

      <xsl:variable name="n" as="xs:integer">
        <xsl:sequence select='count($base-gradient/svg:stop)'/>
      </xsl:variable>
      <xsl:variable name="half" as="xs:double">
        <xsl:sequence select='$n idiv 2'/>
      </xsl:variable>
      <xsl:variable name="delta1" as="xs:double">
        <xsl:sequence select='($peak-opacity - $start-opacity) div ($n idiv 2)'/>
      </xsl:variable>
      <xsl:variable name="delta2" as="xs:double">
        <xsl:sequence select='($end-opacity - $peak-opacity) div (($n + 1) idiv 2)'/>
      </xsl:variable>
      <svg:linearGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:variable name="opacity" as="xs:double">
            <xsl:sequence select='
                                  if ($i le $half)
                                  then util:decimal($start-opacity + $delta1*($i - 1), 2)
                                  else util:decimal($peak-opacity + $delta2*($i - $half - 1), 2)
                                  '/>
          </xsl:variable>
          <svg:stop stop-opacity="{$opacity}">
            <xsl:copy-of select="$stop/(@* except @opacity)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-transform()
    :-->
   <xsl:function name="this:linear-transform" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient"
                 as="element(svg:linearGradient)"/>
      <xsl:param name="transform" as="xs:string"/>
      <svg:linearGradient gradientTransform="{$transform}">
        <xsl:copy-of select="$base-gradient/(@* except @gradientTransform)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-units()
    :-->
   <xsl:function name="this:linear-units" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="objectBoundingBox" as="xs:boolean"/>
      <xsl:variable name="val">
        <xsl:sequence select='if ($objectBoundingBox) then "objectBoundingBox" else "userSpaceOnUse"'/>
      </xsl:variable>
      <svg:linearGradient gradientUnits="{$val}">
        <xsl:copy-of select="$base-gradient/(@* except @gradientUnits)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-spread()
    :-->
   <xsl:function name="this:linear-spread" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="method" as="xs:string"/>
      <svg:linearGradient spreadMethod="{$method}">
        <xsl:copy-of select="$base-gradient/(@* except @spreadMethod)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-layout()
    :-->
   <xsl:function name="this:linear-layout" as="element(svg:linearGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:param name="layout" as="xs:integer*"/>
      <xsl:variable name="x1" as="xs:integer" select='($layout[1],0)[1]'/>
      <xsl:variable name="x2" as="xs:integer" select='($layout[2],100)[1]'/>
      <xsl:variable name="y1" as="xs:integer" select='($layout[3],0)[1]'/>
      <xsl:variable name="y2" as="xs:integer" select='($layout[4],0)[1]'/>
      <svg:linearGradient x1="{$x1}%" x2="{$x2}%" y1="{$y1}%" y2="{$y2}%">
        <xsl:copy-of select="$base-gradient/(@* except (@x1|@x2|@y1|@y2))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:linearGradient>
   </xsl:function>

   <!--
    : this:linear-ref()
    :-->
   <xsl:function name="this:linear-ref" as="element(svg:linearGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="base-gradient-name" as="xs:string"/>
      <svg:linearGradient id="{this:safe($name)}" xlink:href="#{$base-gradient-name}"/>
   </xsl:function>

   <!--
    : this:radial-name()
    :-->
   <xsl:function name="this:radial-name" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="name" as="xs:string"/>
      <svg:radialGradient id="{this:safe($name)}">
        <xsl:copy-of select="$base-gradient/(@* except @id)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>  
   </xsl:function>

   <!--
    : this:radial-outflow()
    :-->
   <xsl:function name="this:radial-outflow" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <svg:radialGradient cx="50%" cy="50%" r="50%">
        <xsl:copy-of select="$base-gradient/(@* except (@r|@cx|@cy))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>  
   </xsl:function>

   <!--
    : this:radial-inflow()
    :-->
   <xsl:function name="this:radial-inflow" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:linearGradient)"/>
      <xsl:variable name="offsets" as="xs:double*">
        <xsl:sequence select='
        reverse(
          for $offset in $base-gradient/svg:stop/@offset return (
            if (contains($offset,"%"))
            then util:decimal(100.0 - number(substring-before($offset,"%")), 2)
            else util:decimal(100.0 - number($offset), 2)
          )
        )
        '/>
      </xsl:variable>
      <xsl:variable name="colours" as="xs:string*">
        <xsl:sequence select='reverse($base-gradient/svg:stop/@stop-color)'/>
      </xsl:variable>
      <xsl:variable name="opacities" as="xs:double*">
        <xsl:sequence select='reverse($base-gradient/svg:stop/@stop-opacity)'/>
      </xsl:variable>
      <svg:radialGradient cx="50%" cy="50%" r="50%">
        <xsl:copy-of select="$base-gradient/(@* except (@r|@cx|@cy))"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <svg:stop offset="{$offsets[$i]}%" stop-color="{$colours[$i]}" stop-opacity="{$opacities[$i]}"/>
        </xsl:for-each>
      </svg:radialGradient>  
   </xsl:function>

   <!--
    : this:radial-reverse()
    :-->
   <xsl:function name="this:radial-reverse" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:variable name="offsets" as="xs:double*">
        <xsl:sequence select='
        reverse(
          for $offset in $base-gradient/svg:stop/@offset return (
            if (contains($offset,"%"))
            then util:decimal(100.0 - number(substring-before($offset,"%")), 2)
            else util:decimal(100.0 - number($offset), 2)
          ))
        '/>
      </xsl:variable>
      <xsl:variable name="colours" as="xs:string*">
        <xsl:sequence select='reverse($base-gradient/svg:stop/@stop-color)'/>
      </xsl:variable>
      <xsl:variable name="opacities" as="xs:double*">
        <xsl:sequence select='reverse($base-gradient/svg:stop/@stop-opacity)'/>
      </xsl:variable>
      <svg:radialGradient cx="50%" cy="50%" r="50%">
        <xsl:copy-of select="$base-gradient/(@* except (@r|@cx|@cy))"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <svg:stop offset="{$offsets[$i]}%" stop-color="{$colours[$i]}" stop-opacity="{$opacities[$i]}"/>
        </xsl:for-each>
      </svg:radialGradient>  
   </xsl:function>

   <!--
    : this:radial-random()
    :-->
   <xsl:function name="this:radial-random" as="element(svg:radialGradient)" new-each-time="yes">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:variable name="stops" as="element()*">
        <xsl:sequence select='rand:shuffle($base-gradient/svg:stop)'/>
      </xsl:variable>
      <xsl:variable name="percentages" as="xs:double*">
        <xsl:sequence select='util:linspace(count($stops), 0, 100)!util:decimal(., 2)'/>
      </xsl:variable>
      <svg:radialGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <svg:stop offset="{$percentages[$i]}%">
            <xsl:copy-of select="$stop/(@* except @offset)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-interleave()
    :-->
   <xsl:function name="this:radial-interleave" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="interleave" as="xs:string"/>
      <svg:radialGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:choose>
            <xsl:when test='$i mod 2 = 0'> 
              <xsl:sequence select='$stop'/>
            </xsl:when>
            <xsl:otherwise>
              <svg:stop stop-color="{$interleave}">
                <xsl:copy-of select="$stop/(@* except @stop-color)"/>
              </svg:stop>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:for-each>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-opacity()
    :-->
   <xsl:function name="this:radial-opacity" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="opacity" as="(function(xs:integer) as xs:double)"/>
      <svg:radialGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:variable name="stop-opacity" as="xs:double">
            <xsl:sequence select='util:decimal(util:clamp($opacity($i),0,1),3)'/>
          </xsl:variable>
          <svg:stop stop-opacity="{$stop-opacity}">
            <xsl:copy-of select="$stop/(@* except @stop-opacity)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-colour()
    :-->
   <xsl:function name="this:radial-colour" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="colour" as="(function(xs:integer) as xs:string)"/>
      <svg:radialGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:variable name="stop-colour" as="xs:string">
            <xsl:sequence select='$colour($i)'/>
          </xsl:variable>
          <svg:stop stop-color="{$stop-colour}">
            <xsl:copy-of select="$stop/(@* except @stop-color)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-sample()
    :-->
   <xsl:function name="this:radial-sample" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="num" as="xs:integer"/>
      <xsl:choose>
        <xsl:when test='$num le 1'>
          <!-- An error, really: just return the gradient back -->
          <xsl:sequence select='$base-gradient'/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:variable name="stops" as="element()*">
            <xsl:sequence select='$base-gradient/svg:stop'/>
          </xsl:variable>
          <xsl:variable name="percentages" as="xs:double*">
            <xsl:sequence select='util:linspace($num, 0, 100)!util:decimal(., 2)'/>
          </xsl:variable>
          <xsl:variable name="sep" as="xs:double">
            <xsl:sequence select='count($stops) div $num'/>
          </xsl:variable>
          <xsl:choose>
            <xsl:when test='count($stops) le $num'>
              <xsl:sequence select='$base-gradient'/>
            </xsl:when>
            <xsl:otherwise>
              <svg:radialGradient>
                <xsl:copy-of select="$base-gradient/@*"/>
                <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
                <xsl:for-each select="1 to $num">
                  <xsl:variable name="i" as="xs:integer" select="."/>
                  <xsl:variable name="stop" as="element()">
                    <xsl:sequence select='$stops[1 + ceiling($sep*($i - 1))]'/>
                  </xsl:variable>
                  <svg:stop offset="{$percentages[$i]}%">
                    <xsl:copy-of select="$stop/(@* except @offset)"/>
                  </svg:stop>
                </xsl:for-each>
              </svg:radialGradient>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:otherwise>
      </xsl:choose>
   </xsl:function>

   <!--
    : this:radial-fade()
    :-->
   <xsl:function name="this:radial-fade" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="start-opacity" as="xs:double"/>
      <xsl:param name="end-opacity" as="xs:double"/>
      <xsl:variable name="n" as="xs:integer">
        <xsl:sequence select='count($base-gradient/svg:stop)'/>
      </xsl:variable>
      <xsl:variable name="delta" as="xs:double">
        <xsl:sequence select='($end-opacity - $start-opacity) div ($n - 1)'/>
      </xsl:variable>
      <svg:radialGradient>
        <xsl:copy-of select="$base-gradient/@*"/>
        <xsl:copy-of select="$base-gradient/(* except svg:stop)"/>
        <xsl:variable name="stops" as="element()*">
          <xsl:sequence select='$base-gradient/svg:stop'/>
        </xsl:variable>
        <xsl:for-each select="1 to count($stops)">
          <xsl:variable name="i" select="."/>
          <xsl:variable name="stop" as="element()">
            <xsl:sequence select='$stops[$i]'/>
          </xsl:variable>
          <xsl:variable name="opacity" as="xs:double">
            <xsl:sequence select='util:decimal($start-opacity + $delta*($i - 1), 3)'/>
          </xsl:variable>
          <svg:stop stop-opacity="{$opacity}">
            <xsl:copy-of select="$stop/(@* except @opacity)"/>
          </svg:stop>
        </xsl:for-each>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-units()
    :-->
   <xsl:function name="this:radial-units" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="objectBoundingBox" as="xs:boolean"/>
      <xsl:variable name="val">
        <xsl:sequence select='if ($objectBoundingBox) then "objectBoundingBox" else "userSpaceOnUse"'/>
      </xsl:variable>
      <svg:radialGradient gradientUnits="{$val}">
        <xsl:copy-of select="$base-gradient/(@* except @gradientUnits)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-center()
    :-->
   <xsl:function name="this:radial-center" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="cx-percent" as="xs:integer"/>
      <xsl:param name="cy-percent" as="xs:integer"/>
      <svg:radialGradient cx="{$cx-percent}%" cy="{$cy-percent}%">
        <xsl:copy-of select="$base-gradient/(@* except (@cx|@cy))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-focus()
    :-->
   <xsl:function name="this:radial-focus" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="fx-percent" as="xs:integer"/>
      <xsl:param name="fy-percent" as="xs:integer"/>
      <svg:radialGradient fx="{$fx-percent}%" fy="{$fy-percent}%">
        <xsl:copy-of select="$base-gradient/(@* except (@fx|@fy))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-radius()
    :-->
   <xsl:function name="this:radial-radius"
                 as="element(svg:radialGradient)"
                >
      <xsl:param name="base-gradient"
                 as="element(svg:radialGradient)"/>
      <xsl:param name="r-percent" as="xs:integer"/>
      <svg:radialGradient r="{$r-percent}%">
        <xsl:copy-of select="$base-gradient/(@* except (@r))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-circle()
    :-->
   <xsl:function name="this:radial-circle" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="radial-circle" as="map(xs:string, item()*)"/>
      <xsl:variable name="r" as="xs:double" select="ellipse:radius($radial-circle)"/>
      <xsl:variable name="center" as="map(*)" select="ellipse:center($radial-circle)"/>
      <xsl:variable name="cx" as="xs:integer" select="point:x($center)"/>
      <xsl:variable name="cy" as="xs:integer" select="point:y($center)"/>
      <svg:radialGradient cx="{$cx}%" cy="{$cy}%" r="{$r}%">
        <xsl:copy-of select="$base-gradient/(@* except (@r|@cx|@cy))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-transform()
    :-->
   <xsl:function name="this:radial-transform" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="transform" as="xs:string"/>
      <svg:radialGradient gradientTransform="{$transform}">
        <xsl:copy-of select="$base-gradient/(@* except @gradientTransform)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-spread()
    :-->
   <xsl:function name="this:radial-spread" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="method" as="xs:string"/>
      <svg:radialGradient spreadMethod="{$method}">
        <xsl:copy-of select="$base-gradient/(@* except @spreadMethod)"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-layout()
    :-->
   <xsl:function name="this:radial-layout" as="element(svg:radialGradient)">
      <xsl:param name="base-gradient" as="element(svg:radialGradient)"/>
      <xsl:param name="layout" as="xs:integer*"/>
      <xsl:variable name="r" as="xs:integer" select='($layout[1],50)[1]'/>
      <xsl:variable name="cx" as="xs:integer" select='($layout[2],50)[1]'/>
      <xsl:variable name="cy" as="xs:integer" select='($layout[3],$cx)[1]'/>
      <svg:radialGradient cx="{$cx}%" cy="{$cy}%" r="{$r}%">      
        <xsl:copy-of select="$base-gradient/(@* except (@cx|@cy|@r))"/>
        <xsl:copy-of select="$base-gradient/*"/>
      </svg:radialGradient>
   </xsl:function>

   <!--
    : this:radial-ref()
    :-->
   <xsl:function name="this:radial-ref" as="element(svg:radialGradient)">
      <xsl:param name="name" as="xs:string"/>
      <xsl:param name="base-gradient-name" as="xs:string"/>
      <svg:radialGradient id="{this:safe($name)}" xlink:href="#{$base-gradient-name}"/>
   </xsl:function>

   <!--
    : this:expand-evenly()
    :-->
   <xsl:function name="this:expand-evenly" as="xs:string*">
      <xsl:param name="num-stops" as="xs:integer"/>
      <xsl:param name="base-rgb-str" as="xs:string*"/>
      <xsl:param name="do-spline" as="xs:boolean"/>
      <xsl:sequence select='
      (: To end up w/ 512 stops, we need to add n interpolations to each edge
       : we have ns stops already, so:
       : (ns - 1)*n + ns = 512
       : n = (512 - ns)/(ns - 1)
       : We can only interpolate an integer number of points so that
       : leaves us a residue which we use as a probability of adding another point
       :)
      let $ns := count($base-rgb-str) return
      let $n := ($num-stops - $ns) idiv ($ns - 1) return
      let $residue := 100 * ($ns - 1) * ((($num-stops - $ns) div ($ns - 1)) - $n) return
      let $base-rgb := $base-rgb-str!rgb:rgb(.) return
      let $base-hsl := $base-rgb!cs:rgb-to-hsluv(.) return
      let $expanded-hsl := (
        $base-hsl[1],
        let $edges :=
          if ($do-spline) then (
            spline:spline($base-hsl)=>path:edges()
          ) else (
            edge:to-edges($base-hsl)
          )
        return
        for $edge in $edges return
        let $n-interp := if (rand:flip($residue)) then $n + 1 else $n return
        let $interp := tail(edge:interpolate($n-interp, $edge))
        return $interp
      ) return
      let $expanded-rgb := $expanded-hsl!cs:hsluv-to-rgb(.) return
      let $expanded-rgb-str := $expanded-rgb!rgb:to-string(.)
      return $expanded-rgb-str
    '/>
   </xsl:function>

   <!--
    : this:expand-scaled()
    : Interpolate in a way to preserve relative scaling 
    :-->
   <xsl:function name="this:expand-scaled" as="xs:string*">
      <xsl:param name="num-stops" as="xs:integer"/>
      <xsl:param name="base-rgb-str" as="xs:string*"/>
      <xsl:param name="base-percents" as="xs:double*"/>
      <xsl:param name="do-spline" as="xs:boolean"/>
      <xsl:sequence select='
      let $ns := count($base-rgb-str) return
      let $edge-fractions := (
        for $i in 2 to count($base-percents)
        return ($base-percents[$i] - $base-percents[$i - 1]) div 100.0
        (:
        for $percent at $i in tail($base-percents)
        return ($percent - $base-percents[$i]) div 100.0
        :)
      ) return
      (: To end up w/ 512 stops, we need to add 512 - ns total
       : allocated in accordance with the percentage differences for each
       : edge
       : Use $num-stops - 1 because the percentages end up including the end
       : points so that works better
       : We can only interpolate an integer number of points so that
       : leaves us a residue which we use as a probability of adding another point
       :)
      let $edge-ns := (
        let $total := $num-stops - 1 (: $ns :) return
        for $fraction in $edge-fractions
        return round($total * $fraction) cast as xs:integer
      ) return
      let $residue := 100 * (($num-stops - (: $ns :) 1) - sum($edge-ns)) div ($ns - 1) return
      let $base-rgb := $base-rgb-str!rgb:rgb(.) return
      let $base-hsl := $base-rgb!cs:rgb-to-hsluv(.) return
      let $expanded-hsl := (
        $base-hsl[1],
        let $edges :=
          if ($do-spline) then (
            spline:spline($base-hsl)=>path:edges()
          ) else (
            edge:to-edges($base-hsl)
          )
        return
        for $i in 1 to count($edges) return
        let $n-interp :=
          if (rand:flip($residue)) then $edge-ns[$i] + 1 else $edge-ns[$i]
        return
        let $interp := tail(edge:interpolate($n-interp, $edges[$i]))
        return $interp
      ) return
      let $expanded-rgb := $expanded-hsl!cs:hsluv-to-rgb(.) return
      let $expanded-rgb-str := $expanded-rgb!rgb:to-string(.)
      return $expanded-rgb-str
    '/>
   </xsl:function>

   <!--
    : this:expand-gradient()
    :-->
   <xsl:function name="this:expand-gradient" as="xs:string*">
      <xsl:param name="num-stops" as="xs:integer"/>
      <xsl:param name="base-gradient" as="xs:string"/>
      <xsl:param name="do-even" as="xs:boolean"/>
      <xsl:param name="do-spline" as="xs:boolean"/>
      <xsl:sequence select='
      let $base-rgb-str := this:gradient-colours($base-gradient) return
      let $base-percents :=
        for $offset in this:gradient-stops($base-gradient)/@offset
        return xs:double(substring-before(string($offset),"%"))
      return
        if ($do-even)
        then this:expand-evenly($num-stops, $base-rgb-str, $do-spline)
        else this:expand-scaled($num-stops, $base-rgb-str, $base-percents, $do-spline)
    '/>
   </xsl:function>

</xsl:package>
<!--gradients.xsl-->
