<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Module with functions for testing
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform" 
             xmlns:this="http://mathling.com/art/test"
             name="http://mathling.com/art/test"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"  
             exclude-result-prefixes="this xs map array math"
             version="3.0">

<xsl:expose component="function" names="this:*" visibility="public"/>

<xsl:function name="this:fail" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:sequence select='
  error(QName("http://mathling.com/errors","ml:FAILURE"), $message)
  '/>
</xsl:function>

<xsl:function name="this:assertThat" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="truth" as="xs:boolean"/>
  <xsl:sequence select='
  if ($truth) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"), "Not true "||$message)
  '/>
</xsl:function>

<xsl:function name="this:assertThat" as="empty-sequence()">
  <xsl:param name="truth" as="xs:boolean"/>
  <xsl:sequence select='
  if ($truth) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Not true")
  '/>
</xsl:function>

<xsl:function name="this:assertTwixt" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="value" as="xs:numeric"/>
  <xsl:param name="low" as="xs:numeric"/>
  <xsl:param name="high" as="xs:numeric"/>
  <xsl:sequence select='
  if (($value ge $low) and ($value le $high)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"), "Value "||$value||" out of bounds ["||$low||","||$high||"]: "||$message)
  '/>
</xsl:function>

<xsl:function name="this:assertTwixt" as="empty-sequence()">
  <xsl:param name="value" as="xs:numeric"/>
  <xsl:param name="low" as="xs:numeric"/>
  <xsl:param name="high" as="xs:numeric"/>
  <xsl:sequence select='
  if (($value ge $low) and ($value le $high)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Value "||$value||" out of bounds ["||$low||","||$high||"]")
  '/>
</xsl:function>

<xsl:function name="this:assertClose" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="value" as="xs:numeric"/>
  <xsl:param name="bound" as="xs:numeric"/>
  <xsl:param name="epsilon" as="xs:numeric"/>
  <xsl:sequence select='
  if (($value ge $bound - $epsilon) and ($value le $bound + $epsilon)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Value "||$value||" too far from "||$bound||": "||$message)
  '/>
</xsl:function>

<xsl:function name="this:assertClose" as="empty-sequence()">
  <xsl:param name="value" as="xs:numeric"/>
  <xsl:param name="bound" as="xs:numeric"/>
  <xsl:param name="epsilon" as="xs:numeric"/>
  <xsl:sequence select='
  if (($value ge $bound - $epsilon) and ($value le $bound + $epsilon)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Value "||$value||" too far from "||$bound)
  '/>
</xsl:function>

<xsl:function name="this:quote" as="xs:string">
  <xsl:param name="items" as="item()*"/>
  <xsl:sequence select='
  string-join(
    for $item in $items return
    if ($item instance of document-node()) then (
      "document{", this:quote($item/node()), "}"
    )
    else if ($item instance of element()) then (
      "&lt;"||local-name($item), this:quote($item/@*), ">", this:quote($item/node()), "&lt;/"||local-name($item)||">"
    )
    else if ($item instance of attribute()) then (
      local-name($item)||"="||string($item)
    )
    else if ($item instance of map(*)) then (
      "{", for $key in $item=>map:keys() return $key||":"||this:quote($item=>map:get($key)), "}"
    )
    else if ($item instance of array(*)) then (
      "[", array:for-each($item, this:quote#1), "]"
    )
    else if ($item instance of empty-sequence()) then (
      "()"
    )
    else (
      string($item)
    )
    ," "
  )
  '/>
</xsl:function>

<xsl:function name="this:assertEmpty" as="empty-sequence()">
  <xsl:param name="items" as="item()*"/>
  <xsl:sequence select='
  if (empty($items)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Non-empty ["||this:quote($items)||"]")
  '/>
</xsl:function>

<xsl:function name="this:assertEmpty" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="items" as="item()*"/>
  <xsl:sequence select='
  if (empty($items)) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"),"Non-empty ["||this:quote($items)||"]: "||$message)
  '/>
</xsl:function>

<xsl:function name="this:same" as="xs:boolean">
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:param name="unordered" as="xs:boolean"/>
  <xsl:choose>
    <xsl:when test="$this instance of document-node()">
      <xsl:sequence select='this:same($this/node(), $that, $unordered)'/>
    </xsl:when>
    <xsl:when test="$that instance of document-node()">
      <xsl:sequence select='this:same($this, $that/node(), $unordered)'/>
    </xsl:when>
    <xsl:when test="count($this) ne count($that)">
      <xsl:sequence select='false()'/>
    </xsl:when>
    <xsl:when test="count($this)=1 and $this instance of element() and $that instance of element()">
      <xsl:variable name="thisKids" as="item()*">
        <xsl:choose>
          <xsl:when test="$unordered">
            <xsl:perform-sort select='$this/element()'>
              <xsl:sort select='string(node-name(.))'/>
            </xsl:perform-sort>
            <xsl:sequence select='
            $this/(node() except element())[
              not(. instance of text()) or not(matches(., "^\s*$"))
            ]
            '/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:sequence select='
            $this/node()[not(. instance of text()) or not(matches(., "^\s*$"))]
            '/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:variable>
      <xsl:variable name="thatKids" as="item()*">
        <xsl:choose>
          <xsl:when test="$unordered">
            <xsl:perform-sort select='$this/element()'>
              <xsl:sort select='string(node-name(.))'/>
            </xsl:perform-sort>
            <xsl:sequence select='
            $this/(node() except element())[
              not(. instance of text()) or not(matches(., "^\s*$"))
            ]
            '/>
          </xsl:when>
          <xsl:otherwise>
            <xsl:sequence select='
            $this/node()[not(. instance of text()) or not(matches(., "^\s*$"))]
            '/>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:variable>
      <xsl:variable name="thisAttr" as="attribute()*">
        <xsl:perform-sort select='$this/@*'>
          <xsl:sort select='string(node-name(.))'/>
        </xsl:perform-sort>
      </xsl:variable>
      <xsl:variable name="thatAttr" as="attribute()*">
        <xsl:perform-sort select='$that/@*'>
          <xsl:sort select='string(node-name(.))'/>
        </xsl:perform-sort>
      </xsl:variable>
      <xsl:sequence select='
      count($this/@*)=count($that/@*) and
      count($thisKids)=count($thatKids) and
      (every $b in (
         for $i in 1 to count($thisAttr)
         return this:same($thisAttr[$i], $thatAttr[$i], $unordered) )
       satisfies $b) 
      and
      (every $b in (
         for $i in 1 to count($thisKids)
         return this:same($thisKids[$i], $thatKids[$i], $unordered) )
       satisfies $b)
      '/>
    </xsl:when>
    <xsl:when test='
      $this instance of map(*) and $that instance of map(*) and
      count(map:keys($this))=count(map:keys($that))
      '>
      <xsl:variable name="thisKeys" as="item()*" select='sort(map:keys($this))'/>
      <xsl:variable name="thatKeys" as="item()*" select='sort(map:keys($that))'/>
      <xsl:sequence select='
        every $i in 1 to count($thisKeys) satisfies (
          ($thisKeys[$i] = $thatKeys[$i]) and
          this:same(
            $this($thisKeys[$i]), $that($thatKeys[$i]), $unordered
          )
        )
      '/>
    </xsl:when>
    <xsl:when test="count($this)=1">
      <xsl:try select="$this eq $that">
        <xsl:catch errors="*">
          <xsl:sequence select='false()'/>
        </xsl:catch>
        <xsl:fallback>
          <xsl:sequence select='$this eq $that'/>
        </xsl:fallback>
      </xsl:try>
    </xsl:when>
    <xsl:when test="$unordered">
      <xsl:variable name="sorted-this" as="item()*">
        <xsl:perform-sort select="$this">
          <xsl:sort select="this:quote(.)"/>
        </xsl:perform-sort>
      </xsl:variable>
      <xsl:variable name="sorted-that" as="item()*">
        <xsl:perform-sort select="$that">
          <xsl:sort select="this:quote(.)"/>
        </xsl:perform-sort>
      </xsl:variable>
      <xsl:sequence select='
      every $i in 1 to count($sorted-this) satisfies (
        this:same($sorted-this[$i], $sorted-that[$i], $unordered)
      )
      '/>
    </xsl:when>
    <xsl:otherwise>
      <xsl:sequence select='
      every $i in 1 to count($this) satisfies (
        this:same($this[$i], $that[$i], $unordered)
      )
      '/>
    </xsl:otherwise>
  </xsl:choose>
</xsl:function>

<xsl:function name="this:same" as="xs:boolean">
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:sequence select='
  this:same($this, $that, false())
  '/>
</xsl:function>


<xsl:function name="this:assertEquals" as="empty-sequence()">
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:sequence select='
  this:assertEquals(
    "["||this:quote($this)||"] not equal to ["||this:quote($that)||"]",
    $this,
    $that,
    false())
  '/>
</xsl:function>

<xsl:function name="this:assertUnorderedEquals" as="empty-sequence()">
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:sequence select='
  this:assertEquals(
    "["||this:quote($this)||"] not equal to ["||this:quote($that)||"]",
    $this,
    $that,
    true())
  '/>
</xsl:function>

<xsl:function name="this:assertEquals" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:sequence select='
  this:assertEquals($message, $this, $that, false())
  '/>
</xsl:function>

<xsl:function name="this:assertEquals" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="this" as="item()*"/>
  <xsl:param name="that" as="item()*"/>
  <xsl:param name="unordered" as="xs:boolean"/>
  <xsl:sequence select='
  if (deep-equal($this,$that) or this:same($this,$that,$unordered)) then ()
  else if ($this instance of document-node())
  then this:assertEquals($message, $this/node(), $that, $unordered)
  else if ($that instance of document-node())
  then this:assertEquals($message, $this, $that/node(), $unordered)
  else (
    error(QName("http://mathling.com/errors","ml:FAILURE"), "["||this:quote($this)||"] not equal to ["||this:quote($that)||"]: "||$message)
  )
  '/>
</xsl:function>

<xsl:function name="this:assertFails" as="empty-sequence()">
  <xsl:param name="message" as="xs:string"/>
  <xsl:param name="f" as="function() as item()*"/>
  <xsl:variable name="ok" as="xs:boolean">
    <xsl:try>
      <xsl:sequence select='
      ($f(), false())=>tail()
      '/>
      <xsl:catch errors="*">
        <xsl:sequence select='true()'/>
      </xsl:catch>
      <xsl:fallback>
        <xsl:sequence select='
        error(QName("http://mathling.com/errors","ml:FAILURE"), "Cannot test for failure: "||$message)
      '/>
      </xsl:fallback>
    </xsl:try>
  </xsl:variable>
  <xsl:sequence select='
  if ($ok) then ()
  else error(QName("http://mathling.com/errors","ml:FAILURE"), "Did not get expected error: "||$message)
  '/>
</xsl:function>

<xsl:function name="this:assertFails" as="empty-sequence()">
  <xsl:param name="f" as="function() as item()*"/>
  <xsl:sequence select='
  this:assertFails("Did not get expected error", $f)
  '/>
</xsl:function>

</xsl:package>
<!-- testlib -->
