<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Reach: a bounded extent
 :
 : Copyright© Mary Holstege 2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/reach"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:config="http://mathling.com/core/config"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             xmlns:svg="http://www.w3.org/2000/svg"
             xmlns:art="http://mathling.com/art"
             version="3.0"
             name="http://mathling.com/type/reach"
             exclude-result-prefixes="this xs math map array config errors util">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>

<!--
 :======================================================================
 : Basic constructors
 :======================================================================
 :-->

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "id", "body")'/>
   </xsl:variable>

   <!--
    : this:reach()
    : Make a reach with the given bounds
    :   $id: id 
    :   $body: contents of the reach
    :   $properties: additional properties
    :-->
   <xsl:function name="this:reach" as="map(xs:string, item()*)">
      <xsl:param name="id" as="xs:string"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($properties,
        map {
          "kind": "reach",
          "id": $id,
          "body": $body
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:reach()
    : Make a reach with the given bounds
    :   $id: id 
    :   $body: contents of the reach
    :-->
   <xsl:function name="this:reach" as="map(xs:string, item()*)">
      <xsl:param name="id" as="xs:string"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      map {
        "kind": "reach",
        "id": $id,
        "body": $body
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $reach: the reach
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $reach("kind")
    '/>
   </xsl:function>

   <!--
    : this:body()
    : Get the contents of the reach
    : 
    : $reach: the reach
    :-->
   <xsl:function name="this:body" as="item()*">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $reach("body")
    '/>
   </xsl:function>

   <!--
    : this:id()
    : Get the ID of the reach 
    : 
    : $reach: the reach
    :-->
   <xsl:function name="this:id" as="xs:string">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $reach("id")
    '/>
   </xsl:function>

   <!--
    : this:ref()
    :-->
   <xsl:function name="this:ref" as="xs:string">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      "url(#"||this:id($reach)||")"
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Setters
 :======================================================================
 :-->

   <!--
    : this:body()
    : Set the contents of the reach, returning the new reach
    : 
    : $reach: the reach
    : $body: the contents
    :-->
   <xsl:function name="this:body" as="map(xs:string, item()*)">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:param name="body" as="item()*"/>
      <xsl:sequence select='
      $reach=>map:put("body", $body)
    '/>
   </xsl:function>

   <!--
    : this:id()
    : Set the id of the reach, returning the new reach
    : 
    : $reach: the reach
    : $id: the ID
    :-->
   <xsl:function name="this:id" as="map(xs:string, item()*)">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:param name="id" as="xs:string"/>
      <xsl:sequence select='
      $reach=>map:put("id", $id)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Property manipulation and access
 :======================================================================
 :-->

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map" as="map(xs:string, item()*)">
      <xsl:param name="reach" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      if (this:kind($reach)="reach") then util:exclude($reach,$this:RESERVED)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="reach" as="item()"/>
      <xsl:sequence select='
      if (this:kind($reach)="reach") then ($reach=>map:keys())[not(. = $this:RESERVED)]
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties" as="map(xs:string, item()*)">
      <xsl:param name="reach" as="item()"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      util:merge-into($reach,
        if (this:kind($reach)="reach") then util:exclude($properties,$this:RESERVED)
        else map {}
      )
    '/>
   </xsl:function>

   <xsl:function name="this:draw-reach" as="element()" visibility="private">
     <xsl:param name="id" as="xs:string"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <art:reach id="{$id}">{
       <xsl:sequence select='util:as-attributes($style-properties)'/>
        <xsl:sequence select='$drawing("draw:draw")($content, map{}, $drawing)'/>
      </art:reach>
   </xsl:function>

   <xsl:function name="this:draw-clip" as="element()" visibility="private">
     <xsl:param name="id" as="xs:string"/>
     <xsl:param name="content" as="item()*"/>
     <xsl:param name="style-properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
      <svg:defs>
        <svg:clipPath id="{$id}">
          <xsl:sequence select='$drawing("draw:draw")($content, $style-properties, $drawing)'/>
        </svg:clipPath>
      </svg:defs>
   </xsl:function>

   <xsl:variable name="this:draw-impl" as="function(*)" visibility="private">
     <xsl:sequence select='
      if ($config:DRAWING-METHOD="art") then (
        function($item, $properties, $drawing) {
          let $style-properties := util:merge-into(($properties, this:property-map($item)))
          return (
            this:draw-reach(this:id($item), this:body($item), $style-properties, $drawing)
          )
        }
      ) else (
        function($item, $properties, $drawing) {
          let $style-properties :=
            util:merge-into(($drawing("draw:svg-style")($properties), $drawing("draw:svg-style")(this:property-map($item))))
          return (
            this:draw-clip(this:id($item), this:body($item), $style-properties, $drawing)
          )
        }
      )
      '/>
   </xsl:variable>
   
   <xsl:function name="this:draw" as="item()*">
     <xsl:param name="item" as="map(xs:string,item()*)"/>
     <xsl:param name="properties" as="map(xs:string,item()*)"/>
     <xsl:param name="drawing" as="map(xs:string,function(*)?)"/>
     <xsl:sequence select='$this:draw-impl($item, $properties, $drawing)'/>
   </xsl:function>
</xsl:package>
<!--reach.xsl-->
