<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Space: axis-aligned rectangular extent with possible margin
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/type/space"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/type/space"
             exclude-result-prefixes="this xs math map array util errors">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>

   <xsl:variable name="this:RESERVED" as="xs:string*">
      <xsl:sequence select='("kind", "height", "width", "edge")'/>
   </xsl:variable>

<!--
 :======================================================================
 : Constructors
 :======================================================================
 :-->

   <!--
    : this:space()
    : space()
    : Construct a rectangular space with no margin
    :
    : $width: width
    : $height: height
    :-->
   <xsl:function name="this:space" as="map(xs:string, item()*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "space",
        "width": $width,
        "height": $height,
        "edge": 0
      }
    '/>
   </xsl:function>

   <!--
    : this:space()
    : Construct a rectangular space with a margin
    :
    : $width: width
    : $height: height
    : $edge: margin on the edges
    :-->
   <xsl:function name="this:space" as="map(xs:string, item()*)">
      <xsl:param name="width" as="xs:double"/>
      <xsl:param name="height" as="xs:double"/>
      <xsl:param name="edge" as="xs:double"/>
      <xsl:sequence select='
      map {
        "kind": "space",
        "width": $width,
        "height": $height,
        "edge": $edge
      }
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Accessors
 :======================================================================
 :-->

   <!--
    : this:width()
    : Accessor for the width of the space
    :
    : $space: the space
    :-->
   <xsl:function name="this:width" as="xs:double">
      <xsl:param name="space" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $space("width")
    '/>
   </xsl:function>

   <!--
    : this:height()
    : Accessor for the height of the space
    :
    : $space: the space
    :-->
   <xsl:function name="this:height" as="xs:double">
      <xsl:param name="space" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $space("height")
    '/>
   </xsl:function>

   <!--
    : this:edge()
    : Accessor for the edge of the space
    :
    : $space: the space
    :-->
   <xsl:function name="this:edge" as="xs:double">
      <xsl:param name="space" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $space("edge")
    '/>
   </xsl:function>

   <!--
    : this:kind()
    : Accessor for the kind of data this is; for debugging and polymorphic functions
    :
    : $space: the space
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="space" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      $space("kind")
    '/>
   </xsl:function>

</xsl:package>
<!--space.xsl-->
