<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Module with functions providing some handy geometric operations.
 :
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 :-->
<xsl:package xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
             xmlns:this="http://mathling.com/geometric"
             xmlns:affine="http://mathling.com/geometric/affine"
             xmlns:array="http://www.w3.org/2005/xpath-functions/array"
             xmlns:box="http://mathling.com/geometric/rectangle"
             xmlns:config="http://mathling.com/core/config"
             xmlns:dist="http://mathling.com/type/distribution"
             xmlns:edge="http://mathling.com/geometric/edge"
             xmlns:ellipse="http://mathling.com/geometric/ellipse"
             xmlns:errors="http://mathling.com/core/errors"
             xmlns:f="http://mathling.com/core/callable"
             xmlns:graph="http://mathling.com/geometric/graph"
             xmlns:map="http://www.w3.org/2005/xpath-functions/map"
             xmlns:mask="http://mathling.com/type/mask"
             xmlns:math="http://www.w3.org/2005/xpath-functions/math"
             xmlns:path="http://mathling.com/geometric/path"
             xmlns:point="http://mathling.com/geometric/point"
             xmlns:rand="http://mathling.com/core/random"
             xmlns:reach="http://mathling.com/type/reach"
             xmlns:roots="http://mathling.com/core/roots"
             xmlns:sdf="http://mathling.com/sdf"
             xmlns:slot="http://mathling.com/type/slot"
             xmlns:solid="http://mathling.com/geometric/solid"
             xmlns:space="http://mathling.com/type/space"
             xmlns:transform="http://mathling.com/geometric/transform"
             xmlns:util="http://mathling.com/core/utilities"
             xmlns:v="http://mathling.com/core/vector"
             xmlns:wrapper="http://mathling.com/type/wrapper"
             xmlns:xs="http://www.w3.org/2001/XMLSchema"
             version="3.0"
             name="http://mathling.com/geometric"
             exclude-result-prefixes="this xs math map array errors config util f v roots rand dist space slot wrapper mask affine point box ellipse edge path graph reach solid transform sdf">

   <xsl:expose component="variable" names="this:*" visibility="public"/>
   <xsl:expose component="function" names="this:*" visibility="public"/>

   <xsl:use-package name="http://mathling.com/core/errors" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/config" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/roots" package-version="*"/>
   <xsl:use-package name="http://mathling.com/core/random" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/distribution" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/space" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/slot" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/wrapper" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/mask" package-version="*"/>
   <xsl:use-package name="http://mathling.com/type/reach" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/affine" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/rectangle" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/graph" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/solid" package-version="*"/>
   <xsl:use-package name="http://mathling.com/geometric/transform" package-version="*"/>
   <xsl:use-package name="http://mathling.com/sdf" package-version="*"/>

   <xsl:variable name="this:π" as="xs:double" select='math:pi()'/>

   <xsl:variable name="this:precision" as="xs:integer" select='math:log10(1 div $config:ε) cast as xs:integer'/>

<!--
 :====================================================================== 
 : Generic accessors
 :======================================================================
 :-->

   <!--
    : this:kind()
    :-->
   <xsl:function name="this:kind" as="xs:string">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      ($region("kind"),"")[1]
    '/>
   </xsl:function>

   <!--
    : this:property-map()
    :-->
   <xsl:function name="this:property-map"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then point:property-map($region)
      else if ($kind="box") then box:property-map($region)
      else if ($kind="block") then box:property-map($region)
      else if ($kind="space") then box:property-map($region)
      else if ($kind="edge") then edge:property-map($region)
      else if ($kind="arc") then edge:property-map($region)
      else if ($kind="ellipse-arc") then edge:property-map($region)
      else if ($kind="quad") then edge:property-map($region)
      else if ($kind="cubic") then edge:property-map($region)
      else if ($kind="path") then path:property-map($region)
      else if ($kind="polygon") then path:property-map($region)
      else if ($kind="graph") then graph:property-map($region)
      else if ($kind="ellipse") then ellipse:property-map($region)
      else if ($kind="slot") then slot:property-map($region)
      else if ($kind="wrapper") then wrapper:property-map($region)
      else if ($kind="mask") then mask:property-map($region)
      else if ($kind="reach") then reach:property-map($region)
      else if ($kind="sphere") then solid:property-map($region)
      else if ($kind="ellipsoid") then solid:property-map($region)
      else if ($kind="face") then solid:property-map($region)
      else if ($kind="tetrahedron") then solid:property-map($region)
      else if ($kind="cube") then solid:property-map($region)
      else if ($kind="octahedron") then solid:property-map($region)
      else if ($kind="icosahedron") then solid:property-map($region)
      else if ($kind="dodecahedron") then solid:property-map($region)
      else if ($kind="polyhedron") then solid:property-map($region)
      else map {}
    '/>
   </xsl:function>

   <!--
    : this:properties()
    :-->
   <xsl:function name="this:properties" as="xs:string*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then point:properties($region)
      else if ($kind="box") then box:properties($region)
      else if ($kind="block") then box:properties($region)
      else if ($kind="space") then box:properties($region)
      else if ($kind="edge") then edge:properties($region)
      else if ($kind="arc") then edge:properties($region)
      else if ($kind="ellipse-arc") then edge:properties($region)
      else if ($kind="quad") then edge:properties($region)
      else if ($kind="cubic") then edge:properties($region)
      else if ($kind="path") then path:properties($region)
      else if ($kind="polygon") then path:properties($region)
      else if ($kind="graph") then graph:properties($region)
      else if ($kind="ellipse") then ellipse:properties($region)
      else if ($kind="slot") then slot:properties($region)
      else if ($kind="wrapper") then wrapper:properties($region)
      else if ($kind="mask") then mask:properties($region)
      else if ($kind="reach") then reach:properties($region)
      else if ($kind="sphere") then solid:properties($region)
      else if ($kind="ellipsoid") then solid:properties($region)
      else if ($kind="face") then solid:properties($region)
      else if ($kind="tetrahedron") then solid:properties($region)
      else if ($kind="cube") then solid:properties($region)
      else if ($kind="octahedron") then solid:properties($region)
      else if ($kind="icosahedron") then solid:properties($region)
      else if ($kind="dodecahedron") then solid:properties($region)
      else if ($kind="polyhedron") then solid:properties($region)
      else ()
    '/>
   </xsl:function>

   <!--
    : this:with-properties()
    :-->
   <xsl:function name="this:with-properties"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then point:with-properties($region, $properties)
      else if ($kind="box") then box:with-properties($region, $properties)
      else if ($kind="block") then box:with-properties($region, $properties)
      else if ($kind="space") then box:with-properties($region, $properties)
      else if ($kind="edge") then edge:with-properties($region, $properties)
      else if ($kind="arc") then edge:with-properties($region, $properties)
      else if ($kind="ellipse-arc") then edge:with-properties($region, $properties)
      else if ($kind="quad") then edge:with-properties($region, $properties)
      else if ($kind="cubic") then edge:with-properties($region, $properties)
      else if ($kind="path") then path:with-properties($region, $properties)
      else if ($kind="polygon") then path:with-properties($region, $properties)
      else if ($kind="graph") then graph:with-properties($region, $properties)
      else if ($kind="ellipse") then ellipse:with-properties($region, $properties)
      else if ($kind="slot") then slot:with-properties($region, $properties)
      else if ($kind="wrapper") then wrapper:with-properties($region, $properties)
      else if ($kind="mask") then mask:with-properties($region, $properties)
      else if ($kind="reach") then reach:with-properties($region, $properties)
      else if ($kind="sphere") then solid:with-properties($region, $properties)
      else if ($kind="ellipsoid") then solid:with-properties($region, $properties)
      else if ($kind="face") then solid:with-properties($region, $properties)
      else if ($kind="tetrahedron") then solid:with-properties($region, $properties)
      else if ($kind="cube") then solid:with-properties($region, $properties)
      else if ($kind="octahedron") then solid:with-properties($region, $properties)
      else if ($kind="icosahedron") then solid:with-properties($region, $properties)
      else if ($kind="dodecahedron") then solid:with-properties($region, $properties)
      else if ($kind="polyhedron") then solid:with-properties($region, $properties)
      else errors:error("GEOM-BADREGION", ($region, "with-properties"))
    '/>
   </xsl:function>

   <!--
    : this:with-edge-properties()
    :-->
   <xsl:function name="this:with-edge-properties"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $edges := $region=>this:edges() return
      let $new-edges :=
        for $edge in $edges return $edge=>this:with-properties($properties)
      return
        let $kind := this:kind($region) return
        if ($kind="path") then path:path($new-edges, $region=>this:property-map())
        else if ($kind="polygon") then path:polygon($new-edges, $region=>this:property-map())
        else if ($kind="graph") then graph:graph($region=>graph:vertices(), $new-edges, $region=>this:property-map())
        else if ($kind="edge") then $new-edges
        else if ($kind="arc") then $new-edges
        else if ($kind="ellipse-arc") then $new-edges
        else if ($kind="quad") then $new-edges
        else if ($kind="cubic") then $new-edges
        else if ($kind="slot") then
          $region=>slot:body(
            for $sub in $region=>slot:body()
            return $sub=>this:with-edge-properties($properties)
          )
        else if ($kind="wrapper") then
          $region=>wrapper:body(
            for $sub in $region=>wrapper:body()
            return $sub=>this:with-edge-properties($properties)
          )
        else if ($kind="mask") then
          $region=>mask:body(
            for $sub in $region=>mask:body()
            return $sub=>this:with-edge-properties($properties)
          )
        else if ($kind="reach") then
          $region=>reach:body(
            for $sub in $region=>reach:body()
            return $sub=>this:with-edge-properties($properties)
          )
        else errors:error("GEOM-BADREGION", ($region, "with-edge-properties"))
    '/>
   </xsl:function>

   <!--
    : this:points()
    : Get 2D points from the region
    :-->
   <xsl:function name="this:points" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then $region=>point:as-dimension(2)
      else if ($kind="box") then box:points($region)
      else if ($kind="block") then box:points($region)
      else if ($kind="space") then box:points($region)
      else if ($kind="path") then path:points($region)
      else if ($kind="polygon") then path:points($region)
      else if ($kind="graph") then graph:points($region)
      else if ($kind="edge") then edge:points($region)
      else if ($kind="arc") then edge:points($region)
      else if ($kind="ellipse-arc") then edge:points($region)
      else if ($kind="quad") then edge:points($region)
      else if ($kind="cubic") then edge:points($region)
      else if ($kind="face") then solid:points($region)
      else if ($kind="tetrahedron") then solid:points($region)
      else if ($kind="cube") then solid:points($region)
      else if ($kind="octahedron") then solid:points($region)
      else if ($kind="icosahedron") then solid:points($region)
      else if ($kind="dodecahedron") then solid:points($region)
      else if ($kind="polyhedron") then solid:points($region)
      else if ($kind="slot") then for $sub in $region=>slot:body() return this:points($sub)
      else if ($kind="wrapper") then for $sub in $region=>wrapper:body() return this:points($sub)
      else if ($kind="mask") then for $sub in $region=>mask:body() return this:points($sub)
      else if ($kind="reach") then for $sub in $region=>reach:body() return this:points($sub)
      else errors:error("GEOM-BADREGION", ($region, "points"))
    '/>
   </xsl:function>

   <!--
    : this:vertices()
    : Get all the region's points.
    :-->
   <xsl:function name="this:vertices"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then $region
      else if ($kind="box") then box:vertices($region)
      else if ($kind="block") then box:vertices($region)
      else if ($kind="space") then box:vertices($region)
      else if ($kind="path") then path:vertices($region)
      else if ($kind="polygon") then path:vertices($region)
      else if ($kind="graph") then graph:vertices($region)
      else if ($kind="edge") then edge:vertices($region)
      else if ($kind="arc") then edge:vertices($region)
      else if ($kind="ellipse-arc") then edge:vertices($region)
      else if ($kind="quad") then edge:vertices($region)
      else if ($kind="cubic") then edge:vertices($region)
      else if ($kind="face") then solid:vertices($region)
      else if ($kind="tetrahedron") then solid:vertices($region)
      else if ($kind="cube") then solid:vertices($region)
      else if ($kind="octahedron") then solid:vertices($region)
      else if ($kind="icosahedron") then solid:vertices($region)
      else if ($kind="dodecahedron") then solid:vertices($region)
      else if ($kind="polyhedron") then solid:vertices($region)
      else if ($kind="slot") then for $sub in $region=>slot:body() return this:vertices($sub)
      else if ($kind="wrapper") then for $sub in $region=>wrapper:body() return this:vertices($sub)
      else if ($kind="mask") then for $sub in $region=>mask:body() return this:vertices($sub)
      else if ($kind="reach") then for $sub in $region=>reach:body() return this:vertices($sub)
      else errors:error("GEOM-BADREGION", ($region, "vertices"))
    '/>
   </xsl:function>

   <!--
    : this:vertex()
    :-->
   <xsl:function name="this:vertex" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="i" as="xs:integer"/>
      <xsl:sequence select='
      this:vertices($region)[$i]
    '/>
   </xsl:function>

   <!--
    : this:edges()
    :   Get the edges of a region
    : $region: The region
    :-->
   <xsl:function name="this:edges" as="map(xs:string, item()*)*">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="box") then box:edges($region)
      else if ($kind="block") then box:edges($region)
      else if ($kind="space") then box:edges($region)
      else if ($kind="path") then path:edges($region)
      else if ($kind="polygon") then path:edges($region)
      else if ($kind="graph") then graph:edges($region)
      else if ($kind="arc") then $region
      else if ($kind="ellipse-arc") then $region
      else if ($kind="edge") then $region
      else if ($kind="cubic") then $region
      else if ($kind="quad") then $region
      else if ($kind="face") then solid:edges($region)
      else if ($kind="tetrahedron") then solid:edges($region)
      else if ($kind="cube") then solid:edges($region)
      else if ($kind="octahedron") then solid:edges($region)
      else if ($kind="icosahedron") then solid:edges($region)
      else if ($kind="dodecahedron") then solid:edges($region)
      else if ($kind="polyhedron") then solid:edges($region)
      else if ($kind="slot") then
        for $sub in $region=>slot:body() return this:edges($sub)
      else if ($kind="wrapper") then
        for $sub in $region=>wrapper:body() return this:edges($sub)
      else if ($kind="mask") then
        for $sub in $region=>mask:body() return this:edges($sub)
      else if ($kind="reach") then
        for $sub in $region=>reach:body() return this:edges($sub)
      else errors:error("GEOM-BADREGION", ($region, "edges"))
    '/>
   </xsl:function>

   <!--
    : this:to-edges()
    : Convert a sequence of points into a sequence of edges.
    :
    : $points: the sequence of points
    :-->
   <xsl:function name="this:to-edges"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      edge:to-edges($points)
    '/>
   </xsl:function>

   <!--
    : this:to-edges()
    : Convert a sequence of points into a sequence of edges.
    :
    : $points: the sequence of points
    : $properties: edge properties
    :-->
   <xsl:function name="this:to-edges"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="points" as="item()*"/>
      <xsl:param name="properties" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      edge:to-edges($points, $properties)
    '/>
   </xsl:function>

<!--
 :====================================================================== 
 : Representation conversions
 :======================================================================
 :-->

   <!--
    : this:as-polygon()
    :-->
   <xsl:function name="this:as-polygon"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="box") then path:polygon(this:edges($region))
      else if ($kind="space") then path:polygon(this:edges($region))
      else if ($kind="edge") then path:polygon($region) 
      else if ($kind="arc") then path:polygon($region) 
      else if ($kind="ellipse-arc") then path:polygon($region) 
      else if ($kind="quad") then path:polygon($region)  
      else if ($kind="cubic") then path:polygon($region) 
      else if ($kind="path") then path:as-polygon($region)
      else if ($kind="polygon") then path:as-polygon($region)
      else if ($kind="face") then solid:polygon($region)
      else errors:error("GEOM-BADREGION", ($region, "as-polygon"))
    '/>
   </xsl:function>

   <!--
    : this:as-box()
    :-->
   <xsl:function name="this:as-box" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then
        util:merge-into($region=>point:property-map(), box:box($region, $region))
      else if ($kind="box") then box:as-box($region)
      else if ($kind="block") then box:as-box($region)
      else if ($kind="space") then box:as-box($region)
      else if ($kind="edge") then
        util:merge-into(
          $region=>edge:property-map(),
          box:box($region=>edge:start(), $region=>edge:end())
        )
      else if ($kind="polygon") then
        (: Only for the special case that it is an axis-aligned box
         : This test is a little incomplete: you could have degenerate
         : cases or butterfly edges, in which case this ends up being
         : the bounding box. Those checks aren&#x27;t worth the bother.
         :)
        let $edges := this:edges($region) return
        let $xs := distinct-values($edges!this:vertices(.)!point:px(.)) return
        let $ys := distinct-values($edges!this:vertices(.)!point:py(.)) return
        let $zs := distinct-values($edges!this:vertices(.)!point:pz(.))
        return (
          if ( (count($edges) eq 4) and
               (count($xs) eq 2) and
               (count($ys) eq 2) and
               (count($zs) eq 1)
             )
          then (
            util:merge-into(
              $region=>path:property-map(),
              box:box(min($xs), min($ys), max($xs), max($ys))
            )
          ) else (
            errors:error("GEOM-BADREGION", ($region, "as-box"))
          )
        )
      else errors:error("GEOM-BADREGION", ($region, "as-box"))
    '/>
   </xsl:function>

<!--
 :====================================================================== 
 : Affine transformations
 : The operations are allowed for all the region types, but they really
 : don't do the "right" thing for some of them, particularly boxes/spaces,
 : ellipses, arcs, and ellipsoids. For example, you can rotate a box,
 : but you don't get a rotated box out of the end of it, because boxes
 : here are axis-aligned boxes of space between minimum and maximum x-y
 : bounds. I allow the operation to proceed in these naive ways because
 : it allows for interesting variations in my art, and allows affine
 : transformations that preserve region types. If you want geometrically
 : correct transformations of the overall shapes, make a polygon from
 : the source region and transform that. For boxes you can make a polygon
 : from the edges, for ellipses you can use interpolate-ellipse() to 
 : get a set of points to construct a polygon from; similarly for arcs.
 : You can scale and translate rectangles as is just fine. It is the
 : rotations and shears that will fail. 
 : 
 : 3D solids will general do weird things when you apply 2D transforms to them
 : and ellipsoids are badly supported because I need to add rotation angles
 : to the representation.
 :
 : The basic rule of this API is that you get out the same kind of thing
 : you put in; so the cases where this can't be done you get "wrong"
 : results. Exception: generally you get generic polyhedra if you do anything
 : to a regular solid, because (naturally!) it is no longer regular so 
 : you can't stick with the simpler representation.
 :======================================================================
 :-->

<!--
 : Public-facing API
 : You can use the lower-level one in affine.xqy, e.g. if you want to 
 : chain a bunch of transforms and only want to calculate the matrix
 : once. Otherwise these are more usable.
 :-->

   <!--
    : this:apply-matrix2()
    : Apply a 2D affine matrix to the region
    : 
    : $region: the region to apply the affine transformation to
    : $affine2: the 2D matrix (6 values)
    :-->
   <xsl:function name="this:apply-matrix2"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="affine2" as="xs:double*"/>
      <xsl:sequence select='
      this:mutate($regions,
        function($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:apply-matrix2($pt, $affine2)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:translate()
    : Translate the regions by the given amount.
    : 
    : $regions: regions to translate
    : $x: how much to translate in the x direction
    : $y: how much to translate in the y direction
    :-->
   <xsl:function name="this:translate"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:translate($pt, $x, $y)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:translate()
    : Translate the regions by the given amount.
    : 
    : $regions: regions to translate
    : $x: how much to translate in the x direction
    : $y: how much to translate in the y direction
    : $z: how much to translate in the z direction
    :-->
   <xsl:function name="this:translate"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:param name="z" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:translate($pt, $x, $y, $z)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Scale the regions by the given amount.
    : 
    : $regions: regions to scale
    : $scaling: how much to scale the point by (all dimensions)
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="scaling" as="xs:double"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:scale($region=>slot:body(), $scaling))
      else if ($kind="wrapper") then $region=>wrapper:body(this:scale($region=>wrapper:body(), $scaling))
      else if ($kind="mask") then $region=>mask:body(this:scale($region=>mask:body(), $scaling))
      else if ($kind="reach") then $region=>reach:body(this:scale($region=>reach:body(), $scaling))
      else if ($kind="polygon") then (
        path:polygon(
          this:scale($region=>path:edges(), $scaling),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:scale($region=>path:edges(), $scaling),
          $region=>this:property-map()
        )
      )
      (: Ellipses need special handling: scale radii too :)
      else if ($kind="ellipse") then (
        ellipse:ellipse(
          this:scale($region=>ellipse:center(), $scaling),
          $region=>ellipse:rx() * $scaling,
          $region=>ellipse:ry() * $scaling,
          $region=>ellipse:rotation(),
          $region=>this:property-map()
        )
      )
      (: Arcs need special handling :)
      else if ($kind="arc") then (
        let $points := $region=>edge:arc-ends() return
        let $angles := $region=>edge:arc-angles() return
        let $circle := this:scale($region=>edge:arc-circle(), $scaling)
        return (
          if (exists($points)) then (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                this:scale($points[1], $scaling), this:scale($points[2], $scaling),
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          ) else (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc-by-angle(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                $angles[1],
                $angles[2],
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          )
        )
      )
      (: Spheres need special handling: need to scale radius as well :)
      else if ($kind="sphere") then (
        util:merge-into(
          $region,
          solid:sphere(
            this:scale(solid:center($region), $scaling),
            solid:radius($region) * $scaling
          )
        )
      )
      else if ($kind="ellipsoid") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale(solid:center($region), $scaling),
            solid:rx($region) * $scaling,
            solid:ry($region) * $scaling,
            solid:rz($region) * $scaling
          )
        )
      )
      (: Uniform scaling of regular solids are OK, special handling :)
      else if ($kind="tetrahedron") then (
        util:merge-into(
          solid:property-map($region),
          solid:tetrahedron(
            this:scale(solid:center($region), $scaling),
            solid:scale($region) * $scaling
          )
        )
      )
      else if ($kind="cube") then (
        util:merge-into(
          solid:property-map($region),
          solid:cube(
            this:scale(solid:center($region), $scaling),
            solid:scale($region) * $scaling
          )
        )
      )
      else if ($kind="octahedron") then (
        util:merge-into(
          solid:property-map($region),
          solid:octahedron(
            this:scale(solid:center($region), $scaling),
            solid:scale($region) * $scaling
          )
        )
      )
      else if ($kind="icosahedron") then (
        util:merge-into(
          solid:property-map($region),
          solid:icosahedron(
            this:scale(solid:center($region), $scaling),
            solid:scale($region) * $scaling
          )
        )
      )
      else if ($kind="dodecahedron") then (
        util:merge-into(
          solid:property-map($region),
          solid:dodecahedron(
            this:scale(solid:center($region), $scaling),
            solid:scale($region) * $scaling
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:scale($pt, $scaling)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Scale the regions by the given amount.
    : 
    : $regions: regions to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:scale($region=>slot:body(), $sx, $sy))
      else if ($kind="wrapper") then $region=>wrapper:body(this:scale($region=>wrapper:body(), $sx, $sy))
      else if ($kind="mask") then $region=>mask:body(this:scale($region=>mask:body(), $sx, $sy))
      else if ($kind="reach") then $region=>reach:body(this:scale($region=>reach:body(), $sx, $sy))
      else if ($kind="polygon") then (
        path:polygon(
          this:scale($region=>path:edges(), $sx, $sy),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:scale($region=>path:edges(), $sx, $sy),
          $region=>this:property-map()
        )
      )
      (: Ellipses need special handling: scale radii too :)
      else if ($kind="ellipse") then (
        ellipse:ellipse(
          this:scale($region=>ellipse:center(), $sx, $sy),
          $region=>ellipse:rx() * $sx,
          $region=>ellipse:ry() * $sy,
          $region=>ellipse:rotation(),
          $region=>this:property-map()
        )
      )
      (: Arcs need special handling :)
      else if ($kind="arc") then (
        let $points := $region=>edge:arc-ends() return
        let $angles := $region=>edge:arc-angles() return
        let $circle := this:scale($region=>edge:arc-circle(), $sx, $sy)
        return (
          if (exists($points)) then (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                this:scale($points[1], $sx, $sy), this:scale($points[2], $sx, $sy),
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          ) else (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc-by-angle(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                $angles[1],
                $angles[2],
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          )
        )
      )
      (: Spheres need special handling: need to scale radius as well :)
      else if ($kind="sphere") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale(solid:center($region), $sx, $sy),
            solid:radius($region) * $sx,
            solid:radius($region) * $sy,
            solid:radius($region)
          )
        )
      )
      else if ($kind="ellipsoid") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale(solid:center($region), $sx, $sy),
            solid:rx($region) * $sx,
            solid:ry($region) * $sy,
            solid:rz($region)
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:scale($pt, $sx, $sy)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:scale3D()
    : Scale the regions by the given amount.
    : 
    : $regions: regions to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $sz: how much to scale the point by in the z dimension
    :-->
   <xsl:function name="this:scale3D"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:scale3D($region=>slot:body(), $sx, $sy, $sz))
      else if ($kind="wrapper") then $region=>wrapper:body(this:scale3D($region=>wrapper:body(), $sx, $sy, $sz))
      else if ($kind="mask") then $region=>mask:body(this:scale3D($region=>mask:body(), $sx, $sy, $sz))
      else if ($kind="reach") then $region=>reach:body(this:scale3D($region=>reach:body(), $sx, $sy, $sz))
      else if ($kind="polygon") then (
        path:polygon(
          this:scale3D($region=>path:edges(), $sx, $sy, $sz),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:scale3D($region=>path:edges(), $sx, $sy, $sz),
          $region=>this:property-map()
        )
      )
      (: Ellipses need special handling: scale radii too :)
      else if ($kind="ellipse") then (
        ellipse:ellipse(
          this:scale3D($region=>ellipse:center(), $sx, $sy, $sz),
          $region=>ellipse:rx() * $sx,
          $region=>ellipse:ry() * $sy,
          $region=>ellipse:rotation(),
          $region=>this:property-map()
        )
      )
      (: Arcs need special handling :)
      else if ($kind="arc") then (
        let $points := $region=>edge:arc-ends() return
        let $angles := $region=>edge:arc-angles() return
        let $circle := this:scale3D($region=>edge:arc-circle(), $sx, $sy, $sz)
        return (
          if (exists($points)) then (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                this:scale3D($points[1], $sx, $sy, $sz), this:scale3D($points[2], $sx, $sy, $sz),
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          ) else (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc-by-angle(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                $angles[1],
                $angles[2],
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          )
        )
      )
      (: Spheres need special handling: need to scale radius as well :)
      else if ($kind="sphere") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale3D(solid:center($region), $sx, $sy, $sz),
            solid:radius($region) * $sx,
            solid:radius($region) * $sy,
            solid:radius($region) * $sz
          )
        )
      )
      else if ($kind="ellipsoid") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale3D(solid:center($region), $sx, $sy, $sz),
            solid:rx($region) * $sx,
            solid:ry($region) * $sy,
            solid:rz($region) * $sz
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:scale3D($pt, $sx, $sy, $sz)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:scale()
    : Scale the regions by the given amount, but keep centered on a particular
    : point. Normally scaling will shift points closer to the origin. This
    : allows you do shift then closer to a different center of scaling instead,
    : so you can, for example, scale a circle about its own center, keeping it
    : in place.
    : 
    : $regions: regions to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $center: center of scaling
    :-->
   <xsl:function name="this:scale" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:scale($region=>slot:body(), $sx, $sy, $center))
      else if ($kind="wrapper") then $region=>wrapper:body(this:scale($region=>wrapper:body(), $sx, $sy, $center))
      else if ($kind="mask") then $region=>mask:body(this:scale($region=>mask:body(), $sx, $sy, $center))
      else if ($kind="reach") then $region=>reach:body(this:scale($region=>reach:body(), $sx, $sy, $center))
      else if ($kind="polygon") then (
        path:polygon(
          this:scale($region=>path:edges(), $sx, $sy, $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:scale($region=>path:edges(), $sx, $sy, $center),
          $region=>this:property-map()
        )
      )
      (: Ellipses need special handling: scale radii too :)
      else if ($kind="ellipse") then (
        ellipse:ellipse(
          this:scale($region=>ellipse:center(), $sx, $sy, $center),
          $region=>ellipse:rx() * $sx,
          $region=>ellipse:ry() * $sy,
          $region=>ellipse:rotation(),
          $region=>this:property-map()
        )
      )
      (: Arcs need special handling :)
      else if ($kind="arc") then (
        let $points := $region=>edge:arc-ends() return
        let $angles := $region=>edge:arc-angles() return
        let $circle := this:scale($region=>edge:arc-circle(), $sx, $sy, $center)
        return (
          if (exists($points)) then (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                this:scale($points[1], $sx, $sy, $center), this:scale($points[2], $sx, $sy, $center),
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          ) else (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc-by-angle(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                $angles[1],
                $angles[2],
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          )
        )
      )
      (: Spheres need special handling: need to scale radius as well :)
      else if ($kind="sphere") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale(solid:center($region), $sx, $sy, $center),
            solid:radius($region) * $sx,
            solid:radius($region) * $sy,
            solid:radius($region)
          )
        )
      )
      else if ($kind="ellipsoid") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale(solid:center($region), $sx, $sy, $center),
            solid:rx($region) * $sx,
            solid:ry($region) * $sy,
            solid:rz($region)
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:scale($pt, $sx, $sy, $center)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:scale-in-place()
    : Scale using the center of the region as point of scaling.
    :-->
   <xsl:function name="this:scale-in-place"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="x" as="xs:double"/>
      <xsl:param name="y" as="xs:double"/>
      <xsl:sequence select='
      for $region in $regions return
      let $location := this:region-center($region)
      return 
        this:translate(
          this:scale($region, $x, $y),
          point:px($location) - $x*point:px($location),
          point:py($location) - $y*point:py($location)
        )
    '/>
   </xsl:function>

   <!--
    : this:scale3D()
    : Scale the regions by the given amount, but keep centered on a particular
    : point. Normally scaling will shift points closer to the origin. This
    : allows you do shift then closer to a different center of scaling instead,
    : so you can, for example, scale a circle about its own center, keeping it
    : in place.
    : 
    : $regions: regions to scale
    : $sx: how much to scale the point by in the x dimension
    : $sy: how much to scale the point by in the y dimension
    : $sz: how much to scale the point by in the z dimension
    : $center: center of scaling
    :-->
   <xsl:function name="this:scale3D"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="sx" as="xs:double"/>
      <xsl:param name="sy" as="xs:double"/>
      <xsl:param name="sz" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:scale3D($region=>slot:body(), $sx, $sy, $sz, $center))
      else if ($kind="wrapper") then $region=>wrapper:body(this:scale3D($region=>wrapper:body(), $sx, $sy, $sz, $center))
      else if ($kind="mask") then $region=>mask:body(this:scale3D($region=>mask:body(), $sx, $sy, $sz, $center))
      else if ($kind="reach") then $region=>reach:body(this:scale3D($region=>reach:body(), $sx, $sy, $sz, $center))
      else if ($kind="polygon") then (
        path:polygon(
          this:scale3D($region=>path:edges(), $sx, $sy, $sz, $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:scale3D($region=>path:edges(), $sx, $sy, $sz, $center),
          $region=>this:property-map()
        )
      )
      (: Ellipses need special handling: scale radii too :)
      else if ($kind="ellipse") then (
        ellipse:ellipse(
          this:scale3D($region=>ellipse:center(), $sx, $sy, $sz, $center),
          $region=>ellipse:rx() * $sx,
          $region=>ellipse:ry() * $sy,
          $region=>ellipse:rotation(),
          $region=>this:property-map()
        )
      )
      (: Arcs need special handling :)
      else if ($kind="arc") then (
        let $points := $region=>edge:arc-ends() return
        let $angles := $region=>edge:arc-angles() return
        let $circle := this:scale3D($region=>edge:arc-circle(), $sx, $sy, $sz, $center)
        return (
          if (exists($points)) then (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                this:scale3D($points[1], $sx, $sy, $sz, $center), this:scale3D($points[2], $sx, $sy, $sz, $center),
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          ) else (
            util:merge-into(
              $region=>this:property-map(),
              edge:arc-by-angle(
                $circle=>ellipse:center(),
                $circle=>ellipse:radius(),
                $angles[1],
                $angles[2],
                $region=>edge:arc-flipped(),
                $region=>edge:arc-large()
              )
            )
          )
        )
      )
      (: Spheres need special handling: need to scale radius as well :)
      else if ($kind="sphere") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale3D(solid:center($region), $sx, $sy, $sz, $center),
            solid:radius($region) * $sx,
            solid:radius($region) * $sy,
            solid:radius($region) * $sz
          )
        )
      )
      else if ($kind="ellipsoid") then (
        util:merge-into(
          $region,
          solid:ellipsoid(
            this:scale3D(solid:center($region), $sx, $sy, $sz, $center),
            solid:rx($region) * $sx,
            solid:ry($region) * $sy,
            solid:rz($region) * $sz
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:scale3D($pt, $sx, $sy, $sz, $center)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:scale-to-unit-circle()
    :-->
   <xsl:function name="this:scale-to-unit-circle"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $region in $regions return
      let $bounds := this:bounding-box($region) return
      let $depth := max(this:vertices($region)!point:pz(.))
      return this:scale-to-unit-circle($region, $bounds=>map:put("depth",$depth))
    '/>
   </xsl:function>

   <!--
    : this:scale-to-unit-circle()
    :-->
   <xsl:function name="this:scale-to-unit-circle"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := box:width($canvas) return
      let $height := box:height($canvas) return
      let $depth := ($canvas("depth"),max(($width,$height)))[1]
      return this:scale3D($regions, 1 div $width, 1 div $height, 1 div $depth)
    '/>
   </xsl:function>

   <!--
    : this:scale-to-canvas()
    :-->
   <xsl:function name="this:scale-to-canvas"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $width := box:width($canvas) return
      let $height := box:height($canvas) 
      return this:scale($regions, $width, $height)
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Rotate the regions by the given amount.
    : 
    : $regions: regions to rotate
    : $degrees: how many degrees to rotate the region
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:sequence select='
      (: Note: ellipsoid is wrong: we need to add rotation angles to support this :)
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:rotate($region=>slot:body(), $degrees))
      else if ($kind="wrapper") then $region=>wrapper:body(this:rotate($region=>wrapper:body(), $degrees))
      else if ($kind="mask") then $region=>mask:body(this:rotate($region=>mask:body(), $degrees))
      else if ($kind="reach") then $region=>reach:body(this:rotate($region=>reach:body(), $degrees))
      else if ($kind="polygon") then (
        path:polygon(
          this:rotate($region=>path:edges(), $degrees),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:rotate($region=>path:edges(), $degrees),
          $region=>this:property-map()
        )
      )
      (: arcs need special handling: convert to angle arc first :)
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:rotate($region=>edge:arc-circle(), $degrees) return
        let $angles := $arc=>edge:arc-angles()
        return (
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] - $degrees,          
              $angles[2] - $degrees,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else if ($kind="ellipse") then (
        if (ellipse:is-circle($region)) then (
          ellipse:circle(
            this:rotate($region=>ellipse:center(), $degrees),
            $region=>ellipse:radius(),
            $region=>this:property-map()
          )
        ) else (
          ellipse:ellipse(
            this:rotate($region=>ellipse:center(), $degrees),
            $region=>ellipse:rx(),
            $region=>ellipse:ry(),
            $region=>ellipse:rotation() + $degrees,
            $region=>this:property-map()
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:rotate($pt, $degrees)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Rotate the regions by the given amount.
    : 
    : $regions: regions to rotate
    : $roll-degrees: how many degrees to rotate the region forwards and back (x)
    : $pitch-degrees: how many degrees to rotate the region up and down (y)
    : $yaw-degrees: how many degrees to rotate the region left and right (z)
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:sequence select='
      (: Note: ellipsoid is wrong: we need to add rotation angles to support this :)
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:rotate($region=>slot:body(), $roll-degrees, $pitch-degrees, $yaw-degrees))
      else if ($kind="wrapper") then $region=>wrapper:body(this:rotate($region=>wrapper:body(), $roll-degrees, $pitch-degrees, $yaw-degrees))
      else if ($kind="mask") then $region=>mask:body(this:rotate($region=>mask:body(), $roll-degrees, $pitch-degrees, $yaw-degrees))
      else if ($kind="reach") then $region=>reach:body(this:rotate($region=>reach:body(), $roll-degrees, $pitch-degrees, $yaw-degrees))
      else if ($kind="polygon") then (
        path:polygon(
          this:rotate($region=>path:edges(), $roll-degrees, $pitch-degrees, $yaw-degrees),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:rotate($region=>path:edges(), $roll-degrees, $pitch-degrees, $yaw-degrees),
          $region=>this:property-map()
        )
      )
      (: arcs need special handling: convert to angle arc first :)
      (: but arcs are 2D, so this is just nonsense, really :)
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:rotate($region=>edge:arc-circle(), $roll-degrees, $pitch-degrees, $yaw-degrees) return
        let $angles := $arc=>edge:arc-angles()
        return (
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] - $roll-degrees,          
              $angles[2] - $pitch-degrees,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else if ($kind="ellipse") then (
        (: More nonsense :)
        if (ellipse:is-circle($region)) then (
          ellipse:circle(
            this:rotate($region=>ellipse:center(), $roll-degrees, $pitch-degrees, $yaw-degrees),
            $region=>ellipse:radius(),
            $region=>this:property-map()
          )
        ) else (
          ellipse:ellipse(
            this:rotate($region=>ellipse:center(), $roll-degrees, $pitch-degrees, $yaw-degrees),
            $region=>ellipse:rx(),
            $region=>ellipse:ry(),
            $region=>ellipse:rotation() + $roll-degrees,
            $region=>this:property-map()
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:rotate($pt, $roll-degrees, $pitch-degrees, $yaw-degrees)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Rotate the regions by the given amount around a center of rotation.
    : Normally rotation is relative to the origin.
    : 
    : $regions: regions to rotate
    : $degrees: how many degrees to rotate the region 
    : $center: center of rotation
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: Note: ellipsoid is wrong: we need to add rotation angles to support this :)
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:rotate($region=>slot:body(), $degrees, $center))
      else if ($kind="wrapper") then $region=>wrapper:body(this:rotate($region=>wrapper:body(), $degrees, $center))
      else if ($kind="mask") then $region=>mask:body(this:rotate($region=>mask:body(), $degrees, $center))
      else if ($kind="reach") then $region=>reach:body(this:rotate($region=>reach:body(), $degrees, $center))
      else if ($kind="polygon") then (
        path:polygon(
          this:rotate($region=>path:edges(), $degrees, $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:rotate($region=>path:edges(), $degrees, $center),
          $region=>this:property-map()
        )
      )
      (: arcs need special handling: convert to angle arc first :)
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:rotate($region=>edge:arc-circle(), $degrees, $center) return
        let $angles := $arc=>edge:arc-angles()
        return (
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] - $degrees,          
              $angles[2] - $degrees,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else if ($kind="ellipse") then (
        if (ellipse:is-circle($region)) then (
          ellipse:circle(
            this:rotate($region=>ellipse:center(), $degrees, $center),
            $region=>ellipse:radius(),
            $region=>this:property-map()
          )
        ) else (
          ellipse:ellipse(
            this:rotate($region=>ellipse:center(), $degrees, $center),
            $region=>ellipse:rx(),
            $region=>ellipse:ry(),
            $region=>ellipse:rotation() + $degrees,
            $region=>this:property-map()
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:rotate($pt, $degrees, $center)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:rotate()
    : Rotate the regions by the given amount around a center of rotation.
    : Normally rotation is relative to the origin.
    : 
    : $regions: regions to rotate
    : $roll-degrees: how many degrees to rotate the region forwards and back (x)
    : $pitch-degrees: how many degrees to rotate the region up and down (y)
    : $yaw-degrees: how many degrees to rotate the region left and right (z)
    : $center: center of rotation
    :-->
   <xsl:function name="this:rotate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: Note: ellipsoid is wrong: we need to add rotation angles to support this :)
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:rotate($region=>slot:body(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center))
      else if ($kind="wrapper") then $region=>wrapper:body(this:rotate($region=>wrapper:body(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center))
      else if ($kind="mask") then $region=>mask:body(this:rotate($region=>mask:body(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center))
      else if ($kind="reach") then $region=>reach:body(this:rotate($region=>reach:body(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center))
      else if ($kind="polygon") then (
        path:polygon(
          this:rotate($region=>path:edges(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:rotate($region=>path:edges(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center),
          $region=>this:property-map()
        )
      )
      (: arcs need special handling: convert to angle arc first :)
      (: but arcs are 2D, so this is just nonsense, really :)
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:rotate($region=>edge:arc-circle(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center) return
        let $angles := $arc=>edge:arc-angles()
        return (
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] - $roll-degrees,          
              $angles[2] - $pitch-degrees,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else if ($kind="ellipse") then (
        (: More nonsense :)
        if (ellipse:is-circle($region)) then (
          ellipse:circle(
            this:rotate($region=>ellipse:center(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center),
            $region=>ellipse:radius(),
            $region=>this:property-map()
          )
        ) else (
          ellipse:ellipse(
            this:rotate($region=>ellipse:center(), $roll-degrees, $pitch-degrees, $yaw-degrees, $center),
            $region=>ellipse:rx(),
            $region=>ellipse:ry(),
            $region=>ellipse:rotation() + $roll-degrees,
            $region=>this:property-map()
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:rotate($pt, $roll-degrees, $pitch-degrees, $yaw-degrees, $center)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:shear()
    : Shear the regions by the given amount.
    : You may wish to think of this as skewing angles instead: 
    : math:tan(util:radians($skew-degrees)) for $xy or $yx 
    : But note that tan approaches infinity near 90°/270°
    : 
    : $regions: regions to shear.
    : $xy: how much to shear x relative to y
    : $yx: how much to shear y relative to x
    :-->
   <xsl:function name="this:shear" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="xy" as="xs:double"/>
      <xsl:param name="yx" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:shear($pt, $xy, $yx)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:shear()
    : Shear the regions by the given amount. 
    : You may wish to think of this as skewing angles instead: 
    : math:tan(math:radians($skew-degrees)) for $xy or $yx 
    : But note that tan approaches infinity near 90°/270°
    : 
    : $regions: regions to shear.
    : $xy: how much to shear x relative to y
    : $yx: how much to shear y relative to x
    : $xz: how much to shear x relative to z
    : $zx: how much to shear z relative to x
    : $yz: how much to shear y relative to z
    : $zy: how much to shear z relative to y
    :-->
   <xsl:function name="this:shear" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="xy" as="xs:double"/>
      <xsl:param name="yx" as="xs:double"/>
      <xsl:param name="xz" as="xs:double"/>
      <xsl:param name="zx" as="xs:double"/>
      <xsl:param name="yz" as="xs:double"/>
      <xsl:param name="zy" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:shear($pt, $xy, $yx, $xz, $zx, $yz, $zy)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:reflect()
    : Reflect the regions across the center.
    : 
    : $regions: regions to reflect.
    : $center: center of reflection.
    :-->
   <xsl:function name="this:reflect"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="center" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:reflect($region=>slot:body(), $center))
      else if ($kind="wrapper") then $region=>wrapper:body(this:reflect($region=>wrapper:body(), $center))
      else if ($kind="mask") then $region=>mask:body(this:reflect($region=>mask:body(), $center))
      else if ($kind="reach") then $region=>reach:body(this:reflect($region=>reach:body(), $center))
      (: arcs need special handling: convert to angle arc first :)
      else if ($kind="polygon") then (
        path:polygon(
          this:reflect($region=>path:edges(), $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:reflect($region=>path:edges(), $center),
          $region=>this:property-map()
        )
      )
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:reflect($region=>edge:arc-circle(), $center) return
        let $angles := $arc=>edge:arc-angles()
        return (
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] + 180,
              $angles[2] + 180,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else if ($kind="ellipse") then (
        if (ellipse:is-circle($region)) then (
          ellipse:circle(
            this:reflect($region=>ellipse:center(), $center),
            $region=>ellipse:radius(),
            $region=>this:property-map()
          )
        ) else (
          ellipse:ellipse(
            this:reflect($region=>ellipse:center(), $center),
            $region=>ellipse:rx(),
            $region=>ellipse:ry(),
            $region=>ellipse:rotation() + 180,
            $region=>this:property-map()
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:reflect($pt, $center)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:reflect()
    : Reflect the regions across the line.
    : 
    : $regions: regions to reflect.
    : $start: start of reflection line
    : $end: end of reflection line
    :-->
   <xsl:function name="this:reflect"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="start" as="map(xs:string, item()*)"/>
      <xsl:param name="end" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="slot") then $region=>slot:body(this:reflect($region=>slot:body(), $start, $end))
      else if ($kind="wrapper") then $region=>wrapper:body(this:reflect($region=>wrapper:body(), $start, $end))
      else if ($kind="mask") then $region=>mask:body(this:reflect($region=>mask:body(), $start, $end))
      else if ($kind="reach") then $region=>reach:body(this:reflect($region=>reach:body(), $start, $end))
      else if ($kind="polygon") then (
        path:polygon(
          this:reflect($region=>path:edges(), $start, $end),
          $region=>this:property-map()
        )
      )
      else if ($kind="path") then (
        path:path(
          this:reflect($region=>path:edges(), $start, $end),
          $region=>this:property-map()
        )
      )
      (: arcs need special handling: convert to angle arc first :)
      else if ($kind="arc") then (
        let $arc := edge:as-angle-arc($region) return
        let $circle := this:reflect($region=>edge:arc-circle(), $start, $end) return
        let $angles := $arc=>edge:arc-angles()
        return ( 
          util:merge-into(
            $region=>this:property-map(),
            edge:arc-by-angle(
              $circle=>ellipse:center(),
              $circle=>ellipse:radius(),
              $angles[1] + 180,
              $angles[2] + 180,
              $arc=>edge:arc-flipped(),
              $arc=>edge:arc-large()
            )
          )
        )
      )
      else (
        this:mutate($region,
          function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
            affine:reflect($pt, $start, $end)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:perspective()
    : Perform a perspective transformation on the regions, rotated relative to 
    : camera angles. Generally speaking camera location is at negative or zero
    : Z and the display surface is at some positive Z.
    : Not, strictly speaking, an affine transformation, because of the remapping
    : of the points back to 2D.
    : 
    : $regions: regions to compute perspective on.
    : $camera: location of the camera
    : $roll-degrees: camera angle (x)
    : $pitch-degrees: camera angle (y)
    : $yaw-degrees: camera angle (z)
    : $display: display surface relative to camera
    :-->
   <xsl:function name="this:perspective"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="roll-degrees" as="xs:double"/>
      <xsl:param name="pitch-degrees" as="xs:double"/>
      <xsl:param name="yaw-degrees" as="xs:double"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:mutate($regions,
        function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:perspective($pt, $camera, $roll-degrees, $pitch-degrees, $yaw-degrees, $display)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:perspective()
    : Perform a perspective transformation on the regions. Generally speaking
    : camera location is at negative or zero Z and the display surface is at
    : some positive Z.
    : Not, strictly speaking, an affine transformation, because of the remapping
    : of the points back to 2D.
    : 
    : $regions: regions to compute perspective on.
    : $camera: location of the camera
    : $display: display surface relative to camera
    :-->
   <xsl:function name="this:perspective"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="camera" as="map(xs:string, item()*)"/>
      <xsl:param name="display" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:perspective($regions, $camera, 0, 0, 0, $display)
    '/>
   </xsl:function>

   <!--
    : this:isometric()
    : Perform a isometric projection of the regions.
    : 
    : $regions: regions to compute isometric projection on
    : $positive: tilt up or down?
    :-->
   <xsl:function name="this:isometric"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="positive" as="xs:boolean"/>
      <xsl:sequence select='
      this:mutate($regions,
        function ($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          affine:isometric($pt, $positive)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:apply-transforms()
    : Apply the transformations to the object. This is if we are going to be doing
    : operations on the slot, like intersections etc. In general none of the
    : operations apply transforms in advance. Look for a transform property or if
    : this is a lot, the various slot transformation properties.
    :-->
   <xsl:function name="this:apply-transforms"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $ok := function($attr) as xs:boolean {
        matches($attr, "^[ ]*matrix[ ]*[(][ 1234567890,.-]+[)][ ]*$")=>trace("ok "||$attr)
      }
      return
      for $region in $regions return 
      let $kind := this:kind($region) return
      if ($kind="slot") then (
        slot:slot(
          let $location := slot:location($region) return
          let $rotation := slot:rotation($region) return
          let $center := slot:center($region) return
          let $scale := slot:scale($region) return
          let $skew := slot:skew($region) return
          let $transform := transform:transform() return
          let $transform :=
            if (exists($location))
            then $transform=>transform:translate(point:px($location), point:py($location))
            else $transform
          return
          let $transform :=
            if (exists($scale)) then (
              if (exists($center))
              then (
                $transform=>
                  transform:translate(-point:px($center), -point:py($center))=>
                  transform:scale(point:px($scale), point:py($scale))=>
                  transform:translate(point:px($center), point:py($center))
              ) else $transform=>transform:scale(point:px($scale), point:py($scale))
            ) else $transform
          return
          let $transform :=
            if (exists($rotation)) then (
              if (exists($center))
              then $transform=>transform:rotate($center, $rotation)
              else $transform=>transform:rotate($rotation)
            )
            else $transform
          return
          let $transform :=
            if (exists($skew))
            then $transform=>transform:skew(point:px($skew), point:py($skew))
            else $transform
          return (
            if (exists($region("transform")) and $ok($region("transform"))) then (
              this:apply-transforms($region=>slot:body())=>
                this:apply-matrix2(transform:parse-matrix($region("transform")))=>
                map:remove("transform")
            ) else (
              this:apply-transforms($region=>slot:body())
            )
          ),
          slot:property-map($region)
        )
      )
      else if ($kind="wrapper") then (
        if (exists($region("transform")) and $ok($region("transform"))) then (
          $region=>wrapper:body(
            this:apply-transforms($region=>wrapper:body())
          )=>this:apply-matrix2(transform:parse-matrix($region("transform")))=>
             map:remove("transform")
        ) else (
          $region=>wrapper:body(this:apply-transforms($region=>wrapper:body()))
        )
      )
      else if ($kind="mask") then (
        if (exists($region("transform")) and $ok($region("transform"))) then (
          $region=>mask:body(
            this:apply-transforms($region=>mask:body())=>
              this:apply-matrix2(transform:parse-matrix($region("transform")))=>
              map:remove("transform")
          )
        ) else (
          $region=>mask:body(this:apply-transforms($region=>mask:body()))
        )
      )
      else if ($kind="reach") then (
        if (exists($region("transform")) and $ok($region("transform"))) then (
          $region=>reach:body(
            this:apply-transforms($region=>reach:body())=>
              this:apply-matrix2(transform:parse-matrix($region("transform")))=>
              map:remove("transform")
          )
        ) else (
          $region=>reach:body(this:apply-transforms($region=>reach:body()))
        )
      )
      else (
        if (exists($region("transform")) and $ok($region("transform"))) then (
          $region=>this:apply-matrix2(transform:parse-matrix($region("transform")))=>
             map:remove("transform")
        ) else (
          $region
        )
      )
    '/>
   </xsl:function>

<!--
 :====================================================================== 
 : Distance metrics
 :======================================================================
 :-->

   <!--
    : this:distance()
    :-->
   <xsl:function name="this:distance" as="xs:double">
      <xsl:param name="p1" as="map(xs:string, item()*)"/>
      <xsl:param name="p2" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:distance($p1, $p2)
    '/>
   </xsl:function>

   <!--
    : this:polar-distance()
    :-->
   <xsl:function name="this:polar-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      abs(point:angle($point:ORIGIN2D, $p) - point:angle($point:ORIGIN2D, $q))
    '/>
   </xsl:function>

   <!--
    : this:taxi-distance()
    :-->
   <xsl:function name="this:taxi-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := point:max-dimension(($p, $q))
      return (
        sum(
          util:zip(
            function ($cp as xs:double, $cq as xs:double) as xs:double {
              abs($cp - $cq)
            },
            point:pcoordinates($p, $d),
            point:pcoordinates($q, $d)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:avg-distance()
    :-->
   <xsl:function name="this:avg-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := point:max-dimension(($p, $q))
      return (
        avg(
          util:zip(
            function ($cp as xs:double, $cq as xs:double) as xs:double {
              abs($cp - $cq)
            },
            point:pcoordinates($p, $d),
            point:pcoordinates($q, $d)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:rail-distance()
    :-->
   <xsl:function name="this:rail-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:distance($p, $point:ORIGIN4D) + this:distance($q, $point:ORIGIN4D)
    '/>
   </xsl:function>

   <!--
    : this:rail-distance()
    :-->
   <xsl:function name="this:rail-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:param name="origin" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:distance($p, $origin) + this:distance($q, $origin)
    '/>
   </xsl:function>

   <!--
    : this:chebyshev-distance()
    :-->
   <xsl:function name="this:chebyshev-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $d := point:max-dimension(($p, $q))
      return (
        max(
          util:zip(
            function ($cp as xs:double, $cq as xs:double) as xs:double {
              abs($cp - $cq)
            },
            point:pcoordinates($p, $d),
            point:pcoordinates($q, $d)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:minkowski-distance()
    : $order=p; p=1=>taxi; p=2=>Euclidean; p=>∞=>Chebyshev; p<1 not a metric 
    :unless remove final pow to 1 div $order
    :-->
   <xsl:function name="this:minkowski-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:param name="order" as="xs:double"/>
      <xsl:sequence select='
      let $d := point:max-dimension(($p, $q))
      return (
        math:pow(
          sum(
            util:zip(
              function ($cp as xs:double, $cq as xs:double) as xs:double {
                math:pow(abs($cp - $cq), $order)
              },
              point:pcoordinates($p, $d),
              point:pcoordinates($q, $d)
            )
          ),
          1 div $order
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:cosine-similarity()
    : similarity = (A·B)/(∥A∥∥B∥) = ΣAiBi/(√Ai²√Bi²)
    : as a metric = acos(similarity)/π => [0,1]
    :-->
   <xsl:function name="this:cosine-similarity" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:dot($p, $q) div (point:magnitude($p)*point:magnitude($q))
    '/>
   </xsl:function>

   <!--
    : this:cosine-distance()
    :-->
   <xsl:function name="this:cosine-distance" as="xs:double">
      <xsl:param name="p" as="map(xs:string, item()*)"/>
      <xsl:param name="q" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      math:acos(point:dot($p, $q) div (point:magnitude($p)*point:magnitude($q))) div $this:π
    '/>
   </xsl:function>

   <!--
    : this:shortest-distance()
    : Distance from a point to a region. Warning: for curves and most solids this
    : is using SDF conversion, which is not terribly efficient if you are going
    : to be comparing a lot of points to the region. You'd be better off doing
    : the conversion once and running the SDF function yourself. Note the
    : subtleties on origin and canvas mapping here if you go that route. 
    : Only some of the solids are supported right now.
    : Returns 0 for interior points.
    :-->
   <xsl:function name="this:shortest-distance" as="xs:double" new-each-time="no">
      <xsl:param name="pt" as="map(xs:string, item()*)"/>
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then this:distance($pt, $region)
      else if ($kind="box") then (
        if (box:contains-point($region,$pt)) then 0
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="block") then (
        if (box:contains-point($region,$pt)) then 0
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="space") then (
        if (box:contains-point($region,$pt)) then 0
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="edge") then (
        edge:point-distance($region, $pt)
      )
      else if ($kind="arc") then (
        let $f := sdf:toSDF($region, map {"width":0})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else if ($kind="quad") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else if ($kind="cubic") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else if ($kind="path") then (
        if (every $edge in this:edges($region) satisfies this:kind($edge)="edge") then (
          min(this:edges($region)!this:shortest-distance($pt, .))
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return max((0, $f(point:pcoordinates($pt))))
        )
      )
      else if ($kind="polygon") then (
        if (every $edge in this:edges($region) satisfies this:kind($edge)="edge") then (
          if (path:polygon-contains-point($region,$pt)) then 0
          else min(this:edges($region)!this:shortest-distance($pt, .))
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return max((0, $f(point:pcoordinates($pt))))
        )
      )
      else if ($kind="ellipse") then (
        if (ellipse:is-circle($region)) then (
          max((0, point:distance(ellipse:center($region), $pt) - ellipse:radius($region)))
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return max((0, $f(point:pcoordinates($pt))))
        )
      )
      else if ($kind="sphere") then (
        max((0, point:distance(solid:center($region), $pt) - solid:radius($region)))
      )
      else if ($kind="ellipsoid") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else if ($kind="tetrahedron") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else if ($kind="cube") then (
        if (solid:contains-point($region,$pt)) then 0
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="octahedron") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return max((0, $f(point:pcoordinates($pt))))
      )
      else errors:error("GEOM-BADREGION", ($region, "shortest-distance"))
    '/>
   </xsl:function>

   <!--
    : this:signed-distance()
    : Distance from a point to a region. Warning: for curves and solids this is
    : using SDF conversion, which is not terribly efficient if you are going to be
    : comparing a lot of points to the region. You'd be better off doing the
    : conversion once and running the SDF function yourself. Note the subtleties
    : on origin and canvas mapping here if you go that route. 
    : Only some of the solids are supported right now.
    : Will return negative values for interior points; 0 for those on the boundary.
    :-->
   <xsl:function name="this:signed-distance" as="xs:double" new-each-time="no">
      <xsl:param name="pt" as="map(xs:string, item()*)"/>
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then this:distance($pt, $region)
      else if ($kind="box") then (
        if (box:contains-point($region,$pt))
        then -min(this:edges($region)!this:shortest-distance($pt, .))
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="block") then (
        if (box:contains-point($region,$pt))
        then -min(this:edges($region)!this:shortest-distance($pt, .))
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="space") then (
        if (box:contains-point($region,$pt))
        then -min(this:edges($region)!this:shortest-distance($pt, .))
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="edge") then (
        if (point:same(edge:start($region), edge:end($region))) then (
          this:distance($pt, edge:start($region))
        ) else (
          let $v := edge:start($region) return
          let $w := edge:end($region) return
          let $l := edge:linear-length($region) return
          let $lsquared := $l*$l return
          let $w_sub_v := point:sub($w, $v) return
          let $t := util:clamp(point:dot(point:sub($pt, $v), $w_sub_v) div $lsquared, 0, 1)
          return (
            this:distance($pt,
              point:point(
                point:px($v) + $t * point:px($w_sub_v),
                point:py($v) + $t * point:py($w_sub_v)
              )
            )
          )
        )
      )
      else if ($kind="arc") then (
        let $f := sdf:toSDF($region, map {"width":0})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else if ($kind="quad") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else if ($kind="cubic") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else if ($kind="path") then (
        if (every $edge in this:edges($region) satisfies this:kind($edge)="edge") then (
          min(this:edges($region)!this:shortest-distance($pt, .))
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return $f(point:pcoordinates($pt))
        )
      )
      else if ($kind="polygon") then (
        if (every $edge in this:edges($region) satisfies this:kind($edge)="edge") then (
          if (path:polygon-contains-point($region,$pt))
          then -min(this:edges($region)!this:shortest-distance($pt, .))
          else min(this:edges($region)!this:shortest-distance($pt, .))
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return $f(point:pcoordinates($pt))
        )
      )
      else if ($kind="ellipse") then
        if (ellipse:is-circle($region)) then (
          this:distance(ellipse:center($region), $pt) - ellipse:radius($region)
        ) else (
          let $f := sdf:toSDF($region, map {})=>f:function()
          return $f(point:pcoordinates($pt))
        )
      else if ($kind="sphere") then (
        point:distance(solid:center($region), $pt) - solid:radius($region)
      )
      else if ($kind="ellipsoid") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else if ($kind="tetrahedron") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else if ($kind="cube") then (
        if (solid:contains-point($region,$pt)) 
        then -min(this:edges($region)!this:shortest-distance($pt, .))
        else min(this:edges($region)!this:shortest-distance($pt, .))
      )
      else if ($kind="octahedron") then (
        let $f := sdf:toSDF($region, map {})=>f:function()
        return $f(point:pcoordinates($pt))
      )
      else errors:error("GEOM-BADREGION", ($region, "signed-distance"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Generic operations
 :======================================================================
 :-->

   <!--
    : this:snap()
    :-->
   <xsl:function name="this:snap" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then point:snap($region)
      else if ($kind="box") then box:snap($region)
      else if ($kind="block") then box:snap($region)
      else if ($kind="space") then box:snap($region)
      else if ($kind="edge") then edge:snap($region)
      else if ($kind="arc") then edge:snap($region)
      else if ($kind="ellipse-arc") then edge:snap($region)
      else if ($kind="quad") then edge:snap($region)
      else if ($kind="cubic") then edge:snap($region)
      else if ($kind="path") then path:snap($region)
      else if ($kind="polygon") then path:snap($region)
      else if ($kind="graph") then graph:snap($region)
      else if ($kind="ellipse") then ellipse:snap($region)
      else if ($kind="slot") then $region=>slot:body($region=>slot:body()=>this:snap())
      else if ($kind="wrapper") then $region=>wrapper:body($region=>wrapper:body()=>this:snap())
      else if ($kind="mask") then $region=>mask:body($region=>mask:body()=>this:snap())
      else if ($kind="reach") then $region=>reach:body($region=>reach:body()=>this:snap())
      else if ($kind="sphere") then solid:snap($region)
      else if ($kind="ellipsoid") then solid:snap($region)
      else if ($kind="face") then solid:snap($region)
      else if ($kind="tetrahedron") then solid:snap($region)
      else if ($kind="cube") then solid:snap($region)
      else if ($kind="octahedron") then solid:snap($region)
      else if ($kind="icosahedron") then solid:snap($region)
      else if ($kind="dodecahedron") then solid:snap($region)
      else if ($kind="polyhedron") then solid:snap($region)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:decimal()
    :-->
   <xsl:function name="this:decimal"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="digits" as="xs:integer"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then point:decimal($region, $digits)
      else if ($kind="box") then box:decimal($region, $digits)
      else if ($kind="block") then box:decimal($region, $digits)
      else if ($kind="space") then box:decimal($region, $digits)
      else if ($kind="edge") then edge:decimal($region, $digits)
      else if ($kind="arc") then edge:decimal($region, $digits)
      else if ($kind="ellipse-arc") then edge:decimal($region, $digits)
      else if ($kind="quad") then edge:decimal($region, $digits)
      else if ($kind="cubic") then edge:decimal($region, $digits)
      else if ($kind="path") then path:decimal($region, $digits)
      else if ($kind="polygon") then path:decimal($region, $digits)
      else if ($kind="graph") then graph:decimal($region, $digits)
      else if ($kind="slot") then $region=>slot:body($region=>slot:body()=>this:decimal($digits))
      else if ($kind="wrapper") then $region=>wrapper:body($region=>wrapper:body()=>this:decimal($digits))
      else if ($kind="mask") then $region=>mask:body($region=>mask:body()=>this:decimal($digits))
      else if ($kind="reach") then $region=>reach:body($region=>reach:body()=>this:decimal($digits))
      else if ($kind="ellipse") then ellipse:decimal($region, $digits)
      else if ($kind="sphere") then solid:decimal($region, $digits)
      else if ($kind="ellipsoid") then solid:decimal($region, $digits)
      else if ($kind="face") then solid:decimal($region, $digits)
      else if ($kind="tetrahedron") then solid:decimal($region, $digits)
      else if ($kind="cube") then solid:decimal($region, $digits)
      else if ($kind="octahedron") then solid:decimal($region, $digits)
      else if ($kind="icosahedron") then solid:decimal($region, $digits)
      else if ($kind="dodecahedron") then solid:decimal($region, $digits)
      else if ($kind="polyhedron") then solid:decimal($region, $digits)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:quote()
    :-->
   <xsl:function name="this:quote" as="xs:string">
      <xsl:param name="regions" as="item()*"/>
      <xsl:sequence select='
      string-join(
        for $region in $regions return 
        if ($region instance of map(*)) then (
          let $kind := this:kind($region) return
          if ($kind="point") then point:quote($region)
          else if ($kind="box") then box:quote($region)
          else if ($kind="block") then box:quote($region)
          else if ($kind="space") then box:quote($region)
          else if ($kind="edge") then edge:quote($region)
          else if ($kind="arc") then edge:quote($region)
          else if ($kind="ellipse-arc") then edge:quote($region)
          else if ($kind="quad") then edge:quote($region)
          else if ($kind="cubic") then edge:quote($region)
          else if ($kind="path") then path:quote($region)
          else if ($kind="polygon") then path:quote($region)
          else if ($kind="graph") then graph:quote($region)
          else if ($kind="ellipse") then ellipse:quote($region)
          else if ($kind="sphere") then solid:quote($region)
          else if ($kind="ellipsoid") then solid:quote($region)
          else if ($kind="face") then solid:quote($region)
          else if ($kind="tetrahedron") then solid:quote($region)
          else if ($kind="cube") then solid:quote($region)
          else if ($kind="octahedron") then solid:quote($region)
          else if ($kind="icosahedron") then solid:quote($region)
          else if ($kind="dodecahedron") then solid:quote($region)
          else if ($kind="polyhedron") then solid:quote($region)
          else errors:quote($region)
        )
        else errors:quote($region)
        ,
        " + "
      )
    '/>
   </xsl:function>

   <!--
    : this:same()
    :-->
   <xsl:function name="this:same" as="xs:boolean">
      <xsl:param name="this" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:kind($this)=this:kind($other) and (
        let $kind := this:kind($this) return
        if ($kind="point") then point:same($this, $other)
        else if ($kind="box") then box:same($this, $other)
        else if ($kind="block") then box:same($this, $other)
        else if ($kind="space") then box:same($this, $other)
        else if ($kind="edge") then edge:same($this, $other)
        else if ($kind="arc") then edge:same($this, $other)
        else if ($kind="ellipse-arc") then edge:same($this, $other)
        else if ($kind="quad") then edge:same($this, $other)
        else if ($kind="cubic") then edge:same($this, $other)
        else if ($kind="path") then path:same($this, $other)
        else if ($kind="polygon") then path:same($this, $other)
        else if ($kind="graph") then graph:same($this, $other)
        else if ($kind="ellipse") then ellipse:same($this, $other)
        else if ($kind="sphere") then solid:same($this, $other)
        else if ($kind="ellipsoid") then solid:same($this, $other)
        else if ($kind="face") then solid:same($this, $other)
        else if ($kind="tetrahedron") then solid:same($this, $other)
        else if ($kind="cube") then solid:same($this, $other)
        else if ($kind="octahedron") then solid:same($this, $other)
        else if ($kind="icosahedron") then solid:same($this, $other)
        else if ($kind="dodecahedron") then solid:same($this, $other)
        else if ($kind="polyhedron") then solid:same($this, $other)
        else deep-equal($this,$other)
      )
    '/>
   </xsl:function>

   <!--
    : this:mutate()
    :-->
   <xsl:function name="this:mutate" as="map(xs:string, item()*)*">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="mutation"
                 as="(function(map(xs:string, item()*)) as map(xs:string, item()*))"/>
      <xsl:sequence select='
      for $region in $regions return let $kind := this:kind($region) return
      if ($kind="point") then point:mutate($region, $mutation)
      else if ($kind="box") then box:mutate($region, $mutation)
      else if ($kind="block") then box:mutate($region, $mutation)
      else if ($kind="space") then box:mutate($region, $mutation)
      else if ($kind="edge") then edge:mutate($region, $mutation)
      else if ($kind="arc") then edge:mutate($region, $mutation)
      else if ($kind="ellipse-arc") then edge:mutate($region, $mutation)
      else if ($kind="quad") then edge:mutate($region, $mutation)
      else if ($kind="cubic") then edge:mutate($region, $mutation)
      else if ($kind="path") then path:mutate($region, $mutation)
      else if ($kind="polygon") then path:mutate($region, $mutation)
      else if ($kind="graph") then graph:mutate($region, $mutation)
      else if ($kind="ellipse") then ellipse:mutate($region, $mutation)
      else if ($kind="slot") then $region=>slot:body(this:mutate($region=>slot:body(), $mutation))
      else if ($kind="wrapper") then $region=>wrapper:body(this:mutate($region=>wrapper:body(), $mutation))
      else if ($kind="mask") then $region=>mask:body(this:mutate($region=>mask:body(), $mutation))
      else if ($kind="reach") then $region=>reach:body(this:mutate($region=>reach:body(), $mutation))
      else if ($kind="sphere") then solid:mutate($region, $mutation)
      else if ($kind="ellipsoid") then solid:mutate($region, $mutation)
      else if ($kind="face") then solid:mutate($region, $mutation)
      else if ($kind="tetrahedron") then solid:mutate($region, $mutation)
      else if ($kind="cube") then solid:mutate($region, $mutation)
      else if ($kind="octahedron") then solid:mutate($region, $mutation)
      else if ($kind="icosahedron") then solid:mutate($region, $mutation)
      else if ($kind="dodecahedron") then solid:mutate($region, $mutation)
      else if ($kind="polyhedron") then solid:mutate($region, $mutation)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:project()
    : Simple projection of regions. Does not account for stroke widths.
    : You will almost certainly need to map from canvas space to the unit circle
    : and back.
    :-->
   <xsl:function name="this:project"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="p" as="xs:double"/>
      <xsl:param name="q" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          let $x := point:px($pt) return
          let $y := point:py($pt) return
          let $z := point:pz($pt) return
          let $d := $p * $x + $q * $y + $r * $z + 1
          return point:point($x div $d, $y div $d)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:project()
    : Simple projection of regions. Does not account for stroke widths.
    : Does the mapping from canvas space to the unit circle and back.
    : $canvas should have a "depth" property
    :-->
   <xsl:function name="this:project"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:param name="p" as="xs:double"/>
      <xsl:param name="q" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      this:mutate($regions,
        function($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          let $scaled := this:scale-to-unit-circle($pt, $canvas) return
          let $x := point:px($scaled) return
          let $y := point:py($scaled) return
          let $z := point:pz($scaled) return
          let $d := $p * $x + $q * $y + $r * $z + 1
          return this:scale-to-canvas(point:point($x div $d, $y div $d), $canvas)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:project-strokes()
    : Project regions. Converts simple strokes into closed paths with the
    : stroke width projected as well. To do this certain circles and ellipses
    : need to be interpolated to polygons: tolerance says how finely that
    : interpolation is done. 
    : You will almost certainly need to map from canvas space to the unit circle
    : and back.
    : Also note that this will turn polygons into a set of polygons (one 
    : per edge), so if you wanted it filled you're out of luck. And if you
    : wanted it filled, why are you projecting stroke width in the first place?
    : Just project it and fill it. 
    :-->
   <xsl:function name="this:project-strokes"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="default-stroke-width" as="xs:integer"/>
      <xsl:param name="p" as="xs:double"/>
      <xsl:param name="q" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $hypoteneuse :=
        util:round(math:sqrt(2*$default-stroke-width*$default-stroke-width)) return
      let $projection := 
        function($pt as map(xs:string,item()*)) as map(xs:string,item()*) {
          let $x := point:px($pt) return
          let $y := point:py($pt) return
          let $z := point:pz($pt) return
          let $d := $p * $x + $q * $y + $r * $z + 1
          return point:point($x div $d, $y div $d)
        }
      return
      for $region in $regions return
      let $kind := this:kind($region)
      return (
        if ($kind="point") then (
          point:mutate($region, $projection)
        ) else if ($kind=("edge","quad","cubic")) then (
          let $leading := edge:mutate($region, $projection) return
          let $trailing := edge:mutate(this:translate(edge:reverse($region), $hypoteneuse, $hypoteneuse), $projection)
          return 
            path:polygon(($leading, edge:edge(edge:end($leading),edge:start($trailing)),$trailing))
        ) else if ($kind="arc") then (
          (: n interpolations : length / tolerance :)
          let $n := edge:arc-length($region) idiv $tolerance return
          let $edges := edge:interpolate-arc($n, $region, false()) return
          for $edge in $edges return (
            this:project-strokes(
              $edge, $default-stroke-width, $p, $q, $r, $tolerance
            )
          )
        ) else if ($kind="ellipse") then (
          (: n interpolations : perimeter / tolerance :)
          let $n := ellipse:perimeter($region) idiv $tolerance return
          let $edges := ellipse:interpolate($n, $region) return
          for $edge in $edges return (
            this:project-strokes(
              $edge, $default-stroke-width, $p, $q, $r, $tolerance
            )
          )
        ) else (
          let $edges := this:edges($region) return (
            for $edge in $edges return (
              this:project-strokes(
                $edge, $default-stroke-width, $p, $q, $r, $tolerance
              )
            ),
            for $i in 2 to count($edges) return (
              if (point:same(edge:end($edges[$i - 1]), edge:start($edges[$i]))) then () else (
                this:project-strokes(
                  edge:edge(edge:end($edges[$i - 1]), edge:start($edges[$i])),
                  $default-stroke-width, $p, $q, $r, $tolerance
                )
              )
            )
            (:
            for $edge at $i in tail($edges) return (
              if (point:same(edge:end($edges[$i]), edge:start($edge))) then () else (
                this:project-strokes(
                  edge:edge(edge:end($edges[$i]), edge:start($edge)),
                  $default-stroke-width, $p, $q, $r, $tolerance
                )
              )
            )
            :)
            ,
            if ($kind="polygon") then (
              if (not(point:same(edge:end($edges[last()]), edge:start($edges[1])))) then (
                this:project-strokes(
                  edge:edge(edge:end($edges[last()]), edge:start($edges[1])),
                  $default-stroke-width, $p, $q, $r, $tolerance
                )
              ) else ()
            ) else ()
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:project-strokes()
    :-->
   <xsl:function name="this:project-strokes"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="default-stroke-width" as="xs:integer"/>
      <xsl:param name="p" as="xs:double"/>
      <xsl:param name="q" as="xs:double"/>
      <xsl:param name="r" as="xs:double"/>
      <xsl:sequence select='
      this:project-strokes($regions, $default-stroke-width, $p, $q, $r, 1)
    '/>
   </xsl:function>

   <!--
    : this:area()
    :-->
   <xsl:function name="this:area" as="xs:double">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      sum (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind="box") then box:area($region)
        else if ($kind="block") then solid:surface-area($region)
        else if ($kind="space") then box:area($region)
        else if ($kind="polygon") then path:area($region)
        else if ($kind="ellipse") then ellipse:area($region)
        else if ($kind="slot") then this:area($region=>slot:body())
        else if ($kind="wrapper") then this:area($region=>wrapper:body())
        else if ($kind="mask") then this:area($region=>mask:body())
        else if ($kind="reach") then this:area($region=>reach:body())
        else if ($kind="sphere") then solid:surface-area($region)
        else if ($kind="ellipsoid") then solid:surface-area($region)
        else if ($kind="face") then solid:surface-area($region)
        else if ($kind="tetrahedron") then solid:surface-area($region)
        else if ($kind="cube") then solid:surface-area($region)
        else if ($kind="octahedron") then solid:surface-area($region)
        else if ($kind="icosahedron") then solid:surface-area($region)
        else if ($kind="dodecahedron") then solid:surface-area($region)
        else if ($kind="polyhedron") then solid:surface-area($region)
        else 0
      )
    '/>
   </xsl:function>

   <!--
    : this:length()
    :-->
   <xsl:function name="this:length" as="xs:double">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      sum (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind="box") then sum(this:edges($region)=>edge:length())
        else if ($kind="space") then sum(this:edges($region)=>edge:length())
        else if ($kind="polygon") then path:length($region)
        else if ($kind="path") then path:length($region)
        else if ($kind="ellipse") then ellipse:perimeter($region)
        else if ($kind="arc") then edge:length($region)
        else if ($kind="quad") then edge:length($region)
        else if ($kind="cubic") then edge:length($region)
        else if ($kind="edge") then edge:length($region)
        else if ($kind="slot") then this:length($region=>slot:body())
        else if ($kind="wrapper") then this:length($region=>wrapper:body())
        else if ($kind="mask") then this:length($region=>mask:body())
        else if ($kind="reach") then this:length($region=>reach:body())
        else 0
      )
    '/>
   </xsl:function>

   <!--
    : this:n-faces()
    :-->
   <xsl:function name="this:n-faces" as="xs:integer">
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      sum (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind="slot") then this:n-faces($region=>slot:body())
        else if ($kind="wrapper") then this:n-faces($region=>wrapper:body())
        else if ($kind="mask") then this:n-faces($region=>mask:body())
        else if ($kind="reach") then this:n-faces($region=>reach:body())
        else solid:n-faces($region)
      ) cast as xs:integer
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Angles, destinations, etc.
 :======================================================================
 :-->

   <!--
    : this:path-angle()
    : Compute the angle (azimuth) from one point to the next, in degrees
    : Return 0 if points are the same
    :
    : $last: previous point; use (0,0) if no previous
    : $curr: current point
    :-->
   <xsl:function name="this:path-angle" as="xs:double">
      <xsl:param name="last" as="map(xs:string, item()*)?"/>
      <xsl:param name="curr" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:angle($last, $curr)
    '/>
   </xsl:function>

   <!--
    : this:path-inclination()
    : Compute the inclination angle from one point in the next, in degrees
    : Return 0 if the points are the same
    :
    :
    : $last: previous point; use (0,0,0) if no previous
    : $curr: current point
    :-->
   <xsl:function name="this:path-inclination" as="xs:double">
      <xsl:param name="last" as="map(xs:string, item()*)?"/>
      <xsl:param name="curr" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      point:inclination($last, $curr)
    '/>
   </xsl:function>

   <!--
    : this:path-length()
    : Compute total length of travel along a path. If path includes cubic
    : edges the result is approximated.
    :
    : $path: the path
    :-->
   <xsl:function name="this:path-length" as="xs:double">
      <xsl:param name="path" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      path:length($path)
    '/>
   </xsl:function>

   <!--
    : this:destination()
    : Point at a particular distance and direction.
    : 
    : $point: starting point
    : $degrees: bearing from point
    : $length: how far along bearing to travel
    :-->
   <xsl:function name="this:destination"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="degrees" as="xs:double"/>
      <xsl:param name="length" as="xs:double"/>
      <xsl:sequence select='
      point:destination($point, $degrees, $length)
    '/>
   </xsl:function>

   <!--
    : this:destination()
    : Point (3D) at a particular distance, azimuth, and inclination.
    : 
    : $point: starting point
    : $azimuth_degrees: angle of azimuth from point
    : $inclination_degrees: angle of inclination from point
    : $length: how far along bearing to travel
    :-->
   <xsl:function name="this:destination"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="azimuth_degrees" as="xs:double"/>
      <xsl:param name="inclination_degrees" as="xs:double"/>
      <xsl:param name="length" as="xs:double"/>
      <xsl:sequence select='
      point:destination($point, $azimuth_degrees, $inclination_degrees, $length)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Containment/intersection (incomplete, especially missing solid geometry variants)
 :======================================================================
 :-->

   <!--
    : this:point-intersects()
    : Point intersects some other region kind
    :-->
   <xsl:function name="this:point-intersects" as="xs:boolean">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then point:same($point, $other, $tolerance)
      else if ($kind="edge") then edge:edge-intersects($other, $point, $tolerance)
      else if ($kind="quad") then edge:edge-intersects($other, $point, $tolerance)
      else if ($kind="cubic") then edge:edge-intersects($other, $point, $tolerance)
      else if ($kind="arc") then edge:edge-intersects($other, $point, $tolerance)
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:point-intersects($point, $edge, $tolerance)
      else if ($kind="box") then this:box-intersects($other, $point, $tolerance)
      else if ($kind="block") then this:block-intersects($other, $point, $tolerance)
      else if ($kind="space") then this:box-intersects($other, $point, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $point, $tolerance)
        else this:ellipse-intersects($other, $point, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $point, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "point-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:point-intersects()
    :-->
   <xsl:function name="this:point-intersects" as="xs:boolean">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:point-intersects($point, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:edge-intersects()
    : Edge (straight) intersects some other region kind
    :-->
   <xsl:function name="this:edge-intersects" as="xs:boolean">
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then edge:edge-intersects($edge, $other, $tolerance)
      else if ($kind="edge") then edge:edge-intersects($edge, $other, $tolerance) 
      else if ($kind="quad") then edge:edge-intersects($other, $edge, $tolerance)
      else if ($kind="cubic") then edge:edge-intersects($other, $edge, $tolerance)
      else if ($kind="arc") then edge:edge-intersects($other, $edge, $tolerance)
      else if ($kind="path") then
        some $e in this:edges($other)
        satisfies this:edge-intersects($edge, $e, $tolerance) (: $e may not be straight edge :)
      else if ($kind="box") then this:box-intersects($other, $edge, $tolerance)
      else if ($kind="block") then this:block-intersects($other, $edge, $tolerance)
      else if ($kind="space") then this:box-intersects($other, $edge, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $edge, $tolerance)
        else this:ellipse-intersects($other, $edge, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $edge, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "edge-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:edge-intersects()
    :-->
   <xsl:function name="this:edge-intersects" as="xs:boolean">
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:edge-intersects($edge, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:quad-intersects()
    : Quad edge intersects some other region kind
    :-->
   <xsl:function name="this:quad-intersects" as="xs:boolean">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then edge:edge-intersects($quad, $other, $tolerance)
      else if ($kind="edge") then edge:edge-intersects($quad, $other, $tolerance)
      else if ($kind="quad") then edge:edge-intersects($quad, $other, $tolerance)
      else if ($kind="cubic") then edge:edge-intersects($other, $quad, $tolerance)
      else if ($kind="arc") then edge:edge-intersects($other, $quad, $tolerance)
      else if ($kind="path") then
        some $e in this:edges($other)
        satisfies this:quad-intersects($quad, $e, $tolerance)
      else if ($kind="box") then this:box-intersects($other, $quad, $tolerance)
      else if ($kind="block") then this:block-intersects($other, $quad, $tolerance)
      else if ($kind="space") then this:box-intersects($other, $quad, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $quad, $tolerance)
        else this:ellipse-intersects($other, $quad, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $quad, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "quad-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:quad-intersects()
    :-->
   <xsl:function name="this:quad-intersects" as="xs:boolean">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:quad-intersects($quad, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:cubic-intersects()
    : Cubic edge intersects some other region kind
    :-->
   <xsl:function name="this:cubic-intersects" as="xs:boolean">
      <xsl:param name="cubic" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then edge:edge-intersects($cubic, $other, $tolerance)
      else if ($kind="edge") then edge:edge-intersects($cubic, $other, $tolerance)
      else if ($kind="quad") then edge:edge-intersects($cubic, $other, $tolerance)
      else if ($kind="cubic") then edge:edge-intersects($cubic, $other, $tolerance)
      else if ($kind="arc") then edge:edge-intersects($cubic, $other, $tolerance)
      else if ($kind="path") then
        some $e in this:edges($other)
        satisfies this:cubic-intersects($cubic, $e, $tolerance)
      else if ($kind="box") then this:box-intersects($other, $cubic, $tolerance)
      else if ($kind="block") then this:block-intersects($other, $cubic, $tolerance)
      else if ($kind="space") then this:box-intersects($other, $cubic, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $cubic, $tolerance)
        else this:ellipse-intersects($other, $cubic, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $cubic, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "cubic-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:cubic-intersects()
    :-->
   <xsl:function name="this:cubic-intersects" as="xs:boolean">
      <xsl:param name="cubic" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:cubic-intersects($cubic, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:arc-intersects()
    : Arc edge intersects some other region kind
    :-->
   <xsl:function name="this:arc-intersects" as="xs:boolean">
      <xsl:param name="arc" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      if (point:same(edge:start($arc), edge:end($arc), $tolerance)) then (
        if (edge:arc-large($arc)) then (
          this:region-intersects(edge:arc-circle(edge:normalize-arc($arc)), $other, $tolerance)
        ) else (
          this:region-intersects(edge:start($arc), $other, $tolerance)
        )
      ) else let $kind := this:kind($other) return
      if ($kind="point") then edge:edge-intersects($arc, $other, $tolerance)
      else if ($kind="edge") then edge:edge-intersects($arc, $other, $tolerance)
      else if ($kind="quad") then edge:edge-intersects($arc, $other, $tolerance)
      else if ($kind="cubic") then edge:edge-intersects($arc, $other, $tolerance)
      else if ($kind="arc") then edge:edge-intersects($arc, $other, $tolerance)
      else if ($kind="path") then
        some $e in this:edges($other)
        satisfies this:arc-intersects($arc, $e, $tolerance)
      else if ($kind="box") then this:box-intersects($other, $arc, $tolerance)
      else if ($kind="space") then this:box-intersects($other, $arc, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $arc, $tolerance)
        else this:ellipse-intersects($other, $arc, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $arc, $tolerance)
      else if ($kind="block") then this:block-intersects($other, $arc, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "arc-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:arc-intersects()
    :-->
   <xsl:function name="this:arc-intersects" as="xs:boolean">
      <xsl:param name="arc" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:arc-intersects($arc, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:box-intersects()
    : Box intersects some other region kind
    :-->
   <xsl:function name="this:box-intersects" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $box=>box:contains-point($other, $tolerance)
      else if ($kind="edge") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="quad") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="cubic") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other))
      else if ($kind="arc") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:box-intersects($box, $edge, $tolerance)
      else if ($kind="box") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="block") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="space") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $box, $tolerance)
        else this:ellipse-intersects($other, $box, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $box, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "box-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:box-intersects()
    :-->
   <xsl:function name="this:box-intersects" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:box-intersects($box, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:block-intersects()
    : Block intersects some other region kind
    :-->
   <xsl:function name="this:block-intersects" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $box=>box:contains-point($other, $tolerance)
      else if ($kind="edge") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="quad") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="cubic") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other))
      else if ($kind="arc") then
        $box=>box:contains-point(edge:start($other), $tolerance) or
        $box=>box:contains-point(edge:end($other), $tolerance) or
        (some $edge in this:edges($box)
         satisfies this:edge-intersects($edge, $other, $tolerance))
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:block-intersects($box, $edge, $tolerance)
      else if ($kind="box") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="block") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="space") then box:intersects-box($box, $other, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:circle-intersects($other, $box, $tolerance)
        else this:ellipse-intersects($other, $box, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($other, $box, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "block-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:block-intersects()
    :-->
   <xsl:function name="this:block-intersects" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:block-intersects($box, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:circle-intersects()
    : Circle intersects some other region kind
    :-->
   <xsl:function name="this:circle-intersects" as="xs:boolean">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $circle=>ellipse:contains-point($other, $tolerance)
      else if ($kind="edge") then
        this:shortest-distance(ellipse:center($circle), $other) &lt;= ellipse:radius($circle) + $tolerance
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-intersects",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-intersects",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-intersects",$other))
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:circle-intersects($circle, $edge, $tolerance)
      else if ($kind="box") then
        (some $point in this:vertices($other)
         satisfies $circle=>ellipse:contains-point($point, $tolerance)) or
        this:region-contains($other, $circle, $tolerance) or
        (: circle tangent to edge :)
        (some $edge in this:edges($other)
         satisfies this:circle-intersects($circle, $edge, $tolerance)
        )
      else if ($kind="space") then
        (some $point in this:vertices($other)
         satisfies $circle=>ellipse:contains-point($point, $tolerance)) or
        this:region-contains($other, $circle, $tolerance) or
        (: circle tangent to edge :)
        (some $edge in this:edges($other)
         satisfies this:circle-intersects($circle, $edge, $tolerance)
        )
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle()) then (
          this:distance(ellipse:center($circle), ellipse:center($other)) &lt;=
          ellipse:radius($circle) + ellipse:radius($other) + $tolerance
        ) else (
          this:ellipse-intersects($other, $circle, $tolerance)
        )
      else if ($kind="polygon") then this:polygon-intersects($other, $circle, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "circle-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:circle-intersects()
    :-->
   <xsl:function name="this:circle-intersects" as="xs:boolean">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:circle-intersects($circle, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:ellipse-intersects()
    : Ellipse intersects some other region kind
    :-->
   <xsl:function name="this:ellipse-intersects" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      if (ellipse:is-circle($ellipse)) then this:circle-intersects($ellipse, $other, $tolerance)
      else
      let $kind :=  this:kind($other) return
      if ($kind="point") then $ellipse=>ellipse:contains-point($other, $tolerance)
      else if ($kind="edge") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:ellipse-intersects($ellipse, $edge, $tolerance)
      else if ($kind="box") then
        (some $point in this:vertices($other)
         satisfies $ellipse=>ellipse:contains-point($point, $tolerance)) or
        this:region-contains($other, $ellipse, $tolerance)
      else if ($kind="space") then
        (some $point in this:vertices($other)
         satisfies $ellipse=>ellipse:contains-point($point, $tolerance)) or
        this:region-contains($other, $ellipse, $tolerance)
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle()) then (
          errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
        ) else (
          errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-intersects",$other))
        )
      else if ($kind="polygon") then this:polygon-intersects($other, $ellipse, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "ellipse-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:ellipse-intersects()
    :-->
   <xsl:function name="this:ellipse-intersects" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:ellipse-intersects($ellipse, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:polygon-intersects()
    : Polygon intersects some other region kind
    :-->
   <xsl:function name="this:polygon-intersects" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then path:polygon-contains-point($polygon, $other, $tolerance)
      else if ($kind="edge") then
        path:polygon-contains-point($polygon, edge:start($other), $tolerance) or
        path:polygon-contains-point($polygon, edge:end($other), $tolerance) or 
        (some $edge in this:edges($polygon)
         satisfies this:edge-intersects($other, $edge, $tolerance))
      else if ($kind="quad") then 
        path:polygon-contains-point($polygon, edge:start($other), $tolerance) or
        path:polygon-contains-point($polygon, edge:end($other), $tolerance) or
        (some $edge in this:edges($polygon)
         satisfies this:quad-intersects($other, $edge, $tolerance))
      else if ($kind="cubic") then
        path:polygon-contains-point($polygon, edge:start($other), $tolerance) or
        path:polygon-contains-point($polygon, edge:end($other), $tolerance) or
        (some $edge in this:edges($polygon)
         satisfies this:cubic-intersects($other, $edge, $tolerance))
      else if ($kind="arc") then 
        path:polygon-contains-point($polygon, edge:start($other), $tolerance) or
        path:polygon-contains-point($polygon, edge:end($other), $tolerance) or
        (some $edge in this:edges($polygon)
         satisfies this:arc-intersects($other, $edge, $tolerance))
      else if ($kind="path") then
        some $edge in this:edges($other)
        satisfies this:polygon-intersects($polygon, $edge, $tolerance)
      else if ($kind="box") then
        (some $point in this:vertices($polygon)
         satisfies $other=>box:contains-point($point, $tolerance)) or
        (some $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance)) or
        (some $e1 in this:edges($other) satisfies (
          some $e2 in this:edges($polygon)
          satisfies this:edge-intersects($e1, $e2, $tolerance)
        ))
      else if ($kind="space") then
        (some $point in this:vertices($polygon)
         satisfies $other=>box:contains-point($point, $tolerance)) or
        (some $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance)) or
        (some $e1 in this:edges($other) satisfies (
          some $e2 in this:edges($polygon)
          satisfies this:edge-intersects($e1, $e2, $tolerance)
        ))
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle()) then (
          (
            some $point in this:vertices($polygon)
            satisfies $other=>ellipse:contains-point($point, $tolerance)
          ) or (
            some $edge in this:edges($polygon)
            satisfies this:region-intersects($edge, $other, $tolerance)
          ) or (
            $polygon=>this:polygon-contains($other)
          )
        ) else (
          (
            some $point in this:vertices($polygon)
            satisfies $other=>ellipse:contains-point($point, $tolerance)
          ) or (
            some $edge in this:edges($polygon)
            satisfies this:region-intersects($edge, $other, $tolerance)
          ) or (
            $polygon=>this:polygon-contains($other, $tolerance)
          )
        )
      else if ($kind="polygon") then (
        (some $point in this:vertices($polygon)
         satisfies $other=>path:polygon-contains-point($point, $tolerance)) or
        (some $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance)) or
        (some $e1 in this:edges($other) satisfies (
          some $e2 in this:edges($polygon)
          satisfies edge:edge-intersects($e1, $e2, $tolerance)
        ))
      )
      else errors:error("GEOM-BADREGION", ($other, "polygon-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:polygon-intersects()
    :-->
   <xsl:function name="this:polygon-intersects" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:polygon-intersects($polygon, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:region-intersects()
    :-->
   <xsl:function name="this:region-intersects" as="xs:boolean">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then this:point-intersects($region, $other, $tolerance)
      else if ($kind="edge") then this:edge-intersects($region, $other, $tolerance)
      else if ($kind="quad") then this:quad-intersects($region, $other, $tolerance)
      else if ($kind="cubic") then this:cubic-intersects($region, $other, $tolerance)
      else if ($kind="arc") then this:arc-intersects($region, $other, $tolerance)
      else if ($kind="path") then
        some $edge in this:edges($region)
        satisfies this:region-intersects($edge, $other, $tolerance)
      else if ($kind="box") then this:box-intersects($region, $other, $tolerance)
      else if ($kind="block") then this:block-intersects($region, $other, $tolerance)
      else if ($kind="space") then this:box-intersects($region, $other, $tolerance)
      else if ($kind="ellipse") then
        if ($region=>ellipse:is-circle())
        then this:circle-intersects($region, $other, $tolerance)
        else this:ellipse-intersects($region, $other, $tolerance)
      else if ($kind="polygon") then this:polygon-intersects($region, $other, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "region-intersects"))
    '/>
   </xsl:function>

   <!--
    : this:region-intersects()
    :-->
   <xsl:function name="this:region-intersects" as="xs:boolean">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:region-intersects($region, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:circle-intersection()
    : Points of intersection, if any, between a circle and another region
    :-->
   <xsl:function name="this:circle-intersection"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      if (not(ellipse:is-circle($circle))) then errors:error("GEOM-BADREGION", ($circle, "circle-intersection"))
      else (),
      let $kind := this:kind($other) return
      if ($kind="point") then
        if (util:twixt(point:distance(ellipse:center($circle), $other), ellipse:radius($circle) - $tolerance, ellipse:radius($circle) + $tolerance))
        then $other
        else ()
      else if ($kind="ellipse") then
        if (ellipse:is-circle($other))
        then ellipse:circle-circle-intersection($circle, $other, $tolerance)
        else errors:error("GEOM-NOTIMPLEMENTED", ("circle-intersection",$other))
      else if ($kind="edge") then
        (: https://math.stackexchange.com/questions/228841/how-do-i-calculate-the-intersections-of-a-straight-line-and-a-circle :)
        (: line y=mx + x; m = (y2-y1)/(x2-x1), c=y1 :)
        let $r := ellipse:radius($circle) return
        let $p := point:px(ellipse:center($circle)) return
        let $q := point:py(ellipse:center($circle))  return
        let $x1 := point:px(edge:start($other)) return
        let $y1 := point:py(edge:start($other)) return
        let $x2 := point:px(edge:end($other)) return
        let $y2 := point:py(edge:end($other)) return
        let $pts := 
           if (abs($x1 - $x2) &lt; $config:ε) then ( (: vertical line :)
             let $k := $x1 return
             let $A :=1  return
             let $B := -2*$q return
             let $C := $p*$p + $q*$q - $r*$r - 2*$k*$p + $k*$k 
             return (
               if ($B*$B >= 4*$A*$C) then (
                 for $y in roots:quadratic-real-roots($A, $B, $C)
                 return point:point($k, $y)
               ) else ()
             )
           ) else (
             let $m := ($y2 - $y1) div ($x2 - $x1) return
             let $c := $y1 - $m*$x1 return
             let $A := $m*$m + 1 return
             let $B := 2*($m*$c - $m*$q - $p) return
             let $C := $q*$q - $r*$r + $p*$p - 2*$c*$q + $c*$c
             return (
               if ($B*$B >= 4*$A*$C) then (
                 for $x in roots:quadratic-real-roots($A, $B, $C)
                 return point:point($x, $m*$x + $c)
               ) else ()
             )
           )
        return (
          $pts[edge:on-segment-colinear($other, ., $tolerance)]
        )
      else errors:error("GEOM-NOTIMPLEMENTED", ("circle-intersection",$other))
    '/>
   </xsl:function>

   <!--
    : this:circle-intersection()
    :-->
   <xsl:function name="this:circle-intersection"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:circle-intersection($circle, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:point-contains()
    : Point contains some other region kind
    : Degenerate cases only
    :-->
   <xsl:function name="this:point-contains" as="xs:boolean">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then point:same($point, $other, $tolerance)
      else if ($kind="edge") then
        point:same($point, edge:start($other), $tolerance) and
        point:same(edge:start($other), edge:end($other), $tolerance)
      else if ($kind="quad") then
        edge:length($other) &lt;= $tolerance and
        point:same($point, edge:start($other), $tolerance)
      else if ($kind="cubic") then
        edge:length($other) &lt;= $tolerance and
        point:same($point, edge:start($other), $tolerance)
      else if ($kind="arc") then
        edge:length($other) &lt;= $tolerance and
        point:same($point, edge:start($other), $tolerance)
      else if ($kind="path") then
        point:same($point, path:start($other), $tolerance) and
        this:path-length($other) &lt;= $tolerance
      else if ($kind="box") then
        point:same($point, box:min-point($other), $tolerance) and
        box:width($other) + box:height($other) &lt;= $tolerance
      else if ($kind="block") then
        point:same($point, box:min-point($other), $tolerance) and
        box:width($other) + box:height($other) + box:depth($other) &lt;= $tolerance
      else if ($kind="space") then 
        point:same($point, box:min-point($other), $tolerance) and
        box:width($other) = 0 and box:height($other) &lt;= $tolerance
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then point:same($point, ellipse:center($other), $tolerance) and
             ellipse:radius($other) &lt;= $tolerance
        else point:same($point, ellipse:center($other), $tolerance) and
             ellipse:rx($other) &lt;= $tolerance and ellipse:ry($other) &lt;= $tolerance
      else if ($kind="polygon") then 
        point:same($point, path:start($other), $tolerance) and
        this:path-length($other) &lt;= $tolerance
      else errors:error("GEOM-BADREGION", ($other, "point-contains"))
    '/>
   </xsl:function>

   <!--
    : this:point-contains()
    :-->
   <xsl:function name="this:point-contains" as="xs:boolean">
      <xsl:param name="point" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:point-contains($point, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:is-straight-edge()
    :-->
   <xsl:function name="this:is-straight-edge" as="xs:boolean">
      <xsl:param name="edge" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $kind := this:kind($edge) return
      if ($kind="edge") then true()
      else if ($kind="quad") then
        point:orientation(edge:start($edge), edge:controls($edge), edge:end($edge))="colinear" or
        edge:length($edge) = 0
      else if ($kind="cubic") then
        (point:orientation(edge:start($edge), edge:controls($edge)[1], edge:end($edge))="colinear" and
        point:orientation(edge:start($edge), edge:controls($edge)[2], edge:end($edge))="colinear") or
        edge:length($edge) = 0
      else if ($kind="arc") then
        edge:length($edge) = 0
      else errors:error("GEOM-BADREGION", ($edge, "is-straight-edge"))
    '/>
   </xsl:function>

   <!--
    : this:is-straight()
    :-->
   <xsl:function name="this:is-straight" as="xs:boolean">
      <xsl:param name="edges" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $n := count($edges)
      return (
        this:is-straight-edge(head($edges)) and
        (
          every $i in 2 to $n satisfies (
            this:is-straight-edge($edges[$i]) and
            point:orientation(edge:start($edges[$i - 1]), edge:start($edges[$i]), edge:end($edges[$i]))="colinear"
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:edge-contains()
    : Edge contains some other region kind
    : Mostly degenerate cases only
    :-->
   <xsl:function name="this:edge-contains" as="xs:boolean">
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $edge=>this:edge-intersects($other, $tolerance)
      else if ($kind="edge") then
        $edge=>this:edge-intersects(edge:start($other), $tolerance) and
        $edge=>this:edge-intersects(edge:end($other), $tolerance)
      else if ($kind="quad") then
        (: XYZZY TODO straightness with tolerance :)
        this:is-straight-edge($other) and 
        $edge=>this:edge-intersects(edge:start($other), $tolerance) and
        $edge=>this:edge-intersects(edge:end($other), $tolerance)
      else if ($kind="cubic") then
        this:is-straight-edge($other) and
        $edge=>this:edge-intersects(edge:start($other), $tolerance) and
        $edge=>this:edge-intersects(edge:end($other), $tolerance)
      else if ($kind="arc") then 
        this:is-straight-edge($other) and
        $edge=>this:edge-intersects(edge:start($other), $tolerance) and
        $edge=>this:edge-intersects(edge:end($other), $tolerance)
      else if ($kind="path") then
        (: Path is degenerate point on this edge :)
        (
          this:path-length($other) &lt;= $tolerance and
          $edge=>this:edge-contains(path:start($other), $tolerance)
        )
        or
        (: Path is just one edge and edge contains it :)
        (
          count(this:edges($other))=1 and
          $edge=>this:edge-contains(this:edges($other), $tolerance)
        )
        or
        (: Path is straight and this edge contains the whole :)
        (
          this:is-straight(this:edges($other)) and
          $edge=>this:edge-intersects(path:start($other), $tolerance) and
          $edge=>this:edge-intersects(path:end($other), $tolerance)
        )
      else if ($kind="box") then
        ((box:width($other) &lt;= $tolerance) or (box:height($other) &lt;= $tolerance)) and
        (some $e in this:edges($other) satisfies $edge=>this:edge-contains($e, $tolerance))
      else if ($kind="block") then
        ((box:width($other) &lt;= $tolerance) or (box:height($other) &lt;= $tolerance) or (box:depth($other) &lt; $tolerance)) and
        (some $e in this:edges($other) satisfies $edge=>this:edge-contains($e, $tolerance))
      else if ($kind="space") then 
        ((box:width($other) &lt;= $tolerance) or (box:height($other) &lt;= $tolerance)) and
        (some $e in this:edges($other) satisfies $edge=>this:edge-contains($e, $tolerance))
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then ellipse:radius($other) &lt;= $tolerance and
             $edge=>this:edge-contains(ellipse:center($other), $tolerance)
        else (
          let $c := ellipse:center($other) return
          let $rx := ellipse:rx($other) return
          let $ry := ellipse:ry($other)
          return (
            ($rx &lt;= $tolerance and $edge=>this:edge-contains(edge:edge($c=>this:translate(0, -$ry), $c=>this:translate(0, $ry))), $tolerance) or
            ($ry &lt;= $tolerance and $edge=>this:edge-contains(edge:edge($c=>this:translate(-$rx, 0), $c=>this:translate($rx, 0))), $tolerance)
          )
        )
      else if ($kind="polygon") then
        (: Flat polygons only :)
        (this:is-straight(this:edges($other)) and
         $edge=>this:edge-intersects(path:start($other), $tolerance) and
         $edge=>this:edge-intersects(path:end($other), $tolerance))
      else errors:error("GEOM-BADREGION", ($other, "edge-contains"))
    '/>
   </xsl:function>

   <!--
    : this:edge-contains()
    :-->
   <xsl:function name="this:edge-contains" as="xs:boolean">
      <xsl:param name="edge" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:edge-contains($edge, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:quad-contains()
    : Quad contains some other region kind
    :-->
   <xsl:function name="this:quad-contains" as="xs:boolean">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind :=  this:kind($other) return
      if ($kind="point") then $quad=>this:edge-intersects($other, $tolerance)
      else if ($kind="edge") then
       (: XYZZY TODO: tolerance on straightness :)
        this:is-straight-edge($quad) and
        $quad=>this:edge-intersects(edge:start($other), $tolerance) and
        $quad=>this:edge-intersects(edge:end($other), $tolerance)
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
    
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="box") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="block") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="space") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="ellipse") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else if ($kind="polygon") then errors:error("GEOM-NOTIMPLEMENTED", ("quad-contains",$other))
      else errors:error("GEOM-BADREGION", ($other, "quad-contains"))
    '/>
   </xsl:function>

   <!--
    : this:quad-contains()
    :-->
   <xsl:function name="this:quad-contains" as="xs:boolean">
      <xsl:param name="quad" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:quad-contains($quad, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:cubic-contains()
    : Cubic contains some other region kind
    :-->
   <xsl:function name="this:cubic-contains" as="xs:boolean">
      <xsl:param name="cubic" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $cubic=>this:edge-intersects($other, $tolerance)
      else if ($kind="edge") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other)) 
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="box") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="block") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="space") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="ellipse") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else if ($kind="polygon") then errors:error("GEOM-NOTIMPLEMENTED", ("cubic-contains",$other))
      else errors:error("GEOM-BADREGION", ($other, "cubic-contains"))
    '/>
   </xsl:function>

   <!--
    : this:cubic-contains()
    :-->
   <xsl:function name="this:cubic-contains" as="xs:boolean">
      <xsl:param name="cubic" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:cubic-contains($cubic, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:arc-contains()
    : Arc contains some other region kind
    :-->
   <xsl:function name="this:arc-contains" as="xs:boolean">
      <xsl:param name="arc" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $arc=>this:edge-intersects($other, $tolerance)
      else if ($kind="edge") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="box") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="block") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="space") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="ellipse") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else if ($kind="polygon") then errors:error("GEOM-NOTIMPLEMENTED", ("arc-contains",$other))
      else errors:error("GEOM-BADREGION", ($other, "arc-contains"))
    '/>
   </xsl:function>

   <!--
    : this:arc-contains()
    :-->
   <xsl:function name="this:arc-contains" as="xs:boolean">
      <xsl:param name="arc" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:arc-contains($arc, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:box-contains()
    : Box contains some other region kind
    :-->
   <xsl:function name="this:box-contains" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $box=>box:contains-point($other, $tolerance)
      else if ($kind="edge") then 
        $box=>box:contains-point(edge:start($other), $tolerance) and 
        $box=>box:contains-point(edge:end($other), $tolerance)
      else if ($kind="quad") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else if ($kind="cubic") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else if ($kind="arc") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else if ($kind="path") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else if ($kind="box") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="block") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="space") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="ellipse") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else if ($kind="polygon") then $box=>this:box-contains(this:bounding-box($other), $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "box-contains"))
    '/>
   </xsl:function>

   <!--
    : this:box-contains()
    :-->
   <xsl:function name="this:box-contains" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:box-contains($box, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:block-contains()
    : Block contains some other region kind
    :-->
   <xsl:function name="this:block-contains" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $box=>box:contains-point($other, $tolerance)
      else if ($kind="edge") then 
        $box=>box:contains-point(edge:start($other), $tolerance) and 
        $box=>box:contains-point(edge:end($other), $tolerance)
      else if ($kind="quad") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else if ($kind="cubic") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else if ($kind="arc") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else if ($kind="path") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else if ($kind="box") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="block") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="space") then
        $box=>box:contains-point(box:min-point($other), $tolerance) and 
        $box=>box:contains-point(box:max-point($other), $tolerance)
      else if ($kind="ellipse") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else if ($kind="polygon") then $box=>this:block-contains(this:bounding-box($other), $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "block-contains"))
    '/>
   </xsl:function>

   <!--
    : this:block-contains()
    :-->
   <xsl:function name="this:block-contains" as="xs:boolean">
      <xsl:param name="box" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:block-contains($box, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:circle-contains()
    : Circle contains some other region kind
    :-->
   <xsl:function name="this:circle-contains" as="xs:boolean">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $circle=>ellipse:contains-point($other, $tolerance)
      else if ($kind="edge") then
        $circle=>ellipse:contains-point(edge:start($other), $tolerance) and
        $circle=>ellipse:contains-point(edge:end($other), $tolerance)
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else if ($kind="box") then
        (every $point in this:vertices($other)
         satisfies $circle=>ellipse:contains-point($point, $tolerance))
      else if ($kind="block") then
        (every $point in this:vertices($other)
         satisfies $circle=>ellipse:contains-point($point, $tolerance))
      else if ($kind="space") then
        (every $point in this:vertices($other)
         satisfies $circle=>ellipse:contains-point($point, $tolerance))
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle())
        then this:distance(ellipse:center($circle), ellipse:center($other)) &lt;= ellipse:radius($circle) - ellipse:radius($other) + $tolerance
        else errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else if ($kind="polygon") then
        if (every $edge in this:edges($other) satisfies this:kind($edge)="edge")
        then every $v in this:vertices($other) satisfies $circle=>this:circle-contains($v, $tolerance)
        else errors:error("GEOM-NOTIMPLEMENTED", ("circle-contains",$other))
      else errors:error("GEOM-BADREGION", ($other, "circle-contains"))
    '/>
   </xsl:function>

   <!--
    : this:circle-contains()
    :-->
   <xsl:function name="this:circle-contains" as="xs:boolean">
      <xsl:param name="circle" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:circle-contains($circle, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:ellipse-contains()
    : Ellipse contains some other region kind
    :-->
   <xsl:function name="this:ellipse-contains" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $ellipse=>ellipse:contains-point($other, $tolerance)
      else if ($kind="edge") then
        $ellipse=>ellipse:contains-point(edge:start($other), $tolerance) and
        $ellipse=>ellipse:contains-point(edge:end($other), $tolerance)
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else if ($kind="box") then
        (every $point in this:vertices($other)
         satisfies $ellipse=>ellipse:contains-point($point, $tolerance))
      else if ($kind="block") then
        (every $point in this:vertices($other)
         satisfies $ellipse=>ellipse:contains-point($point, $tolerance))
      else if ($kind="space") then
        (every $point in this:vertices($other)
         satisfies $ellipse=>ellipse:contains-point($point, $tolerance))
      else if ($kind="ellipse") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else if ($kind="polygon") then errors:error("GEOM-NOTIMPLEMENTED", ("ellipse-contains",$other))
      else errors:error("GEOM-BADREGION", ($other, "ellipse-contains"))
    '/>
   </xsl:function>

   <!--
    : this:ellipse-contains()
    :-->
   <xsl:function name="this:ellipse-contains" as="xs:boolean">
      <xsl:param name="ellipse" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:ellipse-contains($ellipse, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:polygon-contains()
    : Polygon contains some other region kind
    :-->
   <xsl:function name="this:polygon-contains" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($other) return
      if ($kind="point") then $polygon=>path:polygon-contains-point($other, $tolerance)
      else if ($kind="edge") then
        (
          every $point in this:vertices($other)
          satisfies $polygon=>path:polygon-contains-point($point, $tolerance)
        ) and (
          (: If this edge starts/ends at the polygon boundary, count it as in :)
          every $edge in this:edges($polygon) satisfies (
            not(edge:edge-intersects-edge($edge, $other, $tolerance)) or
            edge:on-segment($edge, edge:start($other), $tolerance) or
            edge:on-segment($edge, edge:end($other), $tolerance)
          )
        )
      else if ($kind="quad") then errors:error("GEOM-NOTIMPLEMENTED", ("polygon-contains",$other))
      else if ($kind="cubic") then errors:error("GEOM-NOTIMPLEMENTED", ("polygon-contains",$other))
      else if ($kind="arc") then errors:error("GEOM-NOTIMPLEMENTED", ("polygon-contains",$other))
      else if ($kind="path") then errors:error("GEOM-NOTIMPLEMENTED", ("polygon-contains",$other))
      else if ($kind="box") then
        (every $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance))
      else if ($kind="block") then
        (every $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance))
      else if ($kind="space") then
        (every $point in this:vertices($other)
         satisfies $polygon=>path:polygon-contains-point($point, $tolerance))
      else if ($kind="ellipse") then
        if ($other=>ellipse:is-circle()) then (
          $polygon=>path:polygon-contains-point(ellipse:center($other), $tolerance) and
          (every $edge in this:edges($polygon)
           satisfies this:shortest-distance(ellipse:center($other), $edge) >= ellipse:radius($other) + $tolerance
          )
        ) else errors:error("GEOM-NOTIMPLEMENTED", ("polygon-contains",$other))
      else if ($kind="polygon") then
        (
          every $point in this:vertices($other)
          satisfies $polygon=>path:polygon-contains-point($point, $tolerance)
        ) and (
          every $oedge in this:edges($other) satisfies (
            every $edge in this:edges($polygon) satisfies (
              not(edge:edge-intersects-edge($oedge, $edge, $tolerance)) or
              (: If this edge starts/ends at the polygon boundary, count it as in :)
              edge:on-segment($edge, edge:start($oedge), $tolerance) or
              edge:on-segment($edge, edge:end($oedge), $tolerance)  
            )
          )
        )
      else errors:error("GEOM-BADREGION", ($other, "polygon-contains"))
    '/>
   </xsl:function>

   <!--
    : this:polygon-contains()
    :-->
   <xsl:function name="this:polygon-contains" as="xs:boolean">
      <xsl:param name="polygon" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:polygon-contains($polygon, $other, $config:tolerance)
    '/>
   </xsl:function>

   <!--
    : this:region-contains()
    :-->
   <xsl:function name="this:region-contains" as="xs:boolean">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:param name="tolerance" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then this:point-contains($region, $other, $tolerance)
      else if ($kind="edge") then this:edge-contains($region, $other, $tolerance)
      else if ($kind="quad") then this:quad-contains($region, $other, $tolerance)
      else if ($kind="cubic") then this:cubic-contains($region, $other, $tolerance)
      else if ($kind="arc") then this:arc-contains($region, $other, $tolerance)
      else if ($kind="path") then
        (: XYZZY: incomplete for degenerate cases, e.g. edge+edge align to cover edge :)
        some $edge in this:edges($region)
        satisfies this:region-contains($edge, $other, $tolerance)
      else if ($kind="box") then this:box-contains($region, $other, $tolerance)
      else if ($kind="block") then this:block-contains($region, $other, $tolerance)
      else if ($kind="space") then this:box-contains($region, $other, $tolerance)
      else if ($kind="ellipse") then
        if ($region=>ellipse:is-circle())
        then this:circle-contains($region, $other, $tolerance)
        else this:ellipse-contains($region, $other, $tolerance)
      else if ($kind="polygon") then this:polygon-contains($region, $other, $tolerance)
      else errors:error("GEOM-BADREGION", ($other, "region-contains"))
    '/>
   </xsl:function>

   <!--
    : this:region-contains()
    :-->
   <xsl:function name="this:region-contains" as="xs:boolean">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="other" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:region-contains($region, $other, $config:tolerance)
    '/>
   </xsl:function>

<!--
 :====================================================================== 
 : Centers, midpoints
 :======================================================================
 :-->

   <!--
    : this:region-center()
    : Approximate center of region. A rough approximation for polygons with
    : curved edges. 2D
    :-->
   <xsl:function name="this:region-center"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: Sometimes we set the center we know/care about :)
      if ($region=>map:contains("center")) then $region("center")
      else
      let $kind := this:kind($region) return
      if ($kind="point") then $region
      else if ($kind="box") then $region=>box:center()
      else if ($kind="block") then $region=>box:center()
      else if ($kind="space") then $region=>box:center()
      else if ($kind="path") then (
        let $n := count(this:edges($region))
        return
        if ($n = 0) then $point:ORIGIN (: GIGO :)
        else if ($n = 1) then this:edges($region)[1]=>edge:edge-point(0.5)
        else $region=>path:path-point(0.5)
      )
      else if ($kind="polygon") then $region=>path:polygon-center()
      else if ($kind="arc") then $region=>edge:edge-point(0.5)
      else if ($kind="edge") then $region=>edge:edge-point(0.5)
      else if ($kind="quad") then $region=>edge:edge-point(0.5)
      else if ($kind="cubic") then $region=>edge:edge-point(0.5)
      else if ($kind="ellipse") then $region=>ellipse:center()
      else if ($kind="sphere") then $region=>solid:center()
      else if ($kind="ellipsoid") then $region=>solid:center()
      else if ($kind="face") then $region=>solid:center()
      else if ($kind="tetrahedron") then $region=>solid:center()
      else if ($kind="cube") then $region=>solid:center()
      else if ($kind="octahedron") then $region=>solid:center()
      else if ($kind="icosahedron") then $region=>solid:center()
      else if ($kind="dodecahedron") then $region=>solid:center()
      else if ($kind="polyhedron") then $region=>solid:center()
      else errors:error("GEOM-BADREGION", ($region, "region-center"))
    '/>
   </xsl:function>

<!--
 :======================================================================
 : Interpolations
 : In general interpolate() is edge-by-edge
 : If you want to get a set of interpolated points over a path as a
 : whole use path-point() over a set of $t values, with an augmented
 : path (path:with-edge-ts()) for efficiency
 :======================================================================
 :-->

   <!--
    : this:interpolate-using()
    :-->
   <xsl:function name="this:interpolate-using"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="divisions" as="(function(item()) as xs:double*)"/>
      <xsl:sequence select='
      for $region in $regions return
      let $kind := this:kind($region) return
      if ($kind="edge") then edge:interpolate-edge-using($region, $divisions)
      else if ($kind="arc") then edge:interpolate-arc-using($region, $divisions)
      else if ($kind="quad") then edge:interpolate-quadratic-using($region, $divisions)
      else if ($kind="cubic") then edge:interpolate-cubic-using($region, $divisions)
      else if ($kind="path") then this:interpolate-using($region=>this:edges(), $divisions)
      else if ($kind="polygon") then this:interpolate-using($region=>this:edges(), $divisions)
      else if ($kind="graph") then this:interpolate-using($region=>this:edges(), $divisions)
      else if ($kind="box") then this:interpolate-using($region=>this:edges(), $divisions)
      else if ($kind="space") then this:interpolate-using($region=>this:edges(), $divisions)
      else if ($kind="ellipse") then ellipse:interpolate-using($region, $divisions)
      else $region
    '/>
   </xsl:function>

   <!--
    : this:interpolate()
    : Interpolate the edges of the region; return the interpolated points.
    : Note: when interpolating paths or polygons, exclusivity is still edge-wise
    : so with exclusive interpolation you'll lose the last point and with 
    : inclusive interpolation you'll end up with duplicate points because end
    : of one edge is the start of the next edge. Exclusivity doesn't apply to
    : ellipse interpolation. (Zen koan: what is the end of a closed loop?)
    :
    : $n: number of points of interpolation for each edge
    : $regions: the set of regions
    : $exclusive: include end point? default=false
    :-->
   <xsl:function name="this:interpolate"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="exclusive" as="xs:boolean"/>
      <xsl:sequence select='
      this:interpolate-using($regions,
        if ($exclusive) then (
          function($region as map(xs:string,item()*)) {
            util:linspace($n, 0, 1, not(this:kind($region)="ellipse"))
          }
        ) else (
          function($region as map(xs:string,item()*)) {
            util:linspace($n, 0, 1)
          }
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:interpolate()
    :-->
   <xsl:function name="this:interpolate"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:interpolate($n, $regions, false())
    '/>
   </xsl:function>

   <!--
    : this:interpolate-jittered()
    :-->
   <xsl:function name="this:interpolate-jittered"
                 as="map(xs:string, item()*)*"
                 new-each-time="yes">
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:param name="jitter-percent" as="xs:integer"/>
      <xsl:sequence select='
      this:interpolate-using($regions,
        function($region as map(xs:string,item()*)) as xs:double* {
          let $even-divisions := util:linspace($n, 0, 1) return
          let $jitter :=
            if ($jitter-percent > 0)
            then dist:uniform(1.0 - $jitter-percent div 100, 1.0)
            else dist:constant(1.0)
          return
          let $js := rand:randomize($n, $jitter) return
          for $i in 1 to count($even-divisions) return $even-divisions[$i]*$js[$i]
          (: for $t at $i in $even-divisions return $t * $js[$i] :)
        }
      )
    '/>
   </xsl:function>

   <!--
    : this:region-point()
    :-->
   <xsl:function name="this:region-point"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then $region
      else if ($kind="edge") then edge:edge-point($region, $t)
      else if ($kind="arc") then edge:edge-point($region, $t)
      else if ($kind="ellipse-arc") then errors:error("GEOM-NOTIMPLEMENTED", ("region-point", "ellipse-arc"))
      else if ($kind="quad") then edge:edge-point($region, $t)
      else if ($kind="cubic") then edge:edge-point($region, $t)
      else if ($kind="path") then path:path-point($region, $t)
      else if ($kind="polygon") then path:path-point($region, $t)
      else if ($kind="graph") then errors:error("GEOM-NOTIMPLEMENTED", ("region-point", "ellipse-arc"))
      else if ($kind="ellipse") then ellipse:ellipse-point($region, $t)
      else if ($kind="slot") then this:region-point($region=>slot:body(), $t)
      else if ($kind="wrapper") then this:region-point($region=>wrapper:body(), $t)
      else if ($kind="mask") then this:region-point($region=>mask:body(), $t)
      else if ($kind="reach") then this:region-point($region=>reach:body(), $t)
      else errors:error("GEOM-BADREGION", ($region, "region-point"))
    '/>
   </xsl:function>

   <!--
    : this:tangent()
    : Return the tangent vector to the given point on the edge (2D)
    :-->
   <xsl:function name="this:tangent" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="edge") then edge:tangent($region, $t)
      else if ($kind="quad") then edge:tangent($region, $t)
      else if ($kind="cubic") then edge:tangent($region, $t)
      else if ($kind="path") then path:tangent($region, $t)
      else if ($kind="ellipse") then (
        let $c := ellipse:center($region) return
        let $p := $region=>ellipse:ellipse-point($t) return
        let $angle := point:angle($c, $p)
        return (
          edge:edge(
            point:destination($p, $angle - 90, 1),
            point:destination($p, $angle + 90, 1)
          )=>point:normalize()
        )
      )
      else errors:error("GEOM-NOTIMPLEMENTED", ("tangent", this:kind($region)))
    '/>
   </xsl:function>

   <!--
    : this:normal()
    : Return the normal vector to the given point on the edge/path (2D)
    :-->
   <xsl:function name="this:normal" as="map(xs:string, item()*)">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      this:tangent($region, $t)=>point:perpendicular()
    '/>
   </xsl:function>

   <!--
    : this:tangent-angle()
    :-->
   <xsl:function name="this:tangent-angle" as="xs:double">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      if (this:kind($region)="ellipse") then ellipse:tangent-angle($region, $t)
      else point:angle($point:ORIGIN2D, this:tangent($region, $t))
    '/>
   </xsl:function>

   <!--
    : this:normal-angle()
    :-->
   <xsl:function name="this:normal-angle" as="xs:double">
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:param name="t" as="xs:double"/>
      <xsl:sequence select='
      this:tangent-angle($region, $t) + 90
    '/>
   </xsl:function>

   <!--
    : this:random-points()
    : Generate N random points using the given algorithms.
    :
    : $N: How many points to return
    : $last: Previous value (needed for markov, available to key callback)
    : $x The algorithm to use for x coordinates, a map with various parameter keys
    : $y The algorithm to use for y coordinates, a map with various parameter keys
    :-->
   <xsl:function name="this:random-points"
                 as="map(xs:string, item()*)*"
                 new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="last" as="map(xs:string, item()*)?"/>
      <xsl:param name="x" as="map(xs:string, item()*)"/>
      <xsl:param name="y" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      (: XYZZY we&#x27;d really like to be able to apply algorithms to $last as a point :)
      let $xs := rand:randomize($N, if (empty($last)) then () else point:px($last), $x) return
      let $ys := rand:randomize($N, if (empty($last)) then () else point:py($last), $y) return
      for $i in 1 to $N
      return point:point($xs[$i], $ys[$i])
    '/>
   </xsl:function>

   <!--
    : this:random-points()
    : Generate N random points using the given algorithms.
    :
    : $N: How many points to return
    : $x The algorithm to use for x coordinates, a map with various parameter keys
    : $y The algorithm to use for y coordinates, a map with various parameter keys
    :-->
   <xsl:function name="this:random-points"
                 as="map(xs:string, item()*)*"
                 new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="x" as="map(xs:string, item()*)"/>
      <xsl:param name="y" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
       this:random-points($N, (), $x, $y)
    '/>
   </xsl:function>

   <!--
    : this:random-point()
    : Generate a random point using the given algorithms.
    :
    : $x The algorithm to use for x coordinates, a map with various parameter keys
    : $y The algorithm to use for y coordinates, a map with various parameter keys
    :-->
   <xsl:function name="this:random-point"
                 as="map(xs:string, item()*)"
                 new-each-time="yes">
      <xsl:param name="x" as="map(xs:string, item()*)"/>
      <xsl:param name="y" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
       this:random-points(1, (), $x, $y)
    '/>
   </xsl:function>

   <!--
    : this:random-points-in()
    : Generate N random points within the canvas (uniform distribution)
    :
    : $N: How many points to return
    : $canvas: canvas (space) with height/width/edge parameters
    :-->
   <xsl:function name="this:random-points-in"
                 as="map(xs:string, item()*)*"
                 new-each-time="yes">
      <xsl:param name="N" as="xs:integer"/>
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $edge := box:edge($canvas) return
      let $min-x := box:min-x($canvas) return
      let $max-x := box:max-x($canvas) return
      let $min-y := box:min-y($canvas) return
      let $max-y := box:max-y($canvas) return
      let $x := dist:uniform($min-x + $edge, $max-x - $edge)=>dist:cast("integer") return
      let $y := dist:uniform($min-y + $edge, $max-y - $edge)=>dist:cast("integer") return
      let $xs := rand:randomize($N, (), $x) return
      let $ys := rand:randomize($N, (), $y) return
      for $i in 1 to $N
      return point:point($xs[$i], $ys[$i])
    '/>
   </xsl:function>

   <!--
    : this:random-point-in()
    : Generate a random point within the canvas (uniform distribution)
    :
    : $canvas: canvas (space) with height/width/edge parameters
    :-->
   <xsl:function name="this:random-point-in"
                 as="map(xs:string, item()*)"
                 new-each-time="yes">
      <xsl:param name="canvas" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:random-points-in(1, $canvas)
    '/>
   </xsl:function>

<!--
 :======================================================================
 : More complex algorithms
 :======================================================================
 :-->

   <!--
    : this:convex-hull()
    :-->
   <xsl:function name="this:convex-hull"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $points :=
        point:distinct(
          sort(for $p in $points return this:snap($p), (), function($p) {point:px($p)})
        )
      return
      let $num-points := count($points) return
      let $start := 1 return
      let $hull-pix := 
        fold-left(
          1 to $num-points,
          $start,
          function($hull, $i) {
            $hull
            ,
            let $pix := $hull[last()] return
            let $p := $points[$pix] 
            return (
              if ($i gt 1 and $pix=$start) then ()
              else (
                fold-left(
                  1 to $num-points,
                  ($pix + 1) mod $num-points + 1 (: +1 index origin 1 :),
                  function($qix, $j) {
                    if (point:orientation($p, $points[$j], $points[$qix])="counterclockwise")
                    then $j
                    else $qix
                  }
                )
              )
            )
          }
        )
      return (
        path:polygon(this:to-edges(for $ix in $hull-pix return $points[$ix]))
      )
    '/>
   </xsl:function>

   <!--
    : this:follow()
    :-->
   <xsl:function name="this:follow" as="map(xs:string,item()*)*" visibility="private">
      <xsl:param name="n" as="xs:integer"/>
      <xsl:param name="graph" as="map(xs:string,item()*)"/>
      <xsl:param name="edges" as="map(xs:string,item()*)*"/>
      <xsl:param name="seen" as="map(xs:string,item()*)"/>
      <xsl:sequence select='
      let $last := $edges[last()] return
      let $start := ($edges[1])=>edge:start() return
      let $point := $last=>edge:end() return
      let $path-angle := point:angle(edge:start($last), $point)
      return (
        let $out-edges :=
          for $edge in (
            $graph=>graph:edges-from($point),
            ($graph=>graph:edges-to($point))!edge:reverse(.)
          )
          return (
            if (not(point:same($edge=>edge:end(),$last=>edge:start())) and
                empty($seen=>map:get(this:quote($edge))) and
                empty($seen=>map:get(this:quote(edge:reverse($edge)))))
            then $edge else ()
          )
        return
        let $angles :=
          for $edge in $out-edges return point:angle($point, edge:end($edge))
        return
        let $orientations := 
          for $edge in $out-edges return point:orientation(edge:start($last), $point, edge:end($edge))
        return
        let $next-edge := (
          for $edge in $out-edges return (
            if (point:same($edge=>edge:end(),$start)) then $edge else ()
          )
          ,
          let $out-ixes :=
            sort(
              for $i in 1 to count($out-edges) return (
                if ($orientations[$i]="counterclockwise") then $i else ()
              ), (),
              function($i) {-($path-angle - $angles[$i])}
            )
          return (
            for $i in $out-ixes return $out-edges[$i]
          )
          (:
          for $edge at $i in $out-edges
          where $orientations[$i]="counterclockwise"
          order by $path-angle - $angles[$i] descending
          return $edge
          :)
          ,
          let $out-ixes :=
            sort(
              for $i in 1 to count($out-edges) return (
                if ($orientations[$i]!="counterclockwise") then $i else ()
              ), (),
              function($i) {($path-angle - $angles[$i])}
            )
          return (
            for $i in $out-ixes return $out-edges[$i]
          )
          (: 
          for $edge at $i in $out-edges
          where $orientations[$i]!="counterclockwise"
          order by ($path-angle - $angles[$i]) ascending
          return $edge
          :)
        )[1]
        return (
          if (empty($next-edge)) then ()
          else if (point:same($next-edge=>edge:end(), $start)) then (
            path:polygon(($edges,$next-edge))
          ) else if (some $e in $edges satisfies edge:same(($next-edge,edge:edge($point:ORIGIN,$point:ORIGIN))[1],$e)) then (
            (: XYZZY the edge(origin,origin) thing is a workaround for some sort of
             : SaxonJS bug. It keeps complaining about next-edge being empty, 
             : even though we checked for it above!!
             :)
            (: was a try/catch :)
            util:log("follow() linking to middle "||this:quote($edges)),
            $edges (: Unlinkable edges :)
          ) else (
            let $seen := $seen=>map:put(this:quote($next-edge), true())
            return this:follow($n, $graph, ($edges, $next-edge), $seen)
          )
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:regions()
    : Use a simple edge-following strategy to pluck polygons out of the graph.
    :
    : $graph: the crack graph
    : $prune: prune overlapping polygons (default=false)
    :-->
   <xsl:function name="this:regions"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="prune" as="xs:boolean"/>
      <xsl:sequence select='
      let $candidates :=
        let $edges := $graph=>graph:edges() return
        let $edges := ($edges, for $e in $edges return edge:reverse($e))
        return
          fold-left(1 to count($edges), map {},
            function($data as map(xs:string,item()*)*, $i as xs:integer) as map(xs:string,item()*)* {
              let $pruned := head($data) return
              let $regions := tail($data)
              return (
                if (not($pruned=>map:contains(this:quote($edges[$i]))) and
                    not($pruned=>map:contains(this:quote(edge:reverse($edges[$i])))))
                then (
                  let $seen := map {this:quote($edges[$i]): true()} return
                  let $region := this:follow($i, $graph, $edges[$i], $seen)
                  return (
                    if (count($region) != 1 or this:kind($region)!="polygon") then (
                      $data, $region
                    ) else (
                      fold-left(this:edges($region), $pruned,
                        function($pruned as map(xs:string,item()*), $edge as map(xs:string,item()*)) as map(xs:string,item()*) {
                          $pruned=>map:put(this:quote($edge), $edge)
                        }
                      ),
                      ($regions, $region)
                    )
                  )
                ) else (
                  $data
                )
              )
            }
          )=>tail()
      return
      let $results :=
        if ($prune) then (
          let $polygons := $candidates[this:kind(.) eq "polygon"] return
          (: let $stragglers := $candidates[this:kind(.) ne "polygon"] :)
          (:
          for $polygon at $i in $polygons
          order by path:area($polygon) ascending
          :)
          sort(
            for $i in 1 to count($polygons)
            return (
              if (
                some $smaller in $polygons[position() &lt; $i] satisfies (
                  $polygons[$i]=>this:region-contains($smaller)
                )
              )
              then trace((), "pruned region "||$i)
              else $polygons[$i]
            ), (),
            function($polygon) {path:area($polygon)}
          )
        ) else (
          $candidates
        )
      return $results
    '/>
   </xsl:function>

   <!--
    : this:regions()
    : Construct polygons from the graph edges.
    :-->
   <xsl:function name="this:regions"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      this:regions($graph, false())
    '/>
   </xsl:function>

   <!--
    : this:dangling()
    :-->
   <xsl:function name="this:dangling"
                 as="map(xs:string, item()*)*"
                >
      <xsl:param name="graph" as="map(xs:string, item()*)"/>
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      for $edge in $graph=>this:edges()
      return (
        if (
          some $region in $regions satisfies (
            some $e in ($region=>this:edges()) satisfies (
              (point:same(edge:start($e), edge:start($edge)) and
               point:same(edge:end($e), edge:end($edge))) or
              (point:same(edge:start($e), edge:end($edge)) and
               point:same(edge:end($e), edge:start($edge)))
            )
          )
        )
        then ()
        else $edge
      )
    '/>
   </xsl:function>

   <!--
    : this:trivial-bounding-circle()
    : Construct minimal bounding circle for trivial case; another helper function
    :-->
   <xsl:function name="this:trivial-bounding-circle" as="map(xs:string,item()*)" visibility="private">
      <xsl:param name="points" as="map(xs:string,item()*)*"/>
      <xsl:sequence select='
      util:assert(count($points) &lt;= 3, "Function only for small numbers of points"),
      let $n-points := count($points) return
      if ($n-points=0) then ellipse:circle($point:ORIGIN, 0)
      else if ($n-points=1) then ellipse:circle($points[1], 0)
      else if ($n-points=2) then ellipse:circle(this:midpoint($points[1], $points[2]), this:distance($points[1],$points[2]) div 2)
      else if ($n-points=3) then (
        (: See if we can do the 2pt circle :)
        let $circle := ellipse:circle-from($points[1], $points[2])
        return if (ellipse:contains-point($circle, $points[3])) then $circle
        else let $circle := ellipse:circle-from($points[2], $points[3])
        return if (ellipse:contains-point($circle, $points[1])) then $circle
        else let $circle := ellipse:circle-from($points[3], $points[1])
        return if (ellipse:contains-point($circle, $points[2])) then $circle
        else ellipse:circle-from($points[1], $points[2], $points[3])
      )
      else util:assert(false(), "Cannot happen")
    '/>
   </xsl:function>

   <!--
    : this:weltz()
    : Implement the Weltz algorithm for computing the minimal bounding circle
    : of a set of points.
    :-->
   <xsl:function name="this:weltz" as="map(xs:string,item()*)" visibility="private">
      <xsl:param name="points" as="map(xs:string,item()*)*"/>
      <xsl:param name="others" as="map(xs:string,item()*)*"/>
      <xsl:sequence select='
      if (empty($points) or count($others)=3) then (
        this:trivial-bounding-circle($others)
      ) else (
        let $sub-circle := this:weltz(tail($points), $others)
        return (
          if (ellipse:contains-point($sub-circle, head($points)))
          then $sub-circle
          else this:weltz(tail($points), ($others, head($points)))
        )
      )
    '/>
   </xsl:function>

   <!--
    : this:minimal-bounding-circle()
    : Minimal bounding circle for the given set of points. 
    : If the points are ordered in some way, shuffle them for better convergence.
    :-->
   <xsl:function name="this:minimal-bounding-circle"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="points" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      this:weltz($points, ())
    '/>
   </xsl:function>

   <!--
    : this:bounding-circle()
    : Bounding circle for a region
    : Note: not necessarily minimal for non-linear-edges
    :-->
   <xsl:function name="this:bounding-circle"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="region" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      let $kind := this:kind($region) return
      if ($kind="point") then this:trivial-bounding-circle($region)
      else if ($kind="edge") then this:trivial-bounding-circle(this:points($region))
      else if ($kind="quad") then
        this:minimal-bounding-circle(this:points(this:bounding-box($region)))
      else if ($kind="cubic") then
        this:minimal-bounding-circle(this:points(this:bounding-box($region)))
      else if ($kind="arc") then 
        this:minimal-bounding-circle(this:points(this:bounding-box($region)))
      else if ($kind="path") then
        let $pts := 
          if (every $edge in this:edges($region) satisfies this:kind($edge)="edge")
          then (
            this:points($region)
          ) else (
            let $boxes := this:edges($region)!this:bounding-box(.)
            return $boxes!this:points(.)
          )
        return
          this:minimal-bounding-circle($pts)
      else if ($kind="box") then this:minimal-bounding-circle(this:points($region))
      else if ($kind="space") then this:minimal-bounding-circle(this:points($region))
      else if ($kind="ellipse") then
        if ($region=>ellipse:is-circle())
        then $region
        else ellipse:circle(ellipse:center($region), max((ellipse:rx($region), ellipse:ry($region))))
      else if ($kind="polygon") then 
        let $pts := 
          if (every $edge in this:edges($region) satisfies this:kind($edge)="edge")
          then (
            this:points($region)
          ) else (
            let $boxes := this:edges($region)!this:bounding-box(.)
            return $boxes!this:points(.)
          )
        return
          this:minimal-bounding-circle($pts)
      else errors:error("GEOM-BADREGION", ($region, "bounding-circle"))
    '/>
   </xsl:function>

   <!--
    : this:bounding-box()
    : Minimum box surrounding the region. Some approximation for non-linear
    : edges.
    :-->
   <xsl:function name="this:bounding-box"
                 as="map(xs:string, item()*)"
                >
      <xsl:param name="regions" as="map(xs:string, item()*)*"/>
      <xsl:sequence select='
      let $boxes := (
        for $region in $regions return let $kind := this:kind($region) return
        if ($kind="point") then box:box($region, $region)
        else if ($kind="edge") then edge:bounding-box($region)
        else if ($kind="quad") then edge:bounding-box($region)
        else if ($kind="cubic") then edge:bounding-box($region)
        else if ($kind="arc") then edge:bounding-box($region)
        else if ($kind="path") then path:bounding-box($region)
        else if ($kind="polygon") then path:bounding-box($region)
        else if ($kind="box") then $region
        else if ($kind="space") then box:box(box:min-point($region), box:max-point($region))
        else if ($kind="ellipse") then ellipse:bounding-box($region)
        else if ($kind="wrapper") then this:bounding-box($region=>wrapper:body())
        else if ($kind="slot") then this:bounding-box($region=>slot:body()) (: FIXME: ignores transforms :)
        else errors:error("GEOM-BADREGION", ($region, "bounding-box"))
      )
      return (
        if (empty($boxes)) then box:box(0,0,0,0)
        else if (empty(tail($boxes))) then $boxes
        else (
          let $pts := $boxes!this:points(.)
          return
            box:box(
              min($pts!point:px(.)), min($pts!point:py(.)),
              max($pts!point:px(.)), max($pts!point:py(.))
            )
        )
      )
    '/>
   </xsl:function>


   <!--
    : this:midpoint()
    :-->
   <xsl:function name="this:midpoint" as="map(xs:string, item()*)">
      <xsl:param name="a" as="map(xs:string, item()*)"/>
      <xsl:param name="b" as="map(xs:string, item()*)"/>
      <xsl:sequence select='
      edge:midpoint($a, $b)
    '/>
   </xsl:function>

   <xsl:variable name="this:DRAWING-MAP" as="map(xs:string,function(*)?)">
     <xsl:sequence select='
     map {
       "point": point:draw#3,
       "box": box:draw#3,
       "space": box:draw#3,
       "ellipse": ellipse:draw#3, 
       "edge": edge:draw#3,
       "quad": edge:draw#3,
       "cubic":edge:draw#3,
       "arc": edge:draw#3,
       "ellipse-arc": edge:draw#3,
       "path": path:draw#3,
       "polygon": path:draw#3,
       "graph": graph:draw#3,
       "slot": slot:draw#3,
       "wrapper": wrapper:draw#3,
       "mask": mask:draw#3,
       "reach": reach:draw#3,
       "block": solid:draw#3,
       "sphere": solid:draw#3,
       "ellipsoid": solid:draw#3,
       "face": solid:draw#3,
       "tetrahedron": solid:draw#3,
       "cube": solid:draw#3,
       "octahedron": solid:draw#3,
       "icosahedron": solid:draw#3,
       "dodecahedron" :solid:draw#3,
       "polyhedron": solid:draw#3
     }
     '/>
   </xsl:variable>
</xsl:package>
<!--euclidean.xsl-->
