<?xml version="1.0" encoding="UTF-8"?>
<!--
 : Testing signed distance functions and ray-casting.
 : Copyright© Mary Holstege 2020-2023
 : CC-BY (https://creativecommons.org/licenses/by/4.0/)
 : @since February 2023
 :-->
<xsl:stylesheet xmlns:this="http://mathling.com/tests/sdf"
                xmlns:test="http://mathling.com/art/test"
                xmlns:util="http://mathling.com/core/utilities"
                xmlns:f="http://mathling.com/core/callable"
                xmlns:space="http://mathling.com/type/space"
                xmlns:geom="http://mathling.com/geometric"
                xmlns:v="http://mathling.com/core/vector"
                xmlns:point="http://mathling.com/geometric/point"
                xmlns:path="http://mathling.com/geometric/path"
                xmlns:edge="http://mathling.com/geometric/edge"
                xmlns:ellipse="http://mathling.com/geometric/ellipse"
                xmlns:box="http://mathling.com/geometric/rectangle"
                xmlns:solid="http://mathling.com/geometric/solid"
                xmlns:sdf="http://mathling.com/sdf"
                xmlns:sd2="http://mathling.com/sdf/2d"
                xmlns:sd3="http://mathling.com/sdf/3d"
                xmlns:array="http://www.w3.org/2005/xpath-functions/array"
                xmlns:map="http://www.w3.org/2005/xpath-functions/map"
                xmlns:math="http://www.w3.org/2005/xpath-functions/math"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:err="http://www.w3.org/2005/xqt-errors"
                xmlns:ixsl="http://saxonica.com/ns/interactiveXSLT"
                extension-element-prefixes="ixsl"
                exclude-result-prefixes="this ixsl err util f space geom v point path edge ellipse box solid sdf sd2 sd3 test xs math map array"
                version="3.0">

  <xsl:use-package name="http://mathling.com/art/test" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/utilities" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/callable" package-version="*"/>
  <xsl:use-package name="http://mathling.com/type/space" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric" package-version="*"/>
  <xsl:use-package name="http://mathling.com/core/vector" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/point" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/path" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/edge" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/ellipse" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/rectangle" package-version="*"/>
  <xsl:use-package name="http://mathling.com/geometric/solid" package-version="*"/>
  <xsl:use-package name="http://mathling.com/sdf" package-version="*"/>
  <xsl:use-package name="http://mathling.com/sdf/2d" package-version="*"/>
  <xsl:use-package name="http://mathling.com/sdf/3d" package-version="*"/>

  <xsl:use-package name="http://mathling.com/colour/space" package-version="*"/>

<xsl:param name="inBrowser" static="true" as="xs:boolean" select="true()"/>

<xsl:variable name="ε" as="xs:double" select='xs:double(0.00001)'/>

<xsl:function name="this:testCircle2">
  <xsl:sequence select='"testCircle2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdCircle(10)=>f:function()
  return (
    test:assertThat($f((0,0)) &lt; 0),
    test:assertThat($f((0,10)) = 0),
    test:assertThat($f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $circle := ellipse:circle($point:ORIGIN, 10) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdCircle(10)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testEllipse2">
  <xsl:sequence select='"testEllipse2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdEllipse(10, 5)=>f:function()
  return (
    test:assertThat($f((0,0)) &lt;= 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) = 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) > 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $f := sd2:sdEllipse(5, 10)=>f:function()
  return (
    test:assertThat($f((0,0)) &lt;= 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) = 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $ellipse := ellipse:ellipse($point:ORIGIN, 10, 5) return
  let $ellipse2 := ellipse:ellipse($point:ORIGIN, 5, 10) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdEllipse(10, 5)=>f:function() return
  let $f2 := sd2:sdEllipse(5, 10)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:coordinates($pt)) &lt;= $ε, ellipse:contains-point($ellipse, $pt)),
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f2(point:pcoordinates($pt)), $f2(point:pcoordinates($pt)) &lt;= $ε, ellipse:contains-point($ellipse2, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testBox2">
  <xsl:sequence select='"testBox2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdBox(6, 10)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) = 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $box := box:rectangle(-3, -5, 3, 5) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdBox(6, 10)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  ),
  let $box := box:rectangle(2, 2, 8, 12) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdBox(6, 10)=>sdf:opTranslate(point:point(-5, -7))=>f:function()  return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testPolygon2">
  <xsl:sequence select='"testPolygon2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $points := (
    point:point(0,0), point:point(10,10), point:point(10,-10),
    point:point(2,-5), point:point(8,-1), point:point(0,0)
  )
  return
  let $f := sd2:sdPolygon($points)=>f:function() return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -5 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((5,0)), $f((5,0)) &lt; 0),
    test:assertThat("Fail: "||$f((3,-3)), $f((3,-3)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedBox2">
  <xsl:sequence select='"testRoundedBox2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdRoundedBox(6, 10, point:point(0,0,0,0))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0), 
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) = 0), 
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) &lt; 0), 
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $box := box:rectangle(-3, -5, 3, 5) return
  let $probes :=
    for $x in -7 to 7 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdRoundedBox(6, 10, point:point(0,0,0,0))=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  ),
  let $f := sd2:sdRoundedBox(6, 10, point:point(1,1,1,1))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) &lt; 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testOrientedBox2">
  <xsl:sequence select='"testOrientedBox2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdOrientedBox(point:point(-3,0), point:point(3,0), 10)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) = 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $box := box:rectangle(-3, -5, 3, 5) return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  let $f := sd2:sdOrientedBox(point:point(-3, 0), point:point(3, 0), 10)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSegment2">
  <xsl:sequence select='"testSegment2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdSegment(point:point(1,1), point:point(10,10))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $f := sd2:sdSegment(point:point(-3,0), point:point(3,0))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $box := box:rectangle(-3, 0, 3, 0) return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -2 to 2
    return point:point($x, $y)
  return
  let $f := sd2:sdSegment(point:point(-3, 0), point:point(3, 0))=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRhombus2">
  <xsl:sequence select='"testRhombus2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdRhombus(8, 12)=>f:function() return
  let $points := (
    point:point(-4, 0), point:point(0, 6), point:point(4, 0), point:point(0, -6)
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -12 to 12
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((2,0)), $f((2,0)) &lt; 0),
    test:assertThat("Fail: "||$f((5,5)), $f((5,5)) > 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTrapezoid2">
  <xsl:sequence select='"testTrapezoid2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdTrapezoid(8, 4, 10)=>f:function() return
  let $points := (
    point:point(-2, 5), point:point(2, 5), point:point(4, -5), point:point(-4, -5)
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -7 to 7
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((-2,5)), $f((-2,5)) = 0),
    test:assertThat("Fail: "||$f((2,5)), $f((2,5)) = 0),
    test:assertThat("Fail: "||$f((4,-5)), $f((4,-5)) = 0),
    test:assertThat("Fail: "||$f((-4,-5)), $f((-4,-5)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) = 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testParallelogram2">
  <xsl:sequence select='"testParallelogram2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdParallelogram(4, 6, 0)=>f:function() return
  let $points := (
    point:point(-2, 3), point:point(2, 3), point:point(2, -3), point:point(-2, -3)
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -6 to 6 return
    for $y in -6 to 6
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((-10,3)), $f((-10,3)) > 0),
    test:assertThat("Fail: "||$f((-2,3)), $f((-2,3)) = 0),
    test:assertThat("Fail: "||$f((2,3)), $f((2,3)) = 0),
    test:assertThat("Fail: "||$f((2,-3)), $f((2,-3)) = 0),
    test:assertThat("Fail: "||$f((-2,-3)), $f((-2,-3)) = 0),
    test:assertThat("Fail: "||$f((0,3)), $f((0,3)) = 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, path:polygon-contains-point($polygon, $pt))
    )
  ),
  let $f := sd2:sdParallelogram(4, 6, 45)=>f:function()
  return (
    test:assertThat("Fail: "||$f((-10,3)), $f((-10,3)) > 0),
    test:assertThat("Fail: "||$f((-2,-3)), $f((-2,-3)) = 0),
    test:assertThat("Fail: "||$f((2,-3)), $f((2,-3)) > 0),
    test:assertThat("Fail: "||$f((2,3)), $f((2,3)) = 0),
    test:assertThat("Fail: "||$f((-2,-3)), $f((-2,-3)) = 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testEquilateralTriangle2">
  <xsl:sequence select='"testEquilateralTriangle2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdEquilateralTriangle(2.0)=>f:function() return
  let $s3 := math:sqrt(3.0) div 3.0 return
  let $points := (
    point:point(0, 2*$s3), point:point(1, -$s3), point:point(-1, -$s3), point:point(0, 2*$s3)
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -2 to 2 return
    for $y in -2 to 2
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0.5)), $f((0,0.5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,2*$s3)), $f((0,2*$s3)) = 0),
    test:assertThat("Fail: "||$f((-1,-$s3)), abs($f((-1,-$s3))) &lt;= $ε),
    test:assertThat("Fail: "||$f((1,-$s3)), abs($f((1,-$s3))) &lt;= $ε),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testIsoscelesTriangle2">
  <xsl:sequence select='"testIsoscelesTriangle2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdIsoscelesTriangle(2, 3)=>f:function() return
  let $points := (
    point:point(0, 0), point:point(1, 3), point:point(-1, 3), point:point(0, 0)
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in -4 to 4
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,2)), $f((0,2)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((1,3)), $f((1,3)) = 0),
    test:assertThat("Fail: "||$f((-1,3)), $f((-1,3)) = 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTriangle2">
  <xsl:sequence select='"testTriangle2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $points := (
    point:point(0, 0), point:point(1, 3), point:point(-1, 3), point:point(0, 0)
  ) return
  let $f := sd2:sdTriangle($points[1], $points[2], $points[3])=>f:function() return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in -4 to 4
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,2)), $f((0,2)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((1,3)), $f((1,3)) = 0),
    test:assertThat("Fail: "||$f((-1,3)), $f((-1,3)) = 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testPentagon2">
  <xsl:sequence select='"testPentagon2"'/>
  <xsl:try>
  <xsl:sequence select='
  (:
   : Angles:
   : Interior: 360 / 5 = 72
   : Exterior: sum=(n-2)*180 = 540 divided by 10 to get right triangle as half
   : slice of pie segments from middle
   : sin(54) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(54)
   :   => d(midpoint) = d(vertex) sin(54)
   : cos(54) = ½ side / d(vertex) => side = 2 d(vertex) cos(54)
   :)
  let $f := sd2:sdPentagon(2.0)=>f:function() return
  let $side := 2*2.0*math:cos(util:radians(54)) return
  let $dmid := 2.0*math:sin(util:radians(54)) return
  let $points :=
    for $i in 1 to 5 return
    let $a := -90 + ($i - 1)*72 (: 360 / 5 :)
    return point:decimal(point:destination($point:ORIGIN, $a, 2), 2)
  return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -3 to 3 return
    for $y in -3 to 3
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0.5)), $f((0,0.5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-2)), $f((0,-2)) = 0),
    test:assertThat("Fail: "||$f((-$side div 2, $dmid)), abs($f((-$side div 2, $dmid))) &lt;= $ε),
    test:assertThat("Fail: "||$f(($side div 2, $dmid)), abs($f(($side div 2, $dmid))) &lt;= $ε),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,0)), $f((2,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      (: trace((),"f("||point:quote($pt)||")="||$f(point:pcoordinates($pt))), :)
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testHexagon2">
  <xsl:sequence select='"testHexagon2"'/>
  <xsl:try>
  <xsl:sequence select='
  (:
   : Angles:
   : Interior: 360 / 6 = 60
   : Exterior: sum=(n-2)*180 = 720 divided by 12 to get right triangle as half
   : slice of pie segments from middle = 60
   : sin(60) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(60)
   :   => d(midpoint) = d(vertex) sin(60)
   : cos(60) = ½ side / d(vertex) => side = 2 d(vertex) cos(60)
   : cos(60) = 1/2 sin(60)=½√3
   :)
  let $f := sd2:sdHexagon(2.0)=>f:function() return
  let $side := 2*2.0*math:cos(util:radians(60)) return
  let $dmid := 2.0*math:sin(util:radians(60)) return
  let $points :=
    for $i in 1 to 6 return
    let $a := ($i - 1)*60 (: 360 / 6 :)
    return point:decimal(point:destination($point:ORIGIN, $a, 2), 2)
  (: let $_ := trace((),point:quote($points)) :)
  return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -3 to 3 return
    for $y in -3 to 3
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0.5)), $f((0,0.5)) &lt; 0),
    test:assertThat("Fail: "||$f((2,0)), $f((2,0)) = 0),
    test:assertThat("Fail: "||$f((-2,0)), $f((-2,0)) = 0),
    test:assertThat("Fail: "||$f((1,$dmid)), abs($f((1,$dmid))) &lt;= $ε),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      (: trace((),"f("||point:quote($pt)||")="||$f(point:pcoordinates($pt))), :)
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testOctogon2">
  <xsl:sequence select='"testOctogon2"'/>
  <xsl:try>
  <xsl:sequence select='
  (:
   : Angles:
   : Interior: 360 / 8 = 45
   : Exterior: sum=(n-2)*180 = 1080 divided by 16 to get right triangle as half
   : slice of pie segments from middle = 67.5
   : sin(67.5) = d(midpoint)/d(vertex) => d(vertex) = d(midpoint)/sin(67.5)
   :   => d(midpoint) = d(vertex) sin(67.5)
   :)
  let $f := sd2:sdOctogon(2.0)=>f:function() return
  let $side := 2*2.0*math:cos(util:radians(67.5)) return
  let $dmid := 2.0*math:sin(util:radians(67.5)) return
  let $points :=
    for $i in 1 to 8 return
    let $a := 22.5 + ($i - 1)*45 (: 360 / 8 :)
    return point:decimal(point:destination($point:ORIGIN, $a, 2.0), 2)
  (: let $_ := trace((),point:quote($points)) :)
  return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -3 to 3 return
    for $y in -3 to 3
    return point:point($x, $y)
  return (
    test:assertThat("Fail: "||$f((0,0.5)), $f((0,0.5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,$dmid)), abs($f((0,$dmid))) &lt;= $ε),
    test:assertThat("Fail: "||$f(($dmid,0)), abs($f(($dmid,0))) &lt;= $ε),
    test:assertThat("Fail: "||$f((0,-$dmid)), abs($f((0,-$dmid))) &lt;= $ε),
    test:assertThat("Fail: "||$f((-$dmid,0)), abs($f((-$dmid,0))) &lt;= $ε),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30))),
    for $pt in $probes
    return (
      (: trace((),"f("||point:quote($pt)||")="||$f(point:pcoordinates($pt))), :)
      test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= $ε, path:polygon-contains-point($polygon, $pt))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>


<xsl:function name="this:testQuad2">
  <xsl:sequence select='"testQuad2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $quad := edge:quad(point:point(0,3), point:point(0,0), point:point(3,3)) return
  let $f := sd2:sdQuad($quad)=>f:function() return
  let $fun := sd2:udQuad($quad)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,3)), $f((0,3)) = 0),
    test:assertThat("Fail: "||$fun((0,3)), $fun((0,3)) = 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$fun((0,0)), $fun((0,0)) = 0),
    test:assertThat("Fail: "||$f((3,3)), $f((3,3)) &lt; 0),
    test:assertThat("Fail: "||$fun((3,3)), $fun((3,3)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$fun((-1,0)), $fun((-1,0)) > 0),
    test:assertThat("Fail: "||$f((-2,0)), $f((-2,0)) > 0),
    test:assertThat("Fail: "||$fun((-2,0)), $fun((-2,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0))||"&lt;"||$f((-2,0)), $f((-1,0)) &lt; $f((-2,0))),
    test:assertThat("Fail: "||$fun((-1,0))||"&lt;"||$fun((-2,0)), $fun((-1,0)) &lt; $fun((-2,0))),
    test:assertThat("Fail: "||$f((2,2))||">"||$f((3,2)), $f((2,2)) > $f((3,2))),
    test:assertThat("Fail: "||$fun((2,2))||"&lt;"||$fun((3,2)), $fun((2,2)) &lt; $fun((3,2)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testArc2">
  <xsl:sequence select='"testArc2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $arc := edge:arc-by-angle($point:ORIGIN, 2, -45, 45, false(), false()) return
  let $start := edge:start($arc) return
  let $end := edge:end($arc) return
  let $f := sd2:sdArc(180, 90, 2, 0.7)=>f:function() (: aperture points to opening :) return
  let $probes := (
    for $t in util:linspace(20, 0, 1)
    return $arc=>edge:arc-point($t)
  )
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||")="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" "||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" "||util:decimal($f(point:pcoordinates($pt)),2)) :)
      test:assertThat("Fail: "||$fp, $fp &lt;= 0)
    )
    ,
    test:assertThat("Fail: "||$f(point:pcoordinates($start)), $f(point:pcoordinates($start)) &lt;= 0),
    test:assertThat("Fail: "||$f(point:pcoordinates($end)), $f(point:pcoordinates($end)) &lt;= 0),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((-2,0)), $f((-2,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0))||"&lt;"||$f((-2,0)), $f((-1,0)) &lt; $f((-2,0))),
    test:assertThat("Fail: "||$f((2,0))||"&lt;"||$f((3,0)), $f((2,0)) &lt; $f((3,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>


<xsl:function name="this:testPie2">
  <xsl:sequence select='"testPie2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $arc := edge:arc-by-angle($point:ORIGIN, 2, -45, 45, false(), false()) return
  let $start := edge:start($arc) return
  let $end := edge:end($arc) return
  let $f := sd2:sdPie(90, 2)=>f:function() return
  let $probes := (
    for $t in util:linspace(20, 0, 1)
    return $arc=>edge:arc-point($t)
  )
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f(point:pcoordinates($start)), $f(point:pcoordinates($start)) = 0),
    test:assertThat("Fail: "||$f(point:pcoordinates($end)), $f(point:pcoordinates($end)) = 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((0,2)), $f((0,2)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-1)), $f((0,-1)) > 0),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((0,-1))||"&lt;"||$f((0,-2)), $f((0,-1)) &lt; $f((0,-2))),
    test:assertThat("Fail: "||$f((0,2))||"&lt;"||$f((0,3)), $f((0,2)) &lt; $f((0,3)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCross2">
  <xsl:sequence select='"testCross2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $arm1-points := (
    point:point(-2,4), point:point(2,4), point:point(2,-4), point:point(-2,4)
  ) return
  let $arm2-points := (
    point:point(4,-2), point:point(4,2), point:point(-4,2), point:point(4,-2)
  ) return
  let $arms := (
    path:polygon(edge:to-edges($arm1-points)),
    path:polygon(edge:to-edges($arm2-points))
  ) return
  let $f := sd2:sdCross(4, 2, 0.0)=>f:function() return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -5 to 5
    return point:point($x, $y)
  return (
    for $pt in ($arm1-points, $arm2-points) return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp,
        if ($fp > 0) then (
          every $arm in $arms satisfies not(path:polygon-contains-point($arm, $pt))
        ) else (
          true()
        )
      )
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-1)), $f((0,-1)) &lt; 0),
    test:assertThat("Fail: "||$f((5,0)), $f((5,0)) > 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedCross2">
  <xsl:sequence select='"testRoundedCross2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $edges := (
    edge:arc-by-angle(point:point(2,-2), 2, 90, 180, false(), false()),
    edge:arc-by-angle(point:point(-2,-2), 2, 0, 90, false(), false()),
    edge:arc-by-angle(point:point(-2,2), 2, 270, 0, false(), false()),
    edge:arc-by-angle(point:point(2,2), 2, 180, 270, false(), false())
  ) return
  let $path := path:path($edges)=>path:with-edge-ts() return
  let $f := sd2:sdRoundedCross(2, 1.0)=>f:function()
  return (
    for $t in util:linspace(20, 0, 1) return
    let $pt := $path=>path:path-point($t) return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,2)), $f((0,2)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((1,1)), $f((1,1)) > 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testBlobbyCross2">
  <xsl:sequence select='"testBlobbyCross2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdBlobbyCross(2, 1.0)=>f:function() return
  let $points := (
    point:point(0,2), point:point(0,-2), point:point(2,0), point:point(-2,0)
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((3,0)), $f((3,0)) > 0),
    test:assertThat("Fail: "||$f((1,1)), $f((1,1)) > 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedX2">
  <xsl:sequence select='"testRoundedX2"'/>
  <xsl:try>
  <xsl:sequence select='
  (: w=extent of bounding box => arm length = √(½w²) = w/√2
   : w=4 => a = 4/√2
   : Make a straight cross and then rotate 45°
   :)
  let $a := 4 div math:sqrt(2) return
  let $arm1-points := (
    point:point(-1, $a), point:point(1,$a), point:point(1,-$a), point:point(-1,$a)
  ) return
  let $arm2-points := (
    point:point($a,-1), point:point($a,1), point:point(-$a,1), point:point($a,-1)
  ) return
  let $arms := (
    path:polygon(edge:to-edges($arm1-points)),
    path:polygon(edge:to-edges($arm2-points))
  )!geom:rotate(.,45) return
  let $f := sd2:sdRoundedX(4, 1)=>f:function() return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -5 to 5
    return point:point($x, $y)
  return (
    for $pt in $arms!path:points(.) return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp,
        if ($fp > 0) then (
          every $arm in $arms satisfies not(path:polygon-contains-point($arm, $pt))
        ) else (
          true()
        )
      )
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,1)), $f((1,1)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,-1)), $f((-1,-1)) &lt; 0),
    test:assertThat("Fail: "||$f((2,0)), $f((2,0)) > 0),
    test:assertThat("Fail: "||$f((0,2))||"&lt;"||$f((0,3)), $f((0,2)) &lt; $f((0,3)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testStar2">
  <xsl:sequence select='"testStar2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $points := (
    for $i in 1 to 6
    return (
      point:destination($point:ORIGIN, -90 + ($i - 1)*360 div 6, 2.0),
      point:destination($point:ORIGIN, -90 + 360 div 12 + ($i - 1)*360 div 6, 1.0)
    )
  ) return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $f := sd2:sdStar(6, 2.0, 3)=>f:function() return
  let $probes :=
    for $x in -5 to 5 return
    for $y in -5 to 5
    return point:point($x, $y)
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp,
        if ($fp > $ε) then not(path:polygon-contains-point($polygon, $pt))
        else true()
      )
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,1)), $f((1,1)) = 0),
    test:assertThat("Fail: "||$f((-1,-1)), $f((-1,-1)) = 0),
    test:assertThat("Fail: "||$f((0,3)), $f((0,3)) > 0),
    test:assertThat("Fail: "||$f((0,3))||"&lt;"||$f((0,4)), $f((0,3)) &lt; $f((0,4)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testEgg2">
  <xsl:sequence select='"testEgg2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $circle := ellipse:circle($point:ORIGIN, 2) return
  let $f := sd2:sdEgg(2, 1.0)=>f:function()
  return (
    for $i in 1 to 10 return
    let $a := ($i - 1)*360 div 10 return
    let $pt := point:destination($point:ORIGIN, $a, 2) return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-2)), $f((0,-2)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,3)), $f((0,3)) > 0),
    test:assertThat("Fail: "||$f((0,3))||"&lt;"||$f((0,4)), $f((0,3)) &lt; $f((0,4)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testLens2">
  <xsl:sequence select='"testLens2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $circle := ellipse:circle($point:ORIGIN, 2) return
  let $f := sd2:sdLens(2, 0.5)=>f:function() return
  let $points := (
    point:destination($point:ORIGIN=>geom:translate(-0.5,0), 45, 2),
    point:destination($point:ORIGIN=>geom:translate(-0.5,0), -45, 2),
    point:destination($point:ORIGIN=>geom:translate(0.5,0), 180+45, 2),
    point:destination($point:ORIGIN=>geom:translate(0.5,0), 180-45, 2)
  )!point:decimal(.,3)
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((1.5,0)), $f((1.5,0)) = 0),
    test:assertThat("Fail: "||$f((-1.5,0)), $f((-1.5,0)) = 0),
    test:assertThat("Fail: "||$f((0,3)), $f((0,3)) > 0),
    test:assertThat("Fail: "||$f((0,3))||"&lt;"||$f((0,4)), $f((0,3)) &lt; $f((0,4)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCrescent2">
  <xsl:sequence select='"testCrescent2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $circle1 := ellipse:circle($point:ORIGIN, 2) return
  let $circle2 := ellipse:circle(point:point(0.5, 0), 2) return
  let $f := sd2:sdCrescent(2, 3, 2) =>f:function() return
  let $points := (
    point:point(-3,0), point:point(0,3), point:point(0,-3),
    point:point(0,0), point:point(2,2), point:point(2,-2)
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((0,-1)), $f((0,-1)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,4))||"&lt;"||$f((0,5)), $f((0,4)) &lt; $f((0,5)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testHeart2">
  <xsl:sequence select='"testHeart2"'/>
  <xsl:try>
  <xsl:sequence select='
 (: Heart shape, with point at origin
  : Centers of circular parts at +/- (0.25s,-0.75s), radius = √2/4s
  : Point between circles is (0,-s)
  :)
  let $f := sd2:sdHeart(4)=>f:function() return
  let $k := math:sqrt(2) div 4 return
  let $points := (
    point:point(0,0), point:point(0,-4),
    point:point(1+$k,-3), point:point(-1 -$k,-3)
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt;= $ε)
    ),
    test:assertThat("Fail: "||$f((0,-1)), $f((0,-1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,4))||"&lt;"||$f((0,5)), $f((0,4)) &lt; $f((0,5)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testUnevenCapsule2">
  <xsl:sequence select='"testUnevenCapsule2"'/>
  <xsl:try>
  <xsl:sequence select='
 (: Two circles smoothly joined: center of one is at origin, the other is
  : at (0,height).
  :)
  let $circle1 := ellipse:circle($point:ORIGIN, 4) return
  let $circle2 := ellipse:circle(point:point(0, 4), 2) return
  let $f := sd2:sdUnevenCapsule(4, 2, 4)=>f:function() return
  let $points := (
    point:point(4,0), point:point(-4,0), point:point(0,-4),
    point:point(2,4), point:point(-2,4), point:point(0,2)
  )
  return (
    for $i in 1 to 10 return
    let $a := ($i - 1) * 360 div 10 return
    let $pt := point:destination($point:ORIGIN, $a, 4) return
    let $pt2 := point:destination(point:point(0,4), $a, 2) return
    let $fp := $f(point:pcoordinates($pt)) return
    let $fp2 := $f(point:pcoordinates($pt2))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp,
        if ($circle1=>ellipse:contains-point($pt))
        then $fp &lt; $ε
        else true()
      ),
      (: trace((),"f("||point:quote($pt2)||") d="||util:decimal(point:distance(point:point(0,4),$pt2),2)||" a="||util:decimal(edge:angle(point:point(0,4),$pt2),2)||" f="||util:decimal($fp2,2)), :)
      test:assertThat("Fail: "||$fp2,
        if ($circle2=>ellipse:contains-point($pt2))
        then $fp2 &lt; $ε
        else true()
      )
    ),
    test:assertThat("Fail: "||$f((0,-1)), $f((0,-1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-5)), $f((0,-5)) > 0),
    test:assertThat("Fail: "||$f((0,-5))||"&lt;"||$f((0,-6)), $f((0,-5)) &lt; $f((0,-6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testParabola2">
  <xsl:sequence select='"testParabola2"'/>
  <xsl:try>
  <xsl:sequence select='
 (:
  : Parabola with minimal point at origin
  : f(x) = ax² (a=k)
  : Focus = (0, 1/(4a)), semi-latus rectum p = 1/(2a)
  :)
  let $k := 0.125 return
  let $f := sd2:sdParabola($k)=>f:function() return
  let $points := (
    point:point(1 div $k, 1 div $k), point:point(-1 div $k, 1 div $k),
    for $x in -5 to 5 return point:point($x, $k*$x*$x)
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (:trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)),:)
      test:assertThat("Fail: "||$fp, $fp &lt; $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((0,-5)), $f((0,-5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testParabolaSegment2">
  <xsl:sequence select='"testParabolaSegment2"'/>
  <xsl:try>
  <xsl:sequence select='
 (:
  : Parabola segment, running from (-w,0) to (w,0) with apex in middle at 
  : (0,h)
  :
  : f(x) = ax² + c
  : c = h, a = -h/w²
  : Focus = (0, (4ac + 1)/(4a)), semi-latus rectum p = 1/(2a)
  :)
  let $h := 2 return
  let $w := 4 return
  let $a := -$h div ($w*$w) return
  let $c := $h return
  let $f := sd2:sdParabolaSegment($w,$h)=>f:function() return
  let $points := (
    point:point(-$w, 0), point:point($w, 0), point:point(0, $h),
    for $x in -5 to 5 return point:point($x, $a*$x*$x + $c)
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt; $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((0,-5)), $f((0,-5)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testHorseshoe2">
  <xsl:sequence select='"testHorseshoe2"'/>
  <xsl:try>
  <xsl:sequence select='
 (:
  : Horseshoe: Arced base with straight arms.
  : Angle is angle of arms (degrees)
  : Center of arcs is origin.
  : Radius of center of circle = r
  : Radius of outer circle = r + w/2
  : Radius of inner circle = r - w/2
  : Distance from center to end of arms = l
  :)
  let $r := 3 return
  let $w := 2 return
  let $l := 4 return
  let $f := sd2:sdHorseshoe(0, $r, $w, $l)=>f:function() return
  let $points := (
    point:point(-($r + $w div 2), $l), point:point($r + $w div 2, $l),
    point:point(-($r + $w div 2), $l), point:point($r + $w div 2, $l),
    point:point(-($r - $w div 2), $l), point:point($r - $w div 2, $l),
    point:point(-($r - $w div 2), $l), point:point($r - $w div 2, $l),
    for $i in 1 to 10 return
    let $a := 180 + ($i - 1)*180 div 10 
    return (
      point:destination($point:ORIGIN, $a, $r + $w div 2),
      point:destination($point:ORIGIN, $a, $r - $w div 2)
    )
  )
  return (
    for $pt in $points return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") d="||util:decimal(point:distance($point:ORIGIN,$pt),2)||" a="||util:decimal(edge:angle($point:ORIGIN,$pt),2)||" f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$fp, $fp &lt; $ε)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f(($r, $l div 2)), $f(($r, $l div 2)) &lt; 0),
    test:assertThat("Fail: "||$f((0,-5)), $f((0,-5)) > 0),
    test:assertThat("Fail: "||$f((0,5))||"&lt;"||$f((0,6)), $f((0,5)) &lt; $f((0,6)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testWobble2">
  <xsl:sequence select='"testWobble2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd2:sdWobble(10, 20, 5)=>f:function() return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  return
  for $pt in $probes
  return (
    test:assertThat("Fail: "||point:quote($pt)||"=>"||$f(point:pcoordinates($pt)), util:twixt($f(point:pcoordinates($pt)), -5, 5))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<!--
<xsl:function name="this:testNoise2">
  <xsl:sequence select='"testNoise2"'/>
  <xsl:try>
  <xsl:sequence select='
  let $noise :=
    noise:scale(
      perlin:noise2(perlin:context()),
      noise:constant(1 div 100),
      noise:constant(1 div 100)
    )
  return
  let $f := sd2:sdNoise($noise)=>f:function() return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12
    return point:point($x, $y)
  for $pt in $probes
  return (
    test:assertThat(util:twixt($f(point:pcoordinates($pt)), -1, 1))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>
-->
<xsl:function name="this:test2D">
  <xsl:sequence select='"test2D"'/>
  <xsl:sequence select='util:log("test2D"),
  this:testCircle2(),
  this:testEllipse2(),
  this:testBox2(),
  this:testPolygon2(),
  this:testRoundedBox2(),
  this:testOrientedBox2(),
  this:testSegment2(),
  this:testRhombus2(), 
  this:testTrapezoid2(), 
  this:testParallelogram2(),
  this:testEquilateralTriangle2(),
  this:testIsoscelesTriangle2(), 
  this:testTriangle2(), 
  this:testPentagon2(), 
  this:testHexagon2(), 
  this:testOctogon2(),
  this:testQuad2(), 
  this:testArc2(), 
  this:testPie2(),
  this:testCross2(),
  this:testRoundedCross2(),
  this:testBlobbyCross2(),
  this:testRoundedX2(),
  this:testStar2(), 
  this:testEgg2(), 
  this:testLens2(), 
  this:testCrescent2(),
  this:testHeart2(),
  this:testUnevenCapsule2(),
  this:testParabola2(),
  this:testParabolaSegment2(),
  this:testHorseshoe2(), 
  this:testWobble2()
  (: this:testNoise2() :)
  '/>
</xsl:function>

<xsl:function name="this:testSphere3">
  <xsl:sequence select='"testSphere3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdSphere(10)=>f:function()
  return (
    test:assertThat($f((0,0,0)) &lt; 0),
    test:assertThat($f((0,10,0)) = 0),
    test:assertThat($f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,20))||"&lt;"||$f((30,30,30)), $f((20,20,20)) &lt; $f((30,30,30)))
  ),
  let $circle := ellipse:circle($point:ORIGIN, 5) return
  let $probes :=
    for $x in -6 to 6 return
    for $y in -6 to 6
    return point:point($x, $y, 0)
  return
  let $f := sd3:sdSphere(5)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $pt))
  ),
  let $circle := ellipse:circle($point:ORIGIN, 5) return
  let $probes :=
    for $x in -6 to 6 return
    for $z in -6 to 6
    return point:point($x, 0, $z)
  return
  let $f := sd3:sdSphere(5)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, point:point(point:x($pt),point:z($pt))))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testBox3">
  <xsl:sequence select='"testBox3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdBox(6, 10, 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5,1)), $f((0,5,1)) = 0),
    test:assertThat("Fail: "||$f((2,2,0)), $f((2,2,0)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  ),
  let $box := box:block(-3, -5, -1, 3, 5, 1) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12 return
    for $z in -12 to 12
    return point:point($x, $y, $z)
  return
  let $f := sd3:sdBox(6, 10, 2)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedBox3">
  <xsl:sequence select='"testRoundedBox3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdRoundedBox(6, 10, 2, 0.0)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5,1)), $f((0,5,1)) = 0),
    test:assertThat("Fail: "||$f((2,2,0)), $f((2,2,0)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  ),
  let $box := box:block(-3, -5, -1, 3, 5, 1) return
  let $probes :=
    for $x in -12 to 12 return
    for $y in -12 to 12 return
    for $z in -12 to 12
    return point:point($x, $y, $z)
  return
  let $f := sd3:sdRoundedBox(6, 10, 2, 0.0)=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals("Fail: "||point:quote($pt)||" "||$f(point:pcoordinates($pt)), $f(point:pcoordinates($pt)) &lt;= 0, box:contains-point($box, $pt))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testBoxFrame3">
  <xsl:sequence select='"testBoxFrame3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdBoxFrame(6, 10, 2, 0.5)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) > 0),
    test:assertThat("Fail: "||$f((0,5,1)), $f((0,5,1)) = 0),
    test:assertThat("Fail: "||$f((2.8,4.8,0.8)), $f((2.8,4.8,0.8)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTorus3">
  <xsl:sequence select='"testTorus3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdTorus(10, 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) > 0),
    test:assertThat("Fail: "||$f((10,0,0)), $f((10,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((8,0,0)), $f((8,0,0)) = 0),
    test:assertThat("Fail: "||$f((12,0,0)), $f((12,0,0)) = 0),
    test:assertThat("Fail: "||$f((-10,0,0)), $f((-10,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-8,0,0)), $f((-8,0,0)) = 0),
    test:assertThat("Fail: "||$f((-12,0,0)), $f((-12,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,8)), $f((0,0,8)) = 0),
    test:assertThat("Fail: "||$f((0,0,12)), $f((0,0,12)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((0,8,0)), $f((0,8,0)) > 0),
    test:assertThat("Fail: "||$f((0,12,0)), $f((0,12,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCappedTorus3">
  <xsl:sequence select='"testCappedTorus3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdCappedTorus(180, 10, 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) > 0),
    test:assertThat("Fail: "||$f((10,0,0)), $f((10,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((8,0,0)), $f((8,0,0)) = 0),
    test:assertThat("Fail: "||$f((12,0,0)), $f((12,0,0)) = 0),
    test:assertThat("Fail: "||$f((-10,0,0)), $f((-10,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-8,0,0)), $f((-8,0,0)) = 0),
    test:assertThat("Fail: "||$f((-12,0,0)), $f((-12,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,8)), $f((0,0,8)) = 0),
    test:assertThat("Fail: "||$f((0,0,12)), $f((0,0,12)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((0,8,0)), $f((0,8,0)) > 0),
    test:assertThat("Fail: "||$f((0,12,0)), $f((0,12,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  ),
  let $f := sd3:sdCappedTorus(90, 10, 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) > 0),
    test:assertThat("Fail: "||$f((10,0,0)), $f((10,0,0)) > 0),
    test:assertThat("Fail: "||$f((8,0,0)), $f((8,0,0)) > 0),
    test:assertThat("Fail: "||$f((12,0,0)), $f((12,0,0)) > 0),
    test:assertThat("Fail: "||$f((-10,0,0)), $f((-10,0,0)) > 0),
    test:assertThat("Fail: "||$f((-8,0,0)), $f((-8,0,0)) > 0),
    test:assertThat("Fail: "||$f((-12,0,0)), $f((-12,0,0)) > 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,8)), $f((0,0,8)) = 0),
    test:assertThat("Fail: "||$f((0,0,12)), $f((0,0,12)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((0,8,0)), $f((0,8,0)) > 0),
    test:assertThat("Fail: "||$f((0,12,0)), $f((0,12,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRod3">
  <xsl:sequence select='"testRod3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdRod(point:point(0,0,0), point:point(5,5,5), 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,1,1)), $f((1,1,1)) &lt; 0),
    test:assertThat("Fail: "||$f((5,5,5)), $f((5,5,5)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  ),
  let $f := sd3:sdRod(point:point(0,0,0), point:point(0,0,10), 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,-1,0)), $f((0,-1,0)) = 0),
    test:assertThat("Fail: "||$f((1,0,5)), $f((1,0,5)) = 0)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testVerticalRod3">
  <xsl:sequence select='"testVerticalRod3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdVerticalRod(10, 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((0,0,-1)), $f((0,0,-1)) = 0),
    test:assertThat("Fail: "||$f((-1,10,0)), $f((-1,10,0)) = 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testPlane3">
  <xsl:sequence select='"testPlane3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdPlane(point:point(0.0,1.0,0.0), 0.0)=>f:function()
  return (
    for $x in -1 to 1 return
    for $y in -5 to 5 return
    for $z in -1 to 1
    return test:assertThat("Fail: "||$f(($x,$y,$z)), $f(($x,$y,$z)) = $y)
  ),
  let $f := sd3:sdPlane(point:point(0.0,1.0,0.0), 2.0)=>f:function()
  return (
    for $x in -1 to 1 return
    for $y in -5 to 5 return
    for $z in -1 to 1
    return test:assertThat("Fail: "||$f(($x,$y,$z)), $f(($x,$y,$z)) = $y + 2)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testEllipsoid3">
  <xsl:sequence select='"testEllipsoid3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdEllipsoid(10, 5, 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5,0)), $f((0,5,0)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((10,0,0)), $f((10,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testOctahedron3">
  <xsl:sequence select='"testOctahedron3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdOctahedron(5)=>f:function() return
  let $f2 := sd3:sdApproxOctahedron(5)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5,0)), $f((0,5,0)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((5,0,0)), $f((5,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,5)), $f((0,0,5)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0))),
    test:assertThat("Fail: "||$f2((0,0,0)), $f2((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f2((0,5,0)), $f2((0,5,0)) = 0),
    test:assertThat("Fail: "||$f2((0,10,0)), $f2((0,10,0)) > 0),
    test:assertThat("Fail: "||$f2((5,0,0)), $f2((5,0,0)) = 0),
    test:assertThat("Fail: "||$f2((0,0,5)), $f2((0,0,5)) = 0),
    test:assertThat("Fail: "||$f2((20,20,0)), $f2((20,20,0)) > 0),
    test:assertThat("Fail: "||$f2((20,20,0))||"&lt;"||$f2((30,30,0)), $f2((20,20,0)) &lt; $f2((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testPyramid3">
  <xsl:sequence select='"testPyramid3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdPyramid(4,2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,4,0)), $f((0,4,0)) = 0),
    test:assertThat("Fail: "||$f((0,10,0)), $f((0,10,0)) > 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCone3">
  <xsl:sequence select='"testCone3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdCone(30, 2)=>f:function() return
  let $vertex := (math:tan(util:radians(30)), 0)=>v:times(2) return
  let $circle := ellipse:circle($point:ORIGIN, 1) return
  let $probes :=
    for $x in -2 to 2 return
    for $y in 0 return
    for $z in -2 to 2
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $p2))
    ),
    test:assertThat("Fail: "||$f($vertex), $f($vertex) = 0),
    test:assertThat("Fail: "||$f((0,2,0)), $f((0,2,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCappedCone3">
  <xsl:sequence select='"testCappedCone3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdCappedCone(3, 2, 1)=>f:function() return
  let $circle1 := ellipse:circle($point:ORIGIN, 1) return
  let $circle2 := ellipse:circle($point:ORIGIN, 2) return
  let $probes1 :=
    for $x in -3 to 3 return
    for $y in 3 return
    for $z in -3 to 3
    return point:point($x, $y, $z)
   return
  let $probes2 := 
    for $x in -3 to 3 return
    for $y in 0 return
    for $z in -3 to 3 
    return point:point($x, $y, $z) 
  return (
    for $pt in $probes1 return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle1, $p2))
    ),
    for $pt in $probes2 return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle2, $p2))
    ),
    test:assertThat("Fail: "||$f((0,2,0)), $f((0,2,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedCone3">
  <xsl:sequence select='"testRoundedCone3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdRoundedCone(6, 2, 1)=>f:function() return
  let $circle1 := ellipse:circle($point:ORIGIN, 1) return
  let $circle2 := ellipse:circle($point:ORIGIN, 2) return
  let $probes1 :=
    for $x in -3 to 3 return
    for $y in 5 (: 6 - r2 :) return
    for $z in -3 to 3
    return point:point($x, $y, $z)
  return
  let $probes2 := 
    for $x in -3 to 3 return
    for $y in 2 (: 0 + r1 :) return
    for $z in -3 to 3 
    return point:point($x, $y, $z) 
  return (
    for $pt in $probes1 return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle1, $p2))
    ),
    for $pt in $probes2 return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle2, $p2))
    ),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,6,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) > 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) > 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCylinder3">
  <xsl:sequence select='"testCylinder3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $center := point:point(1,1) return
  let $f := sd3:sdCylinder($center, 2)=>f:function() return
  let $circle := ellipse:circle($center, 2) return
  let $probes :=
    for $x in -3 to 3 return
    for $y in -1 to 1 return
    for $z in -3 to 3
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $p2))
    ),
    test:assertThat("Fail: "||$f((1,0,1)), $f((1,0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((1,0,3)), $f((1,0,3)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCappedCylinder3">
  <xsl:sequence select='"testCappedCylinder3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdCappedCylinder(3, 2)=>f:function() return
  let $circle := ellipse:circle($point:ORIGIN, 2) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in 0 to 3 return
    for $z in -4 to 4
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $p2))
    ),
    test:assertThat("Fail: "||$f((1,-1,1)), $f((1,-1,1)) > 0),
    test:assertThat("Fail: "||$f((1,0,1)), $f((1,0,1)) = 0),
    test:assertThat("Fail: "||$f((1,2,1)), $f((1,2,1)) &lt; 0),
    test:assertThat("Fail: "||$f((1,3,1)), $f((1,3,1)) = 0),
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,1,0)), $f((1,1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,2)), $f((0,0,2)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundedCylinder3">
  <xsl:sequence select='"testRoundedCylinder3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdRoundedCylinder(3, 2, 1)=>f:function() return
  let $circle := ellipse:circle($point:ORIGIN, 2) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in 1 to 2 (: 0 + rb to 3 - rb :) return
    for $z in -4 to 4
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt)) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertEquals($f(point:pcoordinates($pt)) &lt;= 0, ellipse:contains-point($circle, $p2))
    ),
    test:assertThat("Fail: "||$f((1,-1,1)), $f((1,-1,1)) > 0),
    test:assertThat("Fail: "||$f((1,0,1)), $f((1,0,1)) > 0),
    test:assertThat("Fail: "||$f((1,2,1)), $f((1,2,1)) &lt; 0),
    test:assertThat("Fail: "||$f((1,3,1)), $f((1,3,1)) > 0),
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((0,0,10)), $f((0,0,10)) > 0),
    test:assertThat("Fail: "||$f((1,1,0)), $f((1,1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,3,0)), $f((0,3,0)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSolidAngle3">
  <xsl:sequence select='"testSolidAngle3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdSolidAngle(30, 2)=>f:function() return
  let $d := 2 div math:tan(util:radians(30)) (: ~= 3.464 :) return
  let $rsphere := 4 (: 2/sin(30) :) return
  let $circle := ellipse:circle($point:ORIGIN, 2) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in $d return
    for $z in -4 to 4
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := util:decimal($f(point:pcoordinates($pt)), 4) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertEquals($fp &lt;= 0, ellipse:contains-point($circle, $p2))
    ),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,$rsphere,0)), $f((0,$rsphere,0)) = 0),
    test:assertThat("Fail: "||$f((1,3,1)), $f((1,3,1)) &lt; 0),
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((0,0,1)), $f((0,0,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testHexPrism3">
  <xsl:sequence select='"testHexPrism3"'/>
  <xsl:try>
  <xsl:sequence select='
  (: Height is distance from flat side to flat side 
   : Call r = d(center,vertex) and s = half side length
   : Call d = half height
   : So d/r = sin(60°) and d/s = tan(60°) => 
   : r = d/sin(60°) = 4/√3
   : s = d/tan(60°) = 2/√3
   :)
  let $f := sd3:sdHexPrism(4, 2)=>f:function() return
  let $r := 4 div math:sqrt(3) return
  let $s := 2 div math:sqrt(3) return
  let $points :=
    for $i in 1 to 6 return
    let $a := ($i - 1)*60 (: 360 / 6 :)
    return point:destination($point:ORIGIN, $a, $r)
  return
  let $polygon := path:polygon(edge:to-edges(point:decimal($points, 3))) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in -4 to 4 return
    for $z in 1
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := util:decimal($f(point:pcoordinates($pt)), 4) return
    let $p2 := point:point(point:px($pt), point:py($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertEquals("Fail: f("||point:quote($pt)||")="||$fp, $fp &lt;= 0, path:polygon-contains-point($polygon, $p2))
    ),
    for $pt in $points return
    let $p1 := point:point(point:px($pt), point:py($pt), 0) return
    let $p2 := point:point(point:px($pt), point:py($pt), 2) return
    let $fp1 := $f(point:pcoordinates($p1)) return
    let $fp2 := $f(point:pcoordinates($p2))
    return (
      test:assertThat("Fail: f("||point:quote($p1)||")="||$fp1, $fp1 &lt; $ε),
      test:assertThat("Fail: f("||point:quote($p2)||")="||$fp2, $fp2 > 0)
    ),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0), 
    test:assertThat("Fail: "||$f((1,2,0)), $f((1,2,0)) = 0),
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTriPrism3">
  <xsl:sequence select='"testTriPrism3"'/>
  <xsl:try>
  <xsl:sequence select='
  (: Height is distance from center to apex
   : Center of triangle is where lines from vertex to midpoint intersect:
   : 2/3 from vertex 1/3 from midpoint
   : so 1/2 height below total height is 2 + 1 = 3
   : Call s = length of side
   : Call h = total height (h + 1/2h)
   : So d/s = sin(60°) => s = h/sin(60°) = 3/√3
   :)
  let $h := 3 return
  let $f := sd3:sdTriPrism($h, 2)=>f:function() return
  let $s := ($h + $h div 2) div math:sqrt(3) return
  let $points := (point:point(0, $h), point:point($s, -$h div 2), point:point(-$s, -$h div 2))  return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in -4 to 4 return
    for $z in 0 to 1
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := util:decimal($f(point:pcoordinates($pt)), 4) return
    let $p2 := point:point(point:px($pt), point:py($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)||" "||path:polygon-contains-point($polygon, $p2)), :)
      test:assertEquals("Fail: f("||point:quote($pt)||")="||$fp, $fp &lt;= 0, path:polygon-contains-point($polygon, $p2))
    ),
    for $pt in $points return
    let $p1 := point:point(point:px($pt), point:py($pt), 0) return
    let $p2 := point:point(point:px($pt), point:py($pt), 2) return
    let $fp1 := $f(point:pcoordinates($p1)) return
    let $fp2 := $f(point:pcoordinates($p2))
    return (
      test:assertThat("Fail: f("||point:quote($p1)||")="||$fp1, $fp1 &lt; $ε),
      test:assertThat("Fail: f("||point:quote($p2)||")="||$fp2, $fp2 > 0)
    ),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0), 
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRhombus3">
  <xsl:sequence select='"testRhombus3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdRhombus(2, 4, 3, 0)=>f:function() return
  let $points := (
    point:point(1,0), point:point(0,-2), point:point(-1,0), point:point(0,2)
  )
  return
  let $polygon := path:polygon(edge:to-edges($points)) return
  let $probes :=
    for $x in -4 to 4 return
    for $y in 0 return
    for $z in -4 to 4
    return point:point($x, $y, $z)
  return (
    for $pt in $probes return
    let $fp := util:decimal($f(point:pcoordinates($pt)), 4) return
    let $p2 := point:point(point:px($pt), point:pz($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)||" "||path:polygon-contains-point($polygon, $p2)), :)
      test:assertEquals("Fail: f("||point:quote($pt)||")="||$fp, $fp &lt;= 0, path:polygon-contains-point($polygon, $p2))
    ),
    for $pt in $points return
    let $p1 := point:point(point:px($pt), 0, point:pz($pt)) return
    let $p2 := point:point(point:px($pt), 2, point:pz($pt)) return
    let $fp1 := $f(point:pcoordinates($p1)) return
    let $fp2 := $f(point:pcoordinates($p2))
    return (
      test:assertThat("Fail: f("||point:quote($p1)||")="||$fp1, $fp1 &lt; $ε),
      test:assertThat("Fail: f("||point:quote($p2)||")="||$fp2, $fp2 > 0)
    ),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0), 
    test:assertThat("Fail: "||$f((1,4,1)), $f((1,4,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testLink3">
  <xsl:sequence select='"testLink3"'/>
  <xsl:try>
  <xsl:sequence select='
  (: Top of link is at le + r1 + r2 (at z=0)
   : Side of link is at r1 + r2 (at z=0)
   :)
  let $f := sd3:sdLink(2, 4, 3)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) > 0), 
    test:assertThat("Fail: "||$f((0,9,0)), $f((0,9,0)) = 0),
    test:assertThat("Fail: "||$f((0,-9,0)), $f((0,-9,0)) = 0),
    test:assertThat("Fail: "||$f((7,0,0)), $f((7,0,0)) = 0),
    test:assertThat("Fail: "||$f((1,0,0)), $f((1,0,0)) = 0),
    test:assertThat("Fail: "||$f((4,0,3)), $f((4,0,3)) = 0),
    test:assertThat("Fail: "||$f((-1,0,0)), $f((-1,0,0)) = 0),
    test:assertThat("Fail: "||$f((-7,0,0)), $f((-7,0,0)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testUDTriangle3">
  <xsl:sequence select='"testUDTriangle3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sd3:udTriangle(
      point:point(0,0,0),
      point:point(5,0,0),
      point:point(3,0,3)
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((3,0,2)), $f((3,0,2)) = 0),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((2,0,0)), $f((2,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,1,0)), $f((0,1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0,3)), $f((1,0,3)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testUDQuad3">
  <xsl:sequence select='"testUDQuad3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sd3:udQuad(
      point:point(0,0,0),
      point:point(5,0,0),
      point:point(3,0,3),
      point:point(0,0,3)
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((3,0,2)), $f((3,0,2)) = 0),
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) = 0),
    test:assertThat("Fail: "||$f((2,0,0)), $f((2,0,0)) = 0),
    test:assertThat("Fail: "||$f((0,1,0)), $f((0,1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0,3)), $f((1,0,3)) = 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testWobble3">
  <xsl:sequence select='"testWobble3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sd3:sdWobble(5, 10, 2, 5)=>f:function() return
  let $probes :=
    for $x in -6 to 6 return
    for $y in -12 to 12 return
    for $z in -3 to 3
    return point:point($x, $y, $z)
  return
  for $pt in $probes
  return (
    test:assertThat("Fail: "||point:quote($pt)||"=>"||$f(point:pcoordinates($pt)), util:twixt($f(point:pcoordinates($pt)), -5, 5))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<!--
<xsl:function name="this:testNoise3">
  <xsl:sequence select='"testNoise3"'/>
  <xsl:try>
  <xsl:sequence select='
  let $noise :=
    noise:scale(
      perlin:noise4(perlin:context()),
      noise:constant(1 div 100),
      noise:constant(1 div 100)
    )
  return
  let $f := sd2:sdNoise($noise)=>f:function() return
  let $probes :=
    for $x in -3 to 3 return
    for $y in -3 to 3 return
    for $z in -3 to 3
    return point:point($x, $y, $z)
   return
  for $pt in $probes
  return (
    test:assertThat(util:twixt($f(point:pcoordinates($pt)), -1, 1))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>
-->
<xsl:function name="this:test3D">
  <xsl:sequence select='"test3D"'/>
  <xsl:sequence select='util:log("test3D"),
  this:testSphere3(),
  this:testBox3(),
  this:testRoundedBox3(),
  this:testBoxFrame3(),
  this:testTorus3(),
  this:testCappedTorus3(),
  this:testRod3(),
  this:testVerticalRod3(),
  this:testPlane3(),
  this:testEllipsoid3(),
  this:testOctahedron3(),
  this:testPyramid3(),
  this:testCone3(),
  this:testCappedCone3(),
  this:testRoundedCone3(),
  this:testCylinder3(),
  this:testCappedCylinder3(),
  this:testRoundedCylinder3(),
  this:testSolidAngle3(),
  this:testHexPrism3(),
  this:testTriPrism3(),
  this:testRhombus3(),
  this:testLink3(),
  this:testUDTriangle3(),
  this:testUDQuad3(),
  this:testWobble3()
  (: this:testNoise3() :)
  '/>
</xsl:function>

<xsl:function name="this:testConstant">
  <xsl:sequence select='"testConstant"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opConstant(12)=>f:function()
  return (
    test:assertEquals("Fail: "||$f((0,5)), $f((0,5)), 12),
    test:assertEquals("Fail: "||$f((1,2)), $f((1,2)), 12)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testMaterial">
  <xsl:sequence select='"testMaterial"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opMaterial(sdf:opConstant(3), 2)=>f:function()
  return (
    test:assertEquals("Fail: "||v:quote($f((0,5))), $f((0,5)), (3,2)),
    test:assertEquals("Fail: "||v:quote($f((1,2))), $f((1,2)), (3,2))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTranslate">
  <xsl:sequence select='"testTranslate"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opTranslate(sd2:sdSegment(point:point(-3,0), point:point(3,0)), point:point(1,1))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0 -1,0 -1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0 -1,5 -1)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2 -1,2 -1)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20 -1,20 -1)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20 -1,20 -1)) &lt; $f((30 -1,30 -1)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRotate">
  <xsl:sequence select='"testRotate"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRotate(sd2:sdSegment(point:point(-3,0), point:point(3,0)), 90)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f(point:point(0,0)=>geom:rotate(-90)=>point:pcoordinates()) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f(point:point(0,5)=>geom:rotate(-90)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f(point:point(2,2)=>geom:rotate(-90)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f(point:point(20,20)=>geom:rotate(-90)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f(point:point(20,20)=>geom:rotate(-90)=>point:pcoordinates()) &lt; $f(point:point(30,30)=>geom:rotate(-90)=>point:pcoordinates()))
  ),
  let $f :=
    sdf:opRotate(sd2:sdSegment(point:point(-3,0), point:point(3,0)), 90, 45, 30)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f(point:point(0,0)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f(point:point(0,5)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f(point:point(2,2)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f(point:point(20,20)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f(point:point(20,20)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()) &lt; $f(point:point(30,30)=>geom:rotate(-90, -45, -30)=>point:pcoordinates()))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testScale">
  <xsl:sequence select='"testScale"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opScale(sd2:sdSegment(point:point(-3,0), point:point(3,0)), 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f(point:point(0,0)=>geom:scale(0.5)=>point:pcoordinates()) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f(point:point(0,5)=>geom:scale(0.5)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f(point:point(2,2)=>geom:scale(0.5)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f(point:point(20,20)=>geom:scale(0.5)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f(point:point(20,20)=>geom:scale(0.5)=>point:pcoordinates()) &lt; $f(point:point(30,30)=>geom:scale(0.5)=>point:pcoordinates()))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testReflect">
  <xsl:sequence select='"testReflect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opReflect(sd2:sdSegment(point:point(-3,0), point:point(3,0)), point:point(1,1))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f(point:point(0,0)=>geom:reflect(point:point(1,1))=>point:pcoordinates()) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f(point:point(0,5)=>geom:reflect(point:point(1,1))=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f(point:point(2,2)=>geom:reflect(point:point(1,1))=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f(point:point(20,20)=>geom:reflect(point:point(1,1))=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f(point:point(20,20)=>geom:reflect(point:point(1,1))=>point:pcoordinates()) &lt; $f(point:point(30,30)=>geom:reflect(point:point(1,1))=>point:pcoordinates()))
  ),
  let $start := point:point(1,1) return
  let $end := point:point(2,2) return
  let $fxy := sdf:opReflectXY(sd2:sdSegment(point:point(-3,0), point:point(3,0)), $start, $end)=>f:function() return
  let $f := function ($point as xs:double*) as xs:double {v:magnitude($point)} return
  let $fxz := sdf:opReflectXZ($f, $start, $end)=>f:function() return
  let $fyz := sdf:opReflectYZ($f, $start, $end)=>f:function()
  return (
    test:assertThat("Fail: "||$fxy((0,0)), abs($fxy(point:point(0,0)=>geom:reflect($start, $end)=>point:pcoordinates())) &lt; $ε),
    test:assertThat("Fail: "||$fxy((0,5)), $fxy(point:point(0,5)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxy((2,2)), $fxy(point:point(2,2)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxy((20,20)), $fxy(point:point(20,20)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxy((20,20))||"&lt;"||$fxy((30,30)), $fxy(point:point(20,20)=>geom:reflect($start, $end)=>point:pcoordinates()) &lt; $fxy(point:point(30,30)=>geom:reflect($start, $end)=>point:pcoordinates())),
    test:assertThat("Fail: "||$fxz((0,0)), abs($fxz(point:point(0,0,0)=>geom:reflect($start, $end)=>point:pcoordinates())) &lt; $ε),
    test:assertThat("Fail: "||$fxz((0,5)), $fxz(point:point(0,0,5)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxz((2,2)), $fxz(point:point(2,0,2)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxz((20,20)), $fxz(point:point(20,0,20)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fxz((20,20))||"&lt;"||$fxz((30,30)), $fxz(point:point(20,0,20)=>geom:reflect($start, $end)=>point:pcoordinates()) &lt; $fxz(point:point(30,0,30)=>geom:reflect($start, $end)=>point:pcoordinates())),
    test:assertThat("Fail: "||$fyz((0,0)), abs($fyz(point:point(0,0,0)=>geom:reflect($start, $end)=>point:pcoordinates())) &lt; $ε),
    test:assertThat("Fail: "||$fyz((0,5)), $fyz(point:point(0,0,5)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fyz((2,2)), $fyz(point:point(0,2,2)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fyz((20,20)), $fyz(point:point(0,20,20)=>geom:reflect($start, $end)=>point:pcoordinates()) > 0),
    test:assertThat("Fail: "||$fyz((20,20))||"&lt;"||$fyz((30,30)), $fyz(point:point(0,20,20)=>geom:reflect($start, $end)=>point:pcoordinates()) &lt; $fyz(point:point(0,30,30)=>geom:reflect($start, $end)=>point:pcoordinates()))    
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRepeat">
  <xsl:sequence select='"testRepeat"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRepeat(sd2:sdSegment(point:point(-3,0), point:point(3,0)), point:point(3,6))=>f:function() return
  let $probes :=
    for $i in 0 to 6
    return point:point(3 + 3*$i, 0 + 6*$i)
  return (
    for $pt in $probes return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      test:assertThat("Fail: "||$f((0,0)), $fp = 0)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((3,6)), $f((3,6)) = 0),
    test:assertThat("Fail: "||$f((6,12)), $f((6,12)) = 0),
    test:assertThat("Fail: "||$f((9,18)), $f((9,18)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((3,11)), $f((3,11)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((5,8)), $f((5,8)) > 0)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testLimitedRepeat">
  <xsl:sequence select='"testLimitedRepeat"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opLimitedRepeat(sd2:sdSegment(point:point(-3,0), point:point(3,0)), point:point(3,6), point:point(0,0), point:point(2,2))=>f:function()
  return (
    for $i in -1 to 6 return
    let $pt := point:point(3 + 3*$i, 0 + 6*$i) return
    let $fp := $f(point:pcoordinates($pt))
    return (
      (: trace((),"f("||point:quote($pt)||") f="||util:decimal($fp,2)), :)
      if ($i >= 0 and $i &lt;= 2)
      then test:assertThat("Fail: "||$f((0,0)), $fp = 0)      
      else test:assertThat("Fail: "||$f((0,0)), $fp > 0)
    ),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((3,6)), $f((3,6)) = 0),
    test:assertThat("Fail: "||$f((6,12)), $f((6,12)) = 0),
    test:assertThat("Fail: "||$f((9,18)), $f((9,18)) > 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((3,11)), $f((3,11)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((5,8)), $f((5,8)) > 0)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRound">
  <xsl:sequence select='"testRound"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRound(sd2:sdSegment(point:point(-3,0), point:point(3,0)), 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testUnion">
  <xsl:sequence select='"testUnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opUnion(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1)
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSubtract">
  <xsl:sequence select='"testSubtract"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSubtract(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1)
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testIntersect">
  <xsl:sequence select='"testIntersect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opIntersect(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1)
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) = 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSmoothUnion">
  <xsl:sequence select='"testSmoothUnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSmoothUnion(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSmoothSubtract">
  <xsl:sequence select='"testSmoothSubtract"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSmoothSubtract(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSmoothIntersect">
  <xsl:sequence select='"testSmoothIntersect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSmoothIntersect(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundUnion">
  <xsl:sequence select='"testRoundUnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRoundUnion(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundSubtract">
  <xsl:sequence select='"testRoundSubtract"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRoundSubtract(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRoundIntersect">
  <xsl:sequence select='"testRoundIntersect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRoundIntersect(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testChamferUnion">
  <xsl:sequence select='"testChamferUnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opChamferUnion(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testChamferSubtract">
  <xsl:sequence select='"testChamferSubtract"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opChamferSubtract(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testChamferIntersect">
  <xsl:sequence select='"testChamferIntersect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opChamferIntersect(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSteppedUnion">
  <xsl:sequence select='"testSteppedUnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSteppedUnion(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1, 3
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) &lt; 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) = 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSteppedSubtract">
  <xsl:sequence select='"testSteppedSubtract"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSteppedSubtract(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1, 3
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) = 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSteppedIntersect">
  <xsl:sequence select='"testSteppedIntersect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opSteppedIntersect(
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(3,0)), 1),
      sdf:opRound(sd2:sdSegment(point:point(0,0), point:point(0,3)), 1),
      0.1, 3
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-1,0)), $f((-1,0)) > 0),
    test:assertThat("Fail: "||$f((1,0)), $f((1,0)) = 0),
    test:assertThat("Fail: "||$f((4,0)), $f((4,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,1)), $f((0,1)) = 0),
    test:assertThat("Fail: "||$f((0,5)), $f((0,5)) > 0),
    test:assertThat("Fail: "||$f((2,2)), $f((2,2)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testInvert">
  <xsl:sequence select='"testInvert"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opInvert(sd2:sdCircle(10))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20))||">"||$f((30,30)), $f((20,20)) > $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testElongate">
  <xsl:sequence select='"testElongate"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opElongate(sd2:sdCircle(10), point:point(2,0))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) &lt; 0),
    test:assertThat("Fail: "||$f((12,0)), $f((12,0)) = 0),
    test:assertThat("Fail: "||$f((-10,0)), $f((-10,0)) &lt; 0),
    test:assertThat("Fail: "||$f((-12,0)), $f((-12,0)) = 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) = 0),
    test:assertThat("Fail: "||$f((0,12)), $f((0,12)) > 0),
    test:assertThat("Fail: "||$f((0,-10)), $f((0,-10)) = 0),
    test:assertThat("Fail: "||$f((0,-12)), $f((0,-12)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testOnion">
  <xsl:sequence select='"testOnion"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opOnion(sd2:sdCircle(10), 6)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((0,7)), $f((0,7)) = 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) &lt; 0),
    test:assertThat("Fail: "||$f((0,13)), $f((0,13)) = 0),
    test:assertThat("Fail: "||$f((0,14)), $f((0,14)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  ),
  let $f := sdf:opManyOnion(sd2:sdCircle(10), 6, 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,7)), $f((0,7)) = 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) &lt; 0),
    test:assertThat("Fail: "||$f((0,13)), $f((0,13)) = 0),
    test:assertThat("Fail: "||$f((0,14)), $f((0,14)) > 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) &lt; 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  ,
  let $f := sdf:opManyOnion(sd2:sdCircle(10), 6, 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) > 0),
    test:assertThat("Fail: "||$f((0,4)), $f((0,4)) > 0),
    test:assertThat("Fail: "||$f((0,5.5)), $f((0,5.5)) = 0),
    test:assertThat("Fail: "||$f((0,7)), $f((0,7)) &lt; 0),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) > 0),
    test:assertThat("Fail: "||$f((0,13)), $f((0,13)) &lt; 0),
    test:assertThat("Fail: "||$f((0,14)), $f((0,14)) &lt; 0),
    test:assertThat("Fail: "||$f((0,14.5)), $f((0,14.5)) = 0),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )  
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testBend">
  <xsl:sequence select='"testBend"'/>
  <xsl:try>
  <xsl:sequence select='
  let $fseg := sd2:sdSegment(point:point(0,0), point:point(10,0)) return
  let $f := sdf:opBend($fseg, 10)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 0),
    test:assertThat("Fail: "||$f((5,0)), $f((5,0)) > 0),
    test:assertThat("Fail: "||$f((10,0)), $f((10,0)) > 0),
    test:assertThat("Fail: "||$f((3,0))||"&lt;"||$f((4,0)), $f((3,0)) &lt; $f((4,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>


<xsl:function name="this:testAdd">
  <xsl:sequence select='"testAdd"'/>
  <xsl:try>
  <xsl:sequence select='
  let $fseg := sd2:sdSegment(point:point(0,0), point:point(10,0)) return
  let $f := sdf:opAdd($fseg, sdf:opConstant(1))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = 1),
    test:assertThat("Fail: "||$f((5,0)), $f((5,0)) = 1),
    test:assertThat("Fail: "||$f((11,0)), $f((11,0)) = 2),
    test:assertThat("Fail: "||$f((10,10))||"&lt;"||$f((20,20)), $f((10,10)) &lt; $f((20,20)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testAddInto">
  <xsl:sequence select='"testAddInto"'/>
  <xsl:try>
  <xsl:sequence select='
  let $fseg := sdf:opMaterial(sd2:sdSegment(point:point(0,0), point:point(10,0)), 2) return
  let $f := sdf:opAddInto($fseg, sdf:opConstant(1))=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0))[1], $f((0,0))[1] = 1),
    test:assertThat("Fail: "||$f((0,0))[2], $f((0,0))[2] = 2),
    test:assertThat("Fail: "||$f((5,0))[1], $f((5,0))[1] = 1),
    test:assertThat("Fail: "||$f((5,0))[2], $f((5,0))[2] = 2),
    test:assertThat("Fail: "||$f((11,0))[1], $f((11,0))[1] = 2),
    test:assertThat("Fail: "||$f((11,0))[2], $f((11,0))[2] = 2),
    test:assertThat("Fail: "||$f((10,10))[1]||"&lt;"||$f((20,20))[1], $f((10,10))[1] &lt; $f((20,20)))[1],
    test:assertThat("Fail: "||$f((10,10))[2]||"="||$f((20,20))[2], $f((10,10))[2] = $f((20,20))[2])
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRescale">
  <xsl:sequence select='"testRescale"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opRescale(sd2:sdCircle(10), 2)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = -20),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) = 0),
    test:assertThat("Fail: "||$f((0,20)), $f((0,20)) = 20),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testExtrude">
  <xsl:sequence select='"testExtrude"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opExtrude(sd2:sdCircle(10), 2)=>f:function()
  return (
    test:assertThat($f((0,0,0)) &lt; 0),
    test:assertThat($f((0,0,1)) &lt; 0),
    test:assertThat($f((0,0,2)) = 0),
    test:assertThat($f((0,0,3)) > 0),
    test:assertThat($f((0,10,0)) = 0),
    test:assertThat($f((0,10,1)) = 0),
    test:assertThat($f((0,10,2)) = 0),
    test:assertThat($f((0,10,3)) > 0),
    test:assertThat($f((20,20,0)) > 0),
    test:assertThat($f((20,20,1)) > 0),
    test:assertThat($f((20,20,3)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0))),
    test:assertThat("Fail: "||$f((20,20,1))||"&lt;"||$f((30,30,1)), $f((20,20,1)) &lt; $f((30,30,1))),
    test:assertThat("Fail: "||$f((20,20,3))||"&lt;"||$f((30,30,3)), $f((20,20,3)) &lt; $f((30,30,3)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRevolve">
  <xsl:sequence select='"testRevolve"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opRevolve(sd2:sdCircle(5), 7)=>f:function()
  return (
    test:assertThat($f((0,0,0)) > 0),
    test:assertThat($f((12,0,0)) = 0),
    test:assertThat($f((2,0,0)) = 0),
    test:assertThat($f((10,0,0)) &lt; 0),
    test:assertThat($f((0,0,12)) = 0),
    test:assertThat($f((20,20,0)) > 0),
    test:assertThat($f((20,20,1)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testTwist">
  <xsl:sequence select='"testTwist"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:opTwist(sd3:sdBox(6, 10, 2), 1)=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0,0)), $f((0,0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,5,1)), $f((0,5,1)) > 0),
    test:assertThat("Fail: "||$f((2,2,0)), $f((2,2,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0)), $f((20,20,0)) > 0),
    test:assertThat("Fail: "||$f((20,20,0))||"&lt;"||$f((30,30,0)), $f((20,20,0)) &lt; $f((30,30,0)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testModifyDistance">
  <xsl:sequence select='"testModifyDistance"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opModifyDistance(sd2:sdCircle(10),
      function ($d as xs:double) as xs:double {
        $d * 2
      }
    )=>f:function()
  return (
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) &lt; 0),
    test:assertThat("Fail: "||$f((0,0)), $f((0,0)) = -20),
    test:assertThat("Fail: "||$f((0,10)), $f((0,10)) = 0),
    test:assertThat("Fail: "||$f((0,20)), $f((0,20)) = 20),
    test:assertThat("Fail: "||$f((20,20)), $f((20,20)) > 0),
    test:assertThat("Fail: "||$f((20,20))||"&lt;"||$f((30,30)), $f((20,20)) &lt; $f((30,30)))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testModifiers">
  <xsl:sequence select='"testModifiers"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:modAlmostIdentity(2, 1)=>f:function()
  return (
    test:assertEquals($f(2), 2),
    test:assertEquals($f(0), 1),
    test:assertEquals($f(5), 5)
  ),
  let $f := sdf:modExpImpulse(4)=>f:function()
  return (
    test:assertEquals($f(1 div 4), 1),
    test:assertEquals($f(0), 0),
    test:assertThat($f(1) > $f(2))
  ),
  let $f := sdf:modQuadImpulse(4)=>f:function()
  return (
    test:assertEquals($f(math:sqrt(1 div 4)), 1),
    test:assertEquals($f(0), 0),
    test:assertThat($f(1) > $f(2))
  ),
  let $f := sdf:modExpSustainedImpulse(2, 5)=>f:function()
  return (
    test:assertEquals($f(5), 1),
    test:assertEquals($f(0), 0),
    test:assertThat($f(1) &lt; $f(2)),
    test:assertThat($f(7) > $f(8)),
    test:assertClose($f(15), 1, $ε)
  ),
  let $f := sdf:modCubicPulse(2, 3)=>f:function()
  return (
    test:assertEquals($f(5), 0),
    test:assertEquals($f(2), 1),
    test:assertEquals($f(-1), 0),
    test:assertThat($f(1) &lt; $f(2)),
    test:assertThat($f(3) > $f(4))
  ),
  let $f := sdf:modExpStep(math:pow(1,-4)*math:log(2), 4)=>f:function() return
  let $f2 := sdf:modExpStep(math:pow(1,-10)*math:log(2), 10)=>f:function()
  return (
    test:assertEquals($f(0), 1),
    test:assertEquals($f2(0), 1),
    test:assertEquals($f(1), 0.5),
    test:assertEquals($f2(1), 0.5),
    test:assertEquals(util:decimal($f(2),4), 0),
    test:assertEquals(util:decimal($f2(2),4), 0),
    test:assertEquals(util:decimal($f(0.1),3), 1),
    test:assertEquals(util:decimal($f2(0.1),4), 1)
  ),
  let $f := sdf:modGain(1)=>f:function() return
  let $f2 := sdf:modGain(0.5)=>f:function()
  return (
    test:assertEquals($f(0), 0),
    test:assertEquals($f2(0), 0),
    test:assertEquals($f(1), 1),
    test:assertEquals($f2(1), 1),
    test:assertEquals($f(0.5), 0.5),
    test:assertEquals($f2(0.5), 0.5),
    test:assertThat($f(0.1) &lt; $f2(0.1)),
    test:assertThat($f(0.6) > $f2(0.6))
  ),
  let $f := sdf:modPowerCurve(1,2)=>f:function()
  return (
    test:assertEquals($f(0), 0),
    test:assertEquals($f(1), 0),
    for $i in 1 to 9 return test:assertThat($f($i div 10) &lt; 1)
  ),
  let $f := sdf:modSinc(4,2)=>f:function()
  return (
    test:assertEquals($f(0), 0),
    test:assertEquals($f(0.25), 2),
    for $i in 1 to 5 return test:assertThat(abs($f($i div 10)) &lt;= 2),
    for $i in 6 to 10 return test:assertThat(abs($f($i div 10)) &lt;= 1)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testOps">
  <xsl:sequence select='"testOps"'/>
  <xsl:sequence select='util:log("testOps"),
  this:testConstant(),
  this:testMaterial(),
  this:testTranslate(),
  this:testRotate(),
  this:testScale(),
  this:testReflect(), 
  this:testUnion(),
  this:testSubtract(), 
  this:testIntersect(),
  this:testRoundUnion(),
  this:testRoundSubtract(),
  this:testRoundIntersect(), 
  this:testSmoothUnion(),
  this:testSmoothSubtract(),
  this:testSmoothIntersect(), 
  this:testChamferUnion(),
  this:testChamferSubtract(),
  this:testChamferIntersect(), 
  this:testSteppedUnion(),
  this:testSteppedSubtract(),
  this:testSteppedIntersect(), 
  this:testRepeat(), 
  this:testLimitedRepeat(), 
  this:testRound(),
  this:testInvert(),
  this:testElongate(), 
  this:testOnion(),
  this:testBend(),
  this:testAdd(),
  this:testAddInto(), 
  this:testRescale(),
  this:testExtrude(), 
  this:testRevolve(), 
  this:testTwist(),
  this:testModifyDistance(),
  this:testModifiers()
  '/>
</xsl:function>

<!--
<xsl:function name="this:testRenderConstant">
  <xsl:sequence select='"testRenderConstant"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderConstant(rgb:rgb("red"))=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    test:assertEquals(
      $f(1, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderGradient">
  <xsl:sequence select='"testRenderGradient"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderGradient(rgb:rgb("red"), 0.5)=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    let $fp1 := $f(0.5, (0,0,0), (0,0,0), sd2:sdCircle(10)) return
    let $rgb1 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp1))) return
    let $fp2 := $f(1, (0,0,0), (0,0,0), sd2:sdCircle(10)) return
    let $rgb2 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp2)))
    return (
      test:assertThat($rgb1[1] &lt; 255),
      test:assertEquals($rgb1[2], 0),
      test:assertEquals($rgb1[3], 0),
      test:assertThat($rgb2[1] &lt; 255),
      test:assertEquals($rgb2[2], 0),
      test:assertEquals($rgb2[3], 0),
      test:assertThat($rgb1[1] > $rgb2[2])
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderXGradient">
  <xsl:sequence select='"testRenderXGradient"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderXGradient(rgb:rgb("red"), 0.5)=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    let $fp1 := $f(1, (0,0,0), (0.5,0,0), sd2:sdCircle(10)) return
    let $rgb1 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp1))) return
    let $fp2 := $f(1, (0,0,0), (1,0,0), sd2:sdCircle(10)) return
    let $rgb2 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp2)))
    return (
      test:assertThat($rgb1[1] &lt; 255),
      test:assertEquals($rgb1[2], 0),
      test:assertEquals($rgb1[3], 0),
      test:assertThat($rgb2[1] &lt; 255),
      test:assertEquals($rgb2[2], 0),
      test:assertEquals($rgb2[3], 0),
      test:assertThat($rgb1[1] > $rgb2[2])
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderYGradient">
  <xsl:sequence select='"testRenderYGradient"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderYGradient(rgb:rgb("red"), 0.5)=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    let $fp1 := $f(1, (0,0,0), (0,0.5,0), sd2:sdCircle(10)) return
    let $rgb1 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp1))) return
    let $fp2 := $f(1, (0,0,0), (0,1,0), sd2:sdCircle(10)) return
    let $rgb2 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp2)))
    return (
      test:assertThat($rgb1[1] &lt; 255),
      test:assertEquals($rgb1[2], 0),
      test:assertEquals($rgb1[3], 0),
      test:assertThat($rgb2[1] &lt; 255),
      test:assertEquals($rgb2[2], 0),
      test:assertEquals($rgb2[3], 0),
      test:assertThat($rgb1[1] > $rgb2[2])
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderZGradient">
  <xsl:sequence select='"testRenderZGradient"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderZGradient(rgb:rgb("red"), 0.5)=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    let $fp1 := $f(1, (0,0,0), (0,0,0.5), sd2:sdCircle(10)) return
    let $rgb1 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp1))) return
    let $fp2 := $f(1, (0,0,0), (0,0,1), sd2:sdCircle(10)) return
    let $rgb2 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp2)))
    return (
      test:assertThat($rgb1[1] &lt; 255),
      test:assertEquals($rgb1[2], 0),
      test:assertEquals($rgb1[3], 0),
      test:assertThat($rgb2[1] &lt; 255),
      test:assertEquals($rgb2[2], 0),
      test:assertEquals($rgb2[3], 0),
      test:assertThat($rgb1[1] > $rgb2[2])
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderColourPick">
  <xsl:sequence select='"testRenderColourPick"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderColourPick((rgb:rgb("blue"),rgb:rgb("red")))=>f:function() return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red"))) return
  let $xyzblue := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("blue")))
  return (
    test:assertEquals(
      $f((0, 1), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    ),
    test:assertEquals(
      $f((0, 2), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    test:assertEquals(
      $f((0, 3), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderMaterial">
  <xsl:sequence select='"testRenderMaterial"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:renderMaterial((
      sdf:renderConstant(rgb:rgb("blue")),
      sdf:renderConstant(rgb:rgb("red")))
    )=>f:function()
  return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red"))) return
  let $xyzblue := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("blue")))
  return (
    test:assertEquals(
      $f((0, 1), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    ),
    test:assertEquals(
      $f((0, 2), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    ),
    test:assertEquals(
      $f((0, 3), (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderDepth">
  <xsl:sequence select='"testRenderDepth"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderDepth("CET-L13", (0, 10))=>f:function() return
  let $gradient-colours := gradient:gradient-points("CET-L13") return
  let $low := rgb:coordinates(stop:point($gradient-colours[1]), 3) return
  let $middle := rgb:coordinates(stop:point($gradient-colours[count($gradient-colours) idiv 2]), 3) return
  let $high := rgb:coordinates(stop:point($gradient-colours[last()]), 3) return
  let $fp0 := $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)) return
  let $rgb0 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp0))) return
  let $fp5 := $f(5, (0,0,0), (0,0,1), sd2:sdCircle(10)) return
  let $rgb5 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp5))) return
  let $fp10 := $f(10, (0,0,0), (0,0,1), sd2:sdCircle(10)) return
  let $rgb10 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp10)))
  return (
    test:assertEquals($rgb0, $low),
    test:assertEquals($rgb5, $middle),
    test:assertEquals($rgb10, $high)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderPerformance">
  <xsl:sequence select='"testRenderPerformance"'/>
  <xsl:try>
  <xsl:sequence select='
  (: max-iterations = 64 :)
  let $f := sdf:renderPerformance("CET-L13")=>f:function() return
  let $gradient-colours := gradient:gradient-points("CET-L13") return
  let $low := rgb:coordinates(stop:point($gradient-colours[1]), 3) return
  let $middle := rgb:coordinates(stop:point($gradient-colours[count($gradient-colours) idiv 2]), 3) return
  let $high := rgb:coordinates(stop:point($gradient-colours[last()]), 3) return
  let $fp0 := $f((0,0,0), (0,0,0), (0,0,0), sd2:sdCircle(10)) return
  let $rgb0 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp0))) return
  let $fp32 := $f((0,0,32), (0,0,0), (0,0,1), sd2:sdCircle(10)) return
  let $rgb32 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp32))) return
  let $fp64 := $f((0,0,64), (0,0,0), (0,0,1), sd2:sdCircle(10)) return
  let $rgb64 := rgb:coordinates(cs:xyz-to-rgb(point:vector($fp64)))
  return (
    test:assertEquals($rgb0, $low),
    test:assertEquals($rgb32, $middle),
    test:assertEquals($rgb64, $high)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testRenderChoice">
  <xsl:sequence select='"testRenderChoice"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:renderChoice(1,
      sdf:renderConstant(rgb:rgb("blue")),
      sdf:renderConstant(rgb:rgb("red"))
    )=>f:function()
  return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red"))) return
  let $xyzblue := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("blue")))
  return (
    test:assertEquals(
      $f(0, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    ),
    test:assertEquals(
      $f(0.5, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    ),
    test:assertEquals(
      $f(1, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzblue
    ),
    test:assertEquals(
      $f(1.5, (0,0,0), (0,0,0), sd2:sdCircle(10)),
      $xyzred
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:rr">
  <xsl:variable name="f" as="function(*)">
    <xsl:sequence select='sdf:renderRandom((rgb:rgb("blue"),rgb:rgb("red")))=>f:function()'/>
  </xsl:variable>
  <xsl:sequence select='$f(0, (0,0,0), (0,0,0), sd2:sdCircle(10))'/>
</xsl:function>

<xsl:function name="this:testRenderRandom">
  <xsl:sequence select='"testRenderRandom (may fail occasionally)"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f := sdf:renderRandom((rgb:rgb("blue"),rgb:rgb("red")))=>f:function() return
  let $rr := function($i) as function(*) {$f} return
  let $xyzred := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("red"))) return
  let $xyzblue := point:pcoordinates(cs:rgb-to-xyz(rgb:rgb("blue"))) return
  let $renders0 :=
    for $i in 1 to 100 return
    let $f0 := $rr($i)(0, (0,0,0), (0,0,0), sd2:sdCircle(10))
    return
      if (deep-equal($f0, $xyzred)) then "red"
      else if (deep-equal($f0, $xyzblue)) then "blue"
      else "bad"
  return
  let $renders1 :=
    for $i in 1 to 100  return
    let $f1 := $rr($i)(1, (0,0,0), (0,0,0), sd2:sdCircle(10))
    return
      if (deep-equal($f1, $xyzred)) then "red"
      else if (deep-equal($f1, $xyzblue)) then "blue"
      else "bad"
  return (
    test:assertEquals(
      count(($renders0, $renders1)[.="bad"]),
      0
    ),
    test:assertTwixt("r0", count($renders0[.="red"]), 25, 75),
    test:assertTwixt("r1", count($renders1[.="red"]), 25, 75)
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>
-->

<xsl:function name="this:testRender">
  <xsl:sequence select='"testRender"'/>
  <xsl:sequence select='util:log("testRender"),
  (: this:testRenderConstant(),
  this:testRenderGradient(),  
  this:testRenderXGradient(),
  this:testRenderYGradient(),
  this:testRenderZGradient(),
  this:testRenderColourPick(),
  this:testRenderMaterial(),
  this:testRenderDepth(),
  this:testRenderPerformance(),
  this:testRenderChoice(),
  this:testRenderRandom() :) ()
  '/>
</xsl:function>

<!--
<xsl:function name="this:testColourWaves">
  <xsl:sequence select='"testColourWaves"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRenderColourWaves(
      sdf:renderConstant(rgb:rgb("red")),
      -3.0,
      75,
      0.5
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in 0 to 20 return
    let $col := $f($d, (0,0,0), (0,0,0), sd2:sdCircle(10))
    return cs:xyz-to-rgb(point:vector($col))
  
  return (
    test:assertThat(
      every $col in $colours
      satisfies deep-equal(tail(rgb:coordinates($col)), (0,0))
    ),
    test:assertThat(
      some $col in $colours
      satisfies deep-equal(rgb:coordinates($col), $red)
    ),
    test:assertThat(
      some $col in $colours
      satisfies not(deep-equal(rgb:coordinates($col), $red))
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testColourBlend">
  <xsl:sequence select='"testColourBlend"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRenderColourBlend(
      sdf:renderConstant(rgb:rgb("red")),
      rgb:rgb("blue")
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in (for $i in 0 to 10 return $i div 10) return
    let $col := $f($d, (0,0,0), (0,0,0), sd2:sdCircle(10))
    return cs:xyz-to-rgb(point:vector($col))
  return (
    test:assertEquals(rgb:coordinates($colours[1])[3], 0),
    test:assertEquals(rgb:coordinates($colours[last()])[1], 0),
    test:assertThat(
      every $i in count($colours) - 1
      satisfies rgb:coordinates($colours[$i])[1] >= rgb:coordinates($colours[$i+1])[1]
    ),
    test:assertThat(
      every $i in count($colours) - 1
      satisfies rgb:coordinates($colours[$i])[3] &lt;= rgb:coordinates($colours[$i+1])[3]
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testModifyColour">
  <xsl:sequence select='"testModifyColour"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRenderModifyColour(
      sdf:renderConstant(rgb:rgb("red")),
      rgb:rgb("blue"),
      function($d as xs:double) as xs:double {
        $d div 10
      }
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in 0 to 10 return
    let $col := $f($d, (0,0,0), (0,0,0), sd2:sdCircle(10))
    return cs:xyz-to-rgb(point:vector($col))
  return (
    test:assertEquals(rgb:coordinates($colours[1])[3], 0),
    test:assertEquals(rgb:coordinates($colours[last()])[1], 0),
    test:assertThat(
      every $i in count($colours) - 1
      satisfies rgb:coordinates($colours[$i])[1] >= rgb:coordinates($colours[$i+1])[1]
    ),
    test:assertThat(
      every $i in count($colours) - 1
      satisfies rgb:coordinates($colours[$i])[3] &lt;= rgb:coordinates($colours[$i+1])[3]
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testFuzz">
  <xsl:sequence select='"testFuzz"'/>
  <xsl:try>
  <xsl:sequence select='
  let $f :=
    sdf:opRenderFuzz(
      sdf:renderConstant(rgb:rgb("red")),
      (0,10)
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in 0 to 10 return
    let $col := $f($d, (0,0,0), (0,0,0), sd2:sdCircle(10))
    (: let $_ := trace((),"d="||$d||" c="||rgb:to-string(cs:xyz-to-rgb(point:vector($col)))) :)
    return cs:xyz-to-rgb(point:vector($col))
  return (
    test:assertEquals(tail(rgb:coordinates($colours[1])), (0, 0)),
    test:assertEquals(rgb:coordinates($colours[1])[1], $red[1]),
    test:assertEquals(rgb:coordinates($colours[last()])[1], 0),
    test:assertThat(
      every $i in count($colours) - 1
      satisfies rgb:coordinates($colours[$i])[1] >= rgb:coordinates($colours[$i+1])[1]
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testHardShadow">
  <xsl:sequence select='"testHardShadow"'/>
  <xsl:try>
  <xsl:sequence select='
  let $ro := (0,0,-1) return
  let $camera := sdf:camera(point:point(3,1,-3), point:point(0,0,0), 1) return
  let $f :=
    sdf:opRenderHardShadow(
      sdf:renderConstant(rgb:rgb("red")),
      light:diffuse-light(
        point:point(-1,3,6),
        rgb:rgb("yellow"),
        rgb:rgb("yellow"),
        0.4,
        1.0
      ),
      (0.01, 3.0), (: shadow field depth :)
      1.0, (: scale :)
      0.5 (: strength :)
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in -1 return
    for $x in -3 to 3 return
    for $y in -3 to 3 return
    for $z in 0 return
    let $rd := sdf:camera-ray-direction(point:point($x div 10, $y div 10, $z div 10), $camera) return
    let $col := $f($d, $ro, $rd, sd3:sdSphere(0.5)=>f:function())
    (: let $_ := trace((),"d="||$d||" "||$x||","||$y||","||$z||" c="||rgb:to-string(cs:xyz-to-rgb(point:vector($col)))) :)
    return cs:xyz-to-rgb(point:vector($col))
  return
  let $all-red := 255 return
  (: 0.5 = shadow strength :)
  let $shadow-red :=
    rgb:coordinates(
      cs:xyz-to-rgb(point:times(cs:rgb-to-xyz(rgb:rgb("red")), 0.5))
    )[1]
  return (
    test:assertThat(
      every $col in $colours satisfies
      deep-equal(tail(rgb:coordinates($col)), (0, 0))
    ),
    test:assertThat(
      every $col in $colours satisfies
      rgb:coordinates($col)[1] = ($red, $shadow-red)
    ),
    test:assertThat(
      some $col in $colours satisfies
      rgb:coordinates($col)[1] = $red
    ),
    test:assertThat(
      some $col in $colours satisfies
      rgb:coordinates($col)[1] = $shadow-red
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testSoftShadow">
  <xsl:sequence select='"testSoftShadow"'/>
  <xsl:try>
  <xsl:sequence select='
  let $ro := (0,0,-1) return
  let $camera := sdf:camera(point:point(3,1,-3), point:point(0,0,0), 1) return
  let $f :=
    sdf:opRenderSoftShadow(
      sdf:renderConstant(rgb:rgb("red")),
      light:diffuse-light(
        point:point(-1,3,6),
        rgb:rgb("yellow"),
        rgb:rgb("yellow"),
        0.4,
        1.0
      ),
      (0.01, 3.0), (: shadow field depth :)
      1.0, (: scale :)
      0.5, (: strength :)
      3,   (: n-shadows :)
      0.1  (: falloff :)
    )=>f:function()
  return
  let $red := rgb:coordinates(rgb:rgb("red")) return
  let $colours :=
    for $d in -1 return
    for $x in -3 to 3 return
    for $y in -3 to 3 return
    for $z in 0 return
    let $rd := sdf:camera-ray-direction(point:point($x div 10, $y div 10, $z div 10), $camera) return
    let $col := $f($d, $ro, $rd, sd3:sdSphere(0.5)=>f:function())
    (: let $_ := trace((),"d="||$d||" "||$x||","||$y||","||$z||" c="||rgb:to-string(cs:xyz-to-rgb(point:vector($col)))) :)
    return cs:xyz-to-rgb(point:vector($col))
  return
  let $all-red := 255 return
  (: 0.5 = shadow strength :)
  let $shadow-red :=
    rgb:coordinates(
      cs:xyz-to-rgb(point:times(cs:rgb-to-xyz(rgb:rgb("red")), 0.5))
    )[1]
  return (
    test:assertThat(
      every $col in $colours satisfies
      deep-equal(tail(rgb:coordinates($col)), (0, 0))
    ),
    test:assertThat(
      every $col in $colours satisfies
      rgb:coordinates($col)[1] >= $shadow-red
    ),
    test:assertThat(
      some $col in $colours satisfies
      rgb:coordinates($col)[1] = $red
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testDiffuseLight">
  <xsl:sequence select='"testDiffuseLight"'/>
  <xsl:try>
  <xsl:sequence select='
  let $ro := (0,0,-1) return
  let $camera := sdf:camera(point:point(3,1,-3), point:point(0,0,0), 1) return
  let $f :=
    sdf:opRenderDiffuseLight(
      sdf:renderXGradient(rgb:rgb("red"), 0.5),
      light:diffuse-light(
        point:point(-1,3,-6),
        rgb:rgb("blue"),
        rgb:rgb("rgb(100,0,100)"),
        0.4,
        1.0
      )
    )=>f:function()
  return
  let $colours :=
    for $d in -1 return
    for $x in -3 to 3 return
    for $y in -3 to 3 return
    for $z in 0 return
    let $rd := sdf:camera-ray-direction(point:point($x div 10, $y div 10, $z div 10), $camera) return
    let $col := $f($d, $ro, $rd, sd3:sdSphere(0.5)=>f:function())
    return cs:xyz-to-rgb(point:vector($col))
  return (
    test:assertThat(
      every $col in $colours satisfies
      rgb:coordinates($col)[2] = 0
    ),
    test:assertThat(
      not(
        deep-equal(
          rgb:coordinates($colours[1]),
          rgb:coordinates($colours[last()])
        )
      )
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>


<xsl:function name="this:testComplexLighting">
  <xsl:sequence select='"testComplexLighting"'/>
  <xsl:try>
  <xsl:sequence select='
  trace((), "testComplexLighting"),
  let $ro := (0,0,-1) return
  let $camera := sdf:camera(point:point(3,1,3), point:point(0,0,0), 1) return
  let $f :=
    sdf:opRenderComplexLighting(
      sdf:renderXGradient(rgb:rgb("red"), 0.5),
      light:diffuse-light(
        point:point(-1,3,6),
        rgb:rgb("blue"),
        rgb:rgb("rgb(100,0,100)"),
        0.4,
        1.0
      ),
      0.1,
      16
    )=>f:function()
  return
  let $colours :=
    for $d in -1 return
    for $x in -3 to 3 return
    for $y in -3 to 3 return
    for $z in 0 return
    let $rd := sdf:camera-ray-direction(point:point($x div 10, $y div 10, $z div 10), $camera) return
    let $col := $f($d, $ro, $rd, sd3:sdSphere(0.5)=>f:function())
    return cs:xyz-to-rgb(point:vector($col))
  return (
    test:assertThat(
      some $col in $colours satisfies
      not(deep-equal(rgb:coordinates($col), rgb:coordinates(rgb:rgb("red"))))
    ),
    test:assertThat(
      some $col in $colours satisfies
      not(deep-equal(rgb:coordinates($col), rgb:coordinates(rgb:rgb("blue"))))
    ),
    test:assertThat(
      some $col in $colours satisfies
      not(deep-equal(rgb:coordinates($col), (100,0,100)))
    ),
    test:assertThat(
      not(
        deep-equal(
          rgb:coordinates($colours[1]),
          rgb:coordinates($colours[last()])
        )
      )
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>
-->

<xsl:function name="this:testRenderOps">
  <xsl:sequence select='"testRenderOps"'/>
  <xsl:sequence select='util:log("testRenderOps"),
  (: this:testColourWaves(),
  this:testColourBlend(),
  this:testModifyColour(), 
  this:testFuzz(), 
  this:testHardShadow(),
  this:testSoftShadow(),
  this:testDiffuseLight(),
  this:testComplexLighting() :) ()
  '/>
</xsl:function>

<xsl:function name="this:testCastDirect">
  <xsl:sequence select='"testCastDirect"'/>
  <xsl:try>
  <xsl:sequence select='
  let $ro := (0,0,0) return
  let $sdf := sd3:sdSphere(10)=>f:function() return
  let $f := sdf:castDirect()=>f:function()
  return (
    test:assertThat($f($ro,(0,0,0), $sdf) &lt; 0),
    test:assertThat($f($ro,(0,10,0), $sdf) = 0),
    test:assertThat($f($ro,(20,20,0), $sdf) > 0),
    test:assertThat("Fail: "||$f($ro,(20,20,20), $sdf)||"&lt;"||$f($ro,(30,30,30), $sdf), $f($ro,(20,20,20), $sdf) &lt; $f($ro,(30,30,30), $sdf))
  ),
  let $circle := ellipse:circle($point:ORIGIN, 5) return
  let $probes :=
    for $x in -6 to 6 return
    for $y in -6 to 6
    return point:point($x, $y, 0)
  return
  let $ro := (0,0,0) return
  let $sdf := sd3:sdSphere(5)=>f:function() return
  let $f := sdf:castDirect()=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals($f($ro,point:pcoordinates($pt),$sdf) &lt;= 0, ellipse:contains-point($circle, $pt))
  ),
  let $circle := ellipse:circle($point:ORIGIN, 5) return
  let $probes :=
    for $x in -6 to 6 return
    for $z in -6 to 6
    return point:point($x, 0, $z)
  return
  let $ro := (0,0,0) return
  let $sdf := sd3:sdSphere(5)=>f:function() return
  let $f := sdf:castDirect()=>f:function() return
  for $pt in $probes
  return (
    test:assertEquals($f($ro,point:pcoordinates($pt),$sdf) &lt;= 0, ellipse:contains-point($circle, point:point(point:x($pt),point:z($pt))))
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCastDepth">
  <xsl:sequence select='"testCastDepth"'/>
  <xsl:try>
  <xsl:sequence select='
  let $ro := (0,0,1) return
  let $sdf := sd2:sdCircle(5 div 20)=>f:function() return
  let $field := (0.0, 5.0) return
  let $f := sdf:castDepth($field)=>f:function() return
  for $x in 0 return
  for $y in -6 to 6 return
  let $p := ($x div 20, $y div 20, -1)
  return (
    (: trace((),"f("||$x||","||$y||") => "||v:quote($f($ro, $p, $sdf))), :)
    test:assertEquals("f("||$x||","||$y||") => "||$f($ro, $p, $sdf)[1],
      $f($ro, $p, $sdf)[1] >= 0,
      abs($y) &lt;= 5
    )
  )
  ,
  let $ro := (0,0,1) return
  let $camera := sdf:camera(point:point(0,0,1), point:point(0,0,0), 1) return
  let $sdf := sd3:sdSphere(5 div 20)=>f:function() return
  let $field := (0.0, 5.0) return
  let $f := sdf:castDepth($field)=>f:function()
  return ( 
    for $x in -6 to 6 return
    for $y in 0 return
    for $z in 0 return
    let $p := point:point($x div 20,$y div 20,$z div 20) return
    let $rd := sdf:camera-ray-direction($p, $camera)
    return (
      (: trace((),"RD "||$x||","||$y||" => "||v:quote($f($ro, $rd, $sdf))), :)
      test:assertEquals("f("||$x||","||$y||","||$z||")="||$f($ro, $rd, $sdf)[1],
        ($f($ro, $rd, $sdf)[1]>=0),
        abs($x) &lt;= 5
      )
    ),
    for $x in 0 return
    for $y in -6 to 6 return
    for $z in 0 return
    let $p := point:point($x div 20,$y div 20,$z div 20) return
    let $rd := sdf:camera-ray-direction($p, $camera)
    return (
      (: trace((),"RD "||$x||","||$y||" => "||v:quote($f($ro, $rd, $sdf))), :)
      test:assertEquals("f("||$x||","||$y||","||$z||")="||$f($ro, $rd, $sdf)[1],
        ($f($ro, $rd, $sdf)[1]>=0),
        abs($y) &lt;= 5
      )
    ),
    for $x in 5 return
    for $y in -6 to 6 return
    for $z in 0 return
    let $p := point:point($x div 20,$y div 20,$z div 20) return
    let $rd := sdf:camera-ray-direction($p, $camera)
    return (
      (: trace((),"RD "||$x||","||$y||" => "||v:quote($f($ro, $rd, $sdf))), :)
      test:assertEquals("f("||$x||","||$y||","||$z||")="||$f($ro, $rd, $sdf)[1],
        ($f($ro, $rd, $sdf)[1]>=0),
        abs($y) &lt;= 1
      )
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>

<xsl:function name="this:testCast">
  <xsl:sequence select='"testCast"'/>
  <xsl:sequence select='util:log("testCast"),
  this:testCastDirect(),
  this:testCastDepth()
  '/>
</xsl:function>

<xsl:function name="this:testToSDF">
  <xsl:sequence select='"testToSDF"'/>
  <xsl:try>
  <xsl:sequence select='
  let $scale := space:space(100, 100) return
  let $default := map {} return
  let $intersect := map {"op": "intersect"} return
  let $interpolate3 := map{"interpolate": 3} return
  let $use-edge := map{"interpolate": -1} return
  let $radius3 := map{"radius": 3} return
  let $ud := map{"ud": true()} return
  let $probes :=
    for $x in -6 to 6 return
    for $y in -6 to 6
    return point:point($x, $y)
  return
  let $probes3D :=
    for $x in -6 to 6 return
    for $y in -6 to 6 return
    for $z in -6 to 6
    return point:point($x, $y, $z)
  return (
    "circleToSDF",
    let $region := ellipse:circle($point:ORIGIN, 5) return
    let $f := sd2:sdCircle(5)=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    let $region := ellipse:circle($point:ORIGIN=>geom:translate(1,1), 5) return
    let $f := sdf:opTranslate(sd2:sdCircle(5), point:point(-1,-1))=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "ellipseToSDF",
    let $region := ellipse:ellipse($point:ORIGIN, 5, 3) return
    let $f := sd2:sdEllipse(5, 3)=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function() 
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "pointToSDF",
    let $region := $point:ORIGIN return
    let $f := sd2:sdCircle(3)=>f:function() return
    let $fc := sdf:toSDF($region, $radius3)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "boxToSDF",
    let $region := box:box(-1, -2, 1, 2) return
    let $f := sd2:sdBox(2, 4)=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    let $region := box:box(-2, -3, 0, 1) return
    let $f := sdf:opTranslate(sd2:sdBox(2, 4), point:point(1,1))=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "blockToSDF",
    let $region := box:block(-1, -2, -1, 1, 2, 1) return
    let $f := sd3:sdBox(2, 4, 2)=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes3D return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "cubeToSDF",
    let $region := solid:cube(point:point(0,0,0), 2) return
    let $f := sd3:sdBox(2, 2, 2)=>f:function() return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "edgeToSDF",
    let $region := edge:edge(point:point(0,0), point:point(2,2)) return
    let $f := sd2:sdSegment(point:point(0,0), point:point(2,2))=>f:function()
    return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "quadToSDF",
    let $region := edge:quad(point:point(0,0), point:point(2,2), point:point(-1,0)) return
    let $f := sd2:sdQuad($region)=>f:function()
    return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    let $region := edge:quad(point:point(0,0), point:point(2,2), point:point(-1,0)) return
    let $f := sd2:udQuad($region)=>f:function()
    return
    let $fc := sdf:toSDF($region, $ud)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "cubicToSDF",
    let $region := edge:cubic(point:point(0,0), point:point(2,2), point:point(-1,0), point:point(1,0)) return
    let $path := edge:to-edges(geom:interpolate(3, $region)) return
    let $f :=
      fold-left(
        tail($path),
        sd2:sdSegment(
          edge:start(head($path)),
          edge:end(head($path))
        ),
        function($sdf, $edge) {
          sdf:opUnion(
            $sdf,
            sd2:sdSegment(
              edge:start($edge),
              edge:end($edge)
            )
          )
        }
      )=>f:function()
    return
    let $fc := sdf:toSDF($region, $interpolate3)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        ) 
      )
    ),
    let $region := edge:cubic(point:point(0,0), point:point(2,2), point:point(-1,0), point:point(1,0)) return
    let $path := edge:to-edges(geom:interpolate(3, $region)) return
    let $k := 0.1 return
    let $f :=
      fold-left(
        tail($path),
        sd2:sdSegment(
          edge:start(head($path)),
          edge:end(head($path))
        ),
        function($sdf, $edge) {
          sdf:opSmoothUnion(
            $sdf,
            sd2:sdSegment(
              edge:start($edge),
              edge:end($edge)
            ),
            $k
          )
        }
      )=>f:function()
   return
    let $fc := sdf:toSDF($region, $interpolate3)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertThat(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          abs($fp - $fcp) &lt; $k
        ) 
      )
    ),
    "arcToSDF",
    let $region := edge:arc-by-angle($point:ORIGIN, 2, -45, 45, false(), false()) return
    let $ra := 0.5 return
    let $f := sd2:sdArc(180, 90, 2, $ra)=>f:function()
    return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertThat(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          abs($fp - $fcp) &lt;= $ra
        )
      )
    ),
    "pathToSDF",
    let $region :=
      path:path(
        edge:to-edges((point:point(0,0), point:point(1,1), point:point(-1,1)))
      )
    return
    let $f :=
      sdf:opUnion(
        sd2:sdSegment(point:point(0,0), point:point(1,1)),
        sd2:sdSegment(point:point(1,1), point:point(-1,1))
      )=>f:function()
    return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "polygonToSDF",
    let $region :=
      path:polygon(
        edge:to-edges((point:point(0,0), point:point(1,1), point:point(-1,1)))
      ) return
    let $f :=
      sd2:sdPolygon(geom:points($region))=>f:function()
    return
    let $fc := sdf:toSDF($region, $default)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "polygonToSDF",
    let $region :=
      path:polygon(
        edge:to-edges((point:point(0,0), point:point(1,1), point:point(-1,1)))
      ) return
    let $f := sd2:sdPolygon(geom:points($region))=>f:function()
    return
    let $fc := sdf:toSDF($region, $use-edge)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    ),
    "toSDF.intersect",
    let $regions := (
      edge:edge(point:point(0,0), point:point(2,2)),
      edge:edge(point:point(-1,-1), point:point(3,3))
    )
    return
    let $f := sd2:sdSegment(point:point(0,0), point:point(2,2))=>f:function()
    return
    let $fc := sdf:toSDF($regions, $intersect)=>f:function()
    return ( 
      for $pt in $probes return
      let $fp := $f(point:pcoordinates($pt)) return
      let $fcp := $fc(point:pcoordinates($pt))
      return (
        test:assertEquals(point:quote($pt)||" f="||$fp||" fc="||$fcp,
          $fp, $fcp
        )
      )
    )
  )
  '/>
  <xsl:catch errors="*">
  <xsl:sequence select='
    "Failed ("||$err:module||"@"||$err:line-number||","||$err:column-number||") "||$err:code||":"||$err:description||" ("||util:quote($err:value)||")"
  '/>
  </xsl:catch>
  </xsl:try>
</xsl:function>


<xsl:template name="test">
  <!-- <div class="result"><xsl:sequence select='this:test2D()'/></div>
  <div class="result"><xsl:sequence select='this:test3D()'/></div>
  <div class="result"><xsl:sequence select='this:testOps()'/></div>
  <div class="result"><xsl:sequence select='this:testRender()'/></div>
  <div class="result"><xsl:sequence select='this:testRenderOps()'/></div>
  <div class="result"><xsl:sequence select='this:testCast()'/></div>
  <div class="result"><xsl:sequence select='this:testToSDF()'/></div> -->
  <div class="result"><xsl:sequence select='this:testCircle2()'/></div>
  <div>Done</div>
</xsl:template>

  <xsl:template match="*:button[@id='twoD']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:test2D()'/></div>
    </xsl:result-document>
  </xsl:template>
  <xsl:template match="*:button[@id='threeD']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:test3D()'/></div>
    </xsl:result-document>
  </xsl:template>
  <xsl:template match="*:button[@id='render']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:testRender()'/></div>
    </xsl:result-document>
  </xsl:template>
  <xsl:template match="*:button[@id='renderOps']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:testRenderOps()'/></div>
    </xsl:result-document>
  </xsl:template>
  <xsl:template match="*:button[@id='cast']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:testCast()'/></div>
    </xsl:result-document>
  </xsl:template>
  <xsl:template match="*:button[@id='toSDF']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
     <div class="result"><xsl:sequence select='this:testToSDF()'/></div>
    </xsl:result-document>
  </xsl:template>

  <xsl:template match="*:button[@id='go']" mode="ixsl:onclick">
    <xsl:result-document href="#output" method="ixsl:replace-content">
    <xsl:call-template name="test"/>
    </xsl:result-document>
  </xsl:template>

  <xsl:template match="/" name="main">
    <xsl:result-document href="#output" method="ixsl:replace-content">
    <xsl:call-template name="test"/>
    </xsl:result-document>
  </xsl:template>

  <xsl:template name="wait">
  </xsl:template>
</xsl:stylesheet>
